#!/usr/bin/env python
# Uses team configuration files to deploy code to several naos

import sys
import os
# Fix __file__ in case of script is called by windows command using cygwin
if sys.platform =='cygwin':
  __file__ = __file__.replace("\\", os.sep)  
# Add subdir include to PATH, so teamConfigParser can be found
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "include"))
from teamConfigParser import TeamConfigParser
from util import *
from pprint import pprint
from commands import getstatusoutput as runCmdSO
from commands import getstatus as runCmdS
import subprocess
import re

OS_WINDOWS = sys.platform == 'win32' or sys.platform == 'cygwin'
BHUMAN_DIR = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", ".."))
ID_RSA_NAO = os.path.join(BHUMAN_DIR, "Config", "Keys", "id_rsa_nao") + " -o StrictHostKeyChecking=no "

def usage ():
  print """usage: teamDeploy <file> ([-t <team> [<robot>...]...])|-c
  <file>    A file with a team configuration.
  -t        Deploys only specified teams or robots.
  -c        Checks if the specified <file> is a valid team configuration.
  <team>    A team name as specified in <file>.
  <robot>   A robot name as specified in <file>. Must be a member of the
            prior team."""

def runCopyfiles (team, details, robots):
  for r, r_d in details["robots"].iteritems():
    if robots != None and not r in robots:
      continue
    info("Deploying code to %s" % r)
    if OS_WINDOWS:
      cf = "copyfiles.cmd"
    else:
      cf = "copyfiles.sh"
    cmd = os.path.join(BHUMAN_DIR, "Make", cf) + " " + r_d["build"]
    cmd += " " + r_d["ip"]
    cmd += " -l " + details["location"]
    cmd += " -t " + details["color"]
    cmd += " -p " + str(r_d["player"])
    debug(cmd)
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to deploy code to %s" % r)
      return False
    else:
      success("Deployed code to %s" % r)
  return True

def changeSettings (team, details, robots):
  for r, r_d in details["robots"].iteritems():
    if robots != None and not r in robots:
      continue
    info("Changing settings on %s" % r)
    cmd = "ssh -i " + ID_RSA_NAO
    cmd += "root@" + r_d["ip"] + " "
    cmd += "\"cp /media/userdata/Config/settings.cfg "
    cmd += "/media/userdata/Config/settings.cfg.bak && "
    cmd += "cat /media/userdata/Config/settings.cfg.bak | "
    cmd += "sed -e \'s/teamNumber [0-9]*/teamNumber " + str(details["number"])
    cmd += "/\' -e \'s/teamPort [0-9]*/teamPort " + str(details["port"])
    cmd += "/\' > /media/userdata/Config/settings.cfg "
    cmd += "&& rm /media/userdata/Config/settings.cfg.bak\""
    debug(cmd)
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to change settings on %s" % r)
      return False
    else:
      success("Changed settings on %s" % r)
  return True

def changeColortable (team, details, robots):
  for r, r_d in details["robots"].iteritems():
    if robots != None and not r in robots:
      continue
    info("Deploying colortable to %s" % r)
    colorTableFilename = r_d["colorTable"]
    splitted = colorTableFilename.split(".")
    if len(splitted) < 2 or splitted[-1] != "c64":
      colorTableFilename += ".c64"
    colTableFiles = os.listdir(os.path.join(BHUMAN_DIR, "Config", "Locations", details["location"]))
    debug("location=%s" % details["location"])
    if details["location"] != "Default" and (not colorTableFilename in colTableFiles):
      info("Colortable \"%s\" does not exists in location \"%s\". Falling back on same colortable in location Default." % (r_d["colorTable"], details["location"]))
      colTableLocation = "Default"
    else:
      colTableLocation = details['location']
    cmd = "scp -i " + ID_RSA_NAO
    cmd += os.path.join(BHUMAN_DIR, "Config", "Locations", colTableLocation, colorTableFilename)
    cmd += " root@" + r_d["ip"] + ":/media/userdata/Config/Locations/"
    cmd += colTableLocation + "/coltable.c64"
    debug(cmd)
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to deploy colortable to %s" % r)
      return False
    else:
      success("Deployed colortable to %s" % r)
  return True

def changeWirelessConfig (team, details, robots):
  for r, r_d in details["robots"].iteritems():
    if robots != None and not r in robots:
      continue
    info("Changing wireless config on \"%s\" to \"%s\"" % (r, details["wlanConfig"]))
    cmd = "ssh -i " + ID_RSA_NAO
    cmd += "root@" + r_d["ip"]
    cmd += " cp /media/userdata/system/wpa_supplicant.d/wpa_supplicant.conf_" + details["wlanConfig"]
    cmd += " /media/userdata/system/wpa_supplicant.conf"
    debug(cmd)
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to change wireless config on %s (1)" % r)
      return False
    cmd = "ssh -i " + ID_RSA_NAO
    cmd += "root@" + r_d["ip"] + " /root/wireless.sh"
    debug(cmd)
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to change wireless config on %s (2)" % r)
      return False
    cmd = "ssh -i " + ID_RSA_NAO
    cmd += " -o StrictHostKeyChecking=no root@"
    cmd += r_d["ip"] + " /root/wireless.sh"
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Failed to change wireless config on %s (3)" % r)
      return False
    else:
      success("Changed wireless config on %s" % r)
  return True

def checkConfigFile (filename):
  ret = True
  teamConfig = TeamConfigParser(None)
  teamConfig.read(filename)
  teams = teamConfig.getTeams();
  if len(teams) < 1:
    error("No Teams given")
    ret = False
  for t, t_d in teams.iteritems():
    for a in ["location", "color", "number", "port", "robots", "wlanConfig"]:
      if not t_d.has_key(a):
        error("Team \"%s\" has no attribute \"%s\"" % (t, a))
        ret = False
    for r, r_d in t_d["robots"].iteritems():
      for a in ["ip", "build", "colorTable", "player"]:
        if not r_d.has_key(a):
          error("Robot \"%s\" has no attribute \"%s\"" % (r, a))
          ret = False
  if not ret:
    return ret
  for t, t_d in teams.iteritems():
    if t_d["location"] == None or not t_d["location"] in os.listdir(os.path.join(BHUMAN_DIR, "Config", "Locations")):
      error("\"%s\" is not a valid location")
      ret = False
    if t_d["color"] == None or not t_d["color"] in ["red", "blue"]:
      error("Color of \"%s\" is \"%s\" but must be \"red\" or \"blue\"." % (t, t_d["color"]))
      ret = False
    tNumber = t_d["number"]
    if tNumber == None or type(tNumber) != int or tNumber < 0:
      error("Number of team \"%s\" is \"%s\" but must be a positive integer." % (t, t_d["number"]))
      ret = False
    tPort = t_d["port"]
    if t_d["port"] == None or type(tPort) != int or tPort < 0 or tPort > 65535:
      error("Port of team \"%s\" is \"%s\" but must be a positive short." % (t, t_d["number"]))
      ret = False
    wpaPrefix = "wpa_supplicant.conf_"
    wpaFiles = os.listdir(os.path.join(BHUMAN_DIR, "Install", "files"))
    wpaFiles = filter(lambda x : x.startswith(wpaPrefix), wpaFiles)
    wpaFiles = map(lambda x : x[len(wpaPrefix):], wpaFiles)
    if not t_d["wlanConfig"] in wpaFiles:
      if t_d["wlanConfig"] == None:
        warning("No wireless configuration chosen. The configuration on the robot will not be changed.")
      else:
        error("\"%s\" is not a valid wireless configuration." % t_d["wlanConfig"])
        ret = False
    for r, r_d in t_d["robots"].iteritems():
      if r_d["ip"] == None or not is_valid_ip(r_d["ip"]):
        error("\"%s\" for robot \"%s\" is not a valid ip address." % (r_d["ip"], r))
        ret = False
      if r_d["build"] == None or not r_d["build"] in ["Debug", "Optimized", "OptimizedWithoutAssertions", "Release"]:
        error("\"%s\" for robot \"%s\" is not a valid build." % (r_d["build"], r))
        ret = False
      colTableFiles = os.listdir(os.path.join(BHUMAN_DIR, "Config", "Locations", t_d["location"]))
      if r_d["colorTable"] == None or not r_d["colorTable"] in colTableFiles and not r_d["colorTable"] + ".c64" in colTableFiles:
        warning("Colortable \"%s\" of \"%s\" is not present in location \"%s\"" % (r_d["colorTable"], r, t_d["location"]))
        colTableFiles = os.listdir(os.path.join(BHUMAN_DIR, "Config", "Locations", "Default"))
        if t_d["location"] != "Default" and (r_d["colorTable"] == None or not r_d["colorTable"] in colTableFiles and not r_d["colorTable"] + ".c64" in colTableFiles):
          warning("Colortable \"%s\" of \"%s\" is not present in location \"Default\"" % (t_d["colorTable"], r))
      player = r_d["player"]
      if player == None or type(player) != int or player < 1 or player > 3:
        error("Player number of robot \"%s\" is \"%s\" but must be between 1 and 3." % (r, str(player)))
        ret = False

  success("Seems to be a valid file.")
  return ret

def checkAvailability (robots_d, avail_robots):
  if OS_WINDOWS:
    param = "-n 1"
  else:
    param = "-c1"
  for r, r_d in robots_d.iteritems():
    if avail_robots != None and not r in avail_robots:
      continue
    info("Checking availability of robot %s" % r);
    cmd = "ping %s %s" % (param, r_d["ip"])
    cp = subprocess.Popen(args=cmd, shell=True, env=os.environ)
    status = cp.wait()
    if status != 0:
      error("Robot %s is not available" % r)
      return False
    else:
      success("Robot %s online" % r)
  return True

def is_valid_ip(ip):
    """
    Validates IP addresses.
    Only IPv4 is supported.
    """
    return is_valid_ipv4(ip)

def is_valid_ipv4(ip):
    """
    Validates IPv4 addresses.
    Only decimal notation is supported.
    """
    pattern = re.compile(r"""
        ^[1-9][0-9]?|1[0-9]{2}|2[0-4][0-9]|25[0-5]
        (\.([1-9][0-9]?|1[0-9]{2}|2[0-4][0-9]|25[0-5])){3}
    """, re.VERBOSE | re.IGNORECASE)
    return pattern.match(ip) is not None

def main ():
  
  teams = dict()
  lastTeam = None
  if len(sys.argv) < 2:
    usage()
  else:
    filename = sys.argv[1]
    if not os.access(filename, os.F_OK | os.R_OK):
      error("Cannot read file \"%s\"" % (filename))
      sys.exit(1)
    nextTeamName = False
    for arg in sys.argv[2:]:
      if arg == "-c":
          checkConfigFile(filename)
          sys.exit(0)
      elif nextTeamName:
        teams[arg] = list()
        lastTeam = arg
        nextTeamName = False
      elif arg == "-t":
        nextTeamName = True
      elif lastTeam == None:
        error("Wrong syntax")
        usage()
        sys.exit(1)
      else:   # Robot
        teams[lastTeam].append(arg)
    teamConfig = TeamConfigParser(None)
    teamConfig.read(filename)
    for t in teams:
      if not t in teamConfig.getTeams().iterkeys():
        error("Team \"%s\" not specified in config file." % t)
        sys.exit(1)
    for t, d in teamConfig.getTeams().iteritems():
      robots = None
      if lastTeam != None:
        # skip not specified teams
        if not t in teams.iterkeys():
          continue
        if len(teams[t]) > 0:
          robots = teams[t]
          for r in robots:
            if not r in d["robots"]:
              error("Robot \"%s\" is not member of team \"%s\"" % (r, t))
      if not checkAvailability(d["robots"], robots) or \
         not runCopyfiles(t, d, robots) or \
         not changeColortable(t, d, robots) or \
         not changeSettings(t, d, robots):
        sys.exit(1)
      if d["wlanConfig"] != None:
        if not changeWirelessConfig(t, d, robots):
          sys.exit(1)

if __name__ == '__main__':
  main()
