/**
* @file ControllerQt/RemoteRobot.cpp
* Implementation of the base class of processes that communicate with a remote robot.
* @author <a href="mailto:Judith.Mueller@dfki.de">Judith Mller</a>
*/

#include "RemoteRobotWithPuppet.h"
#include "ConsoleRoboCupCtrl.h"
#include "Platform/itoa.h"
#include "RoboCupCtrl.h"



RemoteRobotWithPuppet::RemoteRobotWithPuppet(const char* name, const char* ip) :
RobotConsole(theDebugReceiver,theDebugSender),
theDebugReceiver(this,"Receiver.MessageQueue.O",false),
theDebugSender(this,"Sender.MessageQueue.S",false),
bytesTransfered(0),
transferSpeed(0),
timeStamp(0),
me(0)
{
  strcpy(this->name, name);
  strcpy(this->ip, ip);
  mode = SystemCall::remoteRobotWithPuppet;

  // try to connect for one second
  Thread<RemoteRobotWithPuppet>::start(this, &RemoteRobotWithPuppet::connect);
  Thread<RemoteRobotWithPuppet>::stop();

  std::string robotName("RoboCup.puppets.");

  robotName+=name;

  // get controller
  VERIFY(rctrl = RoboCupCtrl::getController());

  // get the robot
  VERIFY(me = rctrl->getObjectReference(robotName));

  std::vector<std::string> parts;
  parts.resize(2);
  parts[0] = robotName;
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    const std::string& angleName(JointData::getJointName(JointData::Joint(i)));
    parts[1] = angleName;
    spJoints[i] = rctrl->getSensorportId(parts);
    parts[1] += ".angle";
    apJoints[i] = rctrl->getActuatorportId(parts);
    parts[1] = angleName + ".switch";
    psJoints[i] = rctrl->getActuatorportId(parts);
  }
}

void RemoteRobotWithPuppet::connect()
{
  TcpConnection::connect(*ip ? ip : 0, 0xA1BD, TcpConnection::sender);
}

void RemoteRobotWithPuppet::run()
{
  while(isRunning())
    SystemCall::sleep(processMain());
}

int RemoteRobotWithPuppet::main() 
{
  unsigned char* sendData = 0,
    * receivedData;
  int sendSize = 0,
    receivedSize = 0;
  MessageQueue temp;

  {
    // If there is something to send, prepare a package
    if(!theDebugSender.isEmpty())
    {
      SYNC;
      OutBinarySize size;
      size << theDebugSender;
      sendSize = size.getSize();
      sendData = new unsigned char[sendSize];
      OutBinaryMemory stream(sendData);
      stream << theDebugSender;
      // make backup
      theDebugSender.moveAllMessages(temp);
    }
  }

  // exchange data with the router
  if(!sendAndReceive(sendData,sendSize,receivedData,receivedSize) && sendSize)
  { // sending failed, restore theDebugSender
    SYNC;
    // move all messages since cleared (if any)
    theDebugSender.moveAllMessages(temp);
    // restore
    temp.moveAllMessages(theDebugSender);
  }

  // If a package was prepared, remove it
  if(sendSize)
  {
    delete [] sendData;
    sendSize = 0;
  }

  // If a package was received from the router program, add it to receiver queue
  if(receivedSize > 0)
  {
    SYNC;
    InBinaryMemory stream(receivedData,receivedSize);
    stream >> theDebugReceiver;

    delete [] receivedData;
  }

  return receivedSize > 0 ? 1 : 20;
}

void RemoteRobotWithPuppet::announceStop()
{
  {
    SYNC;
    debugOut.out.bin << DebugRequest("disableAll");
    debugOut.out.finishMessage(idDebugRequest);
  }
  SystemCall::sleep(1000);
  Thread<RemoteRobotWithPuppet>::announceStop();
}

void RemoteRobotWithPuppet::update()
{
  RobotConsole::update();

  if(SystemCall::getTimeSince(timeStamp) >= 2000)
  {
    int bytes = this->getOverallBytesSent() + this->getOverallBytesReceived() - bytesTransfered;
    bytesTransfered += bytes;
    timeStamp = SystemCall::getCurrentSystemTime();
    transferSpeed = bytes / 2000.0;
  }

  char buf[20];
  sprintf(buf, "%.1lf kb/s", transferSpeed);
  std::string statusText = robotName.substr(robotName.find_last_of(".") + 1) + ": connected to " + 
    ip + ", " + buf;

  if(logPlayer.getNumberOfMessages() != 0)
  {
    char buf[10];
    statusText += std::string(", recorded ") + itoa_s(logPlayer.numberOfFrames, buf, sizeof(buf), 10);
  }

  if(pollingFor)
  {
    statusText += statusText != "" ? ", polling for " : "polling for ";
    statusText += pollingFor;
  }


  ASSERT(me);

  //You need to put here more things the puppet should do, like fieldposition
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    // Set angles
    const double& targetAngle(jointData.angles[i]);
    if(targetAngle != JointData::off && 
      targetAngle != JointData::ignore &&
      spJoints[i] != -1) // if joint does exist
      ctrl->setActuatorport(apJoints[i], (targetAngle + jointCalibration.joints[i].offset) * jointCalibration.joints[i].sign);
  }

  if(moveOp != noMove)
  {
    if(moveOp == moveBoth)
      ctrl->moveObject(me, movePos * 0.001, moveRot * (pi / 180));
    else if(moveOp == movePosition)
      ctrl->moveObject(me, movePos * 0.001);
    moveOp = noMove;
  }

  ctrl->printStatusText(statusText);
}


