/**
* @file ControllerQt/Views/ImageView.cpp
*
* Implementation of class ImageView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Colin Graf
*/

#include "ImageView.h"
#include "RobotConsole.h"
#include <QPainter>
#include <QApplication>
#include <QMouseEvent>
#include "Representations/Infrastructure/Image.h"
#include "Platform/Thread.h"
#include "Visualization/PaintMethods.h"
#include "ControllerQt/ImageViewAdapter.h"

class ImageWidget : public QWidget
{
public:
  ImageWidget(RobotConsole& console, ImageView& imageView, const std::string& background, const std::string& name, bool segmented) :
    console(console),
    imageView(imageView),
    background(background),
    name(name),
    segmented(segmented), 
    imageData(0), imageWidth(0), imageHeight(0),
    lastImageTimeStamp(0), lastColorTableTimeStamp(0), lastDrawingsTimeStamp(0),
    dragStart(-1, -1), zoom(1.f), offset(0,0)
  {
    setMouseTracking(true);
  }

  ~ImageWidget()
  {
    if(imageData)
      delete imageData;
  }
  
  void update()
  {
    if(needsRepaint())
      QWidget::update();
  }

private:
  RobotConsole& console;
  ImageView& imageView;
  const std::string& background;
  const std::string& name;
  bool segmented;
  QImage* imageData;
  int imageWidth;
  int imageHeight;
  unsigned int lastImageTimeStamp;
  unsigned int lastColorTableTimeStamp;
  unsigned int lastDrawingsTimeStamp;
  QPainter painter;
  QPoint dragStart;
  QPoint dragPos;
  float zoom;
  QPoint offset;

  void paintEvent(QPaintEvent *event)
  {
    painter.begin(this);
    
    {
      SYNC_WITH(console);

      const Image* image = 0;
      RobotConsole::Images::const_iterator i = console.images.find(background);
      if(i != console.images.end())
        image = i->second.image;
      if(image)
      {
        imageWidth = image->cameraInfo.resolutionWidth;
        imageHeight = image->cameraInfo.resolutionHeight;
      }
      else
      {
        imageWidth = cameraResolutionWidth;
        imageHeight = cameraResolutionHeight;
      }

      float scale, xOffset, yOffset;
      {
        const QSize& size(this->size());
        float xScale = float(size.width()) / float(imageWidth);
        float yScale = float(size.height()) / float(imageHeight);
        scale = xScale < yScale ? xScale : yScale;
        scale *= zoom;
        xOffset = (float(size.width()) - float(imageWidth) * scale) * 0.5f + offset.x()*scale;
        yOffset = (float(size.height()) - float(imageHeight) * scale) * 0.5f + offset.y()*scale;
      }

      painter.setTransform(QTransform(scale, 0, 0, scale, xOffset, yOffset));

      if(image)
        paintImage(*image);
      else
        lastImageTimeStamp = 0;
      
      paintDrawings();
    }

    if(dragStart.x() >= 0)
    {
      QPen pen(QColor(128, 128, 128, 128));
      pen.setWidth(1);
      painter.setPen(pen);
      painter.setBrush(QBrush(Qt::NoBrush));
      painter.drawRect(QRect(dragStart, dragPos));
    }
    
    painter.end();
  }

  void paintDrawings()
  {
    const QTransform baseTrans(painter.transform());
    const std::list<std::string>& drawings(console.imageViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.imageDrawings[*i]);
      PaintMethods::paintDebugDrawing(painter, debugDrawing, baseTrans);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        lastDrawingsTimeStamp = debugDrawing.timeStamp;
    }
    painter.setTransform(baseTrans);
  }

  void copyImage(const Image& srcImage)
  {
    int width = srcImage.cameraInfo.resolutionWidth;
    int height = srcImage.cameraInfo.resolutionHeight;

    static const int factor1 = 29016;
    static const int factor2 = 5662;
    static const int factor3 = 22972;
    static const int factor4 = 11706;
    unsigned* p = (unsigned*) imageData->bits();
    int r, g, b;
    int yImage, uImage, vImage;
    for(int y = 0; y < height; ++y) 
    {
      const Image::Pixel* cur = &srcImage.image[y][0];
      const Image::Pixel* end = cur + width;
      for(; cur < end; ++cur)
      {        
        yImage = int(cur->y) << 14;
        uImage = int(cur->cr) - 128;
        vImage = int(cur->cb) - 128;

        r = (yImage + factor3 * uImage) >> 14;
        g = (yImage - factor2 * vImage - factor4 * uImage) >> 14;
        b = (yImage + factor1 * vImage) >> 14;

        *p++ = (r < 0 ? 0 : r > 255 ? 255 : r) << 16 |
               (g < 0 ? 0 : g > 255 ? 255 : g) << 8 |
               (b < 0 ? 0 : b > 255 ? 255 : b) |
               0xff000000;
      }
    }
  }

  void copyImageSegmented(const Image& srcImage)
  {
    int width = srcImage.cameraInfo.resolutionWidth;
    int height = srcImage.cameraInfo.resolutionHeight;
    static const unsigned colorClassColors[] = {
      0xff7f7f7f, // none
      0xffff7f00, // orange
      0xffffff00, // yellow
      0xff0000ff, // blue
      0xffffffff, // white
      0xff00ff00, // green
      0xff000000, // black
      0xffff0000, // red
      0xff00ffff, // robotBlue
    };
    unsigned* p = (unsigned*) imageData->bits();
    unsigned char (*colorClasses)[64][64](console.colorTableHandler.colorTable.colorClasses);
    for(int y = 0; y < height; ++y) 
    {
      const Image::Pixel* cur = &srcImage.image[y][0];
      const Image::Pixel* end = cur + width;
      for(; cur < end; ++cur)
      {
        unsigned char clrclass = colorClasses[cur->y >> 2][cur->cb >> 2][cur->cr >> 2];
        *p++ = colorClassColors[clrclass > 8 ? 0 : clrclass];
      }
    }
  }

  void paintImage(const Image& srcImage)
  {
    // make sure we have a buffer
    if(!imageData || imageWidth != imageData->width() || imageHeight != imageData->height())
    {
      if(imageData)
        delete imageData;
      imageData = new QImage(imageWidth, imageHeight, QImage::Format_RGB32);
    }

    if(srcImage.timeStamp != lastImageTimeStamp || (segmented && console.colorTableHandler.timeStamp != lastColorTableTimeStamp))
    {
      if(segmented)
        copyImageSegmented(srcImage);
      else
        copyImage(srcImage);
      
      lastImageTimeStamp = srcImage.timeStamp;
      if(segmented)
        lastColorTableTimeStamp = console.colorTableHandler.timeStamp;
    }
    
    painter.drawImage(QRectF(0, 0, imageWidth, imageHeight), *imageData);
  }

  bool needsRepaint() const
  {
    SYNC_WITH(console);
    const std::list<std::string>& drawings(console.imageViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.imageDrawings[*i]);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        return true;
    }
    Image* image = 0;
    RobotConsole::Images::const_iterator j = console.images.find(background);
    if(j != console.images.end())
      image = j->second.image;
    return (image && 
           (image->timeStamp != lastImageTimeStamp ||
            (segmented && console.colorTableHandler.timeStamp != lastColorTableTimeStamp))) ||
           (!image && lastImageTimeStamp);
  }

  void window2viewport(QPoint& point)
  {
    float scale, xOffset, yOffset;
    {
      const QSize& size(this->size());
      float xScale = float(size.width()) / float(imageWidth);
      float yScale = float(size.height()) / float(imageHeight);
      scale = xScale < yScale ? xScale : yScale;
      scale *= zoom;
      xOffset = (float(size.width()) - float(imageWidth) * scale) * 0.5f + offset.x()*scale;
      yOffset = (float(size.height()) - float(imageHeight) * scale) * 0.5f + offset.y()*scale;
    }
    point = QPoint(static_cast<int>((point.x() - xOffset) / scale),
                   static_cast<int>((point.y() - yOffset) / scale));
  }

  void mousePressEvent(QMouseEvent* event)
  {
    QWidget::mousePressEvent(event);
    
    if(event->button() == Qt::LeftButton || event->button() == Qt::MidButton)
    {
      bool active;
      {
        SYNC_WITH(console);
        active = console.colorTableHandler.active;
      }
      if(!active)
      {
        //QApplication::beep();
        return;
      }
      dragStart = event->pos();
      window2viewport(dragStart);
      dragPos = dragStart;
      QWidget::update();
    }
  }

  void mouseMoveEvent(QMouseEvent* event)
  {
    QWidget::mouseMoveEvent(event);
    
    if(dragStart.x() >= 0)
    {
      dragPos = event->pos();
      window2viewport(dragPos);
      QWidget::update();
      return;
    }

    {
      SYNC_WITH(console);
      if(!console.colorTableHandler.active)
      {
        QPoint pos(event->pos());
        window2viewport(pos);
        const char* text = 0;
        const std::list<std::string>& drawings(console.imageViews[name]);
        for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
        {
          text = console.imageDrawings[*i].getTip(pos.rx(), pos.ry());
          if(text)
            break;
        }
        if(text)
          setToolTip(QString(text));
        else
          setToolTip(QString());
      }
    }
  }


  void mouseReleaseEvent(QMouseEvent* event)
  {
    QWidget::mouseReleaseEvent(event);
    QPoint posInViewPort = QPoint(event->pos());
    window2viewport(posInViewPort);
    Vector2<int> v = Vector2<int>(posInViewPort.x(), posInViewPort.y());
    ImageViewAdapter::fireClick(name, v);
    
    if(dragStart.x() < 0)
      return;

    Qt::KeyboardModifiers keys = QApplication::keyboardModifiers();
    {
      SYNC_WITH(console);

      Image* image = 0;
      RobotConsole::Images::const_iterator i = console.images.find("raw image");
      if(i != console.images.end())
        image = i->second.image;

      if(image)
      {
        if(keys & Qt::ControlModifier && !(keys & Qt::ShiftModifier))
          console.colorTableHandler.undo();
        else 
        {
          if(dragStart == dragPos) // just a click
          {
            if((keys & Qt::ShiftModifier) && (keys & Qt::ControlModifier))
              console.colorTableHandler.clearPixel(*image, Vector2<int>(dragStart.x(), dragStart.y()));
            else if(keys & Qt::ShiftModifier)
              console.colorTableHandler.selectedColorClass = 
                console.colorTableHandler.colorTable.getColorClass(image->image[dragStart.y()][dragStart.x()].y,
                                                                   image->image[dragStart.y()][dragStart.x()].cb,
                                                                   image->image[dragStart.y()][dragStart.x()].cr);
            else
              console.colorTableHandler.addPixel(*image, Vector2<int>(dragStart.x(), dragStart.y()));
          }
          else if((keys & Qt::ShiftModifier) && (keys & Qt::ControlModifier))
            console.colorTableHandler.clearRectangle(*image, Vector2<int>(std::min<>(dragStart.x(), dragPos.x()), std::min<>(dragStart.y(), dragPos.y())),
                                                             Vector2<int>(std::max<>(dragStart.x(), dragPos.x()), std::max<>(dragStart.y(), dragPos.y())));
          else if(keys & Qt::ShiftModifier)
            console.colorTableHandler.fillRectangle(*image, Vector2<int>(std::min<>(dragStart.x(), dragPos.x()), std::min<>(dragStart.y(), dragPos.y())),
                                                          Vector2<int>(std::max<>(dragStart.x(), dragPos.x()), std::max<>(dragStart.y(), dragPos.y())));
          else if(!(keys & Qt::ShiftModifier))
            console.colorTableHandler.addRectangle(*image, Vector2<int>(std::min<>(dragStart.x(), dragPos.x()), std::min<>(dragStart.y(), dragPos.y())),
                                                           Vector2<int>(std::max<>(dragStart.x(), dragPos.x()), std::max<>(dragStart.y(), dragPos.y())));          
        }
      }
    }

    dragStart = QPoint(-1, -1);
    QWidget::update();
  }

  void keyPressEvent(QKeyEvent* event)
  {
    switch(event->key())
    {
    case Qt::Key_PageUp:
    case Qt::Key_Plus:
      event->accept();
      if(zoom < 3.f)
        zoom += 0.1f;
      if(zoom > 3.f)
        zoom = 3.f;
      QWidget::update();
      break;
    case Qt::Key_PageDown:
    case Qt::Key_Minus:
      event->accept();
      if(zoom > 0.1f)
        zoom -= 0.1f;
      QWidget::update();
      break;
    case Qt::Key_Up:
      offset += QPoint(0,20);
      QWidget::update();
      break;
    case Qt::Key_Down:
      offset += QPoint(0,-20);
      QWidget::update();
      break;
    case Qt::Key_Left:
      offset += QPoint(20, 0);
      QWidget::update();
      break;
    case Qt::Key_Right:
      offset += QPoint(-20, 0);
      QWidget::update();
      break;
    default:
      QWidget::keyPressEvent(event);
      break;
    }
  }

  void wheelEvent(QWheelEvent* event)
  {
    QWidget::wheelEvent(event);

    zoom += 0.1 * event->delta()/120;
    if(zoom > 3.f)
      zoom = 3.f;
    else if(zoom < 0.1f)
      zoom = 0.1f;
    QWidget::update();
  }

  void mouseDoubleClickEvent(QMouseEvent* event)
  {
    zoom = 1.f;
    offset.setX(0);
    offset.setY(0);
    QWidget::update();
  }

  QSize sizeHint () const { return QSize(cameraResolutionWidth, cameraResolutionHeight); }

  friend class ImageView;
};

ImageView::ImageView(RobotConsole& console, const std::string& background, const std::string& name, bool segmented) :
  console(console),
  background(background), name(name),
  segmented(segmented) {}

QWidget* ImageView::createWidget()
{
  return new ImageWidget(console, *this, background, name, segmented);
}

void ImageView::updateWidget(QWidget& widget)
{
  ((ImageWidget*)(&widget))->update();
}

std::string ImageView::getConfiguration(QWidget& widget) const
{
  ImageWidget* imageWidget = (ImageWidget*)&widget;
  QByteArray block;
  QDataStream out(&block, QIODevice::WriteOnly);
  out.setVersion(QDataStream::Qt_4_3);
  out << imageWidget->zoom;
  out << imageWidget->offset;
  return std::string(block.toBase64().constData());
}

void ImageView::setConfiguration(QWidget& widget, const std::string& conf)
{
  ImageWidget* imageWidget = (ImageWidget*)&widget;
  const QByteArray& block(QByteArray::fromBase64(conf.c_str()));
  QDataStream in(block);
  in.setVersion(QDataStream::Qt_4_3);
  float zoom;
  in >> zoom;
  if(zoom)
    imageWidget->zoom = zoom;
  QPoint offset;
  in >> offset;
  imageWidget->offset = offset;
}
