/**
* @file ControllerQt/Views/ViewBike/ViewBikeWidget.h
*
* Declaration of class ViewBikeWidget
*
* @author <a href="mailto:judy@tzi.de">Judith Mueller</a>
*/

#ifndef ViewBikeWidget_H
#define ViewBikeWidget_H

#include "ViewBike.h"
#include "TabWidget.h"
#include "BikeMenuBar.h"
#include "Tools/Math/Vector3.h"
#include "Modules/MotionControl/BIKEParameters.h"
#include <vector>
#include "Tools/Streams/OutStreams.h"

#include <QGLWidget>
#include <QMouseEvent>
#include <QApplication>
#include <QKeyEvent>
#include <QTreeView>
#include <QStandardItem>
#include <QPushButton>
#include <QSignalMapper>

class Selected
{
public:
  int phaseNumber;
  unsigned int pointNumber;
  int limb;
  bool xzRot;

  Selected(){};
};


class ViewBikeWidget : public  QGLWidget
{
  Q_OBJECT

public:
  ViewBikeWidget(ViewBike& viewBike, QWidget* parent);

  private slots:
    //Buttons
    void newButtonClicked();
    void loadButtonClicked();
    void saveButtonClicked();
    void saveAsButtonClicked();
    void setDrawings();
    void setSingleDrawing();
    void setReachedDrawing();
    void setEditor();
    void setTra2d();
    void setTra1d();
    void setVelocity();
    void setAccel();
    void removePhase();
    void addPhaseAfterActual();
    void playWholeMotion();
    void playMotionTilActive();
    void resetRobot();
    void standRobot();
    void setMirrored(int state);
    void undoChanges();
    void redoChanges();
    //Sliders
    void transparencyChanged(const int& i);
    //Tab
    void updateCommonParameters(QStandardItem* item);
    void updatePhaseParameters(QStandardItem* item);
    void movePhase(const int fromIndex, int toIndex);
    void setSelectedFromEditor(const int&);    
    //Contextmenu
    void setDragPlane(const int& plane);
    void splitPhase();

private:
  Vector3d targetPosOffset,
           cameraPosOffset;
  Matrix3d originRot;

  //OpenGL
  void initializeGL();
  void resizeGL(int newWidth, int newHeight);
  void paintGL();
  GLvoid setMatrix(const Vector3d& translation, const Matrix3d& rotation);
  void drawPhases();
  void drawBezierCurves(const int& phaseNumber);
  void drawControlPoint(const Vector3 <>& point, const float& cubeFaktor);
  void draw2dCurves(const float& minA, const float& minB, const float& maxA, const float& maxB, const float& scaleFactorA, 
    const float& scaleFactorB, const float& transA, const float& transB, float *colorA, float *colorB,
    const float& window, const std::vector<Vector2<> >& cp, const std::vector<Vector2<> >& point);
  bool clickControlPoint(const int& x, const int& y);
  void gluUnProjectClick(int x, int y, Vector3d& vecFar, Vector3d& vecNear);

  //QT-Widget-Events
  void keyPressEvent(QKeyEvent* event);
  void wheelEvent(QWheelEvent* event);
  void mouseMoveEvent(QMouseEvent* event);
  void mousePressEvent(QMouseEvent* event); 
  void mouseReleaseEvent(QMouseEvent* event);
  void contextMenuEvent(QContextMenuEvent* event);

  //TreeView, StandardModel and Editor
  QList <QStandardItem*> makeStringAndValueRow(QString string, QVariant value);
  void addStringAndValueToList(QList <QStandardItem*> &list, QString string, QVariant value);
  QStandardItemModel* makeNewModelWithLabels();
  void fillModelWithPhaseData(QStandardItem* item, int i);
  void updateCommon();
  void updateEditorView();
  bool deleteKids(QStandardItem* rootItem);
  void addControlPoints(QStandardItem* item);
  QStandardItem* makeLabel(QString string);
  QStandardItem* makeValue(QVariant var);
  QList <QStandardItem*> addXYZLabel(QStandardItem* item);
  void playMotion(int phase);

  //BikeParameters
  void writeParametersToFile(const std::string& name);
  void makeNewPhaseWithModelAndTree(const int& phaseNumber);
  void showPlane();
  void clipCurve(Vector3d& translationVec);
  bool calcLegJoints(const Vector3<>& footPos, const Vector3<>& footRotAng, const bool& left, double& leg0, double& leg1, double& leg2, double& leg3, double& leg4, double& leg5);
  bool clipLegJointsWithLimits(double& leg1, double& leg2, double& leg3, const JointData::Joint& joint);

  Vector3<> calculateFootPos(const JointData& jointData, JointData::Joint joint);
  Vector3<> calculateHandPos(const JointData& jointData, JointData::Joint joint);

  Vector3<> calcFootPos(const double& leg0, const double& leg1, const double& leg2, const double& leg3, const double& leg4, const double& leg5, JointData::Joint joint);
  //void clipControlPoints(const Vector3<>& footPos, const double& t, const int& phaseNumber, const int& limb);
  ViewBike& viewBike;

  std::string floatToStr(const float& f);
  std::string boolToStr(const bool& b);
  std::string intToStr(const int& i);

  //Editor
  TabWidget *tabber;
  QTreeView *treeViewCommon; //TreeView for the Common Parameters
  std::vector<QTreeView *> treeView; //TreeView for the PhaseParameters
  QStandardItemModel *modelCommon; //Model for the Common Parameters
  std::vector<QStandardItemModel *> phaseTab; //Model for the PhaseParameters
  QPushButton *editor, //show editor or not
    *save,   // save actual Parameters to file
    *saveAs, // save actual Parameters to file with a new name
    *load,   // load actual Parameters from file
    *newMove, //clear parameters
    *addPhase, //add one phase after the actual phase
    *deletePhase; //delete actual phase
  QString fileName; //the actual fileName
  BIKEParameters parameters; //the actual Parameters
  std::vector<BIKEParameters> undo, redo; // list with undo

  //Drawings and openGl
  bool phaseDrawings, //show phaseDrawings
    singleDraw, //show only the drawings of the actual Phase
    reachedDraw, 
    tra2dWindows, //show 2D-View of phaseDrawings for actual Phase and Limb
    tra1dWindows, //show 1D-View of phaseDrawings for actual Phase and Limb
    velocityWindows,//show one curves velocity
    accelWindows; //show one curves acceleration

  int getString,
    ghost; //set the opacitiy of the robot model

  std::vector<std::string> commands;
  std::vector<Vector3<> > reachedPositions[Phase::numOfLimbs];

  MotionRequest::Motion lastMotion;

  QPushButton *drawings, 
    *singleDrawing,
    *reachedDrawing,
    *tra2d, 
    *tra1d,
    *velocity,
    *accel;

  Selected selectedPoint; //Infos about the actual selected Point

  Vector3d dragPlane; //plane for the 3D View where a selected point is moved

  bool moveDrag,  //a Point in the 3DView is moved or not
    moveViewOfViewDragXY, //a Point in the XY(t)View (2D) is moved or not
    moveViewOfViewDragXZ, //a Point in the XZ(t)View (2D)is moved or not
    moveViewOfViewDragYZ, //a Point in the YZ(t)View (2D)is moved or not
    moveViewOfViewDragXP, //a Point in the X:Phase View (1D) is moved or not
    moveViewOfViewDragYP, //a Point in the Y:Phase View (1D) is moved or not
    moveViewOfViewDragZP;  //a Point in the Z:Phase View (1D) is moved or not

  float actualX, //the captured mousePosition when contextMenu is open
    actualY;

  GLubyte stippleMask[17][128]; //The opacity mask

  //Contextmenu
  BikeMenuBar *bikeMenuBar; //the contextMenubar
  QSignalMapper dragPlaneMapper; 

  bool mirror;

};

#endif //ViewBikeWidget_H
