/**
* @file BH2010StableBehaviorControl.cpp
* 
* Implementation of class BH2010StableBehaviorControl.
*
* @author Max Risler
* @author Judith Mller
*/

#include "BH2010StableBehaviorControl.h"
#include "Symbols/BH2010StableBallSymbols.h"
#include "Symbols/BH2010StableBIKESymbols.h"
#include "Symbols/BH2010StableFieldSymbols.h"
#include "Symbols/BH2010StableHeadSymbols.h"
#include "Symbols/BH2010StableLEDSymbols.h"
#include "Modules/BehaviorControl/CommonSymbols/MathSymbols.h"
#include "Symbols/BH2010StableMotionSymbols.h"
#include "Symbols/BH2010StableGoalSymbols.h"
#include "Modules/BehaviorControl/CommonSymbols/SoundSymbols.h"
#include "Modules/BehaviorControl/CommonSymbols/KeySymbols.h"
#include "Symbols/BH2010StableGameSymbols.h"
#include "Symbols/BH2010StableLocatorSymbols.h"
#include "Symbols/BH2010StableFallDownSymbols.h"
#include "Symbols/BH2010StableObstacleSymbols.h"
#include "Tools/Debugging/ReleaseOptions.h"


PROCESS_WIDE_STORAGE BH2010StableBehaviorControl* BH2010StableBehaviorControl::theInstance = 0;

BH2010StableBehaviorControl::BH2010StableBehaviorControl()
: GTXabslEngineExecutor("bh10Stable")
{
  theInstance = this;

  // Common Symbols
  symbols.push_back(new MathSymbols());
  symbols.push_back(new BH2010StableMotionSymbols(behaviorControlOutput.motionRequest, theMotionInfo, theRobotInfo, theRobotName, theFrameInfo));
  symbols.push_back(new SoundSymbols(behaviorControlOutput.soundRequest));
  symbols.push_back(new KeySymbols(theKeyStates));
  symbols.push_back(new BH2010StableBIKESymbols(behaviorControlOutput.motionRequest, theBikeEngineOutput, theMotionInfo));
  symbols.push_back(new BH2010StableBallSymbols(theRobotInfo, theBallModel, theBallHypotheses, theFrameInfo, theRobotPose, theTeamMateData, theFieldDimensions));
  symbols.push_back(new BH2010StableFallDownSymbols(theFallDownState));
  symbols.push_back(new BH2010StableFieldSymbols(theFieldDimensions));
  symbols.push_back(new BH2010StableGameSymbols(behaviorControlOutput.behaviorData, behaviorControlOutput.robotInfo, behaviorControlOutput.ownTeamInfo, theOpponentTeamInfo, behaviorControlOutput.gameInfo, theFrameInfo, theBallModel, theRobotPose, theFieldDimensions));
  symbols.push_back(new BH2010StableGoalSymbols(theFreePartOfOpponentGoalModel, theRobotPose, theFallDownState, theGoalPercept, theGroundContactState, theGameInfo, theFrameInfo, theFieldDimensions));
  symbols.push_back(new BH2010StableHeadSymbols(behaviorControlOutput.headMotionProposal, theHeadJointRequest));
  symbols.push_back(new BH2010StableLEDSymbols(behaviorControlOutput.ledRequest, theFilteredSensorData, theBallModel, theFrameInfo, theTeamMateData));
  symbols.push_back(new BH2010StableLocatorSymbols(theRobotPose, theRobotPoseInfo, behaviorControlOutput.ownTeamInfo, theFrameInfo, theBallModel, theFieldDimensions, theGroundContactState));
  symbols.push_back(new BH2010StableObstacleSymbols(theObstacleModel));

  GTXabslEngineExecutor::init();
  ASSERT(pEngine);

  if (!errorHandler.errorsOccurred)
    currentAgent = pEngine->getSelectedAgentName();
}

void BH2010StableBehaviorControl::init()
{
  // init symbols
  for(std::list<Symbols*>::iterator i = symbols.begin(); i != symbols.end(); ++i)
  (*i)->init();
}

BH2010StableBehaviorControl::~BH2010StableBehaviorControl()
{
  for(std::list<Symbols*>::iterator i = symbols.begin(); i != symbols.end(); ++i)
    delete *i;
}

void BH2010StableBehaviorControl::registerSymbolsAndBasicBehaviors()
{
  for(std::list<Symbols*>::iterator i = symbols.begin(); i != symbols.end(); ++i)
    (*i)->registerSymbols(*pEngine);
}

void BH2010StableBehaviorControl::executeIfEngineCouldNotBeCreated()
{
#ifdef TARGET_ROBOT
  ASSERT(false);
#endif
}

void BH2010StableBehaviorControl::printGeneratedMainActionToString(char* buf) const
{
  behaviorControlOutput.motionRequest.printOut(buf);
}

bool BH2010StableBehaviorControl::handleMessage(InMessage& message)
{
  if (theInstance)
    return theInstance->handleXabslMessage(message);
  else
    return false;
}

void BH2010StableBehaviorControl::update(BehaviorControlOutput& behaviorControlOutput)
{
  this->behaviorControlOutput.ownTeamInfo = theOwnTeamInfo;
  this->behaviorControlOutput.robotInfo = theRobotInfo;
  this->behaviorControlOutput.gameInfo = theGameInfo;

  // update symbols 
  for(std::list<Symbols*>::iterator i = symbols.begin(); i != symbols.end(); ++i)
    (*i)->update();

  // set agent
  if (currentAgent != theBehaviorConfiguration.agent)
  {
    currentAgent = theBehaviorConfiguration.agent;
    if (!setSelectedAgent(currentAgent.c_str()))
    {
      OUTPUT(idText, text, "BH2010StableBehaviorControl: Unable to enable selected agent \"" << currentAgent << "\".");
    }
  }

  // execute the engine
  executeEngine();
  
  behaviorControlOutput = this->behaviorControlOutput;
  TEAM_OUTPUT(idTeamMateBehaviorData, bin, behaviorControlOutput.behaviorData);
}

void BH2010StableBehaviorControl::update(MotionRequest& motionRequest)
{
  motionRequest = theBehaviorControlOutput.motionRequest;
  if(Global::getReleaseOptions().motionRequest)
  {
    TEAM_OUTPUT(idMotionRequest, bin, motionRequest);
  }
}

void BH2010StableBehaviorControl::update(LEDRequest& ledRequest)
{
  ledRequest = theBehaviorControlOutput.ledRequest;
}

void BH2010StableBehaviorControl::update(SimpleXabslInfo& simpleXabslInfo)
{
  simpleXabslInfo.reset();
  
  int i, j;
  xabsl::Array<const xabsl::Action*> activeActions;
  xabsl::Array<int> activeDepth;

  // depth-first-traversal of active options and basic behaviors
  xabsl::Array<const xabsl::Action*> currentActions;
  xabsl::Array<int> currentDepth;
  for (i=pEngine->getRootActions().getSize()-1;i>=0;i--)
  {
    currentActions.append("", pEngine->getRootActions()[i]);
    currentDepth.append("",0);
  }

  while(currentActions.getSize() > 0)
  {
    const xabsl::Action* last = currentActions[currentActions.getSize()-1];
    int lastDepth = currentDepth[currentDepth.getSize()-1];
    activeActions.append("", last);
    activeDepth.append("", lastDepth);
    currentActions.removeLast();
    currentDepth.removeLast();
    if (const xabsl::Option* option = last->getOption())
    {
      const xabsl::Array<xabsl::Action*>& actions = 
        option->activeState->actions;

      for(j=actions.getSize()-1;j>=0;j--)
      {
        currentActions.append("", actions[j]);
        currentDepth.append("", lastDepth + 1);
      }
    }
  }

  // For each active option/basic behavior: 
  // name, depth, duration of activation, active state, duration of state activation, parameters
  for (i=0; i < activeActions.getSize(); i++)
  {
    if (const xabsl::Behavior* behavior = activeActions[i]->getBehavior())
    {
      SimpleXabslInfo::ActiveOption ao;
      ao.depth = activeDepth[i];
      ao.name = behavior->n;
      ao.durationOfActivation = behavior->timeOfExecution;
      const xabsl::Option* option = activeActions[i]->getOption();
      if(option)
      {
        ao.activeState =  option->activeState->n;
        ao.durationOfStateActivation = option->activeState->timeOfStateExecution;
      }
      else
      {
        ao.activeState = "";
        ao.durationOfStateActivation = 0;
      }
      simpleXabslInfo.activeOptions.push_back(ao);
    }
  }
}

MAKE_MODULE(BH2010StableBehaviorControl, Behavior Control)
