/**
* @file BH2010StableBehaviorControl.h
* 
* Implementation of class BehaviorControl.
*
* @author Martin Ltzsch
* @author Matthias Jngel
* @author Max Risler
*/

#ifndef BH2010StableBehaviorControl_H
#define BH2010StableBehaviorControl_H

#include "Tools/Module/Module.h"
#include "Tools/Xabsl/GT/GTXabslEngineExecutor.h"
#include "Modules/BehaviorControl/Symbols.h"
#include "Tools/Team.h"
#include "Tools/RingBufferWithSum.h"
#include "Tools/MessageQueue/MessageQueue.h"

#include "Representations/Configuration/BehaviorConfiguration.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/TeamInfo.h"
#include "Representations/Infrastructure/TeamMateData.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Configuration/RobotName.h"
#include "Representations/Infrastructure/GameInfo.h"
#include "Representations/Infrastructure/LEDRequest.h"
#include "Representations/Infrastructure/JointData.h"
#include "Representations/Infrastructure/SoundRequest.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Representations/Perception/GoalPercept.h"
#include "Representations/Sensing/GroundContactState.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Modeling/FallDownState.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Modeling/RobotsModel.h"
#include "Representations/Modeling/BallModel.h"
#include "Representations/Modeling/BallHypotheses.h"
#include "Representations/Modeling/ObstacleModel.h"
#include "Representations/Modeling/AnotherObstacleModel.h"
#include "Representations/Modeling/FreePartOfOpponentGoalModel.h"
#include "Representations/MotionControl/MotionInfo.h"
#include "Representations/BehaviorControl/BehaviorControlOutput.h"
#include "Representations/BehaviorControl/HeadMotionProposal.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/Debugging/XabslInfo.h"
#include "Representations/Debugging/SimpleXabslInfo.h"
#include "Representations/MotionControl/BikeEngineOutput.h"
#include "Representations/MotionControl/HeadJointRequest.h"
#include "Representations/Configuration/BehaviorParameters.h"
#include "Representations/Configuration/BH2010Parameters.h"
#include "Representations/Configuration/BehaviorTactics.h"

MODULE(BH2010StableBehaviorControl)
  REQUIRES(FrameInfo)
  REQUIRES(AnotherObstacleModel)
  REQUIRES(BehaviorConfiguration) 
  REQUIRES(FieldDimensions)
  REQUIRES(FallDownState)
  REQUIRES(HeadJointRequest)
  REQUIRES(KeyStates)
  REQUIRES(MotionInfo)
  REQUIRES(BallModel)
  REQUIRES(BallHypotheses)
  REQUIRES(BikeEngineOutput)
  REQUIRES(GoalPercept)
  REQUIRES(RobotPose)
  REQUIRES(RobotPoseInfo)
  REQUIRES(RobotsModel)
  REQUIRES(GroundTruthRobotPose)
  REQUIRES(RobotInfo)
  REQUIRES(RobotName)
  REQUIRES(OwnTeamInfo)
  REQUIRES(OpponentTeamInfo)
  REQUIRES(GameInfo)
  REQUIRES(TeamMateData)
  REQUIRES(ObstacleModel)
  REQUIRES(FreePartOfOpponentGoalModel)
  REQUIRES(RobotDimensions)
  REQUIRES(FilteredSensorData)
  REQUIRES(GroundContactState)
  REQUIRES(BehaviorParameters)
  REQUIRES(BH2010Parameters)
  REQUIRES(BehaviorTactics)
  PROVIDES_WITH_MODIFY(BehaviorControlOutput)
  REQUIRES(BehaviorControlOutput)
  PROVIDES_WITH_MODIFY_AND_OUTPUT_AND_DRAW(MotionRequest)
  PROVIDES_WITH_MODIFY(HeadMotionProposal)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(LEDRequest)
  PROVIDES_WITH_MODIFY(SoundRequest)
  PROVIDES(SimpleXabslInfo)
END_MODULE


/**
* @class BH2010StableBehaviorControl
*
* A Behavior based on the xabsl::Engine that is used by B-Human
* for the RoboCup 2010.
*
* @author Martin Ltzsch
* @author Matthias Jngel
* @author Max Risler
* @author Judith Mller
*/ 
class BH2010StableBehaviorControl : public BH2010StableBehaviorControlBase, public GTXabslEngineExecutor
{
public:
  /** 
  * Constructor.
  */
  BH2010StableBehaviorControl();

  /** destructor */
  ~BH2010StableBehaviorControl();

  PROCESS_WIDE_STORAGE_STATIC BH2010StableBehaviorControl* theInstance; /**< Points to the only instance of this class in this process or is 0 if there is none. */

  /** 
  * Is called for every incoming debug message.
  * @param message An interface to read the message from the queue
  * @return if the message was read
  */
  static bool handleMessage(InMessage& message);

private:

  /** Registers symbols and basic behaviors at the engine */
  virtual void registerSymbolsAndBasicBehaviors();
  
  /** Is called if the engine could not be created */
  virtual void executeIfEngineCouldNotBeCreated();

  /** 
  * Prints the main action that was generated by the execution of the engine to a string
  * @param buf the string where to print the action
  */
  virtual void printGeneratedMainActionToString(char* buf) const;

  /** a list of symbol definition classes */
  std::list<Symbols*> symbols;

  /** name of currently selected agent */
  std::string currentAgent;

  /** the output generated by the behavior control */
  BehaviorControlOutput behaviorControlOutput;

  /** updates the behavior control output */
  void update(BehaviorControlOutput& behaviorControlOutput);
  
  /** updates the motion request by copying from behavior control output */
  void update(MotionRequest& motionRequest);

  /** updates the head motion request by copying from behavior control output */
  void update(HeadMotionProposal& headMotionProposal) {headMotionProposal = theBehaviorControlOutput.headMotionProposal;}
  
  /** updates the LED request by copying from behavior control output */
  void update(LEDRequest& ledRequest);
  
  /** updates the sound request by copying from behavior control output */
  void update(SoundRequest& soundRequest) {soundRequest = theBehaviorControlOutput.soundRequest;}

  /** update the xabsl information for logging */
  void update(SimpleXabslInfo& simpleXabslInfo);

  void init();
};

#endif// BH2010StableBehaviorControl_H
