/** 
* @file BH2010StableBallSymbols.cpp
*
* Implementation of class BallSymbols.
*
* @author Max Risler
* \author Colin Graf
*/

#include "BH2010StableBallSymbols.h"
#include "Tools/Debugging/Modify.h"


PROCESS_WIDE_STORAGE BH2010StableBallSymbols* BH2010StableBallSymbols::theInstance = 0;

void BH2010StableBallSymbols::registerSymbols(xabsl::Engine& engine)
{
  engine.registerDecimalInputSymbol("ball.position.field.x", &ballPositionField.x);
  engine.registerDecimalInputSymbol("ball.position.field.y", &ballPositionField.y);

  engine.registerDecimalInputSymbol("ball.x", &ballPositionRel.x);
  engine.registerDecimalInputSymbol("ball.y", &ballPositionRel.y);
  //This is a little bit faster with the Position because it's just using the last percept
  engine.registerDecimalInputSymbol("ball.seen.x", &seenBallPositionRel.x);
  engine.registerDecimalInputSymbol("ball.seen.y", &seenBallPositionRel.y);
  engine.registerDecimalInputSymbol("ball.seen.estimate.x", &staticGetBallLastSeenEstimateX);
  engine.registerDecimalInputSymbol("ball.seen.estimate.y", &staticGetBallLastSeenEstimateY);
  engine.registerDecimalInputSymbol("ball.position.seen.estimate.x", &staticGetBallPositionLastSeenEstimateX);
  engine.registerDecimalInputSymbol("ball.position.seen.estimate.y", &staticGetBallPositionLastSeenEstimateY);
  engine.registerDecimalInputSymbol("ball.seen.angle", &staticGetSeenBallAngle);
  engine.registerDecimalInputSymbol("ball.seen.distance", &staticGetBallSeenDistance);

  engine.registerDecimalInputSymbol("ball.distance", &staticGetBallDistance);
  engine.registerDecimalInputSymbol("ball.angle", &staticGetBallAngle);
  engine.registerBooleanInputSymbol("ball.was_seen", &ballWasSeen);
  engine.registerDecimalInputSymbol("ball.time_since_last_seen", &timeSinceBallWasSeen);
  engine.registerDecimalInputSymbol("ball.time_since_disappeared", &getTimeSinceDisappeared);

  engine.registerDecimalInputSymbol("ball.speed.field.x", &ballSpeedField.x);
  engine.registerDecimalInputSymbol("ball.speed.field.y", &ballSpeedField.y);
  engine.registerDecimalInputSymbol("ball.speed.robot.x", &ballSpeedRel.x);
  engine.registerDecimalInputSymbol("ball.speed.robot.y", &ballSpeedRel.y);

  engine.registerDecimalInputSymbol("ball.time_when_own_y_axis_reached", &staticGetTimeWhenBallReachesOwnYAxis);
  engine.registerDecimalInputSymbol("ball.position_when_ball_reaches_own_y_axis.y", &staticGetYPosWhenBallReachesOwnYAxis);

  engine.registerDecimalInputSymbol("ball.distance.own_goal", &staticGetBallDistanceToOwnGoal);

  engine.registerDecimalInputSymbol("ball.on_field.x", &getOnFieldX);
  engine.registerDecimalInputSymbol("ball.on_field.y", &getOnFieldY);
  engine.registerDecimalInputSymbol("ball.on_field.angle", &getOnFieldAngle);
  engine.registerDecimalInputSymbol("ball.on_field.speed.x", &getOnFieldSpeedX);
  engine.registerDecimalInputSymbol("ball.on_field.speed.y", &getOnFieldSpeedY);
}

void BH2010StableBallSymbols::update()
{
  timeSinceBallWasSeen = (float) frameInfo.getTimeSince(theInstance->ballModel.timeWhenLastSeen);
  ballWasSeen = timeSinceBallWasSeen < 500;
  const BallState& estimate = ballModel.estimate;
  ballPositionRel = estimate.position;
  ballPositionField = estimate.getPositionInFieldCoordinates(robotPose);
  ballSpeedRel = estimate.velocity;
  ballSpeedField = estimate.getVelocityInFieldCoordinates(robotPose);
  seenBallPositionRel = ballModel.lastPerception.position;
  ballLastSeenEstimate = ballModel.lastSeenEstimate.position;
  ballPositionLastSeenEstimate = ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose);
}


float BH2010StableBallSymbols::getBallLastSeenEstimateX()
{
  return ballLastSeenEstimate.x;
}

float BH2010StableBallSymbols::getBallLastSeenEstimateY()
{
  return ballLastSeenEstimate.y;
}

float BH2010StableBallSymbols::getBallPositionLastSeenEstimateX()
{
  return ballPositionLastSeenEstimate.x;
}

float BH2010StableBallSymbols::getBallPositionLastSeenEstimateY()
{
  return ballPositionLastSeenEstimate.y;
}


float BH2010StableBallSymbols::getBallFieldRobotX()
{
  return ballPositionField.x;
}

float BH2010StableBallSymbols::getBallFieldRobotY()
{
  return ballPositionField.y;
}

float BH2010StableBallSymbols::getBallPositionRobotX()
{
  return ballPositionRel.x;
}

float BH2010StableBallSymbols::getBallPositionRobotY()
{
  return ballPositionRel.y;
}

float BH2010StableBallSymbols::getBallAngle()
{
  return toDegrees(ballModel.estimate.getAngle());
}

float BH2010StableBallSymbols::getSeenBallAngle()
{
  return toDegrees(ballModel.lastPerception.getAngle());
}

float BH2010StableBallSymbols::getBallDistance()
{
  return ballModel.estimate.getDistance();
}

float BH2010StableBallSymbols::getBallSeenDistance()
{
  return ballModel.lastPerception.getDistance();
}

float BH2010StableBallSymbols::getTimeWhenBallReachesOwnYAxis()
{
  float result = 1000000;

  float decel = (float)fieldDimensions.ballFriction;
  if(ballModel.estimate.velocity.x == 0.0f)
    return result;
  float decelX = ballModel.estimate.velocity.x * decel / ballModel.estimate.velocity.abs();
  float decelTime = ballModel.estimate.velocity.x / decelX;
  float xPosWhenStopping = ballModel.estimate.position.x + decelTime * ballModel.estimate.velocity.x - 0.5f * decel * sqr(decelTime);
  if(xPosWhenStopping * ballModel.estimate.position.x < 0 )
  {
    float p(ballModel.estimate.velocity.x * 2 / decelX), q(2.0f / decelX * ballModel.estimate.position.x);
    float temp =  p * -0.5f + sqrtf(sqr(p * 0.5f) - q);
    return temp;
  }
  return result;
}

float BH2010StableBallSymbols::getYPosWhenBallReachesOwnYAxis()
{  
  if(ballSpeedRel.x == 0 || ballPositionRel.x * ballSpeedRel.x > 0) // Ball does not move or moves away
  {
    return 0.0f;
  }
  float timeWhenAxisIsReached = fabs(ballPositionRel.x / ballSpeedRel.x);
  Vector2<> finalBallPos = ballPositionRel + (ballSpeedRel * timeWhenAxisIsReached);
  
  MODIFY("behavior symbols:ball:finalBallPos", finalBallPos);
  
  return finalBallPos.y;
}


float BH2010StableBallSymbols::getBallDistanceToOwnGoal()
{ 
  return (ballPositionField - Vector2<>((float) fieldDimensions.xPosOwnGroundline,0.f)).abs();
}



float BH2010StableBallSymbols::getTimeSinceDisappeared()
{
  return float(theInstance->frameInfo.getTimeSince(theInstance->ballHypotheses.timeWhenDisappeared));
}

float BH2010StableBallSymbols::getOnFieldX()
{
  return theInstance->ballModel.estimate.getPositionInFieldCoordinates(theInstance->robotPose).x;
}

float BH2010StableBallSymbols::getOnFieldY()
{
  return theInstance->ballModel.estimate.getPositionInFieldCoordinates(theInstance->robotPose).y;
}

float BH2010StableBallSymbols::getOnFieldAngle()
{
  Vector2<> ballOnField = theInstance->ballModel.estimate.getPositionInFieldCoordinates(theInstance->robotPose);
  return toDegrees((ballOnField - theInstance->robotPose.translation).angle());
}


float BH2010StableBallSymbols::getOnFieldSpeedX()
{
  return theInstance->ballModel.estimate.getVelocityInFieldCoordinates(theInstance->robotPose).x;
}

float BH2010StableBallSymbols::getOnFieldSpeedY()
{
  return theInstance->ballModel.estimate.getVelocityInFieldCoordinates(theInstance->robotPose).y;
}
