/** 
* @file BH2010StableBallSymbols.h
*
* Declaration of class BallSymbols.
*
* @author Max Risler
* @author Colin Graf
*/

#ifndef __BH2010StableBallSymbols_h_
#define __BH2010StableBallSymbols_h_

#include "Modules/BehaviorControl/Symbols.h"
#include "Representations/Modeling/BallModel.h"
#include "Representations/Modeling/BallHypotheses.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/LEDRequest.h"
#include "Representations/Infrastructure/TeamMateData.h"
#include "Representations/Configuration/FieldDimensions.h"


/**
* The Xabsl symbols that are defined in "ball_symbols.xabsl"
*/ 
class BH2010StableBallSymbols : public Symbols
{
public:
  /** Constructor. */
  BH2010StableBallSymbols(const RobotInfo& robotInfo, const BallModel& ballModel, const BallHypotheses& ballHypotheses, const FrameInfo& frameInfo, const RobotPose& robotPose, const TeamMateData& teamMateData, const FieldDimensions& fieldDimensions) : 
    robotInfo(robotInfo),
    frameInfo(frameInfo), 
    ballModel(ballModel), 
    ballHypotheses(ballHypotheses),
    robotPose(robotPose),
    teamMateData(teamMateData),
    fieldDimensions(fieldDimensions),
    ballWasSeen(false), 
    ballPositionRel(0,0), 
    ballPositionField(0,0), 
    ballSpeedRel(0,0), 
    ballSpeedField(0,0),
    timeSinceBallWasSeen(10000.0)
  {
    theInstance = this;

    memset(teamBallPositionsRelTime, 0, sizeof(teamBallPositionsRelTime));
    memset(teamBallPositionsFieldTime, 0, sizeof(teamBallPositionsFieldTime));
  }

  PROCESS_WIDE_STORAGE_STATIC BH2010StableBallSymbols* theInstance; /**< Points to the only instance of this class in this process or is 0 if there is none. */

  /** registers the symbols at an engine */
  void registerSymbols(xabsl::Engine& engine);

  /** updates the symbols */
  void update();

private:
  const RobotInfo& robotInfo;
  const FrameInfo& frameInfo;
  const BallModel& ballModel;
  const BallHypotheses& ballHypotheses;
  const RobotPose& robotPose;
  const TeamMateData& teamMateData;
  const FieldDimensions& fieldDimensions;

  bool ballWasSeen;
  Vector2 <> ballPositionRel;
  Vector2 <> seenBallPositionRel;
  Vector2 <> ballPositionField;
  Vector2 <> ballSpeedRel;
  Vector2 <> ballSpeedField;
  Vector2 <> ballLastSeenEstimate;
  Vector2 <> ballPositionLastSeenEstimate;

  static float getBallSpeedFieldAbs();
  static float getBallSpeedRobotAbs();

  unsigned teamBallPositionsRelTime[TeamMateData::numOfPlayers];
  Vector2 <> teamBallPositionsRel[TeamMateData::numOfPlayers];
  unsigned teamBallPositionsFieldTime[TeamMateData::numOfPlayers];
  Vector2 <> teamBallPositionsField[TeamMateData::numOfPlayers];

  float getBallFieldRobotX();
  float getBallFieldRobotY();
  float getBallPositionRobotX();
  float getBallPositionRobotY();
  float getBallDistance();
  float getBallSeenDistance();
  float getBallAngle();
  float getSeenBallAngle();
  float getBallSpeed();
  float getYPosWhenBallReachesOwnYAxis();
  float getTimeWhenBallReachesOwnYAxis();
  float getBallLastSeenEstimateX();
  float getBallLastSeenEstimateY();
  float getBallPositionLastSeenEstimateX();
  float getBallPositionLastSeenEstimateY();

  bool getBallWasSeenByTeamMate();
  float getBallTimeSinceLastSeenByTeamMate();
  float getBallPositionFieldByTeamMateX();
  float getBallPositionFieldByTeamMateY();
  float getBallPositionRobotByTeamMateX();
  float getBallPositionRobotByTeamMateY();
  float getBallDistanceByTeamMate();
  float getBallAngleByTeamMate();
  float getBallDistanceToOwnGoal();

  static float getOnFieldX();
  static float getOnFieldY();
  static float getOnFieldAngle();
  static float getOnFieldSpeedX();
  static float getOnFieldSpeedY();
  
  //Experimental Goalkeeping Stuff
  bool getBallWouldScore();
  float getBallDistSpeedQuot();

  static bool getIsMovingAnyTeamMate();
  static float getTimeSinceDisappeared();

  int player;
  float timeSinceBallWasSeen;

  static float staticGetBallFieldRobotX() { return theInstance->getBallFieldRobotX(); };
  static float staticGetBallFieldRobotY() { return theInstance->getBallFieldRobotY(); };
  static float staticGetBallLastSeenEstimateX() { return theInstance->getBallLastSeenEstimateX(); };
  static float staticGetBallLastSeenEstimateY() { return theInstance->getBallLastSeenEstimateY(); };
  static float staticGetBallPositionLastSeenEstimateX() { return theInstance->getBallPositionLastSeenEstimateX(); };
  static float staticGetBallPositionLastSeenEstimateY() { return theInstance->getBallPositionLastSeenEstimateY(); };
  static float staticGetBallPositionRobotX() { return theInstance->getBallPositionRobotX(); };
  static float staticGetBallPositionRobotY() { return theInstance->getBallPositionRobotY(); };
  static float staticGetBallDistance() { return theInstance->getBallDistance(); };
  static float staticGetBallSeenDistance() { return theInstance->getBallSeenDistance(); };
  static float staticGetBallAngle() { return theInstance->getBallAngle(); };
  static float staticGetSeenBallAngle() { return theInstance->getSeenBallAngle(); };
  static float staticGetBallSpeed() { return theInstance->getBallSpeed(); };
  static float staticGetYPosWhenBallReachesOwnYAxis() { return theInstance->getYPosWhenBallReachesOwnYAxis(); };
  static float staticGetTimeWhenBallReachesOwnYAxis() { return theInstance->getTimeWhenBallReachesOwnYAxis(); };

  static float staticGetBallDistanceToOwnGoal() { return theInstance->getBallDistanceToOwnGoal(); };

};

#endif // __BallSymbols_h_
