/** 
* @file BH2010StableGameSymbols.h
*
* Declaration of class BH2010StableGameSymbols.
*
* @author Judith Mller
*/

#ifndef __BH2010StableGameSymbols_h_
#define __BH2010StableGameSymbols_h_

#include "Representations/BehaviorControl/BehaviorData.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/TeamInfo.h"
#include "Representations/Infrastructure/GameInfo.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Modeling/BallModel.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Modules/BehaviorControl/Symbols.h"

/**
* The Xabsl symbols that are defined in "game_symbols.xabsl"
*
* @author Max Risler
* @author Judith Mller
*/ 
class BH2010StableGameSymbols : public Symbols
{
public:
  /*
  * Constructor.
  * @param robotInfo A reference to the RobotInfo.
  * @param ownTeamInfo A reference to the OwnTeamInfo.
  * @param opponentTeamInfo A reference to the OpponentTeamInfo.
  * @param gameInfo A reference to the GameInfo.
  */
  BH2010StableGameSymbols(BehaviorData& behaviorData, RobotInfo& robotInfo, OwnTeamInfo& ownTeamInfo, const OpponentTeamInfo& opponentTeamInfo, GameInfo& gameInfo, const FrameInfo& frameInfo, const BallModel& ballModel, const RobotPose& robotPose, const FieldDimensions& fieldDimensions) : 
    behaviorData(behaviorData),
    robotInfo(robotInfo), 
    ownTeamInfo(ownTeamInfo),
    opponentTeamInfo(opponentTeamInfo),
    gameInfo(gameInfo),
    frameInfo(frameInfo),
    ballModel(ballModel),
    robotPose(robotPose),
    fieldDimensions(fieldDimensions),
    startup(true),
    kickOffChanged(false),
    lastTimeInInitial(0),
    disablePreInitialState(false)
  {
    theInstance = this;
    lastGameState = theInstance->gameInfo.state;
  }

  PROCESS_WIDE_STORAGE_STATIC BH2010StableGameSymbols* theInstance;

  BehaviorData& behaviorData;
  RobotInfo& robotInfo;

  OwnTeamInfo& ownTeamInfo;
  const OpponentTeamInfo& opponentTeamInfo;
  GameInfo& gameInfo;
  const FrameInfo& frameInfo;
  const BallModel& ballModel;
  const RobotPose& robotPose;
  const FieldDimensions& fieldDimensions;

  bool startup;
  unsigned char lastGameState;
  unsigned timeStamp;
  unsigned timeSincePen2Play;

  float getTimeSincePlayingState();
  /** registers the symbols at an engine */
  void registerSymbols(xabsl::Engine& engine);

  static int getOwnTeamColor(){return (int)theInstance->ownTeamInfo.teamColor;}
  static void setOwnTeamColor(int color){theInstance->ownTeamInfo.teamColor = color;}

  static int getOppTeamColor()
  {
    if ((int)theInstance->ownTeamInfo.teamColor == TEAM_RED) return TEAM_BLUE;
    else return TEAM_RED;
  }

  static int getPlayerNumber(){return theInstance->robotInfo.number;}

  static int getPenalty(){return (int)theInstance->robotInfo.penalty;}
  static void setPenalty(int penalty){theInstance->robotInfo.penalty = penalty;}

  static float getPenaltyRemainingTime(){return (float)theInstance->robotInfo.secsTillUnpenalised;}

  static int getState(){return (int)theInstance->gameInfo.state;}
  static void setState(int state){theInstance->gameInfo.state = state;}

  static int getKickoffTeam(){return (int)theInstance->gameInfo.kickOffTeam;}
  static void setKickoffTeam(int kickoff){theInstance->gameInfo.kickOffTeam = kickoff;}

  static int getSecondaryState() {return (int)theInstance->gameInfo.secondaryState;}
  static void setSecondaryState(int secondaryState) {theInstance->gameInfo.secondaryState = secondaryState;}

  static float getOwnScore(){return (float)theInstance->ownTeamInfo.score;}

  static float getOpponentScore(){return (float)theInstance->opponentTeamInfo.score;}

  static float getRemainingTime(){return (float)theInstance->gameInfo.secsRemaining;}

  static float getTimeSinceInitial() {return (float)SystemCall::getTimeSince(theInstance->lastTimeInInitial);}

  static float staticGetTimeSincePlayingState(){return theInstance->getTimeSincePlayingState();}

  static float getTimeSinceLastPenalization(){return (float) theInstance->frameInfo.getTimeSince(theInstance->timeSincePen2Play);}

  static float getTimeSinceLastPackageReceived(){return (float) theInstance->gameInfo.timeSinceLastPackageReceived;}

  static int getDropInTeam(){return theInstance->gameInfo.dropInTime > 0 ? (int)theInstance->gameInfo.dropInTeam : (int) TEAM_NONE;}
  static void setDropInTeam(int dropInTeam){theInstance->gameInfo.dropInTeam = dropInTeam;}

  static float getDropInTime(){return (float)theInstance->gameInfo.dropInTime;}
  static void setDropInTime(float dropInTime){theInstance->gameInfo.dropInTime = static_cast<RoboCup::uint16>(dropInTime);}

  void update();

  void init();

private:
  int counter;
  int kickOffTeamBegin;
  bool kickOffChanged;
  unsigned int lastTimeInInitial;
  bool disablePreInitialState;
};


#endif // __GameSymbols_h_

