/** 
 * @file BH2010StableGoalSymbols.cpp
 *
 * Implementation of class BH2010StableGoalSymbols.
 *
 * @author Tim Laue
 * @author Carsten Koenemann
 */

#include "BH2010StableGoalSymbols.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Range.h"

PROCESS_WIDE_STORAGE BH2010StableGoalSymbols* BH2010StableGoalSymbols::theInstance = 0;

void BH2010StableGoalSymbols::registerSymbols(xabsl::Engine& engine)
{
  /*
   * free part angles
   */
  // angle to center of largest free part of opponent goal
  engine.registerDecimalInputSymbol("opponent_goal.free_part.angle_to_center", &getCenterAngleOfFreePart);
  // angle to outer side of largest free part of opponent goal
  engine.registerDecimalInputSymbol("opponent_goal.free_part.angle_to_outer_side", &getOuterAngleOfFreePart);
  // angle to inner side of largest free part of opponent goal
  engine.registerDecimalInputSymbol("opponent_goal.free_part.angle_to_inner_side", &getInnerAngleOfFreePart);
  // angle width of largest free part of opponent goal
  engine.registerDecimalInputSymbol("opponent_goal.free_part.angle_width", &getOpeningAngleOfFreePart);
  // angle tolerance to largest free part of opponent goal
  // (how accurate will the robot be positioned)
  engine.registerDecimalInputSymbol("opponent_goal.free_part.angle_tolerance", &getAngleToleranceToFreePart);
  /*
   * other perceptions
   */
  // time since opponent goal was last seen in ms
  engine.registerDecimalInputSymbol("opponent_goal.time_since_last_seen", &timeSinceOppGoalWasSeen);
  // was the opponent goal recently seen?
  engine.registerBooleanInputSymbol("opponent_goal.was_seen", &oppGoalWasSeen);
  // time since any goal was last seen in ms
  engine.registerDecimalInputSymbol("goal.time_since_last_seen_any_goal", &timeSinceAnyGoalWasSeen);
  engine.registerDecimalInputSymbol("goal.time_since_last_seen", &timeSinceAnyGoalWasSeen);
  // was any goal recently seen?
  engine.registerBooleanInputSymbol("goal.was_seen", &goalWasSeen);
  // count the number of seen goal posts in current ready state 
  engine.registerDecimalInputSymbol("goal.number_of_sightings_in_ready", &goalsSeenInReadyState);
  // count the number of seen goal posts since last pickup
  engine.registerDecimalInputSymbol("goal.number_of_sightings_since_last_pickup", &goalsSeenSinceLastPickup);

  engine.registerDecimalInputSymbol("goal.angle_to_last_seen", &getAngleToLastSeen);
  engine.registerDecimalInputSymbol("goal.distance_to_last_seen", &getDistanceToLastSeen);

}

void BH2010StableGoalSymbols::update()
{
  timeSinceOppGoalWasSeen = (float) theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOppGoalLastSeen);
  timeSinceOwnGoalWasSeen = (float) theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOwnGoalLastSeen);
  timeSinceAnyGoalWasSeen = timeSinceOppGoalWasSeen < timeSinceOwnGoalWasSeen ? timeSinceOppGoalWasSeen : timeSinceOwnGoalWasSeen;
  oppGoalWasSeen          = timeSinceOppGoalWasSeen < 500;
  goalWasSeen             = timeSinceAnyGoalWasSeen < 500;

  /* Calculate seen goals in ready state*/
  if(gameInfo.state == STATE_READY)
  {
    if(theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOppGoalLastSeen) == 0 ||
       theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOwnGoalLastSeen) == 0)
      goalsSeenInReadyState += 1.0;
  }
  else
  {
    goalsSeenInReadyState = 0;
  }
  
  /* Calculate seen goals since last pickup */
  if(groundContactState.noContactSafe == true 
    && (fallDownState.state == FallDownState::upright || fallDownState.state == FallDownState::staggering))
  {
    goalsSeenSinceLastPickup = 0.0; 
  }
  else
  {
    if(theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOppGoalLastSeen) == 0 ||
      theInstance->frameInfo.getTimeSince(goalPercept.timeWhenOwnGoalLastSeen) == 0)
      goalsSeenSinceLastPickup += 1.0;
  }
  PLOT("Behavior:goalsSeenSinceLastPickup", goalsSeenSinceLastPickup);

  /* determine angles to sides of free part of opponent goal */
  Vector2<> centerOfOppGoalAbs = Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGoalpost, (float) (theInstance->fieldDimensions.yPosLeftGoal + theInstance->fieldDimensions.yPosRightGoal) / 2.0f);
  Vector2<> centerOfOppGoalRel = Geometry::fieldCoord2Relative(theInstance->robotPose, centerOfOppGoalAbs);
  float     leftDistance       = fabs(centerOfOppGoalAbs.y - theInstance->freePartOfOpponentGoalModel.leftEnd.y);
  float     rightDistance      = fabs(centerOfOppGoalAbs.y - theInstance->freePartOfOpponentGoalModel.rightEnd.y);
  if(leftDistance < rightDistance)
  {
    innerSide = theInstance->freePartOfOpponentGoalModel.leftEnd;
    outerSide = theInstance->freePartOfOpponentGoalModel.rightEnd;
  }
  else
  {
    innerSide = theInstance->freePartOfOpponentGoalModel.rightEnd;
    outerSide = theInstance->freePartOfOpponentGoalModel.leftEnd;
  }

  /* draw angles to free part into worldstate, see draw()-function below */
  DECLARE_DEBUG_DRAWING("behavior:GoalSymbols:FreePartAnglesOnField", "drawingOnField");
  COMPLEX_DRAWING("behavior:GoalSymbols:FreePartAnglesOnField", drawingOnField());
}

void BH2010StableGoalSymbols::drawingOnField()
{
  float dist = Geometry::distance(theInstance->robotPose.translation, Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGroundline, (float) (theInstance->fieldDimensions.yPosLeftGoal + theInstance->fieldDimensions.yPosRightGoal) / 2.0f));
  Vector2<> targetImage;
  // draw angle to center of free part
  targetImage = Geometry::relative2FieldCoord(theInstance->robotPose, dist * cosf(fromDegrees(getCenterAngleOfFreePart())), dist * sinf(fromDegrees(getCenterAngleOfFreePart())));
  LINE(
    "behavior:GoalSymbols:FreePartAnglesOnField",
    theInstance->robotPose.translation.x, theInstance->robotPose.translation.y, targetImage.x, targetImage.y,
    15, Drawings::ps_solid, ColorClasses::black
  );
  // draw angle to inner side of free part
  targetImage = Geometry::relative2FieldCoord(theInstance->robotPose, dist * cosf(fromDegrees(getInnerAngleOfFreePart())), dist * sinf(fromDegrees(getInnerAngleOfFreePart())));
  LINE(
    "behavior:GoalSymbols:FreePartAnglesOnField",
    theInstance->robotPose.translation.x, theInstance->robotPose.translation.y, targetImage.x, targetImage.y,
    15, Drawings::ps_solid, ColorClasses::red
  );
  // draw angle to outer side of free part
  targetImage = Geometry::relative2FieldCoord(theInstance->robotPose, dist * cosf(fromDegrees(getOuterAngleOfFreePart())), dist * sinf(fromDegrees(getOuterAngleOfFreePart())));
  LINE(
    "behavior:GoalSymbols:FreePartAnglesOnField",
    theInstance->robotPose.translation.x, theInstance->robotPose.translation.y, targetImage.x, targetImage.y,
    15, Drawings::ps_solid, ColorClasses::blue
  );
}

  /*
   * this stuff gets passed to xabsl
   */

float BH2010StableGoalSymbols::getCenterAngleOfFreePart()
{
  Pose2D    poseForOppGoalAngle = theInstance->robotPose;
  Vector2<> leftOppGoalPostAbs  = Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGroundline, (float) theInstance->fieldDimensions.yPosLeftGoal);
  Vector2<> rightOppGoalPostAbs = Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGroundline, (float) theInstance->fieldDimensions.yPosRightGoal);
  if(poseForOppGoalAngle.translation.x > float(theInstance->fieldDimensions.xPosOpponentGroundline - 50))
  {
    poseForOppGoalAngle.translation.x = float(theInstance->fieldDimensions.xPosOpponentGroundline - 50);
  }
  float leftOppGoalPostAngle  = Geometry::angleTo(poseForOppGoalAngle, leftOppGoalPostAbs);
  float rightOppGoalPostAngle = Geometry::angleTo(poseForOppGoalAngle, rightOppGoalPostAbs);
  if(leftOppGoalPostAngle < rightOppGoalPostAngle)
  {
    leftOppGoalPostAngle += pi2;
  }
  Vector2<> centerOfFreePart = (theInstance->freePartOfOpponentGoalModel.leftEnd + theInstance->freePartOfOpponentGoalModel.rightEnd) / 2.0f;
  float angleToCenterOfFreePart = Range<>(rightOppGoalPostAngle - getAngleToleranceToFreePart(), leftOppGoalPostAngle + getAngleToleranceToFreePart()).limit(centerOfFreePart.angle());
  return toDegrees(normalize(angleToCenterOfFreePart));
}

float BH2010StableGoalSymbols::getOuterAngleOfFreePart()
{
  return toDegrees(normalize(theInstance->outerSide.angle()));
}

float BH2010StableGoalSymbols::getInnerAngleOfFreePart()
{
  return toDegrees(normalize(theInstance->innerSide.angle()));
}

float BH2010StableGoalSymbols::getAngleToleranceToFreePart()
{
  return std::max(10.0f, getOpeningAngleOfFreePart() - 10.0f);
}

float BH2010StableGoalSymbols::getOpeningAngleOfFreePart()
{
  float angleToLeftEnd  = theInstance->freePartOfOpponentGoalModel.leftEnd.angle();
  float angleToRightEnd = theInstance->freePartOfOpponentGoalModel.rightEnd.angle();
  if(angleToLeftEnd < angleToRightEnd)
  {
    angleToLeftEnd += pi2;
  }
  return toDegrees(fabs(normalize(angleToLeftEnd - angleToRightEnd)));
}

float BH2010StableGoalSymbols::getAngleToLastSeen()
{
  Vector2<int> position;
  unsigned int maxLastSeen = 0;
  for(int i = 0; i < GoalPercept::NUMBER_OF_GOAL_POSTS; ++i)
    if(theInstance->goalPercept.posts[i].timeWhenLastSeen > maxLastSeen)
    {
      maxLastSeen = theInstance->goalPercept.posts[i].timeWhenLastSeen;
      position = theInstance->goalPercept.posts[i].positionOnField;
    }
  for(int i = 0; i < GoalPercept::NUMBER_OF_UNKNOWN_GOAL_POSTS; ++i)
    if(theInstance->goalPercept.posts[i].timeWhenLastSeen > maxLastSeen)
    {
      maxLastSeen = theInstance->goalPercept.posts[i].timeWhenLastSeen;
      position = theInstance->goalPercept.posts[i].positionOnField;
    }
  return toDegrees(Vector2<>(float(position.x), float(position.y)).angle());
}

float BH2010StableGoalSymbols::getDistanceToLastSeen()
{
  Vector2<int> position;
  unsigned int maxLastSeen = 0;
  for(int i = 0; i < GoalPercept::NUMBER_OF_GOAL_POSTS; ++i)
    if(theInstance->goalPercept.posts[i].timeWhenLastSeen > maxLastSeen)
    {
      maxLastSeen = theInstance->goalPercept.posts[i].timeWhenLastSeen;
      position = theInstance->goalPercept.posts[i].positionOnField;
    }
  for(int i = 0; i < GoalPercept::NUMBER_OF_UNKNOWN_GOAL_POSTS; ++i)
    if(theInstance->goalPercept.posts[i].timeWhenLastSeen > maxLastSeen)
    {
      maxLastSeen = theInstance->goalPercept.posts[i].timeWhenLastSeen;
      position = theInstance->goalPercept.posts[i].positionOnField;
    }
  return Vector2<>(float(position.x), float(position.y)).abs();
}
