/** 
* @file BH2010StableLocatorSymbols.cpp
*
* Implementation of class BH2010StableLocatorSymbols.
*
* @author Judith Mller
*/

#include "BH2010StableLocatorSymbols.h"
//#include "Tools/Debugging/Modify.h"

PROCESS_WIDE_STORAGE BH2010StableLocatorSymbols* BH2010StableLocatorSymbols::theInstance = 0;

void BH2010StableLocatorSymbols::registerSymbols(xabsl::Engine& engine)
{
  // position
  engine.registerDecimalInputSymbol("locator.pose.x",&robotPose.translation.x);
  engine.registerDecimalInputSymbol("locator.pose.y",&robotPose.translation.y);
  engine.registerDecimalInputSymbol("locator.pose.angle", &getPoseAngle);
  engine.registerDecimalInputSymbol("locator.pose.deviation", &robotPose.deviation);
  engine.registerDecimalInputSymbol("locator.time_since_last_pose_reset", &getTimeSinceLastPoseReset);

  engine.registerDecimalInputSymbol("locator.opponent_goal.angle", &angleToOpponentGoal);
  engine.registerDecimalInputSymbol("locator.opponent_goal.angle_width", &angleWidthToOpponentGoal);

  engine.registerDecimalInputSymbol("locator.own_goal.angle", &angleToOwnGoal);
  engine.registerDecimalInputSymbol("locator.own_goal.angle_width", &angleWidthToOwnGoal);

  engine.registerDecimalInputSymbol("locator.ball_robot_with_offset.y", &ballPosRelWithOffset.y);
  engine.registerDecimalInputSymbol("locator.angle_to_ball_robot_with_offset", &angleToBallPosRelWithOffset);

  // "distance_to"
  engine.registerDecimalInputSymbol("locator.distance_to", &distanceTo);
  engine.registerDecimalInputSymbolDecimalParameter("locator.distance_to", "locator.distance_to.x", &x);
  engine.registerDecimalInputSymbolDecimalParameter("locator.distance_to", "locator.distance_to.y", &y);

  // "angle_to"
  engine.registerDecimalInputSymbol("locator.angle_to", &angleTo);
  engine.registerDecimalInputSymbolDecimalParameter("locator.angle_to", "locator.angle_to.x", &x);
  engine.registerDecimalInputSymbolDecimalParameter("locator.angle_to", "locator.angle_to.y", &y);

  // "field_to_relative.x"
  engine.registerDecimalInputSymbol("locator.field_to_relative.x", &getFieldToRelativeX);
  engine.registerDecimalInputSymbolDecimalParameter("locator.field_to_relative.x", "locator.field_to_relative.x.x", &x);
  engine.registerDecimalInputSymbolDecimalParameter("locator.field_to_relative.x", "locator.field_to_relative.x.y", &y);

  // "field_to_relative.y"
  engine.registerDecimalInputSymbol("locator.field_to_relative.y", &getFieldToRelativeY);
  engine.registerDecimalInputSymbolDecimalParameter("locator.field_to_relative.y", "locator.field_to_relative.y.x", &x);
  engine.registerDecimalInputSymbolDecimalParameter("locator.field_to_relative.y", "locator.field_to_relative.y.y", &y);

  engine.registerBooleanInputSymbol("locator.ground_contact", &groundContactState.contact);

  engine.registerDecimalInputSymbol("locator.angle_tolerance", &angleToleranceToOpponentGoal);

  engine.registerDecimalInputSymbol("locator.angle_to_line_through_ball_and_goal", &getAngleToLineThroughBallAndGoal);
}

void BH2010StableLocatorSymbols::update()
{
  // calculate angle to opponent goal

  // if robot pose is behind the groundline it is clipped to a point close to the groundline
  // this is done because bad things happen to the calculated angle if the pose is behind the groundline
  Pose2D poseForOppGoalAngle = theInstance->robotPose;
  if (poseForOppGoalAngle.translation.x > float(theInstance->fieldDimensions.xPosOpponentGroundline - 50))
    poseForOppGoalAngle.translation.x = float(theInstance->fieldDimensions.xPosOpponentGroundline - 50);

  float angleToLeftOpponentGoalPost  = Geometry::angleTo(poseForOppGoalAngle, Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGroundline, (float) theInstance->fieldDimensions.yPosLeftGoal));
  float angleToRightOpponentGoalPost = Geometry::angleTo(poseForOppGoalAngle, Vector2<>((float) theInstance->fieldDimensions.xPosOpponentGroundline, (float) theInstance->fieldDimensions.yPosRightGoal));

  if(angleToLeftOpponentGoalPost < angleToRightOpponentGoalPost)
    angleToLeftOpponentGoalPost += pi2;


  //Pose2D poseForBallToOppGoalAngle(0.0f, theInstance->ballModel.lastPerception.getPositionInFieldCoordinates();
  //float angleFromBallToGoal = Geometry::angleTo(poseForBallToOppGoalAngle, Vector2<>(theInstance->fieldDimensions.xPosOpponentGroundline, 0));
  
  //OUTPUT(idText, text, angleFromBallToGoal);

  angleToOpponentGoal = toDegrees(normalize((angleToLeftOpponentGoalPost + angleToRightOpponentGoalPost) / 2.0f));
  angleWidthToOpponentGoal = toDegrees(fabs(normalize(angleToLeftOpponentGoalPost - angleToRightOpponentGoalPost) / 2.0f));
  angleToleranceToOpponentGoal = std::max(10.0f, angleWidthToOpponentGoal - 10.0f);

  ballPosRelWithOffset = theInstance->ballModel.lastPerception.position;
  angleToBallPosRelWithOffset = toDegrees(theInstance->ballModel.lastPerception.getAngle());


  if (angleToOpponentGoal > 150)
  {
    Vector2<> offset(0,-350);
    ballPosRelWithOffset += offset;
    if (angleToOpponentGoal > 160) ballPosRelWithOffset += Vector2<>(0,700);
    angleToBallPosRelWithOffset = toDegrees(ballPosRelWithOffset.angle());
  }
  else if(angleToOpponentGoal < -150)
  {
    Vector2<> offset(0,350);
    ballPosRelWithOffset += offset;
    angleToBallPosRelWithOffset = toDegrees(ballPosRelWithOffset.angle());
  }
  else
  {
    if (angleToOpponentGoal < -20) 
    {
      Vector2<> offset(0,500);
      ballPosRelWithOffset += offset;
      angleToBallPosRelWithOffset = toDegrees(ballPosRelWithOffset.angle());
    }
    else if(angleToOpponentGoal > 20)
    {
      Vector2<> offset(0,-500);
      ballPosRelWithOffset += offset;
      angleToBallPosRelWithOffset = toDegrees(ballPosRelWithOffset.angle());
    }
  }

  Pose2D poseForOwnGoalAngle = theInstance->robotPose;
  if (poseForOwnGoalAngle.translation.x > float(theInstance->fieldDimensions.xPosOpponentGroundline - 50))
    poseForOwnGoalAngle.translation.x = float(theInstance->fieldDimensions.xPosOpponentGroundline - 50);

  float angleToRightOwnGoalPost  = Geometry::angleTo(poseForOwnGoalAngle, Vector2<>((float) theInstance->fieldDimensions.xPosOwnGroundline, (float) theInstance->fieldDimensions.yPosLeftGoal));
  float angleToLeftOwnGoalPost = Geometry::angleTo(poseForOwnGoalAngle, Vector2<>((float) theInstance->fieldDimensions.xPosOwnGroundline, (float) theInstance->fieldDimensions.yPosRightGoal));
  if(angleToLeftOwnGoalPost < angleToRightOwnGoalPost)
    angleToLeftOwnGoalPost += pi2;

  angleToOwnGoal = toDegrees(normalize((angleToLeftOwnGoalPost + angleToRightOwnGoalPost) / 2.0f));
  angleWidthToOwnGoal = toDegrees(fabs(normalize((angleToRightOwnGoalPost - angleToLeftOwnGoalPost) / 2.0f)));

}

float BH2010StableLocatorSymbols::distanceTo()
{
  return Geometry::distanceTo(theInstance->robotPose,Vector2<>(theInstance->x, theInstance->y));
}

float BH2010StableLocatorSymbols::angleTo()
{
  return toDegrees(Geometry::angleTo(theInstance->robotPose,Vector2<>(theInstance->x, theInstance->y)));
}

float BH2010StableLocatorSymbols::getFieldToRelativeX()
{
  return Geometry::fieldCoord2Relative(theInstance->robotPose,Vector2<>(theInstance->x, theInstance->y)).x;
}

float BH2010StableLocatorSymbols::getFieldToRelativeY()
{
  return Geometry::fieldCoord2Relative(theInstance->robotPose,Vector2<>(theInstance->x, theInstance->y)).y;
}

float BH2010StableLocatorSymbols::getPoseX()
{
  return theInstance->robotPose.translation.x;
}

float BH2010StableLocatorSymbols::getPoseY()
{
  return theInstance->robotPose.translation.y;
}

float BH2010StableLocatorSymbols::getPoseAngle()
{
  return toDegrees(theInstance->robotPose.getAngle());
}
/*
float BH2010StableLocatorSymbols::getOpponentGoalAngle()
{
return theInstance->angleToOpponentGoal;
}
*/

float BH2010StableLocatorSymbols::getAngleToLineThroughBallAndGoal()
{
  const Vector2<> ballAbsPos = theInstance->ballModel.lastSeenEstimate.getPositionInFieldCoordinates(theInstance->robotPose);
  const Vector2<> goalToBall =  ballAbsPos - Vector2<>((float)theInstance->fieldDimensions.xPosOpponentGoal, (float)theInstance->fieldDimensions.yPosCenterGoal);
  const Vector2<> ballToRobot = theInstance->robotPose.translation - ballAbsPos;
  const float angle = toDegrees(acos((goalToBall * ballToRobot) / (goalToBall.abs() * ballToRobot.abs()))); // angle to line between ball and goal

  return angle;
}

float BH2010StableLocatorSymbols::getTimeSinceLastPoseReset()
{
  return float(theInstance->frameInfo.getTimeSince(theInstance->robotPoseInfo.timeLastPoseReset));
}
