/** 
* @file BH2010StableMotionSymbols.cpp
* Implementation of class BH2010StableMotionSymbols.
* @author Max Risler
*/

#include <cstdio>

#include "BH2010StableMotionSymbols.h"
#include "Tools/Xabsl/GT/GTXabslTools.h"

PROCESS_WIDE_STORAGE BH2010StableMotionSymbols* BH2010StableMotionSymbols::theInstance = 0;

void BH2010StableMotionSymbols::registerSymbols(xabsl::Engine& engine)
{
  // motion.type enumeration
  char s[256];
  char* dest = s + sprintf(s, "motion.type.");
  for(int i = 0; i < MotionRequest::numOfMotions; ++i)
  {
    getXabslString(dest, MotionRequest::getMotionName((MotionRequest::Motion)i));
    engine.registerEnumElement("motion.type",s,i);
  }

  // motion.special_action enumeration
  dest = s + sprintf(s, "motion.special_action.");
  for(int i = 0; i < SpecialActionRequest::numOfSpecialActions; ++i)
  {    
    getXabslString(dest, SpecialActionRequest::getSpecialActionName((SpecialActionRequest::SpecialActionID)i));
    engine.registerEnumElement("motion.special_action",s,i);
  }

  // MotionRequest
  engine.registerEnumeratedOutputSymbol("motion.type","motion.type", (int*)&motionRequest.motion);
  
  engine.registerDecimalOutputSymbol("motion.walk_speed.x", &setWalkSpeedX, &getWalkSpeedX);
  engine.registerDecimalOutputSymbol("motion.walk_speed.y", &setWalkSpeedY, &getWalkSpeedY);
  engine.registerDecimalOutputSymbol("motion.walk_speed.rot", &setWalkSpeedRot, &getWalkSpeedRot);
  engine.registerBooleanOutputSymbol("motion.walk_pedantic", &motionRequest.walkRequest.pedantic);

  engine.registerDecimalOutputSymbol("motion.walk_speed.percentage.x", &setPercentageSpeedX, &getPercentageSpeedX);
  engine.registerDecimalOutputSymbol("motion.walk_speed.percentage.y", &setPercentageSpeedY, &getPercentageSpeedY);
  engine.registerDecimalOutputSymbol("motion.walk_speed.percentage.rot", &setPercentageSpeedRot, &getPercentageSpeedRot);

  engine.registerDecimalOutputSymbol("motion.walk_target.x", &setWalkTargetX, &getWalkTargetX);
  engine.registerDecimalOutputSymbol("motion.walk_target.y", &setWalkTargetY, &getWalkTargetY);
  engine.registerDecimalOutputSymbol("motion.walk_target.rot", &setWalkTargetRot, &getWalkTargetRot);
  engine.registerBooleanInputSymbol("motion.walk_target.reached", &getWalkTargetReached);
  engine.registerBooleanInputSymbol("motion.walk_target.in_progress", &getWalkTargetInProgress);
  engine.registerDecimalOutputSymbol("motion.walk_speed", &setWalkSpeed, &getWalkSpeed);

  engine.registerEnumeratedOutputSymbol("motion.special_action","motion.special_action", &setSpecialAction, &getSpecialAction);
  engine.registerBooleanOutputSymbol("motion.special_action.mirror", &motionRequest.specialActionRequest.mirror);
  engine.registerEnumeratedOutputSymbol("motion.last_special_action","motion.special_action", (int*)&lastSpecialAction); //??

  engine.registerDecimalInputSymbol("motion.time_since_last_kick", &getTimeSinceLastKick);

  // SIKE
  dest = s + sprintf(s, "motion.sike.");
  for(int i = 0; i < SIKERequest::numOfMotions; ++i)
  {    
    getXabslString(dest, SIKERequest::getMotionName((SIKERequest::Motion)i));
    engine.registerEnumElement("motion.sike",s,i);
  }
  engine.registerBooleanOutputSymbol("motion.sike.mirror", &motionRequest.sikeRequest.mirror);
  engine.registerEnumeratedOutputSymbol("motion.sike","motion.sike", &setSIKE, &getSIKE);

  // MotionInfo
  engine.registerEnumeratedInputSymbol("executed_motion.type","motion.type", (const int*)&motionInfo.motion);
  engine.registerEnumeratedInputSymbol("executed_motion.special_action","motion.special_action", (const int*)&motionInfo.specialActionRequest.specialAction);
  engine.registerBooleanInputSymbol("executed_motion.special_action.mirror", &motionInfo.specialActionRequest.mirror);
  engine.registerEnumeratedInputSymbol("executed_motion.sike","motion.sike", (const int*)&motionInfo.sikeRequest.motion);
}

void BH2010StableMotionSymbols::update()
{
  motionRequest.walkRequest.pedantic = false; 
  if(motionInfo.motion == MotionRequest::bike || (motionInfo.motion == MotionRequest::specialAction && (motionRequest.specialActionRequest.specialAction == SpecialActionRequest::kickDiagonalNaoFast || motionRequest.specialActionRequest.specialAction == SpecialActionRequest::kickSidewardsNao || motionRequest.specialActionRequest.specialAction == SpecialActionRequest::kickSidewardsNaoMiddle)) || (motionInfo.motion == MotionRequest::walk && (motionRequest.walkRequest.kickType == WalkRequest::sidewardsLeft || motionRequest.walkRequest.kickType == WalkRequest::sidewardsLeft)))
  {
    timeStampOfLastKick = frameInfo.time;
  }
}

void BH2010StableMotionSymbols::setWalkSpeedX(float x)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::speedMode;
  walkRequest.speed.translation.x = x;  
}

float BH2010StableMotionSymbols::getWalkSpeedX()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::speedMode ? walkRequest.speed.translation.x : 0;
}

void BH2010StableMotionSymbols::setWalkSpeedY(float y)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::speedMode;
  walkRequest.speed.translation.y = y;
}

float BH2010StableMotionSymbols::getWalkSpeedY()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::speedMode ? walkRequest.speed.translation.y : 0;
}

void BH2010StableMotionSymbols::setWalkSpeedRot(float rot)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::speedMode;
  walkRequest.speed.rotation = fromDegrees(rot);  
}

float BH2010StableMotionSymbols::getWalkSpeedRot()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::speedMode ? toDegrees(walkRequest.speed.rotation) : 0;
}

void BH2010StableMotionSymbols::setPercentageSpeedX(float percent)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::percentageSpeedMode;
  walkRequest.speed.translation.x = percent / 100.0f;
}

float BH2010StableMotionSymbols::getPercentageSpeedX()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::percentageSpeedMode ? walkRequest.speed.translation.x : 0;
}

void BH2010StableMotionSymbols::setPercentageSpeedY(float percent)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::percentageSpeedMode;
  walkRequest.speed.translation.y = percent / 100.0f;
}

float BH2010StableMotionSymbols::getPercentageSpeedY()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::percentageSpeedMode ? walkRequest.speed.translation.y : 0;
}

void BH2010StableMotionSymbols::setPercentageSpeedRot(float percent)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::percentageSpeedMode;
  walkRequest.speed.rotation = percent / 100.0f;
}

float BH2010StableMotionSymbols::getPercentageSpeedRot()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::percentageSpeedMode ? walkRequest.speed.rotation : 0;
}

void BH2010StableMotionSymbols::setWalkTargetX(float x)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::targetMode;
  walkRequest.target.translation.x = x;
}

float BH2010StableMotionSymbols::getWalkTargetX()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::targetMode ? walkRequest.target.translation.x : 0;
}

void BH2010StableMotionSymbols::setWalkTargetY(float y)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::targetMode;
  walkRequest.target.translation.y = y;
}

float BH2010StableMotionSymbols::getWalkTargetY()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::targetMode ? walkRequest.target.translation.x : 0;
}

void BH2010StableMotionSymbols::setWalkTargetRot(float rot)
{
  WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  walkRequest.mode = WalkRequest::targetMode;
  walkRequest.target.rotation = fromDegrees(rot);
}

float BH2010StableMotionSymbols::getWalkTargetRot()
{
  const WalkRequest& walkRequest(theInstance->motionRequest.walkRequest);
  return walkRequest.mode == WalkRequest::targetMode ? toDegrees(walkRequest.target.rotation) : 0;
}

bool BH2010StableMotionSymbols::getWalkTargetReached()
{
  const MotionInfo& motionInfo(theInstance->motionInfo);
  return motionInfo.motion == MotionRequest::walk && 
    motionInfo.walkRequest.mode == WalkRequest::targetMode &&
    motionInfo.walkRequest.target.translation.squareAbs() < 10.f * 10.f &&
    fabs(motionInfo.walkRequest.target.rotation) < fromDegrees(1.f);
}

bool BH2010StableMotionSymbols::getWalkTargetInProgress()
{
  const MotionInfo& motionInfo(theInstance->motionInfo);
  return motionInfo.motion == MotionRequest::walk && motionInfo.walkRequest.mode == WalkRequest::targetMode && motionInfo.walkRequest.target != Pose2D();
}

void BH2010StableMotionSymbols::setWalkSpeed(float speed)
{
  Pose2D& targetSpeed = theInstance->motionRequest.walkRequest.speed;
  speed /= 100.f;
  if(speed < 0)
    speed = 0.;
  if(speed > 1)
    speed = 1.;
  targetSpeed.translation.x = speed;
  targetSpeed.translation.y = speed;
  targetSpeed.rotation = speed;
}

float BH2010StableMotionSymbols::getWalkSpeed()
{
  return theInstance->motionRequest.walkRequest.speed.translation.x * 100.f;
}

void BH2010StableMotionSymbols::setSpecialAction(int specialAction)
{
  SpecialActionRequest& specialActionRequest(theInstance->motionRequest.specialActionRequest);
  specialActionRequest.specialAction = SpecialActionRequest::SpecialActionID(specialAction);
  specialActionRequest.specialize(theInstance->robotName);
}

int BH2010StableMotionSymbols::getSpecialAction()
{
  return int(theInstance->motionRequest.specialActionRequest.specialAction);
}

void BH2010StableMotionSymbols::setSIKE(int sikeId)
{
  SIKERequest &request(theInstance->motionRequest.sikeRequest);
  request.motion = SIKERequest::Motion(sikeId);
}

int BH2010StableMotionSymbols::getSIKE()
{
  return int(theInstance->motionRequest.sikeRequest.motion);
}

float BH2010StableMotionSymbols::getTimeSinceLastKick()
{
  return (float)theInstance->frameInfo.getTimeSince(theInstance->timeStampOfLastKick);
}

