/** 
* @file MotionSymbols.cpp
* Implementation of class MotionSymbols.
* @author Max Risler
*/

#include <cstdio>

#include "MotionSymbols.h"
#include "Tools/Xabsl/GT/GTXabslTools.h"

PROCESS_WIDE_STORAGE MotionSymbols* MotionSymbols::theInstance = 0;

void MotionSymbols::registerSymbols(xabsl::Engine& engine)
{
  // motion.type enumeration
  char s[256];
  char* dest = s + sprintf(s, "motion.type.");
  for(int i = 0; i < MotionRequest::numOfMotions; ++i)
  {
    getXabslString(dest, MotionRequest::getMotionName((MotionRequest::Motion)i));
    engine.registerEnumElement("motion.type",s,i);
  }

  // motion.special_action enumeration
  dest = s + sprintf(s, "motion.special_action.");
  for(int i = 0; i < SpecialActionRequest::numOfSpecialActions; ++i)
  {    
    getXabslString(dest, SpecialActionRequest::getSpecialActionName((SpecialActionRequest::SpecialActionID)i));
    engine.registerEnumElement("motion.special_action",s,i);
  }

  // MotionRequest
  engine.registerEnumeratedOutputSymbol("motion.type","motion.type", (int*)&motionRequest.motion);
  
  engine.registerDecimalOutputSymbol("motion.walk_target.x", &motionRequest.walkRequest.target.translation.x);
  engine.registerDecimalOutputSymbol("motion.walk_target.y", &motionRequest.walkRequest.target.translation.y);
  engine.registerDecimalOutputSymbol("motion.walk_target.rot", &setWalkTargetRot, &getWalkTargetRot);
  engine.registerBooleanInputSymbol("motion.walk_target.reached", &getWalkTargetReached);
  engine.registerBooleanOutputSymbol("motion.walk_pedantic", &motionRequest.walkRequest.pedantic);

  engine.registerDecimalOutputSymbol("motion.walk_speed", &setWalkSpeed, &getWalkSpeed);

  engine.registerEnumeratedOutputSymbol("motion.special_action","motion.special_action", &setSpecialAction, &getSpecialAction);
  engine.registerBooleanOutputSymbol("motion.special_action.mirror", &motionRequest.specialActionRequest.mirror);
  engine.registerEnumeratedOutputSymbol("motion.last_special_action","motion.special_action", (int*)&lastSpecialAction); //??

  // SIKE
  dest = s + sprintf(s, "motion.sike.");
  for(int i = 0; i < SIKERequest::numOfMotions; ++i)
  {    
    getXabslString(dest, SIKERequest::getMotionName((SIKERequest::Motion)i));
    engine.registerEnumElement("motion.sike",s,i);
  }
  engine.registerBooleanOutputSymbol("motion.sike.mirror", &motionRequest.sikeRequest.mirror);
  engine.registerEnumeratedOutputSymbol("motion.sike","motion.sike", &setSIKE, &getSIKE);

  // MotionInfo
  engine.registerEnumeratedInputSymbol("executed_motion.type","motion.type", (const int*)&motionInfo.motion);
  engine.registerEnumeratedInputSymbol("executed_motion.special_action","motion.special_action", (const int*)&motionInfo.specialActionRequest.specialAction);
  engine.registerBooleanInputSymbol("executed_motion.special_action.mirror", &motionInfo.specialActionRequest.mirror);
  engine.registerEnumeratedInputSymbol("executed_motion.sike","motion.sike", (const int*)&motionInfo.sikeRequest.motion);
}

void MotionSymbols::update()
{
  motionRequest.walkRequest.pedantic = false; 
  motionRequest.walkRequest.mode = WalkRequest::targetMode;
}

void MotionSymbols::setWalkSpeed(float speed)
{
  Pose2D& targetSpeed = theInstance->motionRequest.walkRequest.speed;
  speed /= 100.f;
  if(speed < 0)
    speed = 0.;
  if(speed > 1)
    speed = 1.;
  targetSpeed.translation.x = speed;
  targetSpeed.translation.y = speed;
  targetSpeed.rotation = speed;
}

float MotionSymbols::getWalkSpeed()
{
  return theInstance->motionRequest.walkRequest.speed.translation.x * 100.f;
}

void MotionSymbols::setWalkTargetRot(float rot)
{
  theInstance->motionRequest.walkRequest.target.rotation = fromDegrees(rot);
}

float MotionSymbols::getWalkTargetRot()
{
  return toDegrees(theInstance->motionRequest.walkRequest.target.rotation);
}

bool MotionSymbols::getWalkTargetReached()
{
  const MotionInfo& motionInfo(theInstance->motionInfo);
  return motionInfo.motion == MotionRequest::walk && 
    motionInfo.walkRequest.mode == WalkRequest::targetMode &&
    motionInfo.walkRequest.target.translation.squareAbs() < 10.f * 10.f &&
    fabs(motionInfo.walkRequest.target.rotation) < fromDegrees(1.f);
}

void MotionSymbols::setSpecialAction(int specialAction)
{
  SpecialActionRequest& specialActionRequest(theInstance->motionRequest.specialActionRequest);
  specialActionRequest.specialAction = SpecialActionRequest::SpecialActionID(specialAction);
  specialActionRequest.specialize(theInstance->robotName);
}

int MotionSymbols::getSpecialAction()
{
  return int(theInstance->motionRequest.specialActionRequest.specialAction);
}

void MotionSymbols::setSIKE(int sikeId)
{
  SIKERequest &request(theInstance->motionRequest.sikeRequest);
  request.motion = SIKERequest::Motion(sikeId);
}

int MotionSymbols::getSIKE()
{
  return int(theInstance->motionRequest.sikeRequest.motion);
}

