/**
* @file Logger.cpp
* Declaration of module Logger which collects data during operation and
* can write it to a file if wanted.
*
* @author <a href="mailto:Tim.Laue@dfki.de">Tim Laue</a>
*/

#include "Logger.h"
#include "Tools/Settings.h"


void Logger::update(LoggerOutput& loggerOutput)
{
  // [DONE] Log only if state != FINISHED
  // [DONE] Write always if state == FINISHED and buffer not empty
  // [DONE] Clear buffer after writing to disk
  // [DONE] Game state naming
  DEBUG_RESPONSE("module:Logger:export",exportData("xabsl-log"););
  if(theGameInfo.state != STATE_FINISHED)
  {
    const std::vector<SimpleXabslInfo::ActiveOption>& activeOptions = theSimpleXabslInfo.activeOptions;
    std::vector<std::string> compressedList;
    std::vector<SimpleXabslInfo::ActiveOption>::const_iterator iter = activeOptions.begin();
    while(iter != activeOptions.end())
    {
      std::string optionInfo = iter->name + "(" + iter->activeState +")";
      compressedList.push_back(optionInfo);
      ++iter;
    }
    optionsBuffer->add(compressedList);
  }
  else if(optionsBuffer->getNumberOfEntries())
  {
    exportData(theGameInfo.firstHalf ? "xabsl-log-first-half" : "xabsl-log-second-half");
    optionsBuffer->init();
  }
}

void Logger::exportData(const std::string& fileName)
{
  int lastIndex = optionsBuffer->getNumberOfEntries() - 1;
  const int maxIndex = lastIndex;
  OutTextRawFile stream(Global::getSettings().expandLocationFilename(
    std::string("../../Logs/")+fileName+"-"+theRobotName.name+".html"));
  // Create HTML header
  stream << "<html><body>";
  stream << "<table>";
  // Write lines:
  while (lastIndex >= 0)
  {
    std::vector<std::string>& currentList = (*optionsBuffer)[lastIndex];
    std::vector<std::string>& previousList = (*optionsBuffer)[std::min(lastIndex+1,maxIndex)];
    bool changeOccured(false);
    if(currentList == previousList) // No changes -> row is green
    {
      stream << "<tr style=\"background-color:#DDFFDD\">";
    }
    else                            // Row is highlighted, otherwise
    {
      stream << "<tr style=\"background-color:#FFBB99\">";
      changeOccured = true;
    }
    std::vector<std::string>::const_iterator iter = currentList.begin();
    while(iter != currentList.end())
    {
      // Stronger highlighting for the options which have changed
      if(changeOccured && !listContains(previousList,*iter))
        stream << "<td style=\"background-color:#FF2222\">" << *iter << "</td>";
      else
        stream << "<td>" << *iter << "</td>";
      ++iter;
    }
    stream << "</tr>";
    lastIndex--;
  }
  // Create HTML footer
  stream << "</table>";
  stream << "</body></html>";
}

bool Logger::listContains(const std::vector<std::string>& l, const std::string& e)
{
  std::vector<std::string>::const_iterator i = l.begin();
  while(i != l.end())
  {
    if(*i == e) 
      return true;
    ++i;
  }
  return false;
}

MAKE_MODULE(Logger, Debugging)
