/** 
* @file CameraCalibrator.h
*
* This file implements a module that can provide a semiautomatic camera calibration.
*
* @author <a href="mailto:fynn@tzi.de">Fynn Feldpausch</a>
*/

#include "CameraCalibrator.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Streams/InStreams.h"

CameraCalibrator::CameraCalibrator() : calibrationState(Idle), lastFetchedPoint(-1, -1)
{
  InConfigFile streamDef(Global::getSettings().expandLocationFilename("cameraCalibrator.cfg"), "default");
  if(streamDef.exists())
    streamDef >> config;
  else
  {
    InConfigFile streamDef("cameraCalibrator.cfg", "default");
    ASSERT(streamDef.exists());
    streamDef >> config;
  }

  InConfigFile streamCor(Global::getSettings().expandLocationFilename("cameraCalibrator.cfg"), "corners");
  if(streamCor.exists())
    streamCor >> config;
  else
  {
    InConfigFile streamCor("cameraCalibrator.cfg", "corners");
    ASSERT(streamCor.exists());
    while(!streamCor.eof())
    {
      float corner;
      streamCor >> corner;
      config.corners.push_back(corner);
    }
  }

  ASSERT(config.corners.size()%config.cornerDataSize == 0);
  config.cornersSize = config.corners.size() / config.cornerDataSize;
}

void CameraCalibrator::update(HeadMotionRequest& headMotionRequest)
{
  headMotionRequest.speed = 1.5f;
  headMotionRequest.mode = HeadMotionRequest::panTiltMode;

  if(calibrationState == Setup)
  {
    headMotionRequest.pan = config.corners[setupState*config.cornerDataSize];
    headMotionRequest.tilt = config.corners[setupState*config.cornerDataSize + 1];
  }
  else
  {
    headMotionRequest.pan = 0.0f;
    headMotionRequest.tilt = 0.5f;
  }
}

void CameraCalibrator::update(CameraCalibration& cameraCalibration)
{ 
  Vector2<int> point(-1, -1);
  MODIFY("module:CameraCalibrator:point", point);

  DEBUG_RESPONSE("module:CameraCalibrator:setup",
  {
    calibrationState = Setup;
    setupState = 0;
    setupStateReference = 0;
    lastFetchedPoint = Vector2<int>(-1, -1);
    resetCorners();

    meanError = 0.0f;
    optimizer.uninitialize();
  });

  DEBUG_RESPONSE("module:CameraCalibrator:stop", calibrationState = Idle;);

  DEBUG_RESPONSE("module:CameraCalibrator:error", OUTPUT(idText, text, calculateError(cameraCalibration)););

  DEBUG_RESPONSE("module:CameraCalibrator:reset", {
    cameraCalibration.cameraTiltCorrection = 0.0f;
    cameraCalibration.cameraRollCorrection = 0.0f;
    cameraCalibration.cameraPanCorrection = 0.0f;
    cameraCalibration.bodyTiltCorrection = 0.0f;
    cameraCalibration.bodyRollCorrection = 0.0f;
    cameraCalibration.bodyTranslationCorrection = Vector3<>(0.0f, 0.0f, 0.0f);
  });

  switch(calibrationState)
  {
    case(Idle):
      DEBUG_RESPONSE("module:CameraCalibrator:start",
      {
        meanError = 0.0f;
        calibrationState = Accumulate;
      });
      break;
    case(Setup):
      fetchPoint();
      break;
    case(Accumulate): // calculate the mean error...
      if(accumulatedFrames == 0) //accumulation started
      {
        meanError = 0.0f;
      }
      meanError += calculateError(cameraCalibration); // accumulating
      if(++accumulatedFrames >= config.maxAccumulatedFrames) // accumulation finished
      {
        meanError /= accumulatedFrames;
        accumulatedFrames = 0;
        calibrationState = Optimize;
      }
      break;
    case(Optimize): // optimize the parameters of the cameraCalibration...
      if(!optimizer.isInitialized())
      { 
        resetOptimizer(cameraCalibration);
      }
      if(optimizer.hasReachedMinimum()) // converged
      {
        cameraCalibration.cameraTiltCorrection = parameters[0];
        cameraCalibration.cameraRollCorrection = parameters[1];
        cameraCalibration.cameraPanCorrection = parameters[2];
        cameraCalibration.bodyTiltCorrection = parameters[3];
        cameraCalibration.bodyRollCorrection = parameters[4];
        OUTPUT(idText, text, "optimization converged");
        calibrationState = Idle;
      }
      else 
      {
        optimizer.setReward(meanError);
        optimizer.getNextParameterSet(parameters);
        cameraCalibration.cameraTiltCorrection = parameters[0];
        cameraCalibration.cameraRollCorrection = parameters[1];
        cameraCalibration.cameraPanCorrection = parameters[2];
        cameraCalibration.bodyTiltCorrection = parameters[3];
        cameraCalibration.bodyRollCorrection = parameters[4];
        if(optimizer.wasNewHillDuringLastExecution())
        {
          OUTPUT(idText, text, "meanError: " << meanError <<
            " (camTilt: " << parameters[0] <<
            ", camRoll: " << parameters[1] <<
            ", camPan: " << parameters[2] <<
            ", bodyTilt: " << parameters[3] <<
            ", bodyRoll: " << parameters[4] << ")");
        }
        calibrationState = Accumulate;
      }
      break;
    default:
      ASSERT(false);
  }

  DECLARE_DEBUG_DRAWING("module:CameraCalibrator:corners", "drawingOnImage");
  COMPLEX_DRAWING("module:CameraCalibrator:corners", drawCorners(cameraCalibration););
}

void CameraCalibrator::drawCorners(CameraCalibration& cameraCalibration)
{
  // don't calculate points during accumulation or optimization state
  if(calibrationState == Accumulate || calibrationState == Optimize)
    return;

  for(int i = 0; i < fieldCornersSize; ++i)
  {
    const Vector3<> fieldCorner(fieldCorners[i].x, fieldCorners[i].y, 0);

    Vector2<int> fieldCornerImage;
    Geometry::calculatePointInImage(fieldCorner, theCameraMatrix, theCameraInfo, fieldCornerImage);

    if(calibrationState == Setup && i == setupStateReference)
      MID_DOT("module:CameraCalibrator:corners", fieldCornerImage.x, fieldCornerImage.y, ColorClasses::red, ColorClasses::red);
    else
      DOT("module:CameraCalibrator:corners", fieldCornerImage.x, fieldCornerImage.y, ColorClasses::red, ColorClasses::red);
  }

  for(unsigned int i = 0; i < refCorners.size(); ++i)
  {
    RobotCameraMatrix robotCameraMatrix = RobotCameraMatrix();
    robotCameraMatrix.computeRobotCameraMatrix(theRobotDimensions, refCorners[i].headYaw, refCorners[i].headPitch, cameraCalibration, !theImageInfo.fromLowerCamera());

    CameraMatrix cameraMatrix = CameraMatrix();
    cameraMatrix.computeCameraMatrix(theTorsoMatrix, robotCameraMatrix, cameraCalibration);

    Vector2<> refCorner;
    Geometry::calculatePointOnField(refCorners[i].point.x, refCorners[i].point.y, cameraMatrix, theCameraInfo, refCorner);
    const Vector3<> refCornerWorld(refCorner.x, refCorner.y, 0.0f);

    Vector2<int> refCornerImg;
    Geometry::calculatePointInImage(refCornerWorld, theCameraMatrix, theCameraInfo, refCornerImg);
    DOT("module:CameraCalibrator:corners", refCornerImg.x, refCornerImg.y, ColorClasses::green, ColorClasses::green);
  }
}

void CameraCalibrator::resetCorners()
{
  // reset already defined reference corners
  refCorners.clear();

  // define field corners (clockwise, starting at the midline)
  Vector2<int> fieldCornersAbs[fieldCornersSize];
  fieldCornersAbs[0] = theFieldDimensions.corners[FieldDimensions::xCorner][0];
  fieldCornersAbs[1] = theFieldDimensions.corners[FieldDimensions::tCorner270][2];
  fieldCornersAbs[2] = theFieldDimensions.corners[FieldDimensions::lCorner180][5];
  fieldCornersAbs[3] = theFieldDimensions.corners[FieldDimensions::tCorner180][2];
  fieldCornersAbs[4] = theFieldDimensions.corners[FieldDimensions::lCorner270][6];
  fieldCornersAbs[5] = theFieldDimensions.corners[FieldDimensions::lCorner0][6];
  fieldCornersAbs[6] = theFieldDimensions.corners[FieldDimensions::tCorner180][3];
  fieldCornersAbs[7] = theFieldDimensions.corners[FieldDimensions::lCorner90][5];
  fieldCornersAbs[8] = theFieldDimensions.corners[FieldDimensions::tCorner90][2];
  fieldCornersAbs[9] = theFieldDimensions.corners[FieldDimensions::xCorner][1];

  // calculate field corners relative to the robot
  for(int i = 0; i < fieldCornersSize; ++i)
  {
    const Vector2<> fieldCorner(static_cast<float>(fieldCornersAbs[i].x), static_cast<float>(fieldCornersAbs[i].y));
    fieldCorners[i] = Geometry::fieldCoord2Relative(theRobotPose, fieldCorner);
  }
}

void CameraCalibrator::resetOptimizer(CameraCalibration &cameraCalibration)
{
  optimizer.uninitialize();

  float startVals[] =
  {
    cameraCalibration.cameraTiltCorrection,
    cameraCalibration.cameraRollCorrection,
    cameraCalibration.cameraPanCorrection,
    cameraCalibration.bodyTiltCorrection,
    cameraCalibration.bodyRollCorrection
  };
  float minVals[] =
  {
    -0.15f,
    -0.2f,
    -0.1f,
    -0.15f,
    -0.15f
  };
  float maxVals[] =
  {
    -minVals[0],
    -minVals[1],
    -minVals[2],
    -minVals[3],
    -minVals[4]
  };
  float maxStepSizes[] =
  {
    maxVals[0]/4.0f,
    maxVals[1]/4.0f,
    maxVals[2]/4.0f,
    maxVals[3]/4.0f,
    maxVals[4]/4.0f
  };
  float worstReward = 750000.0f;
  HillClimbing<float, float, 5>::Mode mode = HillClimbing<float, float, 5>::MINIMIZE;
  int numOfRuns = 4;
  int maxUnchangedSteps = 16;

  optimizer.init(startVals, minVals, maxVals, maxStepSizes, worstReward, mode, numOfRuns, maxUnchangedSteps);
}

float CameraCalibrator::calculateError(CameraCalibration &cameraCalibration)
{
  float error = 0.0f;

  for(unsigned int i = 0; i < refCorners.size(); ++i)
  {
    RobotCameraMatrix robotCameraMatrix = RobotCameraMatrix();
    robotCameraMatrix.computeRobotCameraMatrix(theRobotDimensions, refCorners[i].headYaw, refCorners[i].headPitch, cameraCalibration, !theImageInfo.fromLowerCamera());

    CameraMatrix cameraMatrix = CameraMatrix();
    cameraMatrix.computeCameraMatrix(theTorsoMatrix, robotCameraMatrix, cameraCalibration);

    Vector2<> refCorner;
    Geometry::calculatePointOnField(refCorners[i].point.x, refCorners[i].point.y, cameraMatrix, theCameraInfo, refCorner);
    error += Geometry::distance(fieldCorners[refCorners[i].reference], refCorner);
  }

  return error;
}

void CameraCalibrator::fetchPoint()
{
  Vector2<int> point(-1, -1);
  MODIFY("module:CameraCalibrator:point", point);

  if(point.x == lastFetchedPoint.x && point.y == lastFetchedPoint.y)
    return;

  lastFetchedPoint = point;

  struct RefCorner refCorner;
  refCorner.point = point;
  refCorner.reference = setupStateReference;
  refCorner.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
  refCorner.headPitch = theFilteredJointData.angles[JointData::HeadPitch];

  refCorners.push_back(refCorner);

  do {
    ++setupStateReference;
    // next setup state
    if(setupStateReference >= fieldCornersSize) {
      setupStateReference = 0;
      setupState++;
    }
    // setup is complete
    if(setupState >= config.cornersSize) {
      calibrationState = Idle;
      break;
    }
  }
  while (!bitCheck(static_cast<int>(config.corners[config.cornerDataSize*setupState+2]), setupStateReference));
}

MAKE_MODULE(CameraCalibrator, Infrastructure)
