/** 
* @file CameraCalibrator.h
*
* This file implements a module that can provide a semiautomatic camera calibration.
*
* @author <a href="mailto:fynn@tzi.de">Fynn Feldpausch</a>
*/

#ifndef __CameraCalibrator_h_
#define __CameraCalibrator_h_

#include "Tools/Module/Module.h"
#include "Tools/Optimization/HillClimbing.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Configuration/CameraCalibration.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/MotionControl/HeadMotionRequest.h"
#include "Representations/Infrastructure/JointData.h"
#include "Representations/Sensing/TorsoMatrix.h"
#include "Representations/Perception/ImageInfo.h"

MODULE(CameraCalibrator)
  REQUIRES(FieldDimensions)
  REQUIRES(FilteredJointData)
  REQUIRES(RobotDimensions)
  REQUIRES(TorsoMatrix)
  REQUIRES(ImageInfo)
  USES(RobotPose)
  USES(CameraMatrix)
  USES(CameraInfo)
  PROVIDES(HeadMotionRequest)
  PROVIDES_WITH_MODIFY(CameraCalibration)
END_MODULE

class CameraCalibrator : public CameraCalibratorBase
{
private:
  class Configuration : public Streamable
  {
  private:
    void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM(maxAccumulatedFrames);
      STREAM_VECTOR(corners);
      STREAM_REGISTER_FINISH();
    }
  public:
    int maxAccumulatedFrames;            /**< The max. number of frames to be accumulated for mean error calculation. */
    static const int cornerDataSize = 3; /**< The size of one viewpoint specification.*/
    int cornersSize;                     /**< The number of different head positions. */
    std::vector<float> corners;          /**< The visible corners and the corresponding head positions. */
  };

  Configuration config;

  int setupState;
  int setupStateReference;
  enum CalibrationState
  {
    Idle = 0,
    Setup,
    Accumulate,
    Optimize
  } calibrationState;

  struct RefCorner {
    Vector2<int> point;
    int reference;
    float headYaw, headPitch;
  };

  Vector2<int> lastFetchedPoint;            /**< The last point fetched by the CameraCalibratorHandler. */
  std::vector<RefCorner> refCorners;        /**< The already defined reference corners (in image coordinates). */
  static const int fieldCornersSize = 10;   /**< The number of field corners considered during optimization. */
  Vector2<> fieldCorners[fieldCornersSize]; /**< The field corners (relative to the robot). */

  HillClimbing<float, float, 5> optimizer;  /**< The hill climbing optimizer. */
  int accumulatedFrames; /**< The number of frames accumulated during evaluation time. */
  float meanError;       /**< The mean error of one accumulation cycle. */
  float parameters[5];   /**< The camera parameters to be optimized. */

  /**
  * Checks whether a single bit of an integer value is set or not.
  * @param x The integer value to be checked.
  * @param pos The position of the bit.
  */
  inline bool bitCheck(int x, unsigned int pos) {
    return (x & (1 << pos)) != 0;
  }

  /**
  * The method to provide proper head motions for the calibration setup.
  * @param headMotionRequest The requested head motion of the nao.
  */
  void update(HeadMotionRequest& headMotionRequest);

  /**
  * The method to calculate the new camera calibration, depending on the state of the calibrator.
  * @param cameraCalibration The current calibration of the robot's camera.
  */
  void update(CameraCalibration& cameraCalibration);

  /**
  * This method draws debug drawings of the corners into the image view.
  * @param cameraCalibration The current calibration of the robot's camera.
  */
  void drawCorners(CameraCalibration& cameraCalibration);

  /**
  * The method to reset the reference corners and calculate all field corners in relative field coordinates.
  */
  void resetCorners();

  /**
  * This method resets the camera calibrator, i.e. deletes all collected reference corners and resets the optimizer.
  * @param cameraCalibration The current calibration of the robot's camera.
  */
  void resetOptimizer(CameraCalibration &cameraCalibration);

  /**
  * The method to calculate the error of the specified reference corners.
  * @param cameraCalibration The current calibration of the robot's camera.
  */
  float calculateError(CameraCalibration &cameraCalibration);

  /**
  * The method to fetch a point from a click in the image view.
  */
  void fetchPoint();

public:
  /** Default constructor. */
  CameraCalibrator();
};

#endif // __CameraCalibrator_h_
