/** 
* @file CameraProvider.cpp
* This file declares a module that provides camera images.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "CameraProvider.h"
#include "Representations/Perception/JPEGImage.h"
#include "Platform/Camera.h"
#include "Platform/SystemCall.h"

PROCESS_WIDE_STORAGE CameraProvider* CameraProvider::theInstance = 0;

CameraProvider::CameraProvider()
#ifdef CAMERA_INCLUDED
  : camera(new NaoCamera())
#endif
{
  theInstance = this;
}

CameraProvider::~CameraProvider()
{
#ifdef CAMERA_INCLUDED
  delete camera;
#endif
  theInstance = 0;
}

void CameraProvider::update(Image& image)
{
#ifdef CAMERA_INCLUDED
  ASSERT(camera->getImage());
  image.setImage(const_cast<unsigned char*>(camera->getImage()));
  image.timeStamp = camera->getTimeStamp();
  camera->setSettings(theCameraSettings);
  DEBUG_RESPONSE("module:CameraProvider:switchToLowerCamera",
                 {
                   unsigned int startt = SystemCall::getCurrentSystemTime();
                   camera->switchToLower();
                   unsigned int deltat = SystemCall::getCurrentSystemTime() - startt;
                   OUTPUT(idText, text, "Switching to lower camera took " << deltat << "ms");
                 });
  DEBUG_RESPONSE("module:CameraProvider:switchToUpperCamera",
                 {
                   unsigned int startt = SystemCall::getCurrentSystemTime();
                   camera->switchToUpper();
                   unsigned int deltat = SystemCall::getCurrentSystemTime() - startt;
                   OUTPUT(idText, text, "Switching to upper camera took " << deltat << "ms");
                 });
  DEBUG_RESPONSE("module:CameraProvider:assertCameraSettings", camera->assertCameraSettings(); );
  DEBUG_RESPONSE("module:CameraProvider:writeCameraSettings", camera->writeCameraSettings(); );
#endif
  DEBUG_RESPONSE("representation:JPEGImage", OUTPUT(idJPEGImage, bin, JPEGImage(image)); );
}

void CameraProvider::update(ImageInfo& imageInfo)
{
#ifdef CAMERA_INCLUDED
  imageInfo.prevCamera = imageInfo.camera;
  imageInfo.camera = camera->getCurrentCamera();
#endif
  if(imageInfo.prevCamera != imageInfo.camera)
  {
    OUTPUT(idText, text, "Camera provider: Camera changed.");
  }
  return;
}

void CameraProvider::update(FrameInfo& frameInfo)
{
  frameInfo.time = theImage.timeStamp;
}

void CameraProvider::update(CognitionFrameInfo& cognitionFrameInfo)
{
  cognitionFrameInfo.time = theImage.timeStamp;
}

bool CameraProvider::isFrameDataComplete()
{
#ifdef CAMERA_INCLUDED
  if(theInstance)
    return theInstance->camera->getImage() ? true : false;
  else 
#endif
    return true;
}

void CameraProvider::waitForFrameData()
{
#ifdef CAMERA_INCLUDED
  if(theInstance)
    theInstance->camera->captureNew();
#endif
}

MAKE_MODULE(CameraProvider, Infrastructure)
