/**
* @file Modules/Infrastructure/RobotHealthProvider.h
* This file implements a module that provides information about the robot's health.
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#include "RobotHealthProvider.h"
#include "Tools/Settings.h"
#include "Tools/Team.h"
#include "Tools/Debugging/ReleaseOptions.h"
#include "Platform/SoundPlayer.h"

RobotHealthProvider::RobotHealthProvider() :
  startBatteryLow(0),
  lastBatteryLevel(1),
  batteryVoltageFalling(false),
  highTemperatureSince(0),
  lastBodyTemperatureReadTime(0)
{
  p.batteryLow = 0.05f;
  p.temperatureHigh = 79;

  InConfigFile stream("health.cfg");
  if(stream.exists())
    stream >> p;
}

void RobotHealthProvider::update(RobotHealth& robotHealth)
{
  // Transfer information from other process:
  robotHealth = theMotionRobotHealth;

  // Directly transfer temperature and batteryLevel data from SensorData:
  robotHealth.batteryLevel = (unsigned char)((theFilteredSensorData.data[SensorData::batteryLevel] == SensorData::off ? 1.f : theFilteredSensorData.data[SensorData::batteryLevel]) * 100.f);
  unsigned char maxTemperature(0);
  for(int i=0; i<JointData::numOfJoints; ++i)
    if(theFilteredSensorData.temperatures[i] > maxTemperature)
      maxTemperature = theFilteredSensorData.temperatures[i];
  robotHealth.maxJointTemperature = maxTemperature;

  // Add cpu load, memory load and robot name:
  float memoryUsage, load[3];
  SystemCall::getLoad(memoryUsage, load);
  robotHealth.load[0] = (unsigned char)(load[0] * 10.f);
  robotHealth.load[1] = (unsigned char)(load[1] * 10.f);
  robotHealth.load[2] = (unsigned char)(load[2] * 10.f);
  robotHealth.memoryUsage = (unsigned char)(memoryUsage * 100.f);
  robotHealth.robotName = Global::getSettings().robot;
  
  // Compute frame rate of cognition process:
  timeBuffer.add(SystemCall::getCurrentSystemTime());
  unsigned timeDiffSum(0);
  int numOfTimeDiffs(0);
  for(int i=0; i < timeBuffer.getNumberOfEntries() - 1; ++i)
  {
    timeDiffSum += (timeBuffer[i] - timeBuffer[i+1]);
    ++numOfTimeDiffs;
  }
  robotHealth.cognitionFrameRate = timeDiffSum ? 1000.0F / (static_cast<float>(timeDiffSum)/numOfTimeDiffs) : 0.0F;

  // read cpu and mainboard temperature
#ifdef TARGET_ROBOT
  if(theFrameInfo.getTimeSince(lastBodyTemperatureReadTime) > 10 * 1000)
  {
    lastBodyTemperatureReadTime = theFrameInfo.time;
    float cpuTemperature, mbTemperature;
    naoBody.getTemperature(cpuTemperature, mbTemperature);
    robotHealth.cpuTemperature = (unsigned char)cpuTemperature;
    robotHealth.boardTemperature = (unsigned char)mbTemperature;
  }
#endif

  // Communicate health:
  if(Global::getReleaseOptions().robotHealth)
  {
    TEAM_OUTPUT(idRobotHealth, bin, robotHealth);
  }

  if(theFilteredSensorData.timeStamp)
  {
    //battery warning
    if(lastBatteryLevel < robotHealth.batteryLevel)
      batteryVoltageFalling = false;
    else if(lastBatteryLevel > robotHealth.batteryLevel)
      batteryVoltageFalling = true;
    if(robotHealth.batteryLevel < p.batteryLow)
    {
      if(batteryVoltageFalling && theFrameInfo.getTimeSince(startBatteryLow) > 1000)
      {
        SoundPlayer::play("lowbattery.wav");
        //next warning in 90 seconds
        //startBatteryLow = theFrameInfo.time + 90000;
        batteryVoltageFalling = false;
      }
    }
    else if(startBatteryLow < theFrameInfo.time)
      startBatteryLow = theFrameInfo.time;
    lastBatteryLevel = robotHealth.batteryLevel;

    //temperature warning
    if(maxTemperature > p.temperatureHigh)
    {
      if(theFrameInfo.getTimeSince(highTemperatureSince) > 1000)
      {
        SoundPlayer::play("heat.wav");
        highTemperatureSince = theFrameInfo.time + 20000;
      }
    }
    else if(highTemperatureSince < theFrameInfo.time)
      highTemperatureSince = theFrameInfo.time;
  }
}

MAKE_MODULE(RobotHealthProvider, Infrastructure)
