/**
* @file Modules/Infrastructure/RobotHealthProvider.h
* This file declares a module that provides information about the robot's health.
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef RobotHealthProvider_H
#define RobotHealthProvider_H

#include "Tools/Module/Module.h"
#include "Tools/RingBuffer.h"
#include "Representations/Infrastructure/RobotHealth.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Infrastructure/FrameInfo.h"

MODULE(RobotHealthProvider)
  REQUIRES(MotionRobotHealth)
  REQUIRES(FilteredSensorData)
  REQUIRES(FrameInfo)
  PROVIDES_WITH_MODIFY(RobotHealth)
END_MODULE

/**
* @class RobotHealthProvider
* A module that provides information about the robot's health
*/
class RobotHealthProvider : public RobotHealthProviderBase
{
public:
  /** Constructor. */
  RobotHealthProvider();

private:
  /**
  * A collection of parameters for the robot health provider.
  */
  class Parameters : public Streamable
  {
  public:
    /** Default constructor. */
    Parameters() {}

    float batteryLow; /**< The voltage below which the robot is switched off. */
    int temperatureHigh; /**< The temperature the robot starts complaining about the temperature. */

  private:
    /**
    * The method makes the object streamable.
    * @param in The stream from which the object is read.
    * @param out The stream to which the object is written. 
    */
    virtual void serialize(In* in, Out* out)
    {  
      STREAM_REGISTER_BEGIN();
        STREAM(batteryLow);
        STREAM(temperatureHigh);
      STREAM_REGISTER_FINISH();
    }
  };

  Parameters p; /**< Parameters for the module. */
  RingBuffer<unsigned, 30> timeBuffer; /**< Buffered timestamps of previous executions */
  unsigned startBatteryLow; /**< Last time the battery state was not low. */
  float lastBatteryLevel;
  bool batteryVoltageFalling;
  unsigned highTemperatureSince;
#ifdef TARGET_ROBOT
  NaoBody naoBody;
#endif
  unsigned int lastBodyTemperatureReadTime;

  /** The main function, called every cycle
  * @param robotHealth The data struct to be filled
  */
  void update(RobotHealth& robotHealth);
};

#endif // RobotHealthProvider_H
