/**
* @file AnotherObstacleModelProvider.cpp
* Implementation of a module for modelling obstacles from the ultrasonic measurements of the Nao.
* @author Colin Graf
*/

#include "AnotherObstacleModelProvider.h"
#include "Tools/Math/Matrix2x2.h"

MAKE_MODULE(AnotherObstacleModelProvider, Modeling)

AnotherObstacleModelProvider::AnotherObstacleModelProvider() : lastUsTimeStamp(0)
{
  p.sensorPosition = Vector2<>((53.7f + 47.7f) * 0.5f, (34.1f + 41.6f) * 0.5f);
  p.maxValidDistance = 1600;
  p.minValidDistance = 200;
  p.maxAngle = fromDegrees(70);
  p.minAngle = fromDegrees(15);
  p.measurementWidth = 300;

  p.obstacleMinDistance = 400;
  p.obstacleTimeout = 1500;
  p.obstacleMinUpdates = 4;
  p.obstacleShrinking = 2.f;

  obstacles.reserve(10);
}

void AnotherObstacleModelProvider::update(AnotherObstacleModel& obstacleModel)
{
  MODIFY("module:AnotherObstacleModelProvider:parameters", p);

  DECLARE_DEBUG_DRAWING("module:AnotherObstacleModelProvider", "drawingOnField");

  motionUpdate();

  if(theFilteredSensorData.usTimeStamp != lastUsTimeStamp && theGroundContactState.contact && theFallDownState.state == FallDownState::upright)
  {
    if(theFilteredSensorData.usActuatorMode == SensorData::leftToLeft || theFilteredSensorData.usActuatorMode == SensorData::rightToRight)
    {
      if(theFilteredSensorData.usActuatorMode == lastActuatorMode)
      {
        leftToLeft = std::min(leftToLeft, theFilteredSensorData.data[FilteredSensorData::usL]);
        rightToRight = std::min(rightToRight, theFilteredSensorData.data[FilteredSensorData::usR]);
      }
      else
      {
        leftToLeft = theFilteredSensorData.data[FilteredSensorData::usL];
        rightToRight = theFilteredSensorData.data[FilteredSensorData::usR];
      }
    }
    if(theFilteredSensorData.usActuatorMode == SensorData::leftToRight || theFilteredSensorData.usActuatorMode == SensorData::rightToLeft)
    {
      if(theFilteredSensorData.usActuatorMode == lastActuatorMode)
      {
        rightToLeft = std::min(rightToLeft, theFilteredSensorData.data[FilteredSensorData::usL]);
        leftToRight = std::min(leftToRight, theFilteredSensorData.data[FilteredSensorData::usR]);
      }
      else
      {
        rightToLeft = theFilteredSensorData.data[FilteredSensorData::usL];
        leftToRight = theFilteredSensorData.data[FilteredSensorData::usR];
      }
    }

    {
      float sensorDistance = p.sensorPosition.y * 2.f;
      float leftToRight = this->leftToRight * 2.f - leftToLeft;
      float rightToLeft = this->rightToLeft * 2.f - rightToRight;

      if(leftToLeft < p.maxValidDistance && leftToLeft > p.minValidDistance)
      {
        float gamma = p.maxAngle;
        if(leftToRight < p.maxValidDistance && leftToRight > p.minValidDistance)
        {
          float sencos = (sqr(leftToLeft) + sqr(sensorDistance) - sqr(leftToRight)) / (2.f * leftToLeft * sensorDistance);
          gamma = (sencos <= -1.f ? pi : (sencos >= 1.f ? 0.f : acos(sencos))) - pi_2;
          if(gamma > p.maxAngle)
            gamma = p.maxAngle;
          if(gamma < p.minAngle)
            gamma = p.minAngle;
        }
        Vector2<> pos(leftToLeft, 0.f);
        pos.rotate(gamma);
        pos += p.sensorPosition;
        insertMeasurement(pos, theFilteredSensorData.usTimeStamp);
      }

      if(rightToRight < p.maxValidDistance && rightToRight > p.minValidDistance)
      {
        float gamma = p.maxAngle;
        if(rightToLeft < p.maxValidDistance && rightToLeft > p.minValidDistance)
        {
          float sencos = (sqr(rightToRight) + sqr(sensorDistance) - sqr(rightToLeft)) / (2.f * rightToRight * sensorDistance);
          gamma = (sencos <= -1.f ? pi : (sencos >= 1.f ? 0.f : acos(sencos))) - pi_2;
          if(gamma > p.maxAngle)
            gamma = p.maxAngle;
          if(gamma < p.minAngle)
            gamma = p.minAngle;
        }
        Vector2<> pos(rightToRight, 0.f);
        pos.rotate(-gamma);
        pos += Vector2<>(p.sensorPosition.x, -p.sensorPosition.y);
        insertMeasurement(pos, theFilteredSensorData.usTimeStamp);
      }
    }

    lastUsTimeStamp = theFilteredSensorData.usTimeStamp;
    lastActuatorMode = theFilteredSensorData.usActuatorMode;
  }

  generateModel(obstacleModel);

  lastOdometryData = theOdometryData;
}

void AnotherObstacleModelProvider::motionUpdate()
{
  Pose2D odometryOffset = theOdometryData - lastOdometryData;
  float odometryCos = cos(odometryOffset.rotation);
  float odometrySin = sin(odometryOffset.rotation);

  Matrix2x2<> odometryRotation(Vector2<>(odometryCos, -odometrySin), Vector2<>(odometrySin, odometryCos));
  Vector2<> odometryTranslation(-odometryOffset.translation.x, -odometryOffset.translation.y);

  for(std::vector<Obstacle>::iterator it = obstacles.begin(), end = obstacles.end(); it != end; ++it)
  {
    Obstacle& obstacle = *it;
    if(obstacle.persistent)
    {
      // add odometry
      obstacle.position = odometryRotation * obstacle.position + odometryTranslation;
      obstacle.dir = odometryRotation * obstacle.dir;

      // shrinking...
      obstacle.width -= p.obstacleShrinking;
      if(obstacle.width < p.measurementWidth)
        obstacle.width = p.measurementWidth;
    }
  }
}

void AnotherObstacleModelProvider::insertMeasurement(const Vector2<>& measurement, unsigned int timeStamp)
{
  COMPLEX_DRAWING("module:AnotherObstacleModelProvider",
  {
    Vector2<> dir = measurement;
    dir.rotateLeft();
    dir.normalize(p.measurementWidth * 0.5f);
    Vector2<> a = measurement + dir, b = measurement - dir;
    LINE("module:AnotherObstacleModelProvider", a.x, a.y, b.x, b.y, 30, Drawings::ps_solid, ColorRGBA(0xff, 0, 0));
  });

  // create a new obstacle
  Obstacle newObstacle(measurement, p.measurementWidth, timeStamp);

  // try to add the new obstacle to an existing obstacle
  float sqrDistance;
  unsigned int persistentObstacles = 0;
  for(std::vector<Obstacle>::iterator it = obstacles.begin(), end = obstacles.end(); it != end; ++it)
  {
    Obstacle& obstacle = *it;
    if(obstacle.persistent)
    {
      if(theFrameInfo.getTimeSince(obstacle.lastUpdateTime) < p.obstacleTimeout || obstacle.position.squareAbs() < sqr(p.obstacleMinDistance))
      {
        ++persistentObstacles;
        if(obstacle.getSqrDistanceTo(measurement, sqrDistance))
          if(sqrDistance < sqr(p.measurementWidth))
          {
            obstacle.add(newObstacle);
            return;
          }
      }
      else
        obstacle.persistent = false;
    }
  }
  if(persistentObstacles == 0)
    obstacles.resize(0);

  // add the newly created obstacle
  if(persistentObstacles < (unsigned int)obstacles.size())
    for(std::vector<Obstacle>::iterator it = obstacles.begin(), end = obstacles.end(); it != end; ++it)
    {
      Obstacle& obstacle = *it;
      if(!obstacle.persistent)
      {
        obstacle = newObstacle;
        return;
      }
    }
  obstacles.push_back(newObstacle);
}

void AnotherObstacleModelProvider::generateModel(AnotherObstacleModel& obstacleModel) const
{
  obstacleModel.obstacles.resize(0);
  for(std::vector<Obstacle>::const_iterator it = obstacles.begin(), end = obstacles.end(); it != end; ++it)
  {
    const Obstacle& obstacle = *it;
    if(obstacle.persistent && (theFrameInfo.getTimeSince(obstacle.lastUpdateTime) < p.obstacleTimeout || obstacle.position.squareAbs() < sqr(p.obstacleMinDistance)))
      if(obstacle.updates >= p.obstacleMinUpdates)
      {
        Vector2<> dir(obstacle.position.y, -obstacle.position.x);
        dir.normalize(fabs(obstacle.width) * 0.5f);
        obstacleModel.obstacles.push_back(AnotherObstacleModel::Obstacle(obstacle.position - dir, obstacle.position + dir, obstacle.position, obstacle.position));
      }
  }
}

bool AnotherObstacleModelProvider::Obstacle::getSqrDistanceTo(const Vector2<>& measurement, float& sqrDistance) const
{
  float positionToProjectionDistance = (measurement.x - position.x) * dir.x + (measurement.y - position.y) * dir.y;
  if(fabs(positionToProjectionDistance) > width * 0.5)
    return false;
  Vector2<> projection = position + dir * positionToProjectionDistance;
  sqrDistance = (projection - position).squareAbs();
  return true;
}

void AnotherObstacleModelProvider::Obstacle::add(const Obstacle& obstacle)
{
  //ASSERT(obstacle.lastUpdateTime >= lastUpdateTime);
  Range<> range(width * -0.5f, width * 0.5f);
  float positionToProjectionDistance = (obstacle.position.x - position.x) * dir.x + (obstacle.position.y - position.y) * dir.y;
  range.add(positionToProjectionDistance + obstacle.width * 0.5f);
  range.add(positionToProjectionDistance + obstacle.width * -0.5f);
  position += dir * (range.max + range.min) * 0.5f;
  position.normalize(obstacle.position.abs());
  dir = Vector2<>(position.y, -position.x).normalize();
  width = range.max - range.min;
  updates += obstacle.updates;
  lastUpdateTime = obstacle.lastUpdateTime;
}
