/**
* @file AnotherObstacleModelProvider.h
* Declaration of the AnotherObstacleModelProvider module.
* @author Colin Graf
*/

#ifndef AnotherObstacleModelProvider_H
#define AnotherObstacleModelProvider_H

#include "Tools/Module/Module.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Modeling/AnotherObstacleModel.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Modeling/FallDownState.h"
#include "Representations/Sensing/GroundContactState.h"

MODULE(AnotherObstacleModelProvider)
  REQUIRES(OdometryData)
  REQUIRES(FilteredSensorData)
  REQUIRES(FrameInfo)
  REQUIRES(FallDownState)
  REQUIRES(GroundContactState)
  PROVIDES_WITH_MODIFY_AND_DRAW(AnotherObstacleModel)
END_MODULE

/**
* @class AnotherObstacleModelProvider
* A module for modelling obstacles from the ultrasonic measurements of the Nao.
*/
class AnotherObstacleModelProvider : public AnotherObstacleModelProviderBase
{
public:
  /** Default constructor */
  AnotherObstacleModelProvider();

private:
  /**
  * The parameters of the module
  */
  class Parameters: public Streamable
  {

  public:
    Vector2<> sensorPosition;
    float maxValidDistance;
    float minValidDistance;
    float maxAngle;
    float minAngle;
    float measurementWidth;

    float obstacleMinDistance;
    int obstacleTimeout;
    int obstacleMinUpdates;
    float obstacleShrinking;

    /** Constructor */
    Parameters() {}

  private:
    /**
    * Makes the object streamable
    * @param in The stream from which the object is read 
    * @param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM(sensorPosition);
      STREAM(maxValidDistance);
      STREAM(minValidDistance);
      STREAM(maxAngle);
      STREAM(minAngle);
      STREAM(measurementWidth);
      STREAM(obstacleMinDistance);
      STREAM(obstacleTimeout);
      STREAM(obstacleMinUpdates);
      STREAM(obstacleShrinking);
      STREAM_REGISTER_FINISH();
    }
  };

  class Obstacle
  {
  public:
    Vector2<> position;
    Vector2<> dir;
    float width;
    int updates;
    unsigned int lastUpdateTime;
    bool persistent;

    Obstacle() {}
    Obstacle(const Vector2<>& position, float width, unsigned int lastUpdateTime) : position(position), dir(Vector2<>(position.y, -position.x).normalize()), width(width), updates(1), lastUpdateTime(lastUpdateTime), persistent(true) {}

    bool getSqrDistanceTo(const Vector2<>& measurement, float& sqrDistance) const;
    void add(const Obstacle& obstacle);
  };

  Parameters p;
  std::vector<Obstacle> obstacles;

  float leftToLeft;
  float rightToRight;
  float leftToRight;
  float rightToLeft;

  unsigned int lastUsTimeStamp;
  SensorData::UsActuatorMode lastActuatorMode;

  Pose2D lastOdometryData;

  /** 
  * Executes this module
  * @param obstacleModel The data structure that is filled by this module
  */
  void update(AnotherObstacleModel& obstacleModel);

  void motionUpdate();
  void insertMeasurement(const Vector2<>& measurement, unsigned int timeStamp);
  void generateModel(AnotherObstacleModel& obstacleModel) const;
};

#endif // AnotherObstacleModelProvider_H
