/**
* @file ObstacleModelProvider.cpp
*
* This file implements a module that provides information about occupied space in the robot's environment.
* It includes parts of the implementation of the PolygonLocalMapper module of the 
* autonomous wheelchair Rolland.
*
* @author <a href="mailto:Tim.Laue@dfki.de">Tim Laue</a>
* @author <a href="mailto:cman@tzi.de">Christian Mandel</a>
* @author <a href="mailto:marko@tzi.de">Benjamin Markowsky</a>
* @author <a href="mailto:afabisch@tzi.de">Alexander Fabisch</a>
*/

#include "ObstacleModelProvider.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Geometry.h"
#include <algorithm>

ObstacleModelProvider::ObstacleModelProvider():
lastTimePenalized(0),
lastUsActuatorMode(SensorData::UsActuatorMode(-1)),
polygonsComputed(false),
lastUsTimeStamp(0)
{
  InConfigFile stream(Global::getSettings().expandLocationFilename("obstacleModel.cfg"));
  if(stream.exists())
  {
    stream >> parameters.cellFreeInterval;
    stream >> parameters.cellOccupiedThreshold;
    stream >> parameters.cellOccupiedThresholdFast;
    stream >> parameters.fastWalkSpeedThreshold;
    stream >> parameters.imageVerification;
  }
  InConfigFile stream2(Global::getSettings().expandLocationFilename("obstacleModel.cfg"), "image");
  if(stream2.exists())
  {
    stream2 >> imgParams;
  }
}

void ObstacleModelProvider::init()
{
  lastOdometry = theOdometryData;
  const int totalCells = CELLS_X*CELLS_Y;
  for(int i=0; i<totalCells; ++i)
  {
    cells[i].state = Cell::FREE;
  }
  pointExplorer.initFrame(&theImage, &theColorTable64, imgParams.exploreStepSize,
    imgParams.gridStepSize, imgParams.skipOffset, imgParams.minSegSize);
}

void ObstacleModelProvider::update(ObstacleModel& obstacleModel)
{
  MODIFY("parameters:ObstacleModelProvider", parameters);
  MODIFY("parameters:ObstacleModelProvider:imageProcessing", imgParams);
  ageCellState();
  if((theRobotInfo.penalty == PENALTY_NONE) &&
     (theMotionRequest.motion != MotionRequest::specialAction) && 
     (theFrameInfo.getTimeSince(lastTimePenalized) > 3000))
  {
    moveGrid();
    checkUS();
    STOP_TIME_ON_REQUEST("ObstacleImageVerification", 
      if(parameters.imageVerification) verifyObstaclesInImage(););
    if(imgParams.visionOn)
    {
      applyRobotsModel();
    }
  }
  else if(theRobotInfo.penalty != PENALTY_NONE)
  {
    lastTimePenalized = theFrameInfo.time;
  }
  computeModel(obstacleModel);
  draw();
}

void ObstacleModelProvider::update(AnotherObstacleModel& anotherObstacleModel)
{
  anotherObstacleModel.obstacles.clear();
  // When this code becomes executed, "normal" ObstacleModel has already been computed
  currentCluster.init();
  int currentClusterIndex(1);
  for(int y=1; y<CELLS_Y-1; ++y)
  {
    for(int x=1; x<CELLS_X-1; ++x)
    {
      Cell& c = cells[y*CELLS_Y+x];
      if(c.cluster == 0 && c.state >= parameters.cellOccupiedThreshold)
      {
        c.cluster = currentClusterIndex;
        // iterative implementation of floodfill algorithm
        cellStack.push(Vector2<int>(x,y));
        while(!cellStack.empty())
        {
          int x = cellStack.top().x;
          int y = cellStack.top().y;
          cellStack.pop();
          currentCluster.cells.push_back(Vector2<int>(x,y));
          if(x == 0 || y == 0 || x == CELLS_X-1 || y == CELLS_Y-1) //ignore border
            continue;
          // Test all eight neighbors (also stupid check for center cell)
          for(int y2=y-1; y2 <= y+1; ++y2)
          {
            for(int x2=x-1; x2 <= x+1; ++x2)
            {
              Cell& c = cells[y2*CELLS_Y+x2];
              if(c.cluster == 0 && c.state >= parameters.cellOccupiedThreshold)
              {
                c.cluster = currentClusterIndex;
                cellStack.push(Vector2<int>(x2,y2));
              }
            }
          }
        }
        ++currentClusterIndex;
        computeAnotherObstacleModelFromCurrentCluster(anotherObstacleModel);
        currentCluster.init();
      }
    }
  }
}

void ObstacleModelProvider::computeAnotherObstacleModelFromCurrentCluster(AnotherObstacleModel& anotherObstacleModel)
{
  if(currentCluster.cells.size() == 0)
    return;
  // Approximate obstacle as box (might not be cool in some cases, but anyway):
  // Compute center ("of mass")
  // Find point closest to robot
  Vector2<int>& c = currentCluster.cells[0];
  int xMin(c.x);
  int yMin(c.y);
  int xMax(c.x);
  int yMax(c.y);
  Vector2<> centerCells;
  Vector2<int> robotPosition(CELLS_X/2, CELLS_Y/2);
  Vector2<int> closestPoint(CELLS_X,CELLS_Y);
  int closestPointSqrDist(CELLS_X*CELLS_X+CELLS_Y*CELLS_Y);
  for(unsigned int i=0; i < currentCluster.cells.size(); ++i)
  {
    c = currentCluster.cells[i];
    centerCells.x += c.x;
    centerCells.y += c.y;
    int sqrDistToRobot = sqr(c.x - robotPosition.x) + sqr(c.y - robotPosition.y);
    if(sqrDistToRobot < closestPointSqrDist)
    {
      closestPoint = c;
      closestPointSqrDist = sqrDistToRobot;
    }
    if(c.x < xMin)
      xMin = c.x;
    else if(c.x > xMax)
      xMax = c.x;
    if(c.y < yMin)
      yMin = c.y;
    else if(c.y > yMax)
      yMax = c.y;
  }
  centerCells /= static_cast<float>(currentCluster.cells.size());
  Vector2<> closestPointWorld = gridToWorld(Vector2<>(closestPoint.x+0.5f, closestPoint.y+0.5f));
  Vector2<> centerWorld = gridToWorld(centerCells);
  Vector2<> topLeft(xMin+0.f,yMin+0.f);
  Vector2<> topRight(xMax+1.f,yMin+0.f);
  Vector2<> bottomLeft(xMin+0.f,yMax+1.f);
  Vector2<> bottomRight(xMax+1.f,yMax+1.f);
  Vector2<> topLeftWorld = gridToWorld(topLeft);
  Vector2<> topRightWorld = gridToWorld(topRight);
  Vector2<> bottomLeftWorld = gridToWorld(bottomLeft);
  Vector2<> bottomRightWorld = gridToWorld(bottomRight);
  LINE("module:ObstacleModelProvider:anotherObstacleModelStuff",topLeftWorld.x,topLeftWorld.y,topRightWorld.x,topRightWorld.y,
    10, Drawings::ps_solid, ColorRGBA(255,0,0));
  LINE("module:ObstacleModelProvider:anotherObstacleModelStuff",topRightWorld.x,topRightWorld.y,bottomRightWorld.x,bottomRightWorld.y,
    10, Drawings::ps_solid, ColorRGBA(255,0,0));
  LINE("module:ObstacleModelProvider:anotherObstacleModelStuff",bottomRightWorld.x,bottomRightWorld.y,bottomLeftWorld.x,bottomLeftWorld.y,
    10, Drawings::ps_solid, ColorRGBA(255,0,0));
  LINE("module:ObstacleModelProvider:anotherObstacleModelStuff",bottomLeftWorld.x,bottomLeftWorld.y,topLeftWorld.x,topLeftWorld.y,
    10, Drawings::ps_solid, ColorRGBA(255,0,0));
  float topLeftAngle = topLeftWorld.angle();
  float topRightAngle = topRightWorld.angle();
  float bottomLeftAngle = bottomLeftWorld.angle();
  float bottomRightAngle = bottomRightWorld.angle();
  Vector2<> rightCorner = topLeftWorld;
  Vector2<> leftCorner = topLeftWorld;
  float rightAngle = topLeftAngle;
  float leftAngle = topLeftAngle;
  if(topRightAngle < rightAngle)
  {
    rightAngle = topRightAngle;
    rightCorner = topRightWorld;
  }
  else if(topRightAngle > leftAngle)
  {
    leftAngle = topRightAngle;
    leftCorner = topRightWorld;
  }
  if(bottomRightAngle < rightAngle)
  {
    rightAngle = bottomRightAngle;
    rightCorner = bottomRightWorld;
  }
  else if(bottomRightAngle > leftAngle)
  {
    leftAngle = bottomRightAngle;
    leftCorner = bottomRightWorld;
  }
  if(bottomLeftAngle < rightAngle)
  {
    rightAngle = bottomLeftAngle;
    rightCorner = bottomLeftWorld;
  }
  else if(bottomLeftAngle > leftAngle)
  {
    leftAngle = bottomLeftAngle;
    leftCorner = bottomLeftWorld;
  }
  float distToClosest = closestPointWorld.abs();
  leftCorner.normalize(distToClosest);
  rightCorner.normalize(distToClosest);
  anotherObstacleModel.obstacles.push_back(AnotherObstacleModel::Obstacle(leftCorner, rightCorner, 
    centerWorld, closestPointWorld));
}

void ObstacleModelProvider::computeModel(ObstacleModel& obstacleModel)
{
  const int numOfCells(CELLS_X*CELLS_Y);
  if(!polygonsComputed)
  {
    leftPoly[0]  = Vector2<>(0,100);
    leftPoly[1]  = Vector2<>(obstacleModel.maxValidDist,0);
    leftPoly[1].rotate(obstacleModel.outerLeftAngle);
    leftPoly[1] += leftPoly[0];
    leftPoly[2]  = Vector2<>(obstacleModel.maxValidDist,0);
    leftPoly[2].rotate(obstacleModel.centerLeftAngle);
    leftPoly[2] += leftPoly[0];
    rightPoly[0] = Vector2<>(0,-100);
    rightPoly[1] = Vector2<>(obstacleModel.maxValidDist,0);
    rightPoly[1].rotate(obstacleModel.outerRightAngle);
    rightPoly[1] += rightPoly[0];
    rightPoly[2]  = Vector2<>(obstacleModel.maxValidDist,0);
    rightPoly[2].rotate(obstacleModel.centerRightAngle);
    rightPoly[2] += rightPoly[0];
    centerLeftPoly[0] = Vector2<>(0, 0);
    centerLeftPoly[1] = leftPoly[0];
    centerLeftPoly[2] = Vector2<>(obstacleModel.maxValidDist,0);
    centerLeftPoly[2].rotate(obstacleModel.centerLeftAngle);
    centerLeftPoly[2] += centerLeftPoly[1];
    centerLeftPoly[3] = Vector2<>(obstacleModel.maxValidDist,0);
    centerRightPoly[0] = rightPoly[0];
    centerRightPoly[1] = Vector2<>(0, 0);
    centerRightPoly[2] = Vector2<>(obstacleModel.maxValidDist,0);
    centerRightPoly[3] = Vector2<>(obstacleModel.maxValidDist,0);
    centerRightPoly[3].rotate(obstacleModel.centerRightAngle);
    centerRightPoly[3] += centerRightPoly[0];
    polygonsComputed = true;
  }
  obstacleModel.distanceToCenterLeftObstacle = obstacleModel.maxValidDist + 1;
  obstacleModel.distanceToCenterRightObstacle = obstacleModel.maxValidDist + 1;
  obstacleModel.distanceToLeftObstacle = obstacleModel.maxValidDist + 1;
  obstacleModel.distanceToRightObstacle = obstacleModel.maxValidDist + 1;
  
  for(int i=0; i<numOfCells; ++i)
  {
    if((cells[i].state >= parameters.cellOccupiedThreshold) || 
       ((cells[i].state >= parameters.cellOccupiedThresholdFast) && 
        (theWalkingEngineOutput.speed.translation.abs() >= parameters.fastWalkSpeedThreshold)))
    {
      Vector2<int> p = gridToWorld(Vector2<int>(i % CELLS_Y, i / CELLS_Y));
      CIRCLE("module:ObstacleModelProvider:backprojection",p.x,p.y,50,10,
        Drawings::ps_solid, ColorRGBA(255,0,0), Drawings::bs_null, ColorRGBA(255,0,0));
      Vector2<> pFloat((float) p.x, (float) p.y);
      
      if(Geometry::isPointInsideConvexPolygon(leftPoly, 3, pFloat))
      {
        const float dist = (pFloat - leftPoly[0]).abs();
        if(dist < obstacleModel.distanceToLeftObstacle)
          obstacleModel.distanceToLeftObstacle = dist;
      }
      else if(Geometry::isPointInsideConvexPolygon(rightPoly, 3, pFloat))
      {
        const float dist = (pFloat - rightPoly[0]).abs();
        if(dist < obstacleModel.distanceToRightObstacle)
          obstacleModel.distanceToRightObstacle = dist;
      }
      else if(Geometry::isPointInsideConvexPolygon(centerLeftPoly, 4, pFloat))
      {
        const float dist = pFloat.abs();
        if(dist < obstacleModel.distanceToCenterLeftObstacle)
          obstacleModel.distanceToCenterLeftObstacle = dist;
      }
      else if(Geometry::isPointInsideConvexPolygon(centerRightPoly, 4, pFloat))
      {
        const float dist = pFloat.abs();
        if(dist < obstacleModel.distanceToCenterRightObstacle)
          obstacleModel.distanceToCenterRightObstacle = dist;
      }
    }
  }
}

inline Vector2<int> ObstacleModelProvider::worldToGrid(const Vector2<int>& p) const
{
  Pose2D odoRotation(theOdometryData.rotation, 0.0f, 0.0f);
  Vector2<> pFloat((float) p.x, (float) p.y);
  pFloat = odoRotation*pFloat;
  pFloat /= CELL_SIZE;
  float moveX(CELLS_X/2);
  float moveY(CELLS_Y/2);
  pFloat += Vector2<>(moveX, moveY);
  return Vector2<int>(static_cast<int>(pFloat.x), static_cast<int>(pFloat.y));
}

inline Vector2<int> ObstacleModelProvider::gridToWorld(const Vector2<int>& p) const
{
  Vector2<> pFloat((float) p.x, (float) p.y);
  float moveX(CELLS_X/2);
  float moveY(CELLS_Y/2);
  pFloat -= Vector2<>(moveX, moveY);
  pFloat *= CELL_SIZE;
  Pose2D odoRotation(-theOdometryData.rotation, 0.0f, 0.0f);
  pFloat = odoRotation*pFloat;
  return Vector2<int>(static_cast<int>(pFloat.x), static_cast<int>(pFloat.y));
}

inline Vector2<> ObstacleModelProvider::gridToWorld(const Vector2<>& p) const
{
  Vector2<> pWorld(p);
  float moveX(CELLS_X/2);
  float moveY(CELLS_Y/2);
  pWorld -= Vector2<>(moveX, moveY);
  pWorld *= CELL_SIZE;
  Pose2D odoRotation(-theOdometryData.rotation, 0.0f, 0.0f);
  return odoRotation*pWorld;
}

void ObstacleModelProvider::checkUS()
{
  if(theFilteredSensorData.usTimeStamp == lastUsTimeStamp)
    return;
  lastUsTimeStamp = theFilteredSensorData.usTimeStamp;

  bool actuatorChanged = theFilteredSensorData.usActuatorMode != lastUsActuatorMode;
  lastUsActuatorMode = theFilteredSensorData.usActuatorMode;

  switch(theFilteredSensorData.usActuatorMode)
  {
  case SensorData::leftToLeft:
  case SensorData::rightToRight:
    addUsMeasurement(actuatorChanged, theFilteredSensorData.data[SensorData::usL], SensorData::leftToLeft, theFilteredSensorData.usTimeStamp);
    addUsMeasurement(actuatorChanged, theFilteredSensorData.data[SensorData::usR], SensorData::rightToRight, theFilteredSensorData.usTimeStamp);
    break;
  case SensorData::leftToRight:
  case SensorData::rightToLeft:
    addUsMeasurement(actuatorChanged, theFilteredSensorData.data[SensorData::usL], SensorData::rightToLeft, theFilteredSensorData.usTimeStamp);
    addUsMeasurement(actuatorChanged, theFilteredSensorData.data[SensorData::usR], SensorData::leftToRight, theFilteredSensorData.usTimeStamp);
    break;
  default:
    ASSERT(false);
    break;
  }
}

void ObstacleModelProvider::addUsMeasurement(bool actuatorChanged, float m, SensorData::UsActuatorMode mappedActuatorMode, unsigned int timeStamp)
{
  ASSERT(mappedActuatorMode >= 0);
  ASSERT(mappedActuatorMode < 4);

  bool isNew = timeStamp == theFilteredSensorData.usTimeStamp;
  if(timeStamp > bufferedMeasurements[mappedActuatorMode].timeStamp)
  {
    if(actuatorChanged || m < bufferedMeasurements[mappedActuatorMode].value)
      bufferedMeasurements[mappedActuatorMode].value = m;
    bufferedMeasurements[mappedActuatorMode].timeStamp = timeStamp;
  }

  if(m > parameters.maxValidUSDist || m < parameters.minValidUSDist)
    return;
  switch(mappedActuatorMode)
  {
  case SensorData::leftToLeft:
    for(std::vector<UsMeasurement>::iterator it = postponedLeftToRightMeasurements.begin(); it != postponedLeftToRightMeasurements.end();)
    {
      if(timeStamp - it->timeStamp > 1000) // postponed measurement is to old
        it = postponedLeftToRightMeasurements.erase(it);
      else if(fabs(m - it->value) < 200)
      {
        addUsMeasurement(false, it->value, SensorData::leftToRight, it->timeStamp);
        it = postponedLeftToRightMeasurements.erase(it);
      }
      else
        ++it;
    }
    break;
  case SensorData::rightToRight:
    for(std::vector<UsMeasurement>::iterator it = postponedRightToLeftMeasurements.begin(); it != postponedRightToLeftMeasurements.end();)
    {
      if(timeStamp - it->timeStamp > 1000) // postponed measurement is to old
        it = postponedRightToLeftMeasurements.erase(it);
      else if(fabs(m - it->value) < 200)
      {
        addUsMeasurement(false, it->value, SensorData::rightToLeft, it->timeStamp);
        it = postponedRightToLeftMeasurements.erase(it);
      }
      else
        ++it;
    }
    break;
  case SensorData::leftToRight:
    if(isNew && (timeStamp - bufferedMeasurements[SensorData::leftToLeft].timeStamp > 1000 || fabs(bufferedMeasurements[SensorData::leftToLeft].value - m) > 200))
    {
      postponedLeftToRightMeasurements.push_back(UsMeasurement(m, timeStamp));
      return; // Don't fill cells, as leftToRight measurements are only valid if left Sensor measured something
    }
    break;
  case SensorData::rightToLeft:
    if(isNew && (timeStamp - bufferedMeasurements[SensorData::rightToRight].timeStamp > 1000 || fabs(bufferedMeasurements[SensorData::rightToRight].value - m) > 200))
    {
      postponedRightToLeftMeasurements.push_back(UsMeasurement(m, timeStamp));
      return; // Don't fill cells, as rightToLeft measurements are only valid if right Sensor measured something
    }
    break;
  default:
    ASSERT(false);
    break;
  }
 
  // Compute and draw relevant area:
  Vector2<> measurement(m, 0.0f);
  Vector2<> base, leftOfCone(measurement), rightOfCone(measurement);  
  switch(mappedActuatorMode)
  {
  case SensorData::leftToLeft: // left transmitter, left sensor => left
    base = parameters.usLeftPose.translation;
    leftOfCone.rotate(parameters.usOuterOpeningAngle);
    rightOfCone.rotate(parameters.usInnerOpeningAngle);
    leftOfCone = parameters.usLeftPose * leftOfCone;
    rightOfCone = parameters.usLeftPose * rightOfCone;
    break;

  case SensorData::rightToRight: // right transmitter, right sensor => right
    base = parameters.usRightPose.translation;
    leftOfCone.rotate(-parameters.usInnerOpeningAngle);
    rightOfCone.rotate(-parameters.usOuterOpeningAngle);
    leftOfCone = parameters.usRightPose * leftOfCone;
    rightOfCone = parameters.usRightPose * rightOfCone;
    break;

  case SensorData::leftToRight: // left transmitter, right sensor => center left
    base = parameters.usCenterPose.translation;
    leftOfCone.rotate(parameters.usCenterOpeningAngle);
    rightOfCone.rotate(parameters.usCenterOpeningAngle * 0.25f);
    leftOfCone = parameters.usCenterPose * leftOfCone;
    rightOfCone = parameters.usCenterPose * rightOfCone;
    break;
  
  case SensorData::rightToLeft: // right transmitter, left sensor => center right
    base = parameters.usCenterPose.translation;
    leftOfCone.rotate(-parameters.usCenterOpeningAngle * 0.25f);
    rightOfCone.rotate(-parameters.usCenterOpeningAngle);
    leftOfCone = parameters.usCenterPose * leftOfCone;
    rightOfCone = parameters.usCenterPose * rightOfCone;
    break;

  default:
    ASSERT(false);
    break;
  }

  // Draw current (accepted) measurement:
  LINE("module:ObstacleModelProvider:us", base.x, base.y, leftOfCone.x, leftOfCone.y,
    30, Drawings::ps_solid, ColorRGBA(255,0,0));
  LINE("module:ObstacleModelProvider:us", base.x, base.y, rightOfCone.x, rightOfCone.y,
    30, Drawings::ps_solid, ColorRGBA(255,0,0));
  LINE("module:ObstacleModelProvider:us", rightOfCone.x, rightOfCone.y, leftOfCone.x, leftOfCone.y,
    30, Drawings::ps_solid, ColorRGBA(255,0,0));

  // Compute additional cones:
  // *Free is used for clearing cells between the robot and the obstacle
  // *Far is used to add a second obstacle line to the grid (should help in case of high walk speeds and imprecision)
  Vector2<> leftOfConeFree(leftOfCone-base);
  Vector2<> rightOfConeFree(rightOfCone-base);
  Vector2<> leftOfConeFar(leftOfCone-base);
  Vector2<> rightOfConeFar(rightOfCone-base);
  float cellDiameter = sqrt(static_cast<float>(CELL_SIZE*CELL_SIZE + CELL_SIZE*CELL_SIZE));
  leftOfConeFree.normalize(leftOfConeFree.abs() - cellDiameter);
  rightOfConeFree.normalize(rightOfConeFree.abs() - cellDiameter);
  leftOfConeFar.normalize(leftOfConeFar.abs() + cellDiameter);
  rightOfConeFar.normalize(rightOfConeFar.abs() + cellDiameter);
  leftOfConeFree += base;
  rightOfConeFree += base;
  leftOfConeFar += base;
  rightOfConeFar += base;
  // Transfer cones to grid coordinate system:
  Vector2<int> leftOfConeCells = worldToGrid(
    Vector2<int>(static_cast<int>(leftOfCone.x),static_cast<int>(leftOfCone.y)));
  Vector2<int> leftOfConeCellsFree = worldToGrid(
    Vector2<int>(static_cast<int>(leftOfConeFree.x),static_cast<int>(leftOfConeFree.y)));
  Vector2<int> leftOfConeCellsFar = worldToGrid(
    Vector2<int>(static_cast<int>(leftOfConeFar.x),static_cast<int>(leftOfConeFar.y)));
  Vector2<int> rightOfConeCells = worldToGrid(
    Vector2<int>(static_cast<int>(rightOfCone.x),static_cast<int>(rightOfCone.y)));
  Vector2<int> rightOfConeCellsFree = worldToGrid(
    Vector2<int>(static_cast<int>(rightOfConeFree.x),static_cast<int>(rightOfConeFree.y)));
  Vector2<int> rightOfConeCellsFar = worldToGrid(
    Vector2<int>(static_cast<int>(rightOfConeFar.x),static_cast<int>(rightOfConeFar.y)));

  // Free empty space until obstacle:
  polyPoints.clear();
  // Origin (sensor position):
  Vector2<int> p1(static_cast<int>(base.x), static_cast<int>(base.y));
  p1 = worldToGrid(p1);
  polyPoints.push_back(p1);
  // Left corner of "cone":
  polyPoints.push_back(Point(leftOfConeCellsFree, Point::NO_OBSTACLE));
  // Right corner of "cone":
  Vector2<> ll(rightOfConeFree);
  float f1 = ll.abs(),
    f2 = f1 ? leftOfConeFree.abs() / f1 : 0;
  Vector2<> gridPoint(ll);
  gridPoint *= f2;
  Vector2<int> gridPointInt(static_cast<int>(gridPoint.x), static_cast<int>(gridPoint.y));
  polyPoints.push_back(Point(worldToGrid(gridPointInt), polyPoints.back().flags));
  polyPoints.push_back(Point(rightOfConeCellsFree, Point::NO_OBSTACLE));
  // Sensor position again:
  polyPoints.push_back(p1);
  // Clip and fill:
  for(int j = 0; j < (int) polyPoints.size(); ++j)
    clipPointP2(p1, polyPoints[j]);
  fillScanBoundary();

  // Enter obstacle to grid:
  line(leftOfConeCells, rightOfConeCells);
  line(leftOfConeCellsFar, rightOfConeCellsFar);
}


void ObstacleModelProvider::ageCellState()
{
  for(int i=0; i<CELLS_X*CELLS_Y; ++i)
  {
    Cell& c = cells[i];
    if(c.onlyVision)
    {
      c.state = Cell::FREE;
      c.lastUpdate = theFrameInfo.time;
    }
    else if(c.state)
    {
      if(theFrameInfo.getTimeSince(c.lastUpdate) > parameters.cellFreeInterval)
      {
        c.state--;
        c.lastUpdate = theFrameInfo.time;
      }
      c.cluster = 0;
    }
  }
}

void ObstacleModelProvider::moveGrid()
{
  accumulatedOdometry += theOdometryData - lastOdometry;
  lastOdometry = theOdometryData;
  // Move grid backwards in x direction (robot moves forward):
  if(accumulatedOdometry.translation.x >= CELL_SIZE)
  {
    accumulatedOdometry.translation.x -= CELL_SIZE;
    for(int y=0; y<CELLS_Y; ++y)
    {
      Cell* cStartNew = &cells[y*CELLS_X];
      Cell* cStartOld = cStartNew + 1;
      memmove(cStartNew, cStartOld, sizeof(Cell)*(CELLS_X-1));
      cStartNew[CELLS_X-1] = Cell();
    }
  }
  // Move grid forward in x direction (robot moves backwards):
  else if(accumulatedOdometry.translation.x <= -CELL_SIZE)
  {
    accumulatedOdometry.translation.x += CELL_SIZE;
    for(int y=0; y<CELLS_Y; ++y)
    {
      Cell* cStartOld = &cells[y*CELLS_X];
      Cell* cStartNew = cStartOld + 1;
      memmove(cStartNew, cStartOld, sizeof(Cell)*(CELLS_X-1));
      cStartOld[0] = Cell();
    }
  }
  // Move grid backwards in y direction (robot moves to the left):
  if(accumulatedOdometry.translation.y >= CELL_SIZE)
  {
    accumulatedOdometry.translation.y -= CELL_SIZE;
    Cell* cStartOld = &cells[CELLS_X];
    Cell* cStartNew = &cells[0];
    memmove(cStartNew, cStartOld, sizeof(Cell)*CELLS_X*(CELLS_Y-1));
    Cell* c = &cells[(CELLS_Y-1)*CELLS_X];
    for(int x=0; x<CELLS_X; ++x)
      c[x] = Cell();
  }
  // Move grid forward in y direction (robot moves to the right):
  else if(accumulatedOdometry.translation.y <= -CELL_SIZE)
  {
    accumulatedOdometry.translation.y += CELL_SIZE;
    Cell* cStartNew = &cells[CELLS_X];
    Cell* cStartOld = &cells[0];
    memmove(cStartNew, cStartOld, sizeof(Cell)*CELLS_X*(CELLS_Y-1));
    Cell* c = &cells[0];
    for(int x=0; x<CELLS_X; ++x)
      c[x] = Cell();
  }
}

void ObstacleModelProvider::clipPointP2(const Vector2<int>& p1, Point& p2) const
{
  if(p2.x < 0)
  {
    p2.y = p1.y + (p2.y - p1.y) * -p1.x / (p2.x - p1.x);
    p2.x = 0;
  }
  else if(p2.x >= CELLS_X)
  {
    p2.y = p1.y + (p2.y - p1.y) * (CELLS_X - 1 - p1.x) / (p2.x - p1.x);
    p2.x = CELLS_X - 1;
  }

  if(p2.y < 0)
  {
    p2.x = p1.x + (p2.x - p1.x) * -p1.y / (p2.y - p1.y);
    p2.y = 0;
  }
  else if(p2.y >= CELLS_Y)
  {
    p2.x = p1.x + (p2.x - p1.x) * (CELLS_Y - 1 - p1.y) / (p2.y - p1.y);
    p2.y = CELLS_Y - 1;
  }
}

void ObstacleModelProvider::fillScanBoundary()
{
  // generate boundary of polygon
  std::list<Line> lines;
  int j;
  for(j = 1; j < (int) polyPoints.size() - 1; ++j)
  {
    if(polyPoints[j - 1].y < polyPoints[j].y && polyPoints[j + 1].y < polyPoints[j].y)
      polyPoints[j].flags |= Point::PEAK;
    lines.push_back(Line(polyPoints[j - 1], polyPoints[j]));        
  }
  lines.push_back(Line(polyPoints[j - 1], polyPoints[j]));
  
  // sort the lines by their y coordinates
  lines.sort();

  // run through lines in increasing y order
  for(int y = (int) lines.front().a.y; !lines.empty(); ++y)
  {
    inter.clear();
    for(std::list<Line>::iterator it = lines.begin(); it != lines.end() && (*it).a.y <= y;)
      if((*it).b.y > y || ((*it).peak && (*it).b.y == y))
      { // line is not finished yet
        inter.push_back(int((*it).a.x));
        (*it).a.x += (*it).ils;
         ++it;
      }
      else if((*it).b.y == y && (*it).a.y == y && y != 0)
      { // horizontal line, is not drawn when y == 0, because overlapping lines clutter the map
        inter.push_back(int((*it).a.x));
        inter.push_back(int((*it).b.x));
        ++it;
      }
      else if(it != lines.begin())
      { // line is finished -> remove it
        std::list<Line>::iterator it_help = it;
        --it;
        lines.erase(it_help);
        ++it;
      }
      else
      { // special case: remove begin of list
        lines.erase(it);
        it = lines.begin();
      }

    // fill the line on an even/odd basis
    bool paint = false;
    int goalX = -1;
    std::sort(inter.begin(), inter.end());
    std::vector<int>::iterator iIt = inter.begin();
    if(iIt != inter.end())
      for(;;)
      {
        int startX = *iIt;
        if(++iIt == inter.end())
          break;
        paint ^= true;
        if(paint)
        {
          if(startX == goalX)
            ++startX;
          goalX = *iIt;
          Cell* cell = &cells[y*CELLS_X+startX];
          for(int x = startX; x <= goalX; ++x)
          {
            if((*cell).state > 0) 
              (*cell).state--;
            (*cell).lastUpdate = theFrameInfo.time;
            ++cell;
          }
        }
      }
  }
}

void ObstacleModelProvider::line(Vector2<int>& start, Vector2<int>& end) 
{
  Vector2<int> diff = end - start,
               inc(diff.x > 0 ? 1 : -1, (diff.y > 0 ? 1 : -1) * (&cells[CELLS_X] - &cells[0])),
               absDiff(abs(diff.x), abs(diff.y));
  Cell* p = &cells[start.y*CELLS_X + start.x];

  if(absDiff.y < absDiff.x)
  {
    int error = -absDiff.x;
    for(int i = 0; i <= absDiff.x; ++i)
    {
      if((p->state < Cell::MAX_VALUE) && (p->lastUpdate != theFrameInfo.time))
        p->state++;
      p->lastUpdate = theFrameInfo.time;
      p->onlyVision = false;
      p += inc.x;
      error += 2 * absDiff.y;
      if(error > 0)
      {
        p += inc.y;
        error -= 2 * absDiff.x;
      }
    }
  }
  else
  {
    int error = -absDiff.y;
    for(int i = 0; i <= absDiff.y; ++i)
    {
      if((p->state < Cell::MAX_VALUE) && (p->lastUpdate != theFrameInfo.time))
        p->state++;
      p->lastUpdate = theFrameInfo.time;
      p->onlyVision = false;
      p += inc.y;
      error += 2 * absDiff.x;
      if(error > 0)
      {
        p += inc.x;
        error -= 2 * absDiff.y;
      }
    }
  }
}

void ObstacleModelProvider::draw() const
{
  // Test drawing:
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:us", "drawingOnField");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:grid", "drawingOnField");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:backprojection", "drawingOnField");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:pose", "drawingOnField");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:obstaclePositions", "drawingOnImage");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:obstaclesInField", "drawingOnField");
  DECLARE_DEBUG_DRAWING("module:ObstacleModelProvider:anotherObstacleModelStuff", "drawingOnField");
  
  COMPLEX_DRAWING("module:ObstacleModelProvider:grid",
    unsigned char colorOccupiedStep(255/Cell::MAX_VALUE);
    ColorRGBA baseColor(200,200,255);
    const int numOfCells(CELLS_X*CELLS_Y);
    unsigned char cellsForDrawing[CELLS_Y*CELLS_X];
    for(int i=0; i<numOfCells; ++i)
      cellsForDrawing[i] = colorOccupiedStep*cells[i].state;
    GRID_MONO("module:ObstacleModelProvider:grid", CELL_SIZE*2, CELLS_X, CELLS_Y, baseColor, cellsForDrawing);
    int gridWidth(CELLS_X*CELL_SIZE*2);
    int gridHeight(CELLS_Y*CELL_SIZE*2);
    RECTANGLE("module:ObstacleModelProvider:grid", -gridWidth/2, -gridHeight/2, gridWidth/2, gridHeight/2,
      20, Drawings::ps_solid, ColorRGBA(0,0,100));
  );

  COMPLEX_DRAWING("module:ObstacleModelProvider:pose",
    Vector2<> bodyPoints[4] = {Vector2<>(55*2,90*2),
                                     Vector2<>(-55*2,90*2),
                                     Vector2<>(-55*2,-90*2),
                                     Vector2<>(55*2,-90*2)};
    for(int i=0; i<4; i++)
      bodyPoints[i] = Geometry::rotate(bodyPoints[i], theOdometryData.rotation);
    Vector2<> dirVec(200,0*2);
    dirVec.rotate(theOdometryData.rotation);
    LINE("module:ObstacleModelProvider:pose", 0, 0, dirVec.x, dirVec.y, 
      20, Drawings::ps_solid, ColorClasses::white);
    POLYGON("module:ObstacleModelProvider:pose", 4, bodyPoints, 20, Drawings::ps_solid, 
      ColorClasses::green, Drawings::bs_solid, ColorClasses::white);
    CIRCLE("module:ObstacleModelProvider:pose", 0, 0, 42*2, 1, 
      Drawings::ps_solid, ColorClasses::green, Drawings::bs_solid, ColorClasses::green);
  );
}

void ObstacleModelProvider::verifyObstaclesInImage()
{
  const int numOfCells((CELLS_X*CELLS_Y));
      
  for(int i=0; i<numOfCells; ++i)
  {
    bool freeByVision = cells[i].lastVisionFree ? theFrameInfo.getTimeSince(cells[i].lastVisionFree) <= imgParams.imageRememberTime : false;
    if(cells[i].state || freeByVision) 
    {
      //calculate obstacle in world
      Vector2<int> obstacle = gridToWorld(Vector2<int>(i % CELLS_Y, i / CELLS_Y));
      Vector2<> obstacleF((float)obstacle.x,(float)obstacle.y);
      Vector2<> halfObject((float)(imgParams.imageVerificationWidth>>1), 0.0f);
      Vector2<> tmp = obstacleF + (halfObject.rotate(obstacle.angle())).rotateLeft();
      Vector3<> left(tmp.x, tmp.y, 0.0f);
      tmp = obstacleF + halfObject.rotateRight();
      Vector3<> right(tmp.x, tmp.y, (float)imgParams.imageVerificationHeight);
      //check if objects are in front of robot (needed by calculation of image coordinates)
      if(left.x < 0.0f && right.x < 0.0f) 
      {
        if(freeByVision)
        {
          cells[i].state = Cell::FREE;
        }
        continue;
      }
      else if(left.x < 0.0f) 
      {
        left.x = 0.0f;
      }
      else if (right.x < 0.0f)
      {
        right.x = 0.0f;
      }

      COMPLEX_DRAWING("module:ObstacleModelProvider:obstaclesInField",
        Pose2D p = theRobotPose;
        p.translate(Vector2<>(left.x, left.y));
        int x1 = (int)p.translation.x, y1 = (int)p.translation.y;
        CIRCLE("module:ObstacleModelProvider:obstaclesInField",p.translation.x,p.translation.y,25,
          10, Drawings::ps_solid, ColorRGBA(255,0,0), Drawings::bs_null, ColorRGBA(0,255,0));
        p = theRobotPose;
        p.translate(Vector2<>(obstacleF.x, obstacleF.y));
        CIRCLE("module:ObstacleModelProvider:obstaclesInField",p.translation.x,p.translation.y,15,
          10, Drawings::ps_solid, ColorRGBA(255,255,0), Drawings::bs_solid, ColorRGBA(255,255,0));
        p = theRobotPose;
        p.translate(Vector2<>(right.x, right.y));
        CIRCLE("module:ObstacleModelProvider:obstaclesInField",p.translation.x,p.translation.y,25,
          10, Drawings::ps_solid, ColorRGBA(0,255,0), Drawings::bs_null, ColorRGBA(255,0,0));
        LINE("module:ObstacleModelProvider:obstaclesInField",x1,y1,(int)p.translation.x,(int)p.translation.y,
          10, Drawings::bs_solid, ColorRGBA(255,255,0));
      );

      //get points in image
      Vector2<int> lowerLeft, upperRight;
      Geometry::calculatePointInImage(left, theCameraMatrix, theCameraInfo, lowerLeft);
      Geometry::calculatePointInImage(right, theCameraMatrix, theCameraInfo, upperRight);
      
      RECTANGLE(
        "module:ObstacleModelProvider:obstaclePositions",
        lowerLeft.x, lowerLeft.y, upperRight.x, upperRight.y,
        2, Drawings::ps_solid, ColorRGBA(100,0,0)
      );

      //y direction is inverted beacause of direction in image
      //check if there was a free cell detected here before
      if(freeByVision)
      {
        if(lowerLeft.x < 0)
          lowerLeft.x = 0;
        if(upperRight.x > theCameraInfo.resolutionWidth - 1)
          upperRight.x = theCameraInfo.resolutionWidth - 1;
        if(upperRight.y < 0)
          upperRight.y = 0;
        if(lowerLeft.y > theCameraInfo.resolutionHeight - 1)
          lowerLeft.y = theCameraInfo.resolutionHeight - 1;

        //if rectangle is outside the image set the cell (remembered) as free
        if(lowerLeft.x > upperRight.x || upperRight.y > lowerLeft.y)
        {
          cells[i].state = Cell::FREE;
          continue;
        }
      }
      
      //free cell if there is green place in the image 
      float cr = pointExplorer.colorRatio(ColorClasses::green, Boundary<int>(
        Range<int>(lowerLeft.x, upperRight.x), Range<int>(upperRight.y, lowerLeft.y)), 2);
      
      DRAWTEXT(
        "module:ObstacleModelProvider:obstaclePositions", 
        lowerLeft.x + 5, (lowerLeft.y + upperRight.y) / 2, 
        24, ColorClasses::black, "Green: " << cr
      );
      
      if(cr > 0.8f)
      {
        cells[i].state = Cell::FREE;
        cells[i].lastVisionFree = theFrameInfo.time;
      }
      else if(freeByVision)
      {
        cells[i].lastVisionFree = 0;
      }
    }
  }
}

void ObstacleModelProvider::applyRobotsModel()
{
  const int maxXDist = CELL_SIZE * CELLS_X/2;
  const int maxYDist = CELL_SIZE * CELLS_Y/2;
  const int width = imgParams.robotHeight;
  const int cellRadius = width / CELL_SIZE / 2;
  const int cellRadius2 = cellRadius * cellRadius;
  for(RobotsModel::RCIt robot = theRobotsModel.robots.begin(); robot != theRobotsModel.robots.end(); robot++)
  {
    // we add only horizontal robots since the others should be recognized by the ultrasonic sensor
    if(!robot->standing)
    {
      // exaggerated clipping
      if(robot->relPosOnField.x + width < maxXDist && robot->relPosOnField.x - width > -maxXDist
        && robot->relPosOnField.y + width < maxYDist && robot->relPosOnField.y - width > -maxYDist)
      {
        Vector2<int> gridPos = worldToGrid(Vector2<int>((int) robot->relPosOnField.x, (int) robot->relPosOnField.y));
        const int xCellStart = gridPos.x - cellRadius;
        const int xCellEnd = gridPos.x + cellRadius;
        const int yCellStart = gridPos.y - cellRadius;
        const int yCellEnd = gridPos.y + cellRadius;
        for(int x=xCellStart; x<=xCellEnd; x++)
        {
          for(int y=yCellStart; y<=yCellEnd; y++)
          {
            const Vector2<int> curCell(x, y);
            // we insert a circle for every horizontal robot
            if((curCell-gridPos).sqr() <= cellRadius2)
            {
              if(x<0 || x>=CELLS_X || y<0 || y>=CELLS_Y)
                continue; // TODO better solution
              Cell& cell = cells[x + y * CELLS_X];
              // obstacles that only come from the robotsModel will be deleted and added every cognition frame
              if(!cell.state)
                cell.onlyVision = true;
              cell.state = Cell::MAX_VALUE;
            }
          }
        }
      }
    }
  }
}

MAKE_MODULE(ObstacleModelProvider, Modeling)
