/**
* @file ObstacleModelProvider.h
*
* This file declares a module that provides information about occupied space in the robot's environment.
* It includes parts of the implementation of the PolygonLocalMapper module of the 
* autonomous wheelchair Rolland.
*
* @author <a href="mailto:Tim.Laue@dfki.de">Tim Laue</a>
* @author <a href="mailto:cman@tzi.de">Christian Mandel</a>
* @author <a href="mailto:marko@tzi.de">Benjamin Markowsky</a>
*/

#ifndef __ObstacleModelProvider_h_
#define __ObstacleModelProvider_h_

#include "Representations/Modeling/AnotherObstacleModel.h"
#include "Representations/Modeling/ObstacleModel.h"
#include "Representations/MotionControl/MotionRequest.h"
#include "Representations/MotionControl/MotionInfo.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/MotionControl/WalkingEngineOutput.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Representations/Modeling/RobotsModel.h"
#include "Representations/Modeling/RobotPose.h"
#include "Tools/Module/Module.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Infrastructure/CameraInfo.h"
#include "Modules/Perception/PointExplorer.h"
#include "Representations/Infrastructure/Image.h"
#include "Representations/Configuration/ColorTable64.h"
#include <stack>


MODULE(ObstacleModelProvider)
  REQUIRES(OdometryData)
  REQUIRES(FrameInfo)
  REQUIRES(FilteredSensorData)
  REQUIRES(KeyStates)
  REQUIRES(RobotsModel)
  REQUIRES(CameraMatrix)//test
  REQUIRES(CameraInfo)//test
  REQUIRES(Image)
  REQUIRES(ColorTable64)
  REQUIRES(RobotPose)
  REQUIRES(ObstacleModel)
  REQUIRES(WalkingEngineOutput)
  USES(MotionRequest)
  USES(MotionInfo)
  USES(RobotInfo)
  PROVIDES_WITH_MODIFY_AND_DRAW(ObstacleModel)
  PROVIDES_WITH_MODIFY_AND_DRAW(AnotherObstacleModel)
END_MODULE


/**
* @class ObstacleModelProvider
*
* A module for computing the occupied space in the robot's environment
*/
class ObstacleModelProvider: public ObstacleModelProviderBase
{
private:
  /**
  * @class Parameters
  * The parameters of the module
  */
  class Parameters: public Streamable
  {
  private:
    /**
    * The method makes the object streamable.
    * @param in The stream from which the object is read 
    * @param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
        STREAM(cellFreeInterval);
        STREAM(usOuterOpeningAngle);
        STREAM(usInnerOpeningAngle);
        STREAM(usCenterOpeningAngle);
        STREAM(maxValidUSDist);
        STREAM(minValidUSDist);
        STREAM(cellOccupiedThreshold);
        STREAM(cellOccupiedThresholdFast);
        STREAM(fastWalkSpeedThreshold);
        STREAM(usRightPose);
        STREAM(usLeftPose);
        STREAM(usCenterPose);
        STREAM(imageVerification);
      STREAM_REGISTER_FINISH();
    }

  public:
    int cellFreeInterval;                /**< Time (in ms) after which cell value becomes decreased */
    float usOuterOpeningAngle;           /**< Opening angle of US sensor towards side of robot */
    float usInnerOpeningAngle;           /**< Opening angle of US sensor towards center of robot */
    float usCenterOpeningAngle;          /**< Opening angle of Center US sensor to each side */
    int maxValidUSDist;                  /**< Maximum measured distance which is considered for the model */
    int minValidUSDist;                  /**< Minimum measured distance which is considered for the model */
    int cellOccupiedThreshold;           /**< If a cell state is higher than or equal to this value, it is considered as occupied */
    int cellOccupiedThresholdFast;       /**< If a cell state is higher than or equal to this value when walking fast, it is considered as occupied */
    int fastWalkSpeedThreshold;          /**< If walking faster than this threshold, use other cell occupied threshold*/
    Pose2D usRightPose;                  /**< Position and orientation of right US sensor*/
    Pose2D usLeftPose;                   /**< Position and orientation of left US sensor*/
    Pose2D usCenterPose;                 /**< Position and orientation of (virtual) center US sensor */
    bool imageVerification;              /**< Verify obstacles in image, if possible */

    /** Constructor */
    Parameters():cellFreeInterval(3000), maxValidUSDist(1199), minValidUSDist(200), 
      cellOccupiedThreshold(3), cellOccupiedThresholdFast(2), 
      fastWalkSpeedThreshold(180), imageVerification(true)
    {
      usInnerOpeningAngle = fromDegrees(30);
      usOuterOpeningAngle = fromDegrees(70);
      usCenterOpeningAngle = fromDegrees(30);
      usRightPose.translation = Vector2<>(53.7f,-34.1f);
      usLeftPose.translation  = Vector2<>(53.7f, 34.1f);
      usCenterPose.translation = Vector2<>(53.7f, 0);
      usRightPose.rotation = atan2f(-0.34f, 0.93f);
      usLeftPose.rotation  = atan2f( 0.34f, 0.93f);
      usCenterPose.rotation = 0.f;
    }
  };

  /**
  * @class ImageParameters
  * The parameters of the module used to set up the image processing
  */
  class ImageParameters: public Streamable
  {
  private:
    /**
    * The method makes the object streamable.
    * @param in The stream from which the object is read 
    * @param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
        STREAM(imageRememberTime);
        STREAM(imageVerificationWidth);
        STREAM(imageVerificationHeight);
        STREAM(exploreStepSize);
        STREAM(gridStepSize);
        STREAM(skipOffset);
        STREAM_ARRAY(minSegSize);
        STREAM(robotWidth);
        STREAM(robotHeight);
        STREAM(visionOn)
      STREAM_REGISTER_FINISH();
    }

  public:
    int imageRememberTime;               /**< the time which is used to remember free cells, if it was verified by vision in ms*/
    int imageVerificationWidth;          /**< width in mm to check if this could be an obstacle */
    int imageVerificationHeight;         /**< height in mm to check if this could be an obstacle */
    int exploreStepSize;                 /**< Unused but necessary for PointExplorer */
    int gridStepSize;                    /**< Stepsize used by the PointExplorer */
    int skipOffset;                      /**< Offset which Pixels should be skipped by the PointExplorer */
    int minSegSize[ColorClasses::numOfColors]; /**< Minvalues for the different Segements (used by the PointExplorer) */
    int robotWidth;                      /**< The width of a robot. */
    int robotHeight;                     /**< The height of a robot. */
    bool visionOn;                       /**< Do we add horizontal robots to the obstacle model? */

    /** Constructor */
    ImageParameters(): imageVerificationWidth(100), imageVerificationHeight(300),
      exploreStepSize(3), gridStepSize(2), skipOffset(2), robotWidth(400), robotHeight(550), visionOn(false) {};
  };

  /** 
  * @class Cell
  * A cell within the modeled occupancy grid
  */
  class Cell
  {
  public:
    unsigned lastUpdate;                      /** The point of time of the last change of the state*/
    unsigned lastVisionFree;                  /** The point of time when this cell had been checked as free by vision*/
    enum State {FREE = 0, MAX_VALUE = 3};     /** Minimum and maximum value of state */
    int state;                                /** The state of the cell, i.e. the number of positive measurements within the last few frames*/
    int cluster;                              /** The cluster a cell belongs to (for AnotherObstacleModel)*/
    bool onlyVision;                          /** Is the obstacle provided by the robots model? **/

    /** Constructor */
    Cell():lastUpdate(0),lastVisionFree(0),state(FREE),cluster(0), onlyVision(false)
    {}
  };

  /** 
  * @class CellCluster
  * A cluster of cells representing a single obstacle
  */
  class CellCluster
  {
  public:
    std::vector< Vector2<int> > cells;

    /** Constructor*/
    CellCluster()
    {
      cells.reserve(100);
    }

    /** Resets the cluster*/
    void init()
    {
      cells.clear();
    }
  };

  /**
  * @class Point
  * The class represents a grid point together with clipping information.
  */
  class Point : public Vector2<int>
  {
  public:
    /**
    * The enumeration represents different edges of the map 
    * to which scan points can be clipped, and further options.
    */
    enum Flags
    {
      NONE, 
      NO_OBSTACLE = 16,
      PEAK = 32
    };
    int flags; /**< The set of edges to which the point was clipped, and whether it results from a castor wheel. */

    /**
    * Default constructor.
    */
    Point() : flags(NONE) {}

    /**
    * Constructor.
    * @param x The x coordinate of the point.
    * @param y The y coordinate of the point.
    * @param flags The set of edges to which the point was clipped, and whether 
    *              it results from a castor wheel.
    */
    Point(int x, int y, int flags = 0) : Vector2<int>(x, y), flags(flags) {}

    /**
    * Constructor.
    * @param v The x and y coordinates of the point.
    * @param flags The set of edges to which the point was clipped, and whether 
    *              it results from a castor wheel.
    */
    Point(const Vector2<int>& v, int flags = 0) : Vector2<int>(v), flags(flags) {}
  };

  /**
  * @class Line
  * This class represents a single line-segment out of a polyline.
  */
  class Line
  {
  public:
    /** 
    * Constructor.
    * @param ai The first point of the line-segment.
    * @param bi The last point of the line-segment.
    */
    Line(const Point& ai, const Point& bi)
    : peak(ai.flags & Point::PEAK || bi.flags & Point::PEAK)
    { 
      if(ai.y <= bi.y)
      {
        a = Vector2<>( (float)ai.x, (float)ai.y ); 
        b = Vector2<>( (float)bi.x, (float)bi.y );
      }
      else
      {
        b = Vector2<>( (float)ai.x, (float)ai.y ); 
        a = Vector2<>( (float)bi.x, (float)bi.y );
      }
      ils = b.y != a.y ? (b.x - a.x) / (b.y - a.y) : 0;
    }

    /**
    * Operator that compares a given line with this line. 
    * @param other This line is compared to other.
    * @return Is the y coordinate of this line smaller than the one of the other line?
    */
    bool operator<(const Line& other) const
    {
      return a.y < other.a.y;
    }

    Vector2<> a; /**< First point of this line. */
    Vector2<> b; /**< Last point of this line. */
    float ils;   /**< Inverse slope of this line. */
    bool peak;   /**< Is this line part of an upper peak? */
  };

  /**
  * @class UsMeasurement
  * A buffered us measurement.
  */
  class UsMeasurement
  {
  public:
    float value; /**< The measured distance. */
    unsigned int timeStamp; /**< Time stamp of the measurement. */

    /**
    * Default constructor.
    */
    UsMeasurement() : value(0), timeStamp(0) {}

    /**
    * Constructor.
    * @param value The measured distance.
    * @param timeStamp Time stamp of the measurement.
    */
    UsMeasurement(float value, unsigned int timeStamp) : value(value), timeStamp(timeStamp) {}
  };

public:
  /** Constructor */
  ObstacleModelProvider();

private:
  /** Dimensions of the grid */
  enum {CELL_SIZE = 60,
        CELLS_X = 45,
        CELLS_Y = 45};

  Pose2D accumulatedOdometry;                 /**< Storing odometry differences for grid update */
  Pose2D lastOdometry;                        /**< Odometry value at last execution */
  unsigned lastTimePenalized;                 /**< Last point of time the robot was penalized */
  Parameters parameters;                      /**< The parameters of this module */
  Cell cells[CELLS_Y*CELLS_X];                /**< The grid */
  CellCluster currentCluster;                 /**< The current cluster */
  std::stack< Vector2<int> > cellStack;       /**< A stack for the floodfill algorithm */
  std::vector<Point> polyPoints;              /**< Vector of obstacle points to become processed. Clipped and filtered. */
  std::vector<int> inter;                     /**< The intersections per line. */
  UsMeasurement bufferedMeasurements[4];      /**< The last valid us measurements of each actuator mode */
  std::vector<UsMeasurement> postponedLeftToRightMeasurements; /**< leftToRight measurements that were postponed */
  std::vector<UsMeasurement> postponedRightToLeftMeasurements; /**< rightToLeft measurements that were postponed */
  SensorData::UsActuatorMode lastUsActuatorMode; /**< The controlled actuator mode of the last us measurement */
  Vector2<> leftPoly[3];                      /**< Polygon for left output value */
  Vector2<> rightPoly[3];                     /**< Polygon for right output value */
  Vector2<> centerLeftPoly[4];                /**< Polygon for center/left output value */
  Vector2<> centerRightPoly[4];               /**< Polygon for center/right output value */
  bool polygonsComputed;                      /**< Flag to assure that polygons are only computed once */
  unsigned lastUsTimeStamp;                   /**< The time stamp of the last used ultrasonic measurement */

  ImageParameters imgParams;   /**< Parameters used for imageprocesing */
  PointExplorer pointExplorer; /**< Used for verifying obstacles in the image. */

  /** Executes this module
  * @param obstacleModel The data structure that is filled by this module
  */
  void update(ObstacleModel& obstacleModel);

  /** Executes this module
  * @param anotherObstacleModel The second data structure that is filled by this module
  */
  void update(AnotherObstacleModel& anotherObstacleModel);

  /** Initialization after construction */
  void init();

  /** Enter obstacles to the grid */
  void checkUS();

  /** Inserts us measurements into the grid.
  * @param actuatorChanged whether the actuator mode has changed compared to the previous iteration
  * @param m The measured value
  * @param mappedActuatorMode The actual actuator mode that is some how different to the controlled actuator mode
  * @param timeStamp The time stamp of the measurement
  */
  void addUsMeasurement(bool actuatorChanged, float m, SensorData::UsActuatorMode mappedActuatorMode, unsigned int timeStamp);

  /** Ages all cells, i.e. decreases the obstacle value in each cell */
  void ageCellState();

  /** Called by occupyCells, frees the space between the sensor and the obstacle */
  void fillScanBoundary();

  /** Clips point to the grid 
  * @param p1 The position from which the measurement is done
  * @param p2 The measured point
  */
  void clipPointP2(const Vector2<int>& p1, Point& p2) const;

  /** Converts coordinates in the local coordinate system to grid coordinates
  * @param p The position of a point
  * @return The grid coordinates
  */
  Vector2<int> worldToGrid(const Vector2<int>& p) const;

  /** Converts coordinates in grid coordinates to the local coordinate system
  * @param p The position within the grid
  * @return A position in local robot coordinates
  */
  Vector2<int> gridToWorld(const Vector2<int>& p) const;

  /** Converts coordinates in grid coordinates to the local coordinate system
  *   Floating point version, allows to convert coordinates with sub-cell precision
  * @param p The position within the grid
  * @return A position in local robot coordinates
  */
  Vector2<> gridToWorld(const Vector2<>& p) const;

  /** Enters an obstacle line to the grid
  * @param start The start of the line
  * @param end The end of the line
  */
  void line(Vector2<int>& start, Vector2<int>& end);

  /** Computes the model data from the grid
  * @param obstacleModel The model
  */
  void computeModel(ObstacleModel& obstacleModel);

  /** Computes the model data from the grid
  * @param anotherObstacleModel The model
  */
  void computeAnotherObstacleModelFromCurrentCluster(AnotherObstacleModel& anotherObstacleModel);

  /** Moves the grid cells according to the odometry differences*/
  void moveGrid();

  /** Checks if the obstacles detected by the us are robots in 
   * in the camera image. */
  void verifyObstaclesInImage();

  /** Adds robots that are recognized via vision. */
  void applyRobotsModel();

  /** Do debug drawings*/
  void draw() const;

  Vector2<int> projectOnImage(int x, int y, int z) const;
};

#endif// __ObstacleModelProvider_h_
