/** 
* @file KalmanFilter2D.cpp
* Template class for two-dimensional Kalman filtering.
* @author <a href="mailto:afabisch@tzi.de">Alexander Fabisch</a>
*/

#include "KalmanFilter2D.h"
#include "Tools/Math/Common.h"

void KalmanFilter2D::init(
      Vector2<> mean,
      Matrix2x2<> covariance)
{
  this->mean = mean;
  this->covariance = covariance;
}

void KalmanFilter2D::preparePrediction(
    const Matrix2x2<>& stateTransition,
    const Matrix2x2<>& controlTransformation,
    const Matrix2x2<>& predictionNoise)
{
  this->stateTransition = stateTransition;
  this->controlTransformation = controlTransformation;
  this->predictionNoise = predictionNoise;
}

void KalmanFilter2D::prediction(const Vector2<>& control)
{
  mean = stateTransition * mean + controlTransformation * control;
  covariance = stateTransition * covariance * stateTransition.transpose() + predictionNoise;
}

void KalmanFilter2D::prepareSensorUpdate(
    const Matrix2x2<>& measurementTransformation,
    const Matrix2x2<>& measurementNoise)
{
  this->measurementTransformation = measurementTransformation;
  this->measurementNoise = measurementNoise;
}

void KalmanFilter2D::sensorUpdate(const Vector2<> measurement)
{
  static const Matrix2x2<> identity = Matrix2x2<>(Vector2<>(1.0f, 0.0f), Vector2<>(0.0f, 1.0f));
  const Matrix2x2<> kalmanGain = covariance * measurementTransformation *
      (measurementTransformation * covariance * measurementTransformation.transpose() + measurementNoise).invert();
  mean = mean + kalmanGain * (measurement - measurementTransformation * mean);
  covariance = (identity - kalmanGain * measurementTransformation) * covariance;
}

float KalmanFilter2D::squaredMahalanobisDistance(const Vector2<>& measurement) const
{
  // same as (x-mean)T*covariance'*(x-mean)
  Vector2<> diff(measurement - mean);
  return diff * (covariance.invert() * diff);
}

float KalmanFilter2D::approximateProbabilityAtMeanArea(float halfSquareLength) const
{
  return 4.0f * halfSquareLength*halfSquareLength / (pi2*sqrtf(covariance.det()));
}

void KalmanFilter2D::calculateErrorEllipse(float& axis1, float& axis2, float& angle, float factor) const
{
  const float cov012 = covariance.c[0][1]*covariance.c[0][1];
  const float varianceDiff = covariance.c[0][0] - covariance.c[1][1];
  const float varianceDiff2 = varianceDiff*varianceDiff;
  const float varianceSum = covariance.c[0][0] + covariance.c[1][1];
  const float root = sqrtf(varianceDiff2 + 4.0f*cov012);
  const float eigenValue1 = 0.5f * (varianceSum + root);
  const float eigenValue2 = 0.5f * (varianceSum - root);

  angle = 0.5f * atan2f(2.0f * covariance.c[0][1], varianceDiff);
  axis1 = 2.0f * sqrtf(factor * eigenValue1);
  axis2 = 2.0f * sqrtf(factor * eigenValue2);
}

