/** 
* @file KalmanFilter2D.h
* Class for two-dimensional kalman filtering. Assumes that the control and measurement vectors are also 2D.
* @author <a href="mailto:afabisch@tzi.de">Alexander Fabisch</a>
*/

#ifndef __KalmanFilter2D_h_
#define __KalmanFilter2D_h_

#include "Tools/Math/Vector2.h"
#include "Tools/Math/Matrix2x2.h"

class KalmanFilter2D
{
  Matrix2x2<> stateTransition;
  Matrix2x2<> controlTransformation;
  Matrix2x2<> predictionNoise;

  Matrix2x2<> measurementTransformation;
  Matrix2x2<> measurementNoise;

public:
  Vector2<> mean;
  Matrix2x2<> covariance;

  /** 
  * Constructor.
  * @param mean The initial mean.
  * @param covariance The initial covariance.
  */
  KalmanFilter2D(
      Vector2<> mean,
      Matrix2x2<> covariance)
      : mean(mean),
        covariance(covariance) {}
  KalmanFilter2D() {}

  void init(
      Vector2<> mean,
      Matrix2x2<> covariance);

  void preparePrediction(
      const Matrix2x2<>& stateTransition,
      const Matrix2x2<>& controlTransition,
      const Matrix2x2<>& predictionNoise);
  void prediction(
      const Vector2<>& control);
  void prepareSensorUpdate(
      const Matrix2x2<>& measurementTransformation,
      const Matrix2x2<>& measurementNoise);
  void sensorUpdate(
      const Vector2<> measurement);

  float squaredMahalanobisDistance(const Vector2<>& measurement) const;
  float approximateProbabilityAtMeanArea(float halfSquareLength) const;
  void calculateErrorEllipse(float& axis1, float& axis2, float& angle, float factor = 1.0f) const;
};

#endif //__KalmanFilter2D_h_
