/**
 * @class RobotLocator.cpp
 * The implementation of the RobotLocator with a Kalman filter.
 * @author <a href="mailto:afabisch@tzi.de">Alexander Fabisch</a>
 */
#include "RobotLocator.h"
#include "RobotLocatorHelper.h"
#include "Tools/Streams/ConfigHelper.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Math/Matrix2x2.h"
#include "Tools/Math/Vector3.h"
#include "Tools/Math/GaussianDistribution2D.h"
#include "Tools/Math/Common.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Debugging/Stopwatch.h"
#include "Tools/Team.h"
#include <limits>
#include <algorithm>

RobotLocator::RobotLocator()
{
  ConfigHelper::loadConfig(params, "robotPerceptor.cfg", "locator");
}

void RobotLocator::update(RobotsModel& robotsModel)
{
  DECLARE_DEBUG_DRAWING("module:RobotLocator:projection", "drawingOnImage");
  DEBUG_RESPONSE("module:RobotLocator:test", test(theCameraMatrix, theCameraInfo, theImageCoordinateSystem, theImage););
  DEBUG_RESPONSE("module:RobotLocator:clear", robots.clear(););
  MODIFY("parameters:RobotLocator", params);

  motionUpdate();
  sensorUpdate();
  cycle(robotsModel);

  TEAM_OUTPUT(idTeamMateRobotsModel, bin, robotsModel);
}

void RobotLocator::motionUpdate()
{
  Pose2D odometryOffset(theOdometryData - lastOdometryData);
  lastOdometryData = theOdometryData;

  for(RLI r(robots.begin()); r != robots.end(); r++)
  {
    // state transformation matrix
    const float sinNegTheta = sinf(-odometryOffset.rotation);
    const float cosNegTheta = cosf(-odometryOffset.rotation);
    const Matrix2x2<> A_t = Matrix2x2<>(Vector2<>(cosNegTheta, sinNegTheta),
                                        Vector2<>(-sinNegTheta, cosNegTheta));
    // control transformation matrix
    static const Matrix2x2<> B_t = Matrix2x2<>(Vector2<>(-1.0f, 0.0f),
                                               Vector2<>(0.0f, -1.0f));
    // prediction noise covariance matrix
    const float f = std::max(1.0f, (float) fabs(r->kf.mean.angle()));
    ASSERT(f >= 1.0f);
    ASSERT(f <= pi);
    static const Matrix2x2<> R_t = createCovariance(params.motionNoiseStdDev.x, params.motionNoiseStdDev.y, 0.0f) * f;

    // Prediction step: Currently there is actually no motion model of the other robots.
    // So their assumed motion is 0, but we add significant noise.
    r->kf.preparePrediction(A_t, B_t, R_t);
    r->kf.prediction(odometryOffset.translation);
  }
}

void RobotLocator::sensorUpdate()
{
  for(RobotPercept::RCIt r(theRobotPercept.robots.begin()); r != theRobotPercept.robots.end(); r++)
  {
    updateMostLikelyOldRobot(*r);
  }

  // TODO Experimental
  for(RobotPercept::RCIt r(theRobotPercept.unlocalizableRobots.begin()); r != theRobotPercept.unlocalizableRobots.end(); r++)
  { // TODO function
    Vector2<> relPosOnField;
    if(!Geometry::calculatePointOnField(theImageCoordinateSystem.toCorrected(r->lowestPx), theCameraMatrix, theCameraInfo, relPosOnField))
      continue;
    const float angle = relPosOnField.angle(); // TODO more efficient
    for(RLI robot(robots.begin()); robot != robots.end(); robot++)
    {
      if(robot->notUpdatedYet() && robot->team == r->team
        && fabs(robot->kf.mean.angle() - angle) < params.maxAngleDist
        && !shouldBeVisible(theCameraMatrix, theCameraInfo, theImage, theImageCoordinateSystem, robot->kf.mean))
      {
        ASSERT(robot->kf.covariance.det() != 0.0);
        static const Matrix2x2<> I = Matrix2x2<>(Vector2<>(1.0f, 0.0f),
                                                 Vector2<>(0.0f, 1.0f));
        // mean after prediction step
        const Vector2<> mu_s(robot->kf.mean);
        // covariance matrix after prediction step
        const Matrix2x2<> Cov_s(robot->kf.covariance);
        // measurement vector
        const Vector2<> z_t(robot->kf.mean);
        // measurement transformation matrix
        static const Matrix2x2<> C_t = I;
        // measurement noise covariance matrix
        const Matrix2x2<> Q_t = createCovariance(params.angleSensorNoiseStdDev.x,
                                                 params.angleSensorNoiseStdDev.y, z_t.angle());

        // Measurement update:
        robot->kf.prepareSensorUpdate(C_t, Q_t);
        robot->kf.sensorUpdate(z_t);
        robot->cyclesSinceLastUpdate = 0;
        robot->state = Robot::UPDATED;
      }
    }
  }

  for(RLI r(robots.begin()); r != robots.end(); r++)
  {
    if(r->notUpdatedYet())
    {
      if(shouldBeVisible(theCameraMatrix, theCameraInfo, theImage, theImageCoordinateSystem, r->kf.mean)) // TODO dont calculate more than once
      {
        r->kf.covariance += createCovariance(
            params.noSensorNoiseStdDev.x, params.noSensorNoiseStdDev.y, r->kf.mean.angle());
        r->state = Robot::UPDATED;
      }
      if(!theFieldDimensions.isInsideCarpet(Geometry::relative2FieldCoord(theRobotPose, r->kf.mean))) // TODO maybe diagonal length of field?
      {
        r->state = Robot::DELETED;
      }
    }
    if(r->kf.approximateProbabilityAtMeanArea(params.robotWidth) < params.minMeanProbability)
    {
      r->state = Robot::DELETED;
    }
  }
}

void RobotLocator::updateMostLikelyOldRobot(const RobotPercept::Robot& newRobot)
{
  Vector2<> relPosOnField;
  if(!Geometry::calculatePointOnField(theImageCoordinateSystem.toCorrected(newRobot.lowestPx), theCameraMatrix, theCameraInfo, relPosOnField))
    return;
  static const float initial(std::numeric_limits<float>::max());
  float shortestDistance(initial);
  RLI robot;
  const float distance2newRobot = relPosOnField.abs();
  const float maxDistance = params.maxDistOfRobots + std::max(0.0f, distance2newRobot - 2250.0f) * 0.2f; // TODO params
  for(RLI r(robots.begin()); r != robots.end(); r++)
  {
    if(r->team == newRobot.team && Geometry::distance(r->kf.mean, relPosOnField) < maxDistance)
    {
      float distance = sqrtf(r->kf.squaredMahalanobisDistance(relPosOnField));
      if(distance < shortestDistance)
      {
        robot = r;
        shortestDistance = distance;
      }
    }
  }
  if(shortestDistance == initial)
  {
    const Matrix2x2<> initialCovariance = createCovariance(
        params.initialStdDev.x, params.initialStdDev.y, relPosOnField.angle());
    Robot initializedRobot(relPosOnField, initialCovariance, newRobot.standing, newRobot.team);
    ASSERT(initializedRobot.state == Robot::NEW);
    robots.push_back(initializedRobot);
  }
  else if(robot->notUpdatedYet())
  {
    ASSERT(robot->kf.covariance.det() != 0.0);
    static const Matrix2x2<> I = Matrix2x2<>(Vector2<>(1.0f, 0.0f),
                                             Vector2<>(0.0f, 1.0f));
    // mean after prediction step
    const Vector2<> mu_s(robot->kf.mean);
    // covariance matrix after prediction step
    const Matrix2x2<> Cov_s(robot->kf.covariance);
    // measurement vector
    const Vector2<> z_t(relPosOnField);
    // measurement transformation matrix
    static const Matrix2x2<> C_t = I;
    // measurement noise covariance matrix
    const float manhattenDistanceRatio = 0.25f*(fabs(z_t.x) + fabs(z_t.y)); // TODO params
    const Matrix2x2<> Q_t = createCovariance(params.sensorNoiseStdDev.x + manhattenDistanceRatio,
                                             params.sensorNoiseStdDev.y + manhattenDistanceRatio, z_t.angle());

    // Measurement update:
    robot->kf.prepareSensorUpdate(C_t, Q_t);
    robot->kf.sensorUpdate(z_t);
    robot->standing = newRobot.standing;
    robot->cyclesSinceLastUpdate = 0;
    robot->state = Robot::UPDATED;
  }
}

void RobotLocator::cycle(RobotsModel& robotsModel)
{
  std::for_each(robots.begin(), robots.end(), std::mem_fun_ref(&Robot::cycle));
  std::list<Robot> oldRobots(robots);
  robots.clear();
  std::remove_copy_if(oldRobots.begin(), oldRobots.end(), std::back_inserter(robots), deprecationCriterion);
  static const unsigned maxRobots = unsigned(params.maxRobots * 2);
  if(robots.size() > maxRobots)
  {
    robots.sort();
    while(robots.size() > maxRobots) robots.erase(robots.begin());
  }

  robotsModel.robots.clear();
  robotsModel.robots.reserve(maxRobots);
  for(RLI r = robots.begin(); r != robots.end(); r++)
  {
    Vector3<> ellipse;
    r->kf.calculateErrorEllipse(ellipse.x, ellipse.y, ellipse.z);
    robotsModel.robots.push_back(RobotsModel::Robot(r->kf.mean, r->team, r->standing, ellipse));
  }
  ASSERT(robotsModel.robots.size() <= maxRobots);
}

void RobotLocator::Parameters::serialize(In* in, Out* out)
{
  STREAM_REGISTER_BEGIN();
  STREAM(maxDistOfRobots);
  STREAM(maxAngleDist);
  STREAM(maxRobots);
  STREAM(minMeanProbability);
  STREAM(robotWidth);
  STREAM(motionNoiseStdDev);
  STREAM(initialStdDev);
  STREAM(sensorNoiseStdDev);
  STREAM(angleSensorNoiseStdDev);
  STREAM(noSensorNoiseStdDev);
  STREAM_REGISTER_FINISH();
}

bool RobotLocator::DeprecationCriterion::operator()(const Robot& robot)
{
  return robot.state == Robot::DELETED;
}

MAKE_MODULE(RobotLocator, Modeling)

