/**
 * @file RobotLocator.h
 * This module provides the RobotsModel. A Kalman filter is used to smooth the RobotPercept measurements.
 * @author <a href="mailto:afabisch@tzi.de">Alexander Fabisch</a>
 */
#ifndef __RobotLocator_h_
#define __RobotLocator_h_

#include "Tools/Module/Module.h"
#include "Representations/Perception/RobotPercept.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Infrastructure/CameraInfo.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Infrastructure/Image.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Modeling/RobotsModel.h"
#include "KalmanFilter2D.h"
#include <list>

MODULE(RobotLocator)
  REQUIRES(RobotPercept)
  REQUIRES(CameraMatrix)
  REQUIRES(CameraInfo)
  REQUIRES(ImageCoordinateSystem)
  REQUIRES(Image)
  REQUIRES(OdometryData)
  REQUIRES(RobotPose)
  REQUIRES(FieldDimensions)
  PROVIDES_WITH_DRAW(RobotsModel)
END_MODULE

/**
 * @class RobotLocator
 * This module implements the robot localization with a Kalman filter.
 */
class RobotLocator : public RobotLocatorBase
{
  /**
   * @class Parameters
   * The parameters for the RobotLocator.
   */
  class Parameters : public Streamable
  {
  public:
    Parameters() {}
    float maxDistOfRobots;            /**< The maximal euclidean distance of one robot between two cognition frames. */
    float maxAngleDist;               /**< Maximal angle distance (for unlocalizable robots). */
    int maxRobots;                    /**< Maximal number of robots on the field. */
    float minMeanProbability;         /**< Minimal probability in the mean area of the robot's probability function. */
    float robotWidth;                 /**< The approximated width of a standing robot. */
    Vector2<> motionNoiseStdDev;      /**< The standard deviations of the noise of the other robots' motion. */
    Vector2<> initialStdDev;          /**< The standard deviations of an initial distribution of a percepted robot. */
    Vector2<> sensorNoiseStdDev;      /**< The standard deviations of the noise of the measurement. */
    Vector2<> angleSensorNoiseStdDev; /**< The standard deviations of the noise of an imprecise measurement. */
    Vector2<> noSensorNoiseStdDev;    /**< This will be added if there is no measurement update. */
  private:
    void serialize(In* in, Out* out);
  };

  /**
   * @class Robot
   * A Kalman filter state.
   */
  class Robot
  {
  public:
    enum RobotState {NEW, UPDATED, DELETED};

    KalmanFilter2D kf;         /**< Contains the mean (relative position of the robot on the field) and the covariance. */
    bool standing;             /**< Is the robot standing? */
    int team;                  /**< TEAM_RED or TEAM_BLUE? */
    int cyclesSinceLastUpdate; /**< Number of cognition frames since last update. */
    RobotState state;          /**< The robot's state. */

    Robot(Vector2<> relPosOnField, Matrix2x2<> covariance, bool standing, int team)
      : kf(relPosOnField, covariance), standing(standing), team(team), cyclesSinceLastUpdate(0), state(NEW) {}

    inline bool operator<(const Robot& other)
    {
      return this->kf.approximateProbabilityAtMeanArea(0.5f) < other.kf.approximateProbabilityAtMeanArea(0.5f);
    }

    inline void cycle()
    {
      cyclesSinceLastUpdate++;
    }

    inline bool notUpdatedYet() const
    {
      return cyclesSinceLastUpdate != 0;
    }
  };

  class DeprecationCriterion : public std::unary_function<Robot, bool>
  {
  public:
    bool operator()(const Robot& robot);
  };

  typedef std::list<Robot>::const_iterator RLCI;
  typedef std::list<Robot>::iterator RLI;

  Parameters params;
  /** Indicates whether a robot will be deleted. */
  DeprecationCriterion deprecationCriterion;

  /** The currently localized robots. */
  std::list<Robot> robots;
  /** This is used to calculate the odometry offset. */
  OdometryData lastOdometryData;

  /**
   * Processes the current robotPercept and updates the robotsModel.
   * @param robotsModel The percepted robots on the field.
   */
  void update(RobotsModel& robotsModel);
  /**
   * Applies the motion model on the current robots.
   */
  void motionUpdate();
  /**
   * Updates the robotsModel with the percepted robots.
   */
  void sensorUpdate();
  /**
   * Rejects robots that are too uncertain, updates the RobotsModel.
   * @param robotsModel The old RobotsModel.
   */
  void cycle(RobotsModel& robotsModel);

  /**
   * Updates the position of a robot in the model or adds a new robot.
   * @param newRobot The new percepted robot.
   */
  void updateMostLikelyOldRobot(const RobotPercept::Robot& newRobot);
public:
  RobotLocator();
};

#endif // __RobotLocator_h_
