#ifndef __RobotLocatorHelper_cpp_
#define __RobotLocatorHelper_cpp_

#include "Tools/Math/Vector2.h"
#include "Tools/Math/Matrix2x2.h"
#include "Tools/Math/Geometry.h"
#include "Platform/GTAssert.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Infrastructure/CameraInfo.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Infrastructure/Image.h"
#include <cmath>

/**
 * Creates the two-dimensional covariance matrix.
 * @param xDev sigmaX (standard deviation of x)
 * @param yDev sigmaY (standard deviation of y)
 * @param angle rotation angle
 * @return The covariance matrix.
 */
inline const Matrix2x2<> createCovariance(
  const float xDev,
  const float yDev,
  const float angle)
{
  const float sinRotation = sinf(angle);
  const float cosRotation = cosf(angle);
  const Matrix2x2<> r = Matrix2x2<>(Vector2<>(cosRotation, sinRotation), Vector2<>(-sinRotation, cosRotation));
  return r * Matrix2x2<>(Vector2<>(xDev*xDev, 0.0f), Vector2<>(0.0f, yDev*yDev)) * r.transpose();
}

/**
 * Determines whether robot should be visible on given image
 * @param theCameraMatrix CameraMatrix
 * @param theCameraInfo CameraInfo
 * @param theImage Image
 * @param theImageCoordinateSystem Coordinate system of the image
 * @param relPosOnField The relative position of the robot on the field.
 * @return Should the robot be visible on the current image?
 */
inline bool shouldBeVisible(
  const CameraMatrix& theCameraMatrix,
  const CameraInfo& theCameraInfo,
  const Image& theImage,
  const ImageCoordinateSystem& theImageCoordinateSystem,
  const Vector2<>& relPosOnField)
{
  static const float TOLERANCE = 2.0f;
  Vector2<int> pointInImage;
  const bool valid = Geometry::calculatePointInImage(
    Vector2<int>((int) relPosOnField.x, (int) relPosOnField.y),
    theCameraMatrix, theCameraInfo, pointInImage);
  if(!valid) return false;
  Vector2<> rawPointInImage = theImageCoordinateSystem.fromCorrectedApprox(pointInImage);
  CROSS("module:RobotLocator:projection", (int) rawPointInImage.x, (int) rawPointInImage.y,
      2, 2, Drawings::ps_solid, ColorClasses::red);
  const bool isInImage = Geometry::isPointInsideRectangle(
    Vector2<>(-TOLERANCE,
              (float) (theCameraInfo.resolutionHeight / 2) - TOLERANCE), // The team marker should be visible, too.
    Vector2<>((float) (theCameraInfo.resolutionWidth - 1) + TOLERANCE,
              (float) (theCameraInfo.resolutionHeight - 1) + TOLERANCE),
    rawPointInImage);
  return isInImage;
}

inline Vector2<> imageCoords2relFieldCoords(
  const CameraMatrix& theCameraMatrix,
  const CameraInfo& theCameraInfo,
  const Vector2<>& imageCoordinates)
{
  Vector2<> relPosOnField;
  Geometry::calculatePointOnField(int(imageCoordinates.x), int(imageCoordinates.y), theCameraMatrix, theCameraInfo, relPosOnField);
  return relPosOnField;
}

void test(
  const CameraMatrix& theCameraMatrix,
  const CameraInfo& theCameraInfo,
  const ImageCoordinateSystem& theImageCoordinateSystem,
  const Image& theImage)
{
  const int width = theImage.cameraInfo.resolutionWidth;
  const int height = theImage.cameraInfo.resolutionHeight;

  Vector2<> middle(float(width / 2), float(height / 2));
  Vector2<> middleOnField(imageCoords2relFieldCoords(theCameraMatrix, theCameraInfo, middle));
  OUTPUT(idText, text, "middle on field: x=" << middleOnField.x << ", y=" << middleOnField.y);
  ASSERT(shouldBeVisible(theCameraMatrix, theCameraInfo, theImage, theImageCoordinateSystem, middleOnField));

  Vector2<> leftBottom(0, float(height - 1));
  Vector2<> leftBottomOnField(imageCoords2relFieldCoords(theCameraMatrix, theCameraInfo, leftBottom));
  OUTPUT(idText, text, "lb on field: x=" << leftBottomOnField.x << ", y=" << leftBottomOnField.y);
  ASSERT(shouldBeVisible(theCameraMatrix, theCameraInfo, theImage, theImageCoordinateSystem, leftBottomOnField));

  Vector2<> rightBottom(float(width - 1), float(height - 1));
  Vector2<> rightBottomOnField(imageCoords2relFieldCoords(theCameraMatrix, theCameraInfo, rightBottom));
  OUTPUT(idText, text, "rb on field: x=" << rightBottomOnField.x << ", y=" << rightBottomOnField.y);
  ASSERT(shouldBeVisible(theCameraMatrix, theCameraInfo, theImage, theImageCoordinateSystem, rightBottomOnField));
}

#endif
