/* 
* @file BIKE.cpp
* This file implements a module that creates motions.
* @author <A href="mailto:judy@tzi.de">Judith Mller</A>
*/

#include "BIKE.h"

#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Debugging/DebugDrawings3D.h"
#include "Representations/MotionControl/BikeRequest.h"
#include "Tools/Debugging/Modify.h"
#include "BIKEParameters.h"

#include "Tools/Math/Pose3D.h"

void BIKE::update(BikeEngineOutput& bikeOutput)
{
  MODIFY("module:BIKE:px", gyroP.x);
  MODIFY("module:BIKE:dx", gyroD.x);
  MODIFY("module:BIKE:ix", gyroI.x);
  MODIFY("module:BIKE:py", gyroP.y);
  MODIFY("module:BIKE:dy", gyroD.y);
  MODIFY("module:BIKE:iy", gyroI.y);

  int requestedBMotionID = getPosOfRequestedBMotion();
  Vector2<> origin(0.f,0.f);
  Vector3<> correctedCom(0.f,0.f,0.f), comCorr(0.f,0.f,0.f);
  Vector3<> ref(0.f,0.f,0.f);
  Vector3<> actualDiff(0.f,0.f,0.f);

  // float plotCommandedVel = 0.f;

#ifndef RELEASE 
  char saveName[260];
  BIKEParameters dummy;
  if (requestedBMotionID > -1) strcpy(saveName,params[requestedBMotionID].name);
  MODIFY("module:BIKE:actualParameter",  ((requestedBMotionID > -1)? params[requestedBMotionID]:dummy));
  if (requestedBMotionID > -1) strcpy(params[requestedBMotionID].name, saveName);
  MODIFY("module:BIKE:newBMotion", params[theMotionRequest.bikeRequest.getBMotionFromName("noName")]);
  strcpy(params[theMotionRequest.bikeRequest.getBMotionFromName("noName")].name,"noName");
#endif //RELEASE

  //get time from last frame to next (is it 50Hz or 100Hz)
  cycletime = theRobotDimensions.motionCycleTime;

  if(theMotionSelection.ratios[MotionRequest::bike] > 0.f)
  {
    bikeOutput.odometryOffset.translation.x = 0.f;
    bikeOutput.odometryOffset.translation.y = 0.f;
    bikeOutput.odometryOffset.fromAngle(0.f);

    if((!wasActive || ((theMotionRequest.bikeRequest.bMotionType != currentBikeRequest.bMotionType || theMotionRequest.bikeRequest.mirror != currentBikeRequest.mirror) && bikeOutput.isLeavingPossible)) && requestedBMotionID > -1)
    {
      currentBikeRequest = theMotionRequest.bikeRequest;
      
      bikeOutput.executedBikeRequest.mirror = currentBikeRequest.mirror;
      bikeOutput.executedBikeRequest.dynamical = currentBikeRequest.dynamical;
      bikeOutput.executedBikeRequest.bMotionType = currentBikeRequest.bMotionType; 

      actualBMotionID = requestedBMotionID;
      timeStamp = theFrameInfo.time;

      phaseNumber = 0;

      if (!wasActive) // means that no bike motion was played in last Frame 
      {
        toRightSupport = false;
        toLeftSupport = false;
        rSupp = false;
        lSupp = false;

        lastGyroLeft = Vector2<> (0.f, 0.f);
        lastGyroRight = Vector2<> (0.f, 0.f);
        gyro = Vector2<> (0.f, 0.f);

        lastOrigin = 0.f;

        lastCorrectedCom = Vector3<>(0.f,0.f,0.f);
        comErrorSum = Vector3<>(0.f,0.f,0.f);

        bikeOutput.angles[JointData::RHipYawPitch] = 0.f;
        bikeOutput.angles[JointData::LHipYawPitch] = 0.f;
        bikeOutput.isLeavingPossible = false;

        for (int i=JointData::LHipYawPitch; i < JointData::numOfJoints; i++)
        {
          bikeOutput.jointHardness.hardness[i] = 100;
        }

        for (int i=JointData::LShoulderPitch; i < JointData::LHipYawPitch; i++)
        {
          bikeOutput.jointHardness.hardness[i] = 50;
        }

        bikeOutput.angles[JointData::HeadYaw] = JointData::ignore;
        bikeOutput.angles[JointData::HeadPitch] = JointData::ignore;

        for (int i = 0; i < JointData::numOfJoints; i++)
        {
          lastJointRequest.angles[i] = 0.f; 
          lastBalancedJointRequest.angles[i] = 0.f;
          balancedJointRequest.angles[i] = 0.f;
        }

        comOffset = Vector2<>(0.f, 0.f);

        balanceSum = Vector2<>(0.f,0.f);
        gyroErrorRight = Vector2<> (0.f,0.f);
        gyroErrorLeft = Vector2<> (0.f,0.f);

        currentParameters = params[actualBMotionID];

        currentParameters.initFirstPhase();

        if (currentBikeRequest.dynamical && !currentBikeRequest.dynPoints.empty())
        {
          for (unsigned int i=0; i < currentBikeRequest.dynPoints.size(); i++)
          {
            if (currentBikeRequest.dynPoints[i].phaseNumber == 0)
              addDynPoint(currentBikeRequest.dynPoints[i],currentParameters);
          }
          currentParameters.initFirstPhase();
        }

      }
      else
      {    //a different bike motion was played before and you want to go from one motion to another

        Phase lastPhase = currentParameters.phaseParameters[currentParameters.numberOfPhases-1];
        currentParameters = params[actualBMotionID];

        for (int j=0; j < Phase::numOfLimbs; j++)
        {
          currentParameters.phaseParameters[0].controlPoints[j][0] = lastPhase.controlPoints[j][2];
        }

        currentParameters.initFirstPhaseLoop(lastPhase);

        if (currentBikeRequest.dynamical && !currentBikeRequest.dynPoints.empty())
        {
          for (unsigned int i=0; i < currentBikeRequest.dynPoints.size(); i++)
          {
            if (currentBikeRequest.dynPoints[i].phaseNumber == 0)
              addDynPoint(currentBikeRequest.dynPoints[i],currentParameters);
          }
          currentParameters.initFirstPhaseLoop(lastPhase);
        }       
      }
    }
    else if(theMotionRequest.bikeRequest.bMotionType == currentBikeRequest.bMotionType && theMotionRequest.bikeRequest.mirror == currentBikeRequest.mirror)
      currentBikeRequest = theMotionRequest.bikeRequest; // update BikeRequest when it is compatible to the current motion

    if(actualBMotionID > -1)
    {
      if (phaseNumber <  currentParameters.numberOfPhases)
      {
        if (phaseNumber == 0)  bikeOutput.isLeavingPossible = false;

        //if the phase is at the end
        if (theFrameInfo.getTimeSince(timeStamp) > (int)  currentParameters.phaseParameters[phaseNumber].duration) 
        {
          phaseNumber++;
          timeStamp = theFrameInfo.time;
          bikeOutput.odometryOffset.translation.x = currentParameters.phaseParameters[phaseNumber-1].odometryOffset.x;
          bikeOutput.odometryOffset.translation.y = currentParameters.phaseParameters[phaseNumber-1].odometryOffset.y;
          bikeOutput.odometryOffset.fromAngle(currentParameters.phaseParameters[phaseNumber-1].odometryOffset.z);

          //if the motion is in loop
          if (currentParameters.loop && phaseNumber == currentParameters.numberOfPhases) 
          {
            phaseNumber = 0;
            bikeOutput.isLeavingPossible = true;

            Phase lastPhase;
            lastPhase = currentParameters.phaseParameters[currentParameters.numberOfPhases-1];

            for (int j=0; j < Phase::numOfLimbs; j++)
            {
              currentParameters.phaseParameters[0].controlPoints[j][0] = lastPhase.controlPoints[j][2];
            }
            currentParameters.initFirstPhaseLoop(lastPhase);

            if (currentBikeRequest.dynamical && !currentBikeRequest.dynPoints.empty())
            {
              for (unsigned int i=0; i < currentBikeRequest.dynPoints.size(); i++)
              {
                if (currentBikeRequest.dynPoints[i].phaseNumber == phaseNumber)
                  addDynPoint(currentBikeRequest.dynPoints[i],currentParameters);              
              }
              currentParameters.initFirstPhaseLoop(lastPhase);
            }
          }

          //if it's still not the end of the motion
          if (phaseNumber < currentParameters.numberOfPhases && phaseNumber > 0) 
          {
            currentParameters.initPhase(phaseNumber);
            if (currentBikeRequest.dynamical && !currentBikeRequest.dynPoints.empty())
            {
              for (unsigned int i=0; i < currentBikeRequest.dynPoints.size(); i++)
              {
                if (currentBikeRequest.dynPoints[i].phaseNumber == phaseNumber) 
                  addDynPoint(currentBikeRequest.dynPoints[i],currentParameters);
              }
              currentParameters.initPhase(phaseNumber);
            }
          }
        }
      }

      if (phaseNumber < currentParameters.numberOfPhases) 
      {
        phase = (float) theFrameInfo.getTimeSince(timeStamp) / (float) currentParameters.phaseParameters[phaseNumber].duration;

        for (int i=0; i < Phase::numOfLimbs; i++)
        {
          positions[i] = currentParameters.getPosition(phase, phaseNumber, i);
        }

        //switch left and right side when mirror true;
        if (currentBikeRequest.mirror)
        {
          Vector3<> temp =  positions[Phase::rightFootTra];
          positions[Phase::rightFootTra] = Vector3<>( positions[Phase::leftFootTra].x, 
            -positions[Phase::leftFootTra].y, 
            positions[Phase::leftFootTra].z);
          positions[Phase::leftFootTra] = Vector3<> (temp.x, -temp.y, temp.z);

          temp = positions[Phase::leftFootRot];
          positions[Phase::leftFootRot] = positions[Phase::rightFootRot];
          positions[Phase::rightFootRot] = temp;

          temp = positions[Phase::rightArmTra];
          positions[Phase::rightArmTra] = Vector3<>(positions[Phase::leftArmTra].x, 
            -positions[Phase::leftArmTra].y, 
            positions[Phase::leftArmTra].z);
          positions[Phase::leftArmTra] = Vector3<>(temp.x, -temp.y, temp.z);
        }
      }
      else
      {
        bikeOutput.isLeavingPossible = true;
      }

      calcJoints(positions[Phase::rightFootTra], positions[Phase::leftFootTra], 
        positions[Phase::leftFootRot], positions[Phase::rightFootRot], 
        positions[Phase::rightArmTra], positions[Phase::leftArmTra], comOffset, bikeOutput); 

      robotModel.setJointData(theFilteredJointData, theRobotDimensions, theMassCalibration);


      calculatePreviewCom(phaseNumber, (float)theFrameInfo.getTimeSince(timeStamp), currentParameters.preview, ref, origin);

      if((currentParameters.supportLegMode == BIKEParameters::auto_ && fabs(positions[Phase::leftFootTra].y - standLegPos.getAverage().y) < 1.f) ||
        (currentParameters.supportLegMode == BIKEParameters::left && !currentBikeRequest.mirror) ||
        (currentParameters.supportLegMode == BIKEParameters::right && currentBikeRequest.mirror))
      {
        lSupp = true;
        rSupp = false;
      }
      if((currentParameters.supportLegMode == BIKEParameters::auto_ && fabs(positions[Phase::rightFootTra].y - standLegPos.getAverage().y) < 1.f) ||
        (currentParameters.supportLegMode == BIKEParameters::left && currentBikeRequest.mirror) ||
        (currentParameters.supportLegMode == BIKEParameters::right && !currentBikeRequest.mirror))
      {
        rSupp = true;
        lSupp = false;
      }
      if((currentParameters.supportLegMode == BIKEParameters::auto_ && fabs(origin.y) < 1.f) ||
        (currentParameters.supportLegMode == BIKEParameters::both))
      {
        rSupp = false;
        lSupp = false;
      }

      actualDiff = robotModel.centerOfMass - ref;

      balanceSum+= Vector2<> (actualDiff.x, actualDiff.y);

      Vector2<> balance;
      balance.y = -currentParameters.kpx* (actualDiff.y) + -currentParameters.kix*balanceSum.y
        + -currentParameters.kdx*((actualDiff.y-lastCorrectedCom.y)/cycletime);
      balance.x = currentParameters.kpy*(actualDiff.x)  + currentParameters.kiy*balanceSum.x
        + currentParameters.kdy*((actualDiff.x-lastCorrectedCom.x)/cycletime);

      if (robotModel.centerOfMass.z-ref.z != 0.)
      {
 
        comOffset.x = atan2f((balance.y-2),(robotModel.centerOfMass.z-ref.z)); 
        comOffset.y = atan2f((balance.x),(robotModel.centerOfMass.z-ref.z));
      }

      //if (!theGroundContactState.contactSafe)
      //{
      //bikeOutput.isLeavingPossible = true;
      //}

      lastCorrectedCom = actualDiff;

      //recalculate joints
      calcJoints(positions[Phase::rightFootTra], positions[Phase::leftFootTra], 
        positions[Phase::leftFootRot], positions[Phase::rightFootRot], 
        positions[Phase::rightArmTra], positions[Phase::leftArmTra], comOffset, bikeOutput); 

      //some clipping
      for (int i=JointData::LHipYawPitch; i<JointData::numOfJoints; i++ )
      {
        if (bikeOutput.angles[i] > theJointCalibration.joints[i].maxAngle) bikeOutput.angles[i] = theJointCalibration.joints[i].maxAngle;
        if (bikeOutput.angles[i] < theJointCalibration.joints[i].minAngle) bikeOutput.angles[i] = theJointCalibration.joints[i].minAngle;
      }

      balancedJointRequest = (JointRequest) bikeOutput;

      if (theMotionSelection.ratios[MotionRequest::bike] == 1.f)
      {
        //Predict next gyrodata
        gyro.y = theFilteredSensorData.data[SensorData::gyroY]*0.5f + 0.5f*gyro.y;
        gyro.x = theFilteredSensorData.data[SensorData::gyroX]*0.5f + 0.5f*gyro.x;


        //calculate the commandedVelocity
        float commandedVelocity[4];
        //y-velocity if left leg is support
        commandedVelocity[0] = (balancedJointRequest.angles[JointData::LHipPitch] - lastBalancedJointRequest.angles[JointData::LHipPitch])/cycletime;
        //y-velocity if right leg is support
        commandedVelocity[1] = (balancedJointRequest.angles[JointData::RHipPitch] - lastBalancedJointRequest.angles[JointData::RHipPitch])/cycletime;
        //x-velcocity if left leg is support
        commandedVelocity[2] = (balancedJointRequest.angles[JointData::LHipRoll] - lastBalancedJointRequest.angles[JointData::LHipRoll])/cycletime;
        //x-velocity if right leg is support
        commandedVelocity[3] = (balancedJointRequest.angles[JointData::RHipRoll] - lastBalancedJointRequest.angles[JointData::RHipRoll])/cycletime;

        //   plotCommandedVel = commandedVelocity;


        //calculate disturbance from meseaured velocity and commanded velocity
        // y-velocity if left leg is support
        float gyroVelyLeft = (gyro.y+commandedVelocity[0] - lastGyroLeft.y)/cycletime;
        lastGyroLeft.y = gyro.y+commandedVelocity[0];

        //y-velocity if right leg is support
        float gyroVelyRight = (gyro.y+commandedVelocity[1] - lastGyroRight.y)/cycletime;
        lastGyroRight.y = gyro.y+commandedVelocity[1];
        //y-velocity if right leg is support
        float gyroVelxLeft = (gyro.x-commandedVelocity[2] - lastGyroLeft.x)/cycletime;
        lastGyroLeft.x = gyro.x-commandedVelocity[2]; 
        //x-velocity if right leg is support
        float gyroVelxRight = (gyro.x + commandedVelocity[3] - lastGyroRight.x)/cycletime;
        lastGyroRight.x = gyro.x+commandedVelocity[3];


        //calculate "Stellgre" with PID-Control
        float calcVelocity[4];
        //y if left supprt
        calcVelocity[0] = -gyroP.y*(gyro.y+commandedVelocity[0])- gyroD.y*gyroVelyLeft - gyroI.y*(gyroErrorLeft.y);
        //y if right support
        calcVelocity[1] = -gyroP.y*(gyro.y+commandedVelocity[1])- gyroD.y*gyroVelyRight - gyroI.y*(gyroErrorRight.y);
        //x if left support
        calcVelocity[2] = -gyroP.x*(gyro.x - commandedVelocity[2]) - gyroD.x*gyroVelxLeft - gyroI.x*gyroErrorLeft.y;
        //x if right support
        calcVelocity[3] = gyroP.x*(gyro.x - commandedVelocity[3]) + gyroD.x*gyroVelxRight + gyroI.x*gyroErrorRight.y;


        if((currentParameters.supportLegMode == BIKEParameters::auto_ && origin.y < -10.f) || //Right Leg Support
          (currentParameters.supportLegMode == BIKEParameters::right && !currentBikeRequest.mirror) ||
          (currentParameters.supportLegMode == BIKEParameters::left && currentBikeRequest.mirror))
        {
          toRightSupport = true;
          toLeftSupport = false;
          lastGyroLeft = lastGyroRight;
          
          //y
          bikeOutput.angles[JointData::RHipPitch] +=  calcVelocity[1]*cycletime;
          bikeOutput.angles[JointData::LHipPitch] += calcVelocity[1]*cycletime;
          //x
       //   bikeOutput.angles[JointData::LAnkleRoll] -= -calcVelocity[3]*cycletime;
          bikeOutput.angles[JointData::LHipRoll] += calcVelocity[3]*cycletime;
         // bikeOutput.angles[JointData::RAnkleRoll] -= calcVelocity[3]*cycletime;
          bikeOutput.angles[JointData::RHipRoll] += -calcVelocity[3]*cycletime;

          //OUTPUT(idText, text, "right");

        }
        else if((currentParameters.supportLegMode == BIKEParameters::auto_ && origin.y > 10.f) || //Left Leg Support
          (currentParameters.supportLegMode == BIKEParameters::right && currentBikeRequest.mirror) ||
          (currentParameters.supportLegMode == BIKEParameters::left && !currentBikeRequest.mirror))
        {
          toLeftSupport = true;
          toRightSupport = false;
          lastGyroRight = lastGyroLeft;

          //y 
          bikeOutput.angles[JointData::LHipPitch] += calcVelocity[0]*cycletime;
          bikeOutput.angles[JointData::RHipPitch] += calcVelocity[0]*cycletime;
          //x
     //     bikeOutput.angles[JointData::LAnkleRoll] -= -calcVelocity[2]*cycletime;
          bikeOutput.angles[JointData::LHipRoll] += -calcVelocity[2]*cycletime;
       //   bikeOutput.angles[JointData::RAnkleRoll] -= calcVelocity[2]*cycletime;
          bikeOutput.angles[JointData::RHipRoll] += calcVelocity[2]*cycletime;

          //OUTPUT(idText, text, "left");
        }
        else
        {
          if (toLeftSupport) //last support Leg was left
          {
            lastGyroRight = lastGyroLeft;
            //y
            bikeOutput.angles[JointData::RHipPitch] += calcVelocity[0]*cycletime;
            bikeOutput.angles[JointData::LHipPitch] += calcVelocity[0]*cycletime;
            //x
      //      bikeOutput.angles[JointData::LAnkleRoll] -= -calcVelocity[2]*cycletime;
            bikeOutput.angles[JointData::LHipRoll] += -calcVelocity[2]*cycletime;
        //    bikeOutput.angles[JointData::RAnkleRoll] -= calcVelocity[2]*cycletime;
            bikeOutput.angles[JointData::RHipRoll] += calcVelocity[2]*cycletime;
          }
          else if (toRightSupport)
          {
            lastGyroLeft = lastGyroRight;
            //y
            bikeOutput.angles[JointData::RHipPitch] += calcVelocity[1]*cycletime;
            bikeOutput.angles[JointData::LHipPitch] += calcVelocity[1]*cycletime;        
            //x
//            bikeOutput.angles[JointData::LAnkleRoll] -= -calcVelocity[3]*cycletime;
            bikeOutput.angles[JointData::LHipRoll] += calcVelocity[3]*cycletime;
//            bikeOutput.angles[JointData::RAnkleRoll] -= calcVelocity[3]*cycletime;
            bikeOutput.angles[JointData::RHipRoll] += -calcVelocity[3]*cycletime;
          }
          else
          {
            //y
            bikeOutput.angles[JointData::LHipPitch] += calcVelocity[0]*cycletime;
            bikeOutput.angles[JointData::RHipPitch] += calcVelocity[1]*cycletime;

            //x
 //           bikeOutput.angles[JointData::LAnkleRoll] -= -calcVelocity[2]*cycletime;
            bikeOutput.angles[JointData::LHipRoll] += -calcVelocity[2]*cycletime;
 //           bikeOutput.angles[JointData::RAnkleRoll] -= calcVelocity[3]*cycletime;
            bikeOutput.angles[JointData::RHipRoll] += calcVelocity[3]*cycletime;

           // OUTPUT(idText, text, "double");
          }
        }

        gyroErrorLeft+=lastGyroLeft;
        gyroErrorRight+=lastGyroRight;
      }  

      lastBalancedJointRequest = balancedJointRequest;
    }
  }
  //Plot com stabilizing
  PLOT("module:Bike:comy",robotModel.centerOfMass.y);
  PLOT("module:Bike:comCorry", toDegrees(comOffset.x));
  PLOT("module:Bike:refy",ref.y);

  PLOT("module:Bike:comx", robotModel.centerOfMass.x);
  PLOT("module:Bike:comCorrx", -balanceSum.x);
  PLOT("module:Bike:refx",ref.x);

  //    PLOT("module:Bike:comy",toDegrees(bikeOutput.angles[JointData::LHipPitch]));
  //    PLOT("module:Bike:comCorry",toDegrees(gyro.y));
  //    PLOT("module:Bike:refy",toDegrees(lastGyroLeft.y));

//     PLOT("module:Bike:comx", toDegrees(bikeOutput.angles[JointData::RHipRoll]));
//     PLOT("module:Bike:comCorrx", toDegrees(gyro.x));
//     PLOT("module:Bike:refx",toDegrees(lastGyroLeft.x));



  wasActive = theMotionSelection.ratios[MotionRequest::bike] != 0.f;
}


void BIKE::calcJoints(const Vector3<>& rFootPos, const Vector3<>& lFootPos, const Vector3<>& lFootRotAng,
                      const Vector3<>& rFootRotAng, const Vector3<>& rightArm, const Vector3<>& leftArm, 
                      const Vector2<>& body, JointRequest& jointRequest)
{

  //Calculate Legs
  calcLegJoints(lFootPos, lFootRotAng, JointData::LHipYawPitch, body.x, jointRequest);
  calcLegJoints(rFootPos, rFootRotAng, JointData::RHipYawPitch, body.x, jointRequest);

  jointRequest.angles[JointData::LHipRoll] -= body.x;
  jointRequest.angles[JointData::RHipRoll] += body.x;

  jointRequest.angles[JointData::RHipPitch] += body.y; 
  jointRequest.angles[JointData::LHipPitch] +=  body.y; 

  calcArmJoints(rightArm, JointData::RShoulderPitch, jointRequest);
  calcArmJoints(leftArm, JointData::LShoulderPitch, jointRequest);
}

void BIKE::calcLegJoints(const Vector3<>& footPos, const Vector3<>& footRotAng, const JointData::Joint& joint, float bodyRoll, JointRequest& jointRequest)
{
  float sign = joint == JointData::LHipYawPitch ? 1.f : -1.f;
  float leg0,leg1,leg2,leg3,leg4,leg5;


  RotationMatrix rotateBodyTilt = RotationMatrix().rotateX(bodyRoll);
  Vector3<> anklePos = rotateBodyTilt*Vector3<>(footPos.x, footPos.y,footPos.z);
  //we need just the leg length x and y have to stay the same
  anklePos.y = footPos.y;
  anklePos.x = footPos.x;
  // for the translation of the footpos we only need to translate the anklepoint, which is the intersection of the axis leg4 and leg5
  // the rotation of the foot will be made later by rotating the footpos around the anklepoint
  anklePos -= Vector3<>(0.f,  0.f, -theRobotDimensions.heightLeg5Joint);

  anklePos -= Vector3<>(0.f, sign * (theRobotDimensions.lengthBetweenLegs / 2), 0.f);

  RotationMatrix rotateBecauseOfHip = RotationMatrix().rotateZ(sign*footRotAng.z).rotateX(-sign*pi_4);

  anklePos = rotateBecauseOfHip * anklePos;
  leg0 = footRotAng.z;

  leg2 = -atan2f(anklePos.x, Vector2<>(anklePos.y, anklePos.z).abs());

  float diagonal = anklePos.abs();

  // upperLegLength, lowerLegLength, and diagonal form a triangle, use cosine theorem
  float a1 = (theRobotDimensions.upperLegLength * theRobotDimensions.upperLegLength - 
    theRobotDimensions.lowerLegLength * theRobotDimensions.lowerLegLength + diagonal * diagonal) / 
    (2 * theRobotDimensions.upperLegLength * diagonal);
  a1 = fabs(a1) > 1.f ? 0.f : acosf(a1);

  float a2 = (theRobotDimensions.upperLegLength * theRobotDimensions.upperLegLength +
    theRobotDimensions.lowerLegLength * theRobotDimensions.lowerLegLength - diagonal * diagonal) / 
    (2 * theRobotDimensions.upperLegLength * theRobotDimensions.lowerLegLength);
  a2 = fabs(a2) > 1.f ? pi : acosf(a2);

  leg1 = -sign*atan2f(anklePos.y,  Vector2<>(anklePos.x, anklePos.z).abs());
  leg2 -= a1;
  leg3 = pi - a2;

  RotationMatrix footRot = RotationMatrix().rotateX(leg1 * -sign).rotateY(leg2 + leg3);
  footRot = footRot.invert() * rotateBecauseOfHip;

  leg5 = footRot.getXAngle() * -sign;
  leg4 = footRot.getYAngle();

  leg4 += footRotAng.y;
  leg5 += footRotAng.x;

  jointRequest.angles[joint] = leg0;
  jointRequest.angles[joint + 1] = -(pi/4) +leg1;
  jointRequest.angles[joint + 2] = leg2;
  jointRequest.angles[joint + 3] = leg3;
  jointRequest.angles[joint + 4] = leg4;
  jointRequest.angles[joint + 5] = leg5;

}

void BIKE::calcArmJoints(const Vector3<>& armPos, const JointData::Joint& joint, JointRequest& jointRequest)
{
  float sign = joint == JointData::LShoulderPitch ? 1.f : -1.f;
  float arm0,arm1,arm2=0,arm3;

  Vector3<> handPos = armPos - Vector3<>(0.f,sign * (theRobotDimensions.lengthBetweenLegs / 2.f), (theRobotDimensions.lowerArmLength + theRobotDimensions.upperArmLength));

  arm0 = atan2f(handPos.z, handPos.x);
  arm1 = atan2f(handPos.y * sign, -handPos.z);

  float diagonal = handPos.abs();

  float a1 = (theRobotDimensions.upperArmLength * theRobotDimensions.upperArmLength - 
    theRobotDimensions.lowerArmLength * theRobotDimensions.lowerArmLength + diagonal * diagonal) / 
    (2 * theRobotDimensions.upperArmLength * diagonal);
  a1 = fabs(a1) > 1.f ? 0.f : acosf(a1);

  float a2 = (theRobotDimensions.upperArmLength * theRobotDimensions.upperArmLength - diagonal * diagonal +
    theRobotDimensions.lowerArmLength * theRobotDimensions.lowerArmLength) / 
    (2 * theRobotDimensions.upperArmLength * theRobotDimensions.lowerArmLength);
  a2 = fabs(a2) > 1.f ? pi : acosf(a2);


  arm1 += a1;
  arm3 = pi - a2;

  jointRequest.angles[joint] = arm0;
  jointRequest.angles[joint + 1] = arm1;
  jointRequest.angles[joint + 2] = arm2;
  jointRequest.angles[joint + 3] = -arm3;
}

void BIKE::getCOMReference(const Vector3<>& lFootPos, const Vector3<>& rFootPos, Vector3<>& comRef, Vector2<>& origin)
{
  origin = Vector2<>(rFootPos.x+10.f,0.f);
  comRef = Vector3<>(origin.x,0.f,rFootPos.z); //double support

  if (fabs(rFootPos.y + currentParameters.footOrigin.y ) > 1.f ||
    fabs(lFootPos.y - currentParameters.footOrigin.y ) > 1.f)
  {
    if(rFootPos.y > -35.f && !rSupp && !lSupp)
    {
      origin.y = -currentParameters.footOrigin.y;
      comRef = rFootPos - Vector3<>(-10.f,0.f,0.f); //R-Support
      if (comRef.y < 0.f) comRef.y = 0.f;
      standLegPos.add(Vector2<int>((int)comRef.x, (int)comRef.y));
    }
    else if (lFootPos.y < 35. && !lSupp && !rSupp)
    {
      origin.y = currentParameters.footOrigin.y;
      comRef = lFootPos - Vector3<>(-10.f, 0.f,0.f); //L-Support
      if (comRef.y > 0.f) comRef.y = 0.f;
      standLegPos.add(Vector2<int>((int)comRef.x, (int)comRef.y));
    } 

    if (lSupp)
    {
      origin.y = currentParameters.footOrigin.y;
      comRef = Vector3<>((float)standLegPos.getAverage().x,(float)standLegPos.getAverage().y,lFootPos.z);
    }
    if (rSupp)
    {
      origin.y = -currentParameters.footOrigin.y;
      comRef = Vector3<>((float)standLegPos.getAverage().x,(float)standLegPos.getAverage().y,rFootPos.z);
    }

  }
}


void BIKE::addDynPoint(const DynPoint& dynPoint, BIKEParameters& currentParameters)
{
  Vector3 <> cubePoint, diff;

  int phaseNumber = dynPoint.phaseNumber;
  int limb =  dynPoint.limb;

  currentParameters.phaseParameters[phaseNumber].odometryOffset = dynPoint.odometryOffset;

  cubePoint = currentParameters.phaseParameters[phaseNumber].originPos[limb] + 
    currentParameters.phaseParameters[phaseNumber].controlPoints[limb][2] - 
    currentParameters.phaseParameters[phaseNumber].originOffset[limb];

  diff = cubePoint - dynPoint.translation;

  currentParameters.phaseParameters[phaseNumber].controlPoints[limb][2] -= diff;

  currentParameters.phaseParameters[phaseNumber].controlPoints[limb][1] -= diff;

  Vector3<> point1 = currentParameters.phaseParameters[phaseNumber].controlPoints[limb][2] -
    currentParameters.phaseParameters[phaseNumber].controlPoints[limb][1];

  Vector3<> absAngle;

  absAngle.x = atan2f(point1.y,point1.z);
  absAngle.y = atan2f(point1.x,point1.z);
  absAngle.z = atan2f(point1.x,point1.y);

  RotationMatrix rot = RotationMatrix().rotateX(dynPoint.angle.x-absAngle.x).
    rotateY(dynPoint.angle.y-absAngle.y).
    rotateZ(dynPoint.angle.z-absAngle.z);

  currentParameters.phaseParameters[phaseNumber].controlPoints[limb][1] = 
    (rot*point1) + currentParameters.phaseParameters[phaseNumber].controlPoints[limb][1];

  if (phaseNumber < currentParameters.numberOfPhases-1)
  {
    currentParameters.phaseParameters[phaseNumber+1].controlPoints[limb][0] = 
      currentParameters.phaseParameters[phaseNumber].controlPoints[limb][2] - 
      currentParameters.phaseParameters[phaseNumber].controlPoints[limb][1];

    float factor = (float)currentParameters.phaseParameters[phaseNumber+1].duration /  
      (float)currentParameters.phaseParameters[phaseNumber].duration;
    currentParameters.phaseParameters[phaseNumber+1].controlPoints[limb][0] *= factor; 

    currentParameters.phaseParameters[phaseNumber+1].controlPoints[limb][0] += 
      currentParameters.phaseParameters[phaseNumber].controlPoints[limb][2];
  }
}


void BIKE::calculatePreviewCom(int futPhaNum, float currPhaseTime, int preview, Vector3<> &ref, Vector2<> &origin)
{
  float lastPhaseTime = 0.f;
  float future = 0.f;

  ref=Vector3<>(0,0,0);
  origin=Vector2<>(0,0);

  for (int j=1; j<preview+1; j++)
  { 
    if (futPhaNum < currentParameters.numberOfPhases)
    {
      future = (float) (currPhaseTime+j*cycletime*1000 - lastPhaseTime) / (float) currentParameters.phaseParameters[futPhaNum].duration;

      if (future > 1.f &&  futPhaNum < currentParameters.numberOfPhases-1)
      {
        future -= 1.f;
        lastPhaseTime = j*cycletime*1000;
        currPhaseTime = 0.f;
        futPhaNum++;
        currentParameters.initPhase(futPhaNum);
      }

      if (future > 1.f && futPhaNum+1 > currentParameters.numberOfPhases-1)

      {
        if(currentParameters.loop)
        {
          future = 0.f;
          lastPhaseTime = j*cycletime*1000;
          currPhaseTime = 0.f;
          futPhaNum = 0;
        }
        else
        {
          future = 1.f;
          futPhaNum = currentParameters.numberOfPhases-1;
        }
      }
    }
    else
    {
      future = 1.f;
      futPhaNum = currentParameters.numberOfPhases-1;
    }

    Vector3<> futPos[Phase::numOfLimbs];

    for (int i=0; i < Phase::numOfLimbs; i++)
    {
      futPos[i] = currentParameters.getPosition(future, futPhaNum,i);
    }

    Vector3<> comRef(0,0,0);
    Vector2<> fOrigin(0,0);

    if (currentBikeRequest.mirror)
    {
      Vector3<> temp =  futPos[Phase::rightFootTra];
      futPos[Phase::rightFootTra] = Vector3<>( futPos[Phase::leftFootTra].x, 
        -futPos[Phase::leftFootTra].y, 
        futPos[Phase::leftFootTra].z);
      futPos[Phase::leftFootTra] = Vector3<> (temp.x, -temp.y, temp.z);
    }

    getCOMReference(futPos[Phase::leftFootTra],  futPos[Phase::rightFootTra], comRef, fOrigin);

    ref+=comRef;
    origin+=fOrigin; 

  }
  ref/= (float)preview; 
  origin/= (float)preview;
}

MAKE_MODULE(BIKE, Motion Control)
