/** 
* @file Modules/MotionControl/BIKE.h
* This file declares a module that creates the walking motions.
* @author <A href="mailto:judy@tzi.de">Judith Mller</A>
*/

#ifndef __BIKE_h_
#define __BIKE_h_

#ifdef LINUX
#include <dirent.h>
#else
#include <winsock.h>
#endif

#include "BIKEParameters.h"
#include "Tools/Module/Module.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/Configuration/JointCalibration.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/MotionControl/MotionSelection.h"
#include "Representations/MotionControl/BikeEngineOutput.h"
#include "Representations/MotionControl/MotionRequest.h"
#include "Representations/Sensing/GroundContactState.h"
#include "Tools/Math/Pose3D.h"
#include "Tools/Streams/InStreams.h"
#include "Platform/GTAssert.h"
#include "Representations/Infrastructure/SensorData.h"
#include <vector>
#include "Representations/Configuration/MassCalibration.h"
#include "Representations/Sensing/RobotModel.h"
//#include "Representations/Perception/InertiaMatrix.h"
#include "Tools/RingBufferWithSum.h"


MODULE(BIKE)
REQUIRES(RobotDimensions)
REQUIRES(JointCalibration)
REQUIRES(FrameInfo)
REQUIRES(FilteredSensorData)
REQUIRES(SensorData)
REQUIRES(MotionSelection)
REQUIRES(MotionRequest)
REQUIRES(FilteredJointData)
REQUIRES(MassCalibration)
REQUIRES(GroundContactState)
PROVIDES_WITH_MODIFY(BikeEngineOutput)
END_MODULE


class BIKE : public BIKEBase
{
private:
  std::vector<BIKEParameters> params;
  BIKEParameters currentParameters;
  BikeRequest currentBikeRequest;
  RobotModel robotModel;

  int actualBMotionID, 
    phaseNumber;

  unsigned int timeStamp;

  Vector3<> 
    positions[6], 
    lastCorrectedCom, 
    comErrorSum,
      lastRef;

  Vector2<>  balanceSum,
    gyroErrorLeft,
    gyroErrorRight,
    comOffset, 
    lastGyroLeft, 
    lastGyroRight,
    gyro; 


  JointRequest lastJointRequest, 
    lastBalancedJointRequest, 
    balancedJointRequest;

  float lastOrigin,  
    phase, 
    cycletime;

  bool toRightSupport, 
    toLeftSupport,
    rSupp,
    lSupp,
    wasActive;

  Vector2<> gyroP,
    gyroI,
    gyroD;

  RingBufferWithSum<Vector2<int>,10> standLegPos;


  void calcJoints(const Vector3<>& rFootPos, const Vector3<>& lFootPos, const Vector3<>& lFootRotAng, 
    const Vector3<>& rFootRotAng, const Vector3<>& rightArm, const Vector3<>& leftArm, 
    const Vector2<>& body, JointRequest& jointRequest);

  void calcLegJoints(const Vector3<>& footPos, const Vector3<>& footRotAng, const JointData::Joint& joint, 
    float bodyRoll, JointRequest& jointRequest);

  void calcArmJoints(const Vector3<>& armPos, const JointData::Joint& joint, JointRequest& jointRequest);

  void getCOMReference(const Vector3<>& lFootPos, const Vector3<>& rFootPos, Vector3<>& comRef, Vector2<>& origin);

    void calculatePreviewCom(int futPhaNum, float currPhaseTime, int preview, Vector3<> &ref, Vector2<> &origin);

  int getPosOfRequestedBMotion()
  {
    int actualId = -1;
    for (unsigned int i = 0; i < params.size(); i++)
      if (theMotionRequest.bikeRequest.getBMotionFromName(&params[i].name[0]) == theMotionRequest.bikeRequest.bMotionType) 
        actualId = i;

    return actualId;   
  }

  void addDynPoint(const DynPoint& dynPoint, BIKEParameters& currentParameters);

public:

  void update(BikeEngineOutput& BIKEOutput);

  /**
  * Default constructor.
  */
  BIKE():
  actualBMotionID(-1),
    phaseNumber(0),
    timeStamp(0),
    lastCorrectedCom(0.f, 0.f, 0.f), 
    comErrorSum(0.f, 0.f, 0.f),
    lastRef(0.f, 0.f, 0.f), 
    balanceSum(0.f, 0.f),  
    gyroErrorLeft(0.f, 0.f),
    gyroErrorRight(0.f, 0.f),
    comOffset(0.f, 0.f), 
    lastGyroLeft(0.f, 0.f), 
    lastGyroRight(0.f, 0.f),
    gyro(0.f, 0.f),
    lastOrigin(0.f), 
    phase(0.f), 
    cycletime(0.f),
    toRightSupport(false), 
    toLeftSupport(false),
    rSupp(false), 
    lSupp(false),
    wasActive(false),

    //Parameter for P, I and D for gyro PID Contol
    gyroP(-2.f,-2.f),
    gyroI(0.1f,0.f),
    gyroD(0.01f,0.01f)
  {
    params.reserve(10);

    char dirname[260]; 


#ifndef LINUX
    sprintf(dirname, "%s/Config/Bike/*.bmc", File::getGTDir());
    WIN32_FIND_DATA findFileData;
    std::string fileName;
    HANDLE hFind = FindFirstFile(dirname, &findFileData);


    while(hFind != INVALID_HANDLE_VALUE)
    {
      char name[512];

      fileName = findFileData.cFileName;    

      sprintf(name, "Bike/%s", fileName.c_str());


      InConfigFile stream(name);
      ASSERT(stream.exists());

      BIKEParameters parameters;
      stream >> parameters;

      char temp[260];

      sprintf(temp, "%s", fileName.c_str());


      for (int i=0; i<260; i++)
      {
        if (temp[i]== '.') temp[i]=0;
      }

      strcpy(parameters.name,temp);

      params.push_back(parameters);

      if(!FindNextFile(hFind, &findFileData))break;
    }

#else //LINUX
    sprintf(dirname, "%s/Config/Bike/", File::getGTDir());
    DIR * dir = opendir(dirname);
    ASSERT(dir);
    struct dirent *file = readdir(dir);

    while(file != NULL)
    {
      char name[260];
      sprintf(name, "Bike/%s", file->d_name);

      if(strstr(name, ".bmc"))
      {
        InConfigFile stream(name);
        ASSERT(stream.exists());

        BIKEParameters parameters;
        stream >> parameters;

        sprintf(name, "%s", file->d_name);
        for (int i=0; i<260; i++)
        {
          if (name[i]== '.') name[i]=0;
        }
        strcpy(parameters.name,name);
        params.push_back(parameters);
      }
      file = readdir(dir);
    }
    closedir(dir);


#endif //LINUX
    //This is needed for adding new kicks 
#ifndef RELEASE
    BIKEParameters newBMotion;
    strcpy(newBMotion.name, "noName");
    params.push_back(newBMotion);
#endif

  };

  ~BIKE()
  {
  }
};




#endif // __BIKE_h_
