/**
* @file BIKEParameters.cpp
* @author <a href="mailto:judy@informatik.uni-bremen.de">Judith Mller</a>
*/

#include "BIKEParameters.h"
#include "Representations/MotionControl/BikeRequest.h"
#include "BIKE.h"

void Phase::serialize(In* in, Out* out)
{
  STREAM_REGISTER_BEGIN(); 

  STREAM(duration);

  STREAM_ARRAY(controlPoints[Phase::leftFootTra]);
  STREAM_ARRAY(controlPoints[Phase::leftFootRot]);

  STREAM_ARRAY(controlPoints[Phase::rightFootTra]);
  STREAM_ARRAY(controlPoints[Phase::rightFootRot]);

  STREAM_ARRAY(controlPoints[Phase::leftArmTra]);
  STREAM_ARRAY(controlPoints[Phase::rightArmTra]);

  STREAM(odometryOffset);

  STREAM_REGISTER_FINISH();    
}

void BIKEParameters::serialize(In* in, Out* out)
{
  STREAM_REGISTER_BEGIN(); 

  STREAM(footOrigin);
  STREAM(armOrigin);

  STREAM(kpx);
  STREAM(kix);
  STREAM(kdx);

  STREAM(kpy);
  STREAM(kiy);
  STREAM(kdy);

  STREAM(preview)

  STREAM(loop);

  if(in)
  {
    phaseParameters.clear();
  }
  STREAM_VECTOR(phaseParameters);
  numberOfPhases = phaseParameters.size();

  STREAM_ENUM(supportLegMode, numOfSupportLegModes, getSupportLegModeName);

  STREAM_REGISTER_FINISH();    
}

Vector3<> BIKEParameters::getPosition(float phase, int phaseNumber, int limb)
{
  Vector3<> p(0.f,0.f,0.f);

  //StartPoint is the endingPoint of the last phase or origin
  //p += p*(1.0-phase);
  Vector3<> offset = phaseParameters[phaseNumber].originOffset[limb];
  for (unsigned i = 1; i < NUM_OF_POINTS+1; i++)
  {
    p += (phaseParameters[phaseNumber].controlPoints[limb][i-1]-offset)*(BikeExtraMath::polyKo((float)NUM_OF_POINTS, (float)i) * pow(phase,(float)i) * pow((1.f-phase),((float)NUM_OF_POINTS-i)));
  }
  return phaseParameters[phaseNumber].originPos[limb] + p;
}

void BIKEParameters::initFirstPhase()
{
  phaseParameters[0].originPos[Phase::leftFootTra] = Vector3<> (this->footOrigin.x, this->footOrigin.y, this->footOrigin.z);
  phaseParameters[0].originPos[Phase::rightFootTra] = Vector3<>(this->footOrigin.x, -this->footOrigin.y, this->footOrigin.z);

  phaseParameters[0].originPos[Phase::leftFootRot] = Vector3<>(0.f,0.f,0.f);
  phaseParameters[0].originPos[Phase::rightFootRot] = Vector3<>(0.f,0.f,0.f);

  phaseParameters[0].originPos[Phase::leftArmTra] = Vector3<> (this->armOrigin.x, this->armOrigin.y, this->armOrigin.z);
  phaseParameters[0].originPos[Phase::rightArmTra]= Vector3<>(this->armOrigin.x, -this->armOrigin.y, this->armOrigin.z);


  //set the Offset for the first Phase to zero, because all calculations based on the startOrigin

  for (int j=0; j < Phase::numOfLimbs; j++)
  {
    phaseParameters[0].originOffset[j] = Vector3<>(0.f,0.f,0.f);
  }
}

void BIKEParameters::initFirstPhaseLoop(const Phase& lastParameters)
{
  for (int j=0; j < Phase::numOfLimbs; j++)
  {
    phaseParameters[0].originPos[j] = lastParameters.originPos[j] + 
      lastParameters.controlPoints[j][NUM_OF_POINTS-1]
    - lastParameters.originOffset[j];
  }

  phaseParameters[0].originOffset[Phase::leftFootTra] = phaseParameters[0].originPos[Phase::leftFootTra] - footOrigin;
  phaseParameters[0].originOffset[Phase::rightFootTra] = phaseParameters[0].originPos[Phase::rightFootTra] - Vector3<> (footOrigin.x, -footOrigin.y, footOrigin.z);

  phaseParameters[0].originOffset[Phase::leftFootRot] =  phaseParameters[0].originPos[Phase::leftFootRot];
  phaseParameters[0].originOffset[Phase::rightFootRot] =  phaseParameters[0].originPos[Phase::rightFootRot];

  phaseParameters[0].originOffset[Phase::leftArmTra] = phaseParameters[0].originPos[Phase::leftArmTra] - armOrigin;
  phaseParameters[0].originOffset[Phase::rightArmTra] = phaseParameters[0].originPos[Phase::rightArmTra] - Vector3<> (armOrigin.x, -armOrigin.y, armOrigin.z);
}


void BIKEParameters::initPhase(int i)
{
  if(i > 0)
  {
    for (int j=0; j < Phase::numOfLimbs; j++)
    {
      phaseParameters[i].originPos[j] = phaseParameters[i-1].originPos[j] + 
        phaseParameters[i-1].controlPoints[j][NUM_OF_POINTS-1]
      - phaseParameters[i-1].originOffset[j];
    }

    phaseParameters[i].originOffset[Phase::leftFootTra] = phaseParameters[i].originPos[Phase::leftFootTra] - footOrigin;
    phaseParameters[i].originOffset[Phase::rightFootTra] = phaseParameters[i].originPos[Phase::rightFootTra] - Vector3<> (footOrigin.x, -footOrigin.y, footOrigin.z);

    phaseParameters[i].originOffset[Phase::leftFootRot] =  phaseParameters[i].originPos[Phase::leftFootRot];
    phaseParameters[i].originOffset[Phase::rightFootRot] =  phaseParameters[i].originPos[Phase::rightFootRot];

    phaseParameters[i].originOffset[Phase::leftArmTra] = phaseParameters[i].originPos[Phase::leftArmTra] - armOrigin;
    phaseParameters[i].originOffset[Phase::rightArmTra] = phaseParameters[i].originPos[Phase::rightArmTra] - Vector3<> (armOrigin.x, -armOrigin.y, armOrigin.z);
  }
}

void BIKEParameters::init()
{
  if(numberOfPhases > -1)
  {
    for (int i = 0; i < numberOfPhases; i++)
    {
      if(i == 0)
      {
        phaseParameters[i].originPos[Phase::leftFootTra] = Vector3<> (this->footOrigin.x, this->footOrigin.y, this->footOrigin.z);
        phaseParameters[i].originPos[Phase::rightFootTra] = Vector3<>(this->footOrigin.x, -this->footOrigin.y, this->footOrigin.z);

        phaseParameters[i].originPos[Phase::leftFootRot] = Vector3<>(0.f,0.f,0.f);
        phaseParameters[i].originPos[Phase::rightFootRot] = Vector3<>(0.f,0.f,0.f);

        phaseParameters[i].originPos[Phase::leftArmTra] = Vector3<> (this->armOrigin.x, this->armOrigin.y, this->armOrigin.z);
        phaseParameters[i].originPos[Phase::rightArmTra]= Vector3<>(this->armOrigin.x, -this->armOrigin.y, this->armOrigin.z);


        //set the Offset for the first Phase to zero, because all calculations based on the startOrigin

        for (int j=0; j < Phase::numOfLimbs; j++)
        {
          phaseParameters[i].originOffset[j] = Vector3<>(0.f,0.f,0.f);
        }
      }
      else
      {
        for (int j=0; j < Phase::numOfLimbs; j++)
        {
          phaseParameters[i].originPos[j] = phaseParameters[i-1].originPos[j] + 
            phaseParameters[i-1].controlPoints[j][NUM_OF_POINTS-1]
          - phaseParameters[i-1].originOffset[j];
        }

        //set the Offset for the first Phase to zero, because all calculations based on the startOrigin
        phaseParameters[i].originOffset[Phase::leftFootTra] = phaseParameters[i].originPos[Phase::leftFootTra] - footOrigin;
        phaseParameters[i].originOffset[Phase::rightFootTra] = phaseParameters[i].originPos[Phase::rightFootTra] - Vector3<> (footOrigin.x, -footOrigin.y, footOrigin.z);

        phaseParameters[i].originOffset[Phase::leftFootRot] =  phaseParameters[i].originPos[Phase::leftFootRot];
        phaseParameters[i].originOffset[Phase::rightFootRot] =  phaseParameters[i].originPos[Phase::rightFootRot];

        phaseParameters[i].originOffset[Phase::leftArmTra] = phaseParameters[i].originPos[Phase::leftArmTra] - armOrigin;
        phaseParameters[i].originOffset[Phase::rightArmTra] = phaseParameters[i].originPos[Phase::rightArmTra] - Vector3<> (armOrigin.x, -armOrigin.y, armOrigin.z);
      }
    }
  }
}
