/**
* @file HeadMotionEngine.cpp
* This file implements a module that creates head joint angles from desired head motion.
* @author <a href="allli@informatik.uni-bremen.de">Alexander Hrtl</a>
* @author Colin Graf
*/

#include "HeadMotionEngine.h"
#include "Tools/Math/Common.h"
#include "Tools/Range.h"
#include "Tools/Debugging/DebugDrawings.h" // PLOT

MAKE_MODULE(HeadMotionEngine, Motion Control)

void HeadMotionEngine::init()
{
  lastSpeed = Vector2<>();
  lastIteration = theFrameInfo.time - (unsigned int)(theRobotDimensions.motionCycleTime * 1000.f);

  deathPoints[0] = Geometry::getCircle(
    Vector2<int>(45, -25), Vector2<int>(120, -20), Vector2<int>(90, -17));
  deathPoints[1] = Geometry::getCircle(
    Vector2<int>(-45, -25), Vector2<int>(-120, -20), Vector2<int>(-90, -17));
  deathPoints[2] = Geometry::getCircle(
    Vector2<int>(17, 46), Vector2<int>(85, 23), Vector2<int>(120, 27));
  deathPoints[3] = Geometry::getCircle(
    Vector2<int>(-17, 46), Vector2<int>(-85, 23), Vector2<int>(-120, 27));

  for(int i = 0; i < 4; ++i)
  {
    deathPoints[i].center.x = fromDegrees(deathPoints[i].center.x);
    deathPoints[i].center.y = fromDegrees(deathPoints[i].center.y);
    deathPoints[i].radius = fromDegrees(deathPoints[i].radius) * 1.015f;
  }
}

void HeadMotionEngine::update(HeadJointRequest& headJointRequest)
{
  switch(theHeadMotionRequest.mode)
  {
    case HeadMotionRequest::panTiltMode:
      updatePanTiltMode(headJointRequest, theHeadMotionRequest.pan, theHeadMotionRequest.tilt);
      break;
    case HeadMotionRequest::targetMode:
      updateTargetMode(headJointRequest);
      break;
    case HeadMotionRequest::targetOnGroundMode:
      updateTargetOnGroundMode(headJointRequest);
      break;
    default:
      ASSERT(false);
  }
}

void HeadMotionEngine::updateTargetMode(HeadJointRequest& headJointRequest)
{
  if(theHeadMotionRequest.target != lastTarget)
  {
    const Vector3<> target = theHeadMotionRequest.target;
    const float x = sqrtf(sqr(target.x) + sqr(target.y)),
                 y = target.z - theRobotDimensions.zLegJoint1ToHeadPan;

    const float cx = theRobotDimensions.xHeadTiltToCamera;
    const float cy = theRobotDimensions.zHeadTiltToCamera;
    const float c = sqrtf(sqr(cx) + sqr(cy));
    const float d  = sqrtf(sqr(x) + sqr(y));

    const float alpha = pi_2 - atan2f(cy, cx);
    const float mue = atan2f(y,x) + pi_2;
    const float epsilon = pi_2 - alpha;
    const float gamma = pi_2 + (pi_2 - epsilon - theRobotDimensions.headTiltToCameraTilt);

    const float delta = asinf(c*sinf(gamma)/d);

    const float beta = pi - gamma - delta;

    targetPanTilt.x = atan2f(target.y, target.x),
    targetPanTilt.y = - (pi - beta - alpha - mue);
    lastTarget = theHeadMotionRequest.target;
  }

  updatePanTiltMode(headJointRequest, targetPanTilt.x, targetPanTilt.y);
}

void HeadMotionEngine::updateTargetOnGroundMode(HeadJointRequest& headJointRequest)
{
  if(theHeadMotionRequest.target != lastOnGroundTarget)
  {
    float targetAbs = Vector2<>(theHeadMotionRequest.target.x, theHeadMotionRequest.target.y).abs();
    if(targetAbs < 50.f) // clip distances close to and inside of the robot
      targetAbs = 50.f;

    // yes, i know this calculation is imprecise... (but fast)
    Pose3D simplifiedCameraMatrix = theTorsoMatrix;
    simplifiedCameraMatrix.translate(theRobotDimensions.xHeadTiltToCamera, 0.f, theRobotDimensions.zLegJoint1ToHeadPan + theRobotDimensions.zHeadTiltToCamera);
    float beta = atan2f(simplifiedCameraMatrix.translation.z - theHeadMotionRequest.target.z, targetAbs - theRobotDimensions.xHeadTiltToCamera);
    onGroundTargetPanTilt.y = -beta + theRobotDimensions.headTiltToCameraTilt;
    onGroundTargetPanTilt.x = atan2f(theHeadMotionRequest.target.y, theHeadMotionRequest.target.x);
    lastOnGroundTarget = theHeadMotionRequest.target;
  }
  updatePanTiltMode(headJointRequest, onGroundTargetPanTilt.x, onGroundTargetPanTilt.y);
}

void HeadMotionEngine::updatePanTiltMode(HeadJointRequest& headJointRequest, float unclippedPan, float unclippedTilt)
{
  // arbitrary value that seems to be good...
  float maxAcc = 7.f;
  MODIFY("module:HeadMotionEngine:maxAcceleration", maxAcc);

  float pan = unclippedPan == JointData::off ? JointData::off : Range<>(theJointCalibration.joints[JointData::HeadYaw].minAngle, theJointCalibration.joints[JointData::HeadYaw].maxAngle).limit(unclippedPan);
  float tilt = unclippedTilt == JointData::off ? JointData::off : Range<>(theJointCalibration.joints[JointData::HeadPitch].minAngle, theJointCalibration.joints[JointData::HeadPitch].maxAngle).limit(unclippedTilt);

  const float deltaTime = float(theFrameInfo.time - lastIteration) * 0.001f;
  const Vector2<> position(headJointRequest.pan == JointData::off ? theFilteredJointData.angles[JointData::HeadYaw] : headJointRequest.pan,
    headJointRequest.tilt == JointData::off ? theFilteredJointData.angles[JointData::HeadPitch] : headJointRequest.tilt);
  const Vector2<> target(pan == JointData::off ? 0 : pan, tilt == JointData::off ? 0 : tilt);
  Vector2<> offset(target - position);
  const float distanceToTarget = offset.abs();
  
  // calculate max speed
  const float maxSpeedForDistance = sqrtf(2.f * distanceToTarget * maxAcc * 0.8f);
  const float maxSpeed = std::min(maxSpeedForDistance, theHeadMotionRequest.speed);

  // max speed clipping
  if(distanceToTarget / deltaTime > maxSpeed)
    offset *= maxSpeed * deltaTime / distanceToTarget; //<=> offset.normalize(maxSpeed * deltaTime);

  // max acceleration clipping
  Vector2<> speed(offset / deltaTime);
  Vector2<> acc((speed - lastSpeed) / deltaTime);
  const float accSquareAbs = acc.squareAbs();
  if(accSquareAbs > maxAcc * maxAcc)
  {
    acc *= maxAcc * deltaTime / sqrt(accSquareAbs);
    speed = acc + lastSpeed;
    offset = speed * deltaTime;
  }
  /* <=>
  Vector2<> speed(offset / deltaTime);
  Vector2<> acc((speed - lastSpeed) / deltaTime);
  if(acc.squareAbs() > maxAcc * maxAcc)
  {
    speed = acc.normalize(maxAcc * deltaTime) + lastSpeed;
    offset = speed * deltaTime;
  }
  */
  PLOT("module:HeadMotionEngine:speed", toDegrees(speed.abs()));

  // calculate new position
  Vector2<> newPosition(position + offset);

  // make sure we don't get to close to the evil points of death
  if(pan != JointData::off && tilt != JointData::off)
    for(int i = 0; i < 4; ++i)
    {
      Vector2<> deathPointToPosition(newPosition - deathPoints[i].center);
      const float deathPointToPositionSquareAbs = deathPointToPosition.squareAbs();
      if(deathPointToPositionSquareAbs < sqr(deathPoints[i].radius))
      {
        const float deathPointToPositionAbs = sqrt(deathPointToPositionSquareAbs);
        deathPointToPosition *= (deathPoints[i].radius - deathPointToPositionAbs) / deathPointToPositionAbs;
        newPosition += deathPointToPosition;
      }
    }
  
  // set new position
  headJointRequest.pan = pan == JointData::off ? JointData::off : newPosition.x;
  headJointRequest.tilt = tilt == JointData::off ? JointData::off : newPosition.y;
  headJointRequest.moving = pan != JointData::off && tilt != JointData::off && ((newPosition - position) / deltaTime).squareAbs() > sqr(maxAcc * deltaTime * 0.5f);

  // check reachability
  headJointRequest.reachable = true;
  if(pan != unclippedPan || tilt != unclippedTilt)
    headJointRequest.reachable = false;
  else
    for(int i = 0; i < 4; ++i)
      if((target - deathPoints[i].center).squareAbs() < sqr(deathPoints[i].radius))
        headJointRequest.reachable = false;
  
  // store some values for the next iteration
  lastSpeed = speed;
  lastIteration = theFrameInfo.time;
}
