/** 
* @file Modules/MotionControl/MotionCombinator.cpp
* This file implements a module that combines the motions created by the different modules.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#include "MotionCombinator.h"
#include "Tools/Debugging/DebugDrawings.h"

MotionCombinator::MotionCombinator()
{
  InConfigFile stream(Global::getSettings().expandLocationFilename("motionCombinator.cfg"), "default");
  if(stream.exists())
    stream >> parameters;
  else
  {
    InConfigFile stream("motionCombinator.cfg", "default");
    ASSERT(stream.exists());
    stream >> parameters;
  }

  recoveryTime = parameters.recoveryTime + 1;
  headJawInSavePosition = false;
  headPitchInSavePosition = false; 
}

void MotionCombinator::update(UnstableJointRequest& jointRequest)
{
  MODIFY("parameters:MotionCombinator", parameters);

  const JointRequest* jointRequests[MotionRequest::numOfMotions] = 
  {
    &theWalkingEngineOutput,
    &theBikeEngineOutput,
    &theSpecialActionsOutput,
    &theWalkingEngineStandOutput,
    &theSIKEngineOutput
  };

  jointRequest.angles[JointData::HeadYaw] = theHeadJointRequest.pan;
  jointRequest.angles[JointData::HeadPitch] = theHeadJointRequest.tilt;  

  copy(*jointRequests[theMotionSelection.targetMotion], jointRequest);

  int i;
  for(i = 0; i < MotionRequest::numOfMotions; ++i)
    if(theMotionSelection.ratios[i] == 1.f)
    {
      // default values
      motionInfo.motion = MotionRequest::Motion(i);
      motionInfo.isMotionStable = true;

      lastJointData = theFilteredJointData;

      if(theMotionSelection.ratios[MotionRequest::walk] == 1.f)
      {
        odometryData += theWalkingEngineOutput.odometryOffset;
        motionInfo.walkRequest = theWalkingEngineOutput.executedWalk;
      }
      else if(theMotionSelection.ratios[MotionRequest::bike] == 1.f)
      {
        odometryData += theBikeEngineOutput.odometryOffset;
        motionInfo.bikeRequest = theBikeEngineOutput.executedBikeRequest;
        motionInfo.isMotionStable = true;
      }
      else if(theMotionSelection.ratios[MotionRequest::sike] == 1.f)
      {
        motionInfo.sikeRequest.motion = theSIKEngineOutput.executedMotion;
        motionInfo.isMotionStable = false;
      }
      else if(theMotionSelection.ratios[MotionRequest::specialAction] == 1.)
      {
        odometryData += theSpecialActionsOutput.odometryOffset;
        motionInfo.specialActionRequest = theSpecialActionsOutput.executedSpecialAction;
        motionInfo.isMotionStable = theSpecialActionsOutput.isMotionStable;
      }
      else if(theMotionSelection.ratios[MotionRequest::stand] == 1.f)
      {
        motionInfo.motion = MotionRequest::stand;
      }
      break;
    }

    if(i == MotionRequest::numOfMotions)
    {
      for(i = 0; i < MotionRequest::numOfMotions; ++i)
        if(i != theMotionSelection.targetMotion && theMotionSelection.ratios[i] > 0.)
        {
          bool interpolateHardness = !(i == MotionRequest::specialAction && theMotionSelection.specialActionRequest.specialAction == SpecialActionRequest::playDead); // do not interpolate from play_dead
          interpolate(*jointRequests[i], *jointRequests[theMotionSelection.targetMotion], theMotionSelection.ratios[i], jointRequest, interpolateHardness);
        }    
    }

#ifndef RELEASE
    JointDataDeg jointRequestDeg(jointRequest);
#endif
    MODIFY("representation:JointRequestDeg", jointRequestDeg);  
}

void MotionCombinator::update(JointRequest& jointRequest)
{
  jointRequest = theUnstableJointRequest;

  if (parameters.emergencyOffEnabled && theFallDownState.state == FallDownState::falling){
    saveFall(jointRequest);
    centerHead(jointRequest);
    recoveryTime = 0;
  }
  else if (parameters.emergencyOffEnabled
    && theFallDownState.state != FallDownState::upright
    && theFallDownState.state != FallDownState::undefined){
    centerHead(jointRequest);
  }
  else 
  {
    if (parameters.emergencyOffEnabled && theFallDownState.state == FallDownState::upright){
      headJawInSavePosition = false;
      headPitchInSavePosition = false;
    }

    if (parameters.emergencyOffEnabled && recoveryTime < parameters.recoveryTime){
      recoveryTime += 1;
      float ratio = (1.f/float(parameters.recoveryTime))*recoveryTime;
      for (int i = 0; i < JointData::numOfJoints; i ++){       
        jointRequest.jointHardness.hardness[i] = 30 + int (ratio * float(jointRequest.jointHardness.hardness[i] - 30));
      }     
    }
  }

#ifndef RELEASE
  float sum(0);
  int count(0);
  for(int i = JointData::LHipYawPitch; i < JointData::numOfJoints; i++)
  {
    if(jointRequest.angles[i] != JointData::off && jointRequest.angles[i] != JointData::ignore && lastJointRequest.angles[i] != JointData::off && lastJointRequest.angles[i] != JointData::ignore)
    {
      sum += fabs(jointRequest.angles[i] - lastJointRequest.angles[i]);
      count++;
    }
  }
  PLOT("module:MotionCombinator:deviations:JointRequest:legsOnly", sum / count);
  for(int i = 0; i < JointData::LHipYawPitch; i++)
  {
    if(jointRequest.angles[i] != JointData::off && jointRequest.angles[i] != JointData::ignore && lastJointRequest.angles[i] != JointData::off && lastJointRequest.angles[i] != JointData::ignore)
    {
      sum += fabs(jointRequest.angles[i] - lastJointRequest.angles[i]);
      count++;
    }
  }
  PLOT("module:MotionCombinator:deviations:JointRequest:all", sum / count);

  sum = 0;
  count = 0;
  for(int i = JointData::LHipYawPitch; i < JointData::numOfJoints; i++)
  {
    if(lastJointRequest.angles[i] != JointData::off && lastJointRequest.angles[i] != JointData::ignore)
    {
      sum += fabs(lastJointRequest.angles[i] - theFilteredJointData.angles[i]);
      count++;
    }
  }
  PLOT("module:MotionCombinator:differenceToJointData:legsOnly", sum / count);

  for(int i = 0; i < JointData::LHipYawPitch; i++)
  {
    if(lastJointRequest.angles[i] != JointData::off && lastJointRequest.angles[i] != JointData::ignore)
    {
      sum += fabs(lastJointRequest.angles[i] - theFilteredJointData.angles[i]);
      count++;
    }
  }
  lastJointRequest = jointRequest;
  PLOT("module:MotionCombinator:differenceToJointData:all", sum / count);
#endif    
}

void MotionCombinator::copy(const JointRequest& source, JointRequest& target) const
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    if(source.angles[i] != JointData::ignore)
      target.angles[i] = source.angles[i];
    target.jointHardness.hardness[i] = source.angles[i] != JointData::off ? source.jointHardness.getHardness(i) : 0;
  }
}

void MotionCombinator::interpolate(const JointRequest& from, const JointRequest& to,
                                   float fromRatio, JointRequest& target, bool interpolateHardness) const
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    float f = from.angles[i];
    float t = to.angles[i];

    if(t == JointData::ignore && f == JointData::ignore)
      continue;

    if(t == JointData::ignore)
      t = target.angles[i];
    if(f == JointData::ignore)
      f = target.angles[i];

    const float fHardness = f != JointData::off ? (float) from.jointHardness.getHardness(i) : 0;
    const float tHardness = t != JointData::off ? (float) to.jointHardness.getHardness(i) : 0;

    if(t == JointData::off || t == JointData::ignore)
      t = lastJointData.angles[i];
    if(f == JointData::off || f == JointData::ignore)
      f = lastJointData.angles[i];
    if(target.angles[i] == JointData::off || target.angles[i] == JointData::ignore)
      target.angles[i] = lastJointData.angles[i];

    ASSERT(target.angles[i] != JointData::off && target.angles[i] != JointData::ignore);
    ASSERT(t != JointData::off && t != JointData::ignore);
    ASSERT(f != JointData::off && f != JointData::ignore);

    target.angles[i] += -fromRatio * t + fromRatio * f;
    if(interpolateHardness)
      target.jointHardness.hardness[i] += int(-fromRatio * tHardness + fromRatio * fHardness);
    else
      target.jointHardness.hardness[i] = int(tHardness);
  }
}

void MotionCombinator::update(OdometryData& odometryData)
{
  odometryData = this->odometryData;

#ifndef RELEASE
  Pose2D odometryOffset(odometryData);
  odometryOffset -= lastOdometryData;
  PLOT("module:MotionCombinator:odometryOffsetX", odometryOffset.translation.x);
  PLOT("module:MotionCombinator:odometryOffsetY", odometryOffset.translation.y);
  PLOT("module:MotionCombinator:odometryOffsetRotation", toDegrees(odometryOffset.rotation));
  lastOdometryData = odometryData;
#endif
}

void MotionCombinator::saveFall(JointRequest &jointRequest)
{
  for (int i = 0; i < JointData::numOfJoints; i++)
    jointRequest.jointHardness.hardness[i] = 30;    
}

void MotionCombinator::centerHead(JointRequest &jointRequest)
{
  jointRequest.angles[JointData::HeadYaw] = 0;
  jointRequest.angles[JointData::HeadPitch] = 0;
  if (fabs(theFilteredJointData.angles[JointData::HeadYaw]) > 0.1 && !headJawInSavePosition)
    jointRequest.jointHardness.hardness[JointData::HeadYaw] = 100;    
  else{
    headJawInSavePosition = true;
    jointRequest.jointHardness.hardness[JointData::HeadYaw] = 30;    
  }
  if (fabs(theFilteredJointData.angles[JointData::HeadPitch]) > 0.1 && !headPitchInSavePosition)    
    jointRequest.jointHardness.hardness[JointData::HeadPitch] = 100;
  else{
    headPitchInSavePosition = true;
    jointRequest.jointHardness.hardness[JointData::HeadPitch] = 30;    
  }
}

MAKE_MODULE(MotionCombinator, Motion Control)
