/**
* @file SIKEMotion.cpp
* author jeff
*/

#include "SIKEMotion.h"

#include "Representations/Configuration/MassCalibration.h"
#include "Tools/InverseKinematic.h"

bool SIKEMotion::getMotionAt(unsigned timeSinceMotionStarted, bool mirror, const KeyFrame &startFrame, KeyFrame &result) const
{
  //find last and next frame
  unsigned lastFrameTime = 0;
  const KeyFrame *nextFrame = NULL,
                 *lastFrame = NULL;

  for(size_t i = 0; i < frames.size(); i++)
  {
    const KeyFrame &f = frames.at(i);
    if(lastFrameTime + f.time >= timeSinceMotionStarted)
    {
      nextFrame = &f;
      break;
    }
    lastFrameTime += f.time;
    lastFrame = &f;
  }

  //mirror and use startFrame if necessary
  KeyFrame last, next;
  if(nextFrame)
    next = nextFrame->mirrored(mirror);

  if(lastFrame)
    last = lastFrame->mirrored(mirror);
  else //lastFrame == NULL -> interpolate from startFrame to nextFrame
    last = startFrame;


  //interpolate between frames if necessary
  if(!nextFrame) //motion done, hold last frame
  {
    result = last;
    return true;
  }
  else //interpolate between last and next
  {
    const unsigned timeSinceLastFrame = timeSinceMotionStarted - lastFrameTime,
                   timeBetweenLastAndNext = nextFrame->time;

    const float ratio = timeBetweenLastAndNext == 0 ? 1 : timeSinceLastFrame / float(timeBetweenLastAndNext);

    result = KeyFrame::interpolateBetweenFrames(last, next, ratio);
    return false;
  }
}

bool SIKEMotion::loadFromFile(std::string filename, const RobotDimensions &theRobotDimensions)
{
  InConfigFile stream("SIKE/" + filename + ".sike");
  if(!stream.exists())
  {
    printf("SIKEMotion: can't open file SIKE/%s.sike\n", filename.c_str());
    return false;
  }

  KeyFrame f;
  while(!stream.eof())
  {
    //read frame
    JointData jd;
    unsigned time;
    for(int i = 0; i < 22; i++)
    {
      stream >> jd.angles[i];
      ASSERT(i < 2 || jd.angles[i] < 1000); //off and ignore not supported yet
    }
    stream >> time;

    //transform to KeyFrame
    RobotModel rm(jd, theRobotDimensions, MassCalibration());
    f = KeyFrame::fromRobotModel(rm, theRobotDimensions);
    f.headYaw = jd.angles[JointData::HeadYaw];
    f.headPitch = jd.angles[JointData::HeadPitch];
    f.time = time;

    //add to frames
    frames.push_back(f);
  }
  return true;
}

void SIKEMotion::saveToFile(std::string filename, const RobotDimensions &theRobotDimensions, const JointCalibration &theJointCalibration)
{
  // TODO: remove this assert after testing
  ASSERT(frames.size() && "Number of Frames must at least one!");

  OutTextFile stream("SIKE/" + filename + ".sike");
  OutTextFile bikeFile("SIKE/" + filename + ".bmc.draft");

  KeyFrame initialFrame = frames[0];
  Vector3<float> bikeFootOriginL = initialFrame.lfoot.translation;
  Vector3<float> bikeFootOriginR = bikeFootOriginL;
  bikeFootOriginR.y = -bikeFootOriginR.y;
  Vector3<float> bikeArmOriginL = initialFrame.lhand.translation;
  Vector3<float> bikeArmOriginR = bikeArmOriginL;
  bikeArmOriginR.y = -bikeArmOriginR.y;

  Vector3<float> lastRFootRelative(0.0f, 0.0f, 0.0f);
  Vector3<float> lastLFootRelative(0.0f, 0.0f, 0.0f);
  Vector3<float> lastRFootRot(0.0f, 0.0f, 0.0f);
  Vector3<float> lastLFootRot(0.0f, 0.0f, 0.0f);
  Vector3<float> lastRArm(0.0f, 0.0f, 0.0f);
  Vector3<float> lastLArm(0.0f, 0.0f, 0.0f);

  bikeFile << "#****" << "SIKE2BIKE" << "****" << endl;
  bikeFile << "#Common" << "Parameters" << endl << endl;
  bikeFile << "#FootOrigin" << "XYZ" << endl;
  bikeFile << bikeFootOriginL << endl << endl;
  bikeFile << "#ArmOrigin" << "XYZ" << endl;
  bikeFile << bikeArmOriginL << endl << endl;
  bikeFile << "#COMBalanceX" << endl;
  bikeFile << 0.1f << 0.1f << 0.0f << endl << endl;
  bikeFile << "#COMBalancY" << endl;
  bikeFile << 0.04f << 0.04f << 0.0f << endl << endl;
  bikeFile << "#Preview" << "(in" << "Frames)" << endl;
  bikeFile << 100 << endl << endl;
  bikeFile << "#loop" << endl;
  bikeFile << 0 << endl << endl;

  bikeFile << "#NumberOfPhases" << endl;
  bikeFile << (frames.size() - 1) << endl << endl;
  for(size_t i = 0; i < frames.size(); i++)
  {
    KeyFrame &f = frames.at(i);
    
    //transform to jointAngleSpace
    JointData jd;
    InverseKinematic::calcLegJoints(f.lfoot, f.rfoot, jd, theRobotDimensions);
    InverseKinematic::calcArmJoints(f.lhand, f.rhand, jd, theRobotDimensions, theJointCalibration);
    jd.angles[JointData::HeadYaw] = f.headYaw;
    jd.angles[JointData::HeadPitch] = f.headPitch;

    for(int j = 0; j < 22; j++)
      stream << jd.angles[j];
    stream << endl;
    stream << f.time;
    stream << endl << endl;

    if (i == 0) {
      // the first frame only provides the bike origins
      continue;
    }
    //write poses
    bikeFile << "#######" << "Phase" << (i - 1) <<"########################################" << endl << endl;
    bikeFile << "#duration" << "Timings" << "in" << "ms" << endl;
    bikeFile << f.time << endl << endl;
    bikeFile << "#Feet" << endl << endl;
    bikeFile << "#LeftFoot" << endl << endl;
    bikeFile << "#LFootTranslation" << endl;
    Vector3<float> relative = f.lfoot.translation - bikeFootOriginL;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastLFootRelative << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << relative << endl;
      lastLFootRelative = relative;
    }
    bikeFile << endl << "#LFootRotation" << endl;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastLFootRot << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << f.lfoot.rotation.getAngleAxis() << endl;
    }
    bikeFile << endl << "#RightFoot" << endl << endl;
    bikeFile << "#RFootTranslation" << endl;
    relative = f.rfoot.translation - bikeFootOriginR;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastRFootRelative << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << relative << endl;
      lastRFootRelative = relative;
    }
    bikeFile << endl << "#RFootRotation" << endl;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastRFootRot << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << f.rfoot.rotation.getAngleAxis() << endl;
    }
    bikeFile << endl << "#LeftArm" << endl;
    relative = f.lhand.translation - bikeArmOriginL;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastLArm << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << relative << endl;
    }
    bikeFile << endl << "#RightArm" << endl;
    relative = f.rhand.translation - bikeArmOriginR;
    bikeFile << "#Point" << 2 << endl;
    bikeFile << lastRArm << endl;
    for(size_t k = 3; k < 5; ++k) {
      bikeFile << "#Point" << k << endl;
      bikeFile << relative << endl;
    }
    bikeFile << endl << "#odometryOffset" << "x" << "y" << "rotation" << endl;
    bikeFile << 0 << 0 << 0 << endl << endl;
  }
}

SIKEMotion::KeyFrame SIKEMotion::KeyFrame::mirrored(bool mirror) const
{
  KeyFrame f = *this;

  if(mirror)
  {
    //mirror translatin
    f.lfoot = this->rfoot;
    f.lfoot.translation.y *= -1;

    f.rfoot = this->lfoot;
    f.rfoot.translation.y *= -1;

    f.lhand = this->rhand;
    f.lhand.translation.y *= -1;

    f.rhand = this->lhand;
    f.rhand.translation.y *= -1;

    //mirror rotation
    Vector3<> axis;
    float angle;

    f.lfoot.rotation.getAngleAxis(axis, angle);
    f.lfoot.rotation = RotationMatrix(Vector3<>(-axis.x, axis.y, -axis.z), angle);

    f.rfoot.rotation.getAngleAxis(axis, angle);
    f.rfoot.rotation = RotationMatrix(Vector3<>(-axis.x, axis.y, -axis.z), angle);

    f.lhand.rotation.getAngleAxis(axis, angle);
    f.lhand.rotation = RotationMatrix(Vector3<>(-axis.x, axis.y, -axis.z), angle);

    f.rhand.rotation.getAngleAxis(axis, angle);
    f.rhand.rotation = RotationMatrix(Vector3<>(-axis.x, axis.y, -axis.z), angle);

    //mirror head
    if(f.headYaw != JointData::off && f.headYaw != JointData::ignore)
      f.headYaw *= -1;
  }

  return f;
}

SIKEMotion::KeyFrame SIKEMotion::KeyFrame::fromRobotModel(const RobotModel &theRobotModel, const RobotDimensions &theRobotDimensions)
{
  KeyFrame f;

  f.lfoot = theRobotModel.limbs[RobotModel::footLeft];
  f.rfoot = theRobotModel.limbs[RobotModel::footRight];

  f.lhand = theRobotModel.limbs[RobotModel::foreArmLeft];
  f.lhand.translate(theRobotDimensions.lowerArmLength,0,0);

  f.rhand = theRobotModel.limbs[RobotModel::foreArmRight];
  f.rhand.translate(theRobotDimensions.lowerArmLength,0,0);

  return f;
}

SIKEMotion::KeyFrame SIKEMotion::KeyFrame::interpolateBetweenFrames(const KeyFrame &f1, const KeyFrame &f2, float ratio)
{
  ASSERT(ratio >= 0);
  ASSERT(ratio <= 1);

  KeyFrame result;

  //interpolate translation
  const float ratioInv = 1 - ratio;
  result.lfoot.translation = f1.lfoot.translation * ratioInv + f2.lfoot.translation * ratio;
  result.rfoot.translation = f1.rfoot.translation * ratioInv + f2.rfoot.translation * ratio;
  result.lhand.translation = f1.lhand.translation * ratioInv + f2.lhand.translation * ratio;
  result.rhand.translation = f1.rhand.translation * ratioInv + f2.rhand.translation * ratio;

  //interpolate rotation
  float angle;
  Vector3<> axis;

  RotationMatrix deltaLFoot(f2.lfoot.rotation * f1.lfoot.rotation.transpose());
  deltaLFoot.getAngleAxis(axis, angle);
  result.lfoot.rotation = RotationMatrix(axis, angle * ratio) * f1.lfoot.rotation;

  RotationMatrix deltaRFoot(f2.rfoot.rotation * f1.rfoot.rotation.transpose());
  deltaRFoot.getAngleAxis(axis, angle);
  result.rfoot.rotation = RotationMatrix(axis, angle * ratio) * f1.rfoot.rotation;

  RotationMatrix deltaLHand(f2.lhand.rotation * f1.lhand.rotation.transpose());
  deltaLHand.getAngleAxis(axis, angle);
  result.lhand.rotation = RotationMatrix(axis, angle * ratio) * f1.lhand.rotation;

  RotationMatrix deltaRHand(f2.rhand.rotation * f1.rhand.rotation.transpose());
  deltaRHand.getAngleAxis(axis, angle);
  result.rhand.rotation = RotationMatrix(axis, angle * ratio) * f1.rhand.rotation;

  //interpolate head joints
  if(f2.headYaw == JointData::ignore || f2.headYaw == JointData::off)
    result.headYaw = f2.headYaw;
  else if(f1.headYaw == JointData::ignore || f1.headYaw == JointData::off)
    result.headYaw = f2.headYaw;
  else
    result.headYaw = f1.headYaw * ratioInv + f2.headYaw * ratio;

  if(f2.headPitch == JointData::ignore || f2.headPitch == JointData::off)
    result.headPitch = f2.headPitch;
  else if(f1.headPitch == JointData::ignore || f1.headPitch == JointData::off)
    result.headPitch = f2.headPitch;
  else
    result.headPitch = f1.headPitch * ratioInv + f2.headPitch * ratio;

  return result;
}

