/**
* @file SIKEngine.cpp
* S(imple)I(inverse)K(inematic)Engine
* author jeff
*/

#include "SIKEngine.h"

#include "Tools/InverseKinematic.h"

void SIKEngine::init()
{
  //load motions
  for(int i = 0; i < SIKERequest::numOfMotions; i++)
  {
    SIKEMotion m;
    VERIFY(m.loadFromFile(SIKERequest::getMotionName((SIKERequest::Motion)i), theRobotDimensions));
    motions.push_back(m);
  }
}

void SIKEngine::update(SIKEngineOutput &sikengineOutput)
{
  ASSERT(motions.size() == SIKERequest::numOfMotions);

  output = &sikengineOutput;
  output->jointHardness.resetToDefault();

  //start a new motion if we are done with the last (and the SIKERequest changed)
  if(state.done)
  {
    const SIKERequest &request = theMotionSelection.sikeRequest;

    if(request.motion != state.actMotion || request.mirror != state.mirror)
    {
      state.actMotion = request.motion;
      state.mirror = request.mirror;
      state.actMotionStartFrame = Frame::fromRobotModel(theRobotModel, theRobotDimensions);
      state.actMotionStartFrame.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
      state.actMotionStartFrame.headPitch = theFilteredJointData.angles[JointData::HeadPitch];
      state.actMotionStartTime = theFrameInfo.time;

      output->executedMotion = request.motion;
    }
  }

  if(state.actMotion == SIKERequest::numOfMotions)
    return;

  DEBUG_RESPONSE("module:SIKEditor",
      editor(state);
      return;
      );
  //"else"
  {
    editorState.lastFrameId = -1;
    //would be nice to reset frameId, but the modify would overwrite it
    //editorState.frameId = 0;
  }

  //play motion and set the SIKEngineOutput
  playMotion(state);

  if(!state.done)
    output->isLeavingPossible = false;
  else
    output->isLeavingPossible = true;
}

void SIKEngine::playMotion(SIKEState &state)
{
  ASSERT(state.actMotion >= 0);
  ASSERT(state.actMotion <= SIKERequest::numOfMotions);

  const unsigned timeSinceMotionStarted = theFrameInfo.getTimeSince(state.actMotionStartTime);

  const SIKEMotion &motion = motions.at(state.actMotion);
  Frame frame;
  state.done = motion.getMotionAt(timeSinceMotionStarted, state.mirror, state.actMotionStartFrame, frame);
  setOutput(frame);
}

void SIKEngine::setOutput(const Frame &frame)
{
  InverseKinematic::calcLegJoints(frame.lfoot, frame.rfoot, *output, theRobotDimensions);
  InverseKinematic::calcArmJoints(frame.lhand, frame.rhand, *output, theRobotDimensions, theJointCalibration);
  output->angles[JointData::HeadYaw] = frame.headYaw;
  output->angles[JointData::HeadPitch] = frame.headPitch;
}

void SIKEngine::editor(SIKEState &state)
{
  int interpolationLength = 1000;
  MODIFY("SIKEditor:interpolationLength", interpolationLength);

  bool lfootOff, rfootOff, hipOff, lhandOff, rhandOff, headOff;
  lfootOff = rfootOff = hipOff = lhandOff = rhandOff = headOff = false;
  MODIFY("SIKEditor:lfootOff", lfootOff);
  MODIFY("SIKEditor:rfootOff", rfootOff);
  MODIFY("SIKEditor:hipOff", hipOff);
  MODIFY("SIKEditor:lhandOff", lhandOff);
  MODIFY("SIKEditor:rhandOff", rhandOff);
  MODIFY("SIKEditor:headOff", headOff);

  MODIFY("SIKEditor:actFrame", editorState.frameId);

  SIKEMotion &motion = motions.at(state.actMotion);

  //handle editorState.frameId changed
  if(editorState.frameId != editorState.lastFrameId)
  {
    if(editorState.frameId < 0 || editorState.frameId >= (int)motion.frames.size())
      return;
    else
    {
      editorState.interpolationTime = interpolationLength;
      {
        editorState.lastFrame = Frame::fromRobotModel(theRobotModel, theRobotDimensions);
        editorState.lastFrame.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
        editorState.lastFrame.headPitch = theFilteredJointData.angles[JointData::HeadPitch];
      }
      editorState.lastFrameId = editorState.frameId;
    }
  }

  //interpolate from on frame to another
  if(editorState.interpolationTime > 0)
  {
    const Frame &f = motion.frames.at(editorState.frameId);

    float ratio = 1.0f - float(editorState.interpolationTime)/float(interpolationLength);
    //float imprecision =(((
    if(ratio < 0 || ratio > 1)
    {
      ASSERT(fabs(ratio) < 0.001 || ratio - 1.0f < 0.001);
      ratio = Range<>(0.0f,1.0f).limit(ratio);
    }
    const Frame &interFrame = Frame::interpolateBetweenFrames(editorState.lastFrame, f, ratio);

    setOutput(interFrame);
    editorState.interpolationTime -= (int)(1000.0f * theRobotDimensions.motionCycleTime);

    return;
  }

  Frame &frameFromMotion = motion.frames.at(editorState.frameId);
  Frame frame = frameFromMotion;

  //turn of certain bodyparts if requested and overwrite frame with actual position
  {
    Frame actFrame = Frame::fromRobotModel(theRobotModel, theRobotDimensions);
    actFrame.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
    actFrame.headPitch = theFilteredJointData.angles[JointData::HeadPitch];

    if(lfootOff)
    {
      frame.lfoot = actFrame.lfoot;
      for(size_t i = JointData::LHipRoll; i < JointData::RHipYawPitch; i++)
        output->jointHardness.hardness[i] = 0;
    }
    if(rfootOff)
    {
      frame.rfoot = actFrame.rfoot;
      for(size_t i = JointData::RHipRoll; i < JointData::numOfJoints; i++)
        output->jointHardness.hardness[i] = 0;
    }
    if(hipOff)
    {
      frame.rfoot = actFrame.rfoot;
      frame.lfoot = actFrame.lfoot;
      output->jointHardness.hardness[JointData::LHipYawPitch] = 0;
      output->jointHardness.hardness[JointData::RHipYawPitch] = 0;
    }
    if(lhandOff)
    {
      frame.lhand = actFrame.lhand;
      for(size_t i = JointData::LShoulderPitch; i < JointData::RShoulderPitch; i++)
        output->jointHardness.hardness[i] = 0;
    }
    if(rhandOff)
    {
      frame.rhand = actFrame.rhand;
      for(size_t i = JointData::RShoulderPitch; i < JointData::LHipYawPitch; i++)
        output->jointHardness.hardness[i] = 0;
    }
    if(headOff)
    {
      frame.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
      frame.headPitch = theFilteredJointData.angles[JointData::HeadPitch];

      for(int i = JointData::HeadYaw; i < JointData::LShoulderPitch; i++)
        output->jointHardness.hardness[i] = 0;
    }
  }

  //set resulting frame
  setOutput(frameFromMotion);

  //save frame && save motion
  DEBUG_RESPONSE("SIKEditor:saveFrame", frameFromMotion = frame;);
#ifdef TARGET_ROBOT
  DEBUG_RESPONSE("SIKEditor:saveMotion",
      motion.saveToFile("editorSave", theRobotDimensions, theJointCalibration);
      OUTPUT(idText, text, "SIKEMotion saved ON THE NAO(!). If you start copyfiles, it will be deleted!!!");
    );
#else
  DEBUG_RESPONSE("SIKEditor:saveMotion",
      motion.saveToFile("editorSave", theRobotDimensions, theJointCalibration);
    );
#endif
  DEBUG_RESPONSE("SIKEditor:duplicateFrame",
      Frame newFrame = frameFromMotion;
      motion.frames.insert(motion.frames.begin() + editorState.frameId + 1, newFrame);
      OUTPUT(idText, text, "duplicated frame " << editorState.frameId);
      );

  DEBUG_RESPONSE("SIKEditor:deleteFrame",
      motion.frames.erase(motion.frames.begin() + editorState.frameId);
      editorState.interpolationTime = interpolationLength;
      {
        editorState.lastFrame = Frame::fromRobotModel(theRobotModel, theRobotDimensions);
        editorState.lastFrame.headYaw = theFilteredJointData.angles[JointData::HeadYaw];
        editorState.lastFrame.headPitch = theFilteredJointData.angles[JointData::HeadPitch];
      }
      /*
      editorState.frameId -= 1;
      if(editorState.frameId < 0)
        editorState.frameId = 0;
      */ //grrrr modify overwrites it =(
      editorState.lastFrameId = editorState.frameId;
      );
}

MAKE_MODULE(SIKEngine, Motion Control)

