/** 
* @file WalkingEngine.cpp
* Implementation of a module that creates the walking motions
* @author Colin Graf
*/

#include "WalkingEngine.h"
#include "Tools/Debugging/DebugDrawings.h" // PLOT
#include "Tools/InverseKinematic.h"
#include "Tools/Math/MVTools.h"

#ifdef RELEASE

#define BH_CHECK_NAN(value) ((void)0)
#define BH_CHECK_INF(value) ((void)0)
#define BH_CHECK_ZERO(value) ((void)0)
#define BH_CHECK_NAN_AND_INF(value) ((void)0)

#else

#define BH_CHECK_NAN(value) do { \
  if(MVTools::isNaN(value)) { \
    BH_TRACE_MSG("nan"); \
    ASSERT(!MVTools::isNaN(value)); \
  } \
  break; \
} while(1)

#define BH_CHECK_INF(value) do { \
  if(MVTools::isNearInf(value)) { \
    BH_TRACE_MSG("inf"); \
    ASSERT(!MVTools::isNearInf(value)); \
  } \
  break; \
} while(1)

#define BH_CHECK_ZERO(value) do { \
  if(MVTools::isNearZero(value)) { \
    BH_TRACE_MSG("zero"); \
    ASSERT(!MVTools::isNearZero(value)); \
  } \
  break; \
} while(1)

#define BH_CHECK_NAN_AND_INF(value) do { \
  BH_CHECK_NAN(value); \
  BH_CHECK_INF(value); \
  break; \
} while(1)

#endif

inline float saveAsinh(float xf)
{
  double x = xf; // yes, we need double here
#ifdef _MSC_VER
  return float(log(x + sqrt(x * x + 1.)));
#else
  return float(asinh(x));
#endif
}

inline float saveAcosh(float xf)
{
  //ASSERT(xf >= 1.f);
  double x = xf; // yes, we need double here
  if(x < 1.)
    return 0.000001f;
#ifdef _MSC_VER
  x = log(x + sqrt(x * x - 1.));
#else
  x = acosh(x);
#endif
  if(x < 0.000001)
    return 0.000001f;
  return float(x);
}

MAKE_MODULE(WalkingEngine, Motion Control)

#ifdef _MSC_VER
#pragma warning(disable:4355) // VC++: "this" in initializer list is ok
#endif

WalkingEngine::WalkingEngine() : calibrated(false), enforceStandTime(0), wasActive(false), coMShape(*this), lastActivationTime(0), lastKickTime(0), isLeavingPossible(true), kickState(0)
{
  p.originFoot = Vector3<>(-25.f, 50.f, -195.f);
  p.originFootRotation = Vector3<>(0.f, 0.f, 0.f);
  p.originBodyTilt = 0.f;
  p.originArmRotation = Vector2<>(0.4f, 0.f);

  p.modelWidth = 50.f;
  p.modelWidthAtFullSpeed = Vector2<>(45.f, 55.f);
  p.modelSwingWidth = 15.5f;
  p.modelSwingWidthAtFullSpeed = Vector2<>(12.5, p.modelSwingWidth);
  p.modelShiftAtFullSpeed = -10.f;

  p.bodyRotation = -0.02f;

  p.liftPhases = Vector2<>(0.f, 1.f);
  p.liftPhasesAtFullSpeed = Vector2<>(0.f, 1.f);
  p.liftOffset = Vector3<>(0.f, 0.f, 17.f);
  p.liftOffsetAtFullSpeedY = Vector3<>(-3.f, 0.f, 20.f);

  p.movePhases = Vector2<>(0.2f, 0.6f);
  p.movePhasesAtFullSpeed = Vector2<>(0.f, 1.f);
  p.moveArmRotation = Vector2<>(0.f, 0.5f);
  p.moveBodyRotationFactor = Vector2<>();

  p.hardness = 75;
  p.hardnessAnklePitch = 75;
  p.hardnessAnkleRoll = 30;

  p.speedMax = Pose2D(0.75f, 130, 50);
  p.speedBackwardMax = -80;
  p.speedDashMax = Pose2D(0.f, 0.f, 0.f);
  p.speedChangeMax = Pose2D(0.45f, 35.f, 25.f);
  p.speedUseEllipsoidClipping = true;

  p.measurementDelay = 50;
  p.measurementUseRobotModel = false;

  p.errorCoMMax = Vector3<>(10.f, 10.f, 10.f);
  p.errorCoMVelocityMax = Vector3<>(100.f, 100.f, 100.f);
  p.errorRotationMax = Vector2<>(0.3f, 0.3f);

  p.balance = true;
  p.balanceStepSize = Vector2<>(0.08f, 0.04f);
  p.balanceRefZMP = Vector2<>(0.04f, 0.08f);
  p.balanceModelCoM = Vector2<>(0.0f, -0.008f);
  p.balanceModelCoMVelocity = Vector2<>(0.0f, -0.008f);
  p.balanceCoM = Vector3<>(0.f, -0.2f, 0.f);
  p.balanceBodyRotation = Vector2<>();
  p.balanceFootRotation = Vector2<>();
  p.balanceArm = Vector2<>();

  p.cutout = true;
  p.cutoutInstability = 200.f;
  p.cutoutDuration = 1500;

  p.odometryScale = Pose2D(1.f, 1.1f, 1.1f);
  p.odometryUseTorsoMatrix = true;

  p.kickStrikeOut = Vector3<>(-30.f, 0.f, 20.f);
  p.kickDestination = Vector3<>(90.f, 0.f, 40.f);
  p.kickSwingWidth = 30.f;
  p.kickStepSize = Vector3<>(60.f, 0.f, 0.f);

  p.kickSidewardsPreStepSize = Vector3<>(0.f, 0.f, 0.f);
  p.kickSidewardsStrikeOut = Vector3<>(30.f, 30.f, 40.f);
  p.kickSidewardsDestination = Vector3<>(30.f, -20.f, 40.f);
  p.kickSidewardsStepSize = Vector3<>(60.f, 0.f, 0.f);
  p.kickDuration = 2000;
}

void WalkingEngine::init()
{
  InConfigFile stream(Global::getSettings().expandRobotFilename("walking.cfg"));
  if(stream.exists())
    stream >> p;
  else
  {
    InConfigFile stream("walking.cfg");
    if(stream.exists())
      stream >> p;
  }

#ifdef TARGET_SIM
  p.measurementDelay = 100;
#endif
}

void WalkingEngine::calculateShapes()
// => double& leftFadeIn, double& rightFadeIn, double& leftLift, double& rightLift, double& leftMoveFadeIn, double& rightMoveFadeIn
{
  const float phase2 = phase * 2;

  if(phase < 0.5f)
    leftFadeIn = (1.f - cos(phase2 * pi)) / 2.f;
  else
    leftFadeIn = 0.f;

  if(phase >= 0.5f)
    rightFadeIn = (1.f - cos((phase2 - 1.f) * pi)) / 2.f;
  else
    rightFadeIn = 0.f;

  if(leftStep.kick)
  {
    leftLift = 0.f;
    if(phase2 > 0.f && phase2 < 0.5f)
      leftStrikeOut = (1.f - cos(phase2 * pi2)) / 2.f;
    else
      leftStrikeOut = 0.f;
    if(phase2 >= 0.5f && phase2 < 1.f)
      leftKick = 1.f - (1.f - cos((phase2 - 0.5f) * pi2)) / 2.f;
    else
      leftKick = 0.f;
  }
  else
  {
    if(phase2 > leftStep.liftPhases.x && phase2 < leftStep.liftPhases.x + leftStep.liftPhases.y)
      leftLift = (1.f - cos((phase2 - leftStep.liftPhases.x) / leftStep.liftPhases.y * pi2)) / 2.f;
    else
      leftLift = 0.f;
    leftStrikeOut = 0.f;
    leftKick = 0.f;
  }

  if(rightStep.kick)
  {
    rightLift = 0.f;
    if(phase2 - 1.f > 0.f && phase2 - 1.f < 0.5f)
      rightStrikeOut = (1.f - cos((phase2 - 1.f) * pi2)) / 2.f;
    else
      rightStrikeOut = 0.f;
    if((phase2 - 1.f) >= 0.5f && (phase2 - 1.f) < 1.f)
      rightKick = 1.f - (1.f - cos(((phase2 - 1.f) - 0.5f) * pi2)) / 2.f;
    else
      rightKick = 0.f;
  }
  else
  {
    if(phase2 - 1.f > rightStep.liftPhases.x && phase2 - 1. < rightStep.liftPhases.x + rightStep.liftPhases.y)
      rightLift = (1.f - cos((phase2 - 1.f - rightStep.liftPhases.x) / rightStep.liftPhases.y * pi2)) / 2.f;
    else
      rightLift = 0.f;
    rightStrikeOut = 0.f;
    rightKick = 0.f;
  }

  if(phase2 > leftStep.movePhases.x && phase2 < 1.f)
    leftMoveFadeIn = (phase2 < leftStep.movePhases.x + leftStep.movePhases.y) ? ((1.f - cos((phase2 - leftStep.movePhases.x) / leftStep.movePhases.y * pi)) / 2.f) : 1.f;
  else
    leftMoveFadeIn = 0.f;

  if(phase2 - 1.f > rightStep.movePhases.x && phase2 - 1.f < 1.f)
    rightMoveFadeIn = (phase2 - 1.f < rightStep.movePhases.x + rightStep.movePhases.y) ? ((1.f - cos((phase2 - 1.f - rightStep.movePhases.x) / rightStep.movePhases.y * pi)) / 2.f) : 1.f;
  else
    rightMoveFadeIn = 0.f;
  
  PLOT("module:WalkingEngine:leftFadeIn", leftFadeIn);
  PLOT("module:WalkingEngine:rightFadeIn", -rightFadeIn);
  PLOT("module:WalkingEngine:leftMoveFadeIn", leftMoveFadeIn);
  PLOT("module:WalkingEngine:rightMoveFadeIn", -rightMoveFadeIn);
  PLOT("module:WalkingEngine:leftLift", leftLift);
  PLOT("module:WalkingEngine:rightLift", -rightLift);
}

void WalkingEngine::calculateTmpRobotFootVectors() // in robot coordinates
// => Vector3<>& tmpLeftFoot, Vector3<>& tmpRightFoot, double& leftRotationZ, double& rightRotationZ, Vector2<>& leftArm, Vector2<>& rightArm, Vector2<>& bodyStepRotation
{
  // start with stepOrigin
  tmpLeftFoot = Vector3<>(p.originFoot.x, p.originFoot.y, p.originFoot.z);
  tmpRightFoot = Vector3<>(p.originFoot.x, -p.originFoot.y, p.originFoot.z);
  leftRotationZ = 0.f;
  rightRotationZ = 0.f;
  leftArm = p.originArmRotation;
  rightArm = p.originArmRotation;
  bodyStepRotation = Vector2<>();

  // add air leg lift
  tmpLeftFoot += leftStep.liftOffset * leftLift;
  tmpRightFoot += rightStep.liftOffset * rightLift;
  tmpLeftFoot += leftStep.kickStrikeOut * leftStrikeOut;
  tmpRightFoot += rightStep.kickStrikeOut * rightStrikeOut;
  tmpLeftFoot += leftStep.kickDestination * leftKick;
  tmpRightFoot += rightStep.kickDestination * rightKick;

  // add / remove step offsets
  if(phase < 0.5) // left is air foot
  {
    // basic foot shifting

    tmpLeftFoot.x -= rightStep.size.x * (1.f - leftMoveFadeIn) * leftFadeIn; // decrement step size from the other foot
    tmpLeftFoot.y -= rightStep.size.y * (1.f - leftMoveFadeIn) * leftFadeIn;
    leftRotationZ -= rightStep.rotation * (1.f - leftMoveFadeIn) * leftFadeIn;
    bodyStepRotation -= rightStep.bodyRotation * (1.f - leftMoveFadeIn) * leftFadeIn;

    tmpLeftFoot += leftStep.size * leftMoveFadeIn; // add step size
    leftRotationZ += leftStep.rotation * leftMoveFadeIn;
    bodyStepRotation += leftStep.bodyRotation * leftMoveFadeIn;

    tmpRightFoot += rightStep.size * (1.f - leftFadeIn);
    rightRotationZ += rightStep.rotation * (1.f - leftFadeIn);
    bodyStepRotation += rightStep.bodyRotation * (1.f - leftFadeIn);

    rightArm += p.moveArmRotation * (leftStep.size.x / 100.f * leftMoveFadeIn);
    leftArm += p.moveArmRotation * (rightStep.size.x / 100.f * (1.f - leftFadeIn));
  }
  else // right is air foot
  {
    // basic foot shifting

    tmpRightFoot.x -= leftStep.size.x * (1.f - rightMoveFadeIn) * rightFadeIn; // decrement step size from the other foot
    tmpRightFoot.y -= leftStep.size.y * (1.f - rightMoveFadeIn) * rightFadeIn;
    rightRotationZ -= leftStep.rotation * (1.f - rightMoveFadeIn) * rightFadeIn;
    bodyStepRotation -= leftStep.bodyRotation * (1.f - rightMoveFadeIn) * rightFadeIn;

    tmpRightFoot += rightStep.size * rightMoveFadeIn; // add step size
    rightRotationZ += rightStep.rotation * rightMoveFadeIn;
    bodyStepRotation += rightStep.bodyRotation * rightMoveFadeIn;

    tmpLeftFoot += leftStep.size * (1.f - rightFadeIn);
    leftRotationZ += leftStep.rotation * (1.f - rightFadeIn);
    bodyStepRotation += leftStep.bodyRotation * (1.f - rightFadeIn);

    leftArm += p.moveArmRotation * (rightStep.size.x / 100.f * rightMoveFadeIn);
    rightArm += p.moveArmRotation * (leftStep.size.x / 100.f * (1.f - rightFadeIn));
  }

  if(p.balance)
  {
    Vector2<> armCorrection(rotationError.x * p.balanceArm.x, rotationError.y * p.balanceArm.y);
    leftArm += armCorrection;
    rightArm += armCorrection;
  }
}

void WalkingEngine::calculateBodyRotation()
// => Vector2<>& bodyRotation, RotationMatrix& world2Body, RotationMatrix& body2World
{
  bodyRotation = bodyStepRotation + Vector2<>(p.bodyRotation * coMShape.getOscillation(), p.originBodyTilt);

  if(p.balance)
    bodyRotation += Vector2<>(rotationError.x * -p.balanceBodyRotation.x, rotationError.y * -p.balanceBodyRotation.y);

  world2Body = RotationMatrix(Vector3<>(bodyRotation.x, bodyRotation.y, 0.));
  body2World = world2Body.invert();
}

void WalkingEngine::calculateReferenceCoM()
// => CoMSet& referenceCoMSet
{
  if(wasActive)
    lastReferenceCoMSet = referenceCoMSet;

  if(supportLeg == left)
  {
    referenceCoMSet.leftFoot = Vector3<>(p.originFoot.x, p.originFoot.y, p.originFoot.z) - standRobotModel.centerOfMass;
    referenceCoMSet.leftFoot -= coMShape.getRefZMP() + coMShape.getPosition();
    referenceCoMSet.rightFoot = referenceCoMSet.leftFoot + (tmpRightFoot - tmpLeftFoot);
  }
  else // (supportLeg == right)
  {
    referenceCoMSet.rightFoot = Vector3<>(p.originFoot.x, -p.originFoot.y, p.originFoot.z) - standRobotModel.centerOfMass;
    referenceCoMSet.rightFoot -= coMShape.getRefZMP() + coMShape.getPosition();
    referenceCoMSet.leftFoot = referenceCoMSet.rightFoot + (tmpLeftFoot - tmpRightFoot);
  }

  referenceCoMSet.bodyRotation = bodyRotation;
  referenceCoMSet.timeStamp = theFrameInfo.time;

  if(wasActive)
  {
    referenceCoMSet.velocity = ((referenceCoMSet.leftFoot - lastReferenceCoMSet.leftFoot) + (referenceCoMSet.rightFoot - lastReferenceCoMSet.rightFoot)) * 0.5f;
    referenceCoMSet.velocity /= float(referenceCoMSet.timeStamp - lastReferenceCoMSet.timeStamp) * 0.001f;
  }
  else
    referenceCoMSet.velocity = Vector3<>();

  oldCoMSets.add(referenceCoMSet);

  PLOT("module:WalkingEngine:referenceLeftFootX", referenceCoMSet.leftFoot.x);
  PLOT("module:WalkingEngine:referenceLeftFootY", referenceCoMSet.leftFoot.y);
  PLOT("module:WalkingEngine:referenceRightFootX", referenceCoMSet.rightFoot.x);
  PLOT("module:WalkingEngine:referenceRightFootY", referenceCoMSet.rightFoot.y);

  PLOT("module:WalkingEngine:referenceAngleX", referenceCoMSet.bodyRotation.x);
  PLOT("module:WalkingEngine:referenceAngleY", referenceCoMSet.bodyRotation.y);
  PLOT("module:WalkingEngine:referenceCoMVelocityY", referenceCoMSet.velocity.y);
  PLOT("module:WalkingEngine:referenceCoMX", (referenceCoMSet.leftFoot.x + referenceCoMSet.rightFoot.x) * 0.5);
  PLOT("module:WalkingEngine:referenceCoMY", (referenceCoMSet.leftFoot.y + referenceCoMSet.rightFoot.y) * 0.5);
}

void WalkingEngine::generateNextStepSize(SupportLeg supportLeg)
// => lastSelectedSpeed, leftStep or rightStep, isLeavingPossible
{
  bool left = supportLeg == right;
  Step& step(left ? leftStep : rightStep);
  Step& otherStep(left ? rightStep : leftStep);

  // choose target speed
  bool usingWalkTarget = false, kick = false, preKick = false;
  Pose2D walkTarget(this->walkTarget);
  Pose2D selectedSpeed(theMotionSelection.walkRequest.speed);
  if(MVTools::isNearInf(selectedSpeed.translation.x) || MVTools::isNaN(selectedSpeed.translation.x))
    selectedSpeed.translation.x = 0.;
  if(MVTools::isNearInf(selectedSpeed.translation.y) || MVTools::isNaN(selectedSpeed.translation.y))
    selectedSpeed.translation.y = 0.;
  if(MVTools::isNearInf(selectedSpeed.rotation) || MVTools::isNaN(selectedSpeed.rotation))
    selectedSpeed.rotation = 0.;
  switch(theMotionSelection.walkRequest.mode)
  {
  case WalkRequest::targetMode:
    {
      usingWalkTarget = true;

      // remove pending odometry offset from walk target
      walkTarget -= Pose2D(otherStep.rotation * 0.5f, 
        otherStep.size.x * 0.5f, otherStep.size.y * 0.5f);

      selectedSpeed.translation.x = walkTarget.translation.x * 2.f;
      selectedSpeed.translation.y = walkTarget.translation.y * 2.f;
      selectedSpeed.rotation = walkTarget.rotation * 2.f;
    }
    break;
  default:
    break;
  }

  // get maxspeed
  Pose2D maxSpeed(p.speedMax);
  if(selectedSpeed.translation.x < 0)
    maxSpeed.translation.x = -p.speedBackwardMax;
  if(selectedSpeed.translation.y == 0 && selectedSpeed.rotation == 0 && selectedSpeed.translation.x > 0 && p.speedDashMax.translation.x > maxSpeed.translation.x)
    maxSpeed.translation.x = p.speedDashMax.translation.x;
  if(selectedSpeed.translation.x == 0 && selectedSpeed.rotation == 0 && p.speedDashMax.translation.y > maxSpeed.translation.y)
    maxSpeed.translation.y = p.speedDashMax.translation.y;
  if(selectedSpeed.translation.x == 0 && selectedSpeed.translation.y == 0 && p.speedDashMax.rotation > maxSpeed.rotation)
    maxSpeed.rotation = p.speedDashMax.rotation;

  // scale percentage speeds
  if(theMotionSelection.walkRequest.mode == WalkRequest::percentageSpeedMode)
  {
    selectedSpeed.translation.x *= maxSpeed.translation.x;
    selectedSpeed.translation.y *= maxSpeed.translation.y;
    selectedSpeed.rotation *= maxSpeed.rotation;
  }

  // max speed clipping
  if(usingWalkTarget)
  {
    Pose2D maxSpeedForTarget(
      sqrtf(2.f * fabs(walkTarget.rotation) * p.speedChangeMax.rotation),
      sqrtf(2.f * fabs(walkTarget.translation.x) * p.speedChangeMax.translation.x),
      sqrtf(2.f * fabs(walkTarget.translation.y) * p.speedChangeMax.translation.y));
    maxSpeed.rotation = std::min(std::max(maxSpeedForTarget.rotation, p.speedChangeMax.rotation), maxSpeed.rotation);
    maxSpeed.translation.x = std::min(std::max(maxSpeedForTarget.translation.x, p.speedChangeMax.translation.x), maxSpeed.translation.x);
    maxSpeed.translation.y = std::min(std::max(maxSpeedForTarget.translation.y, p.speedChangeMax.translation.y), maxSpeed.translation.y);
  }
  if(p.speedUseEllipsoidClipping) // clip target speeds to a "three-dimensional ellipse"
  {
    float ratioY = maxSpeed.translation.x / maxSpeed.translation.y;
    float ratioRot = maxSpeed.translation.x / maxSpeed.rotation;
    Vector3<> tempSpeed(selectedSpeed.translation.x, selectedSpeed.translation.y * ratioY, selectedSpeed.rotation * ratioRot);
    float speedVectorLength = tempSpeed.abs();
    if(speedVectorLength > maxSpeed.translation.x)
    {
      tempSpeed /= speedVectorLength;
      tempSpeed *= maxSpeed.translation.x;
      selectedSpeed.translation.x = tempSpeed.x;
      selectedSpeed.translation.y = tempSpeed.y / ratioY;
      selectedSpeed.rotation = tempSpeed.z / ratioRot;
    }
  }
  else
  {
    selectedSpeed.translation.x = Range<>(-maxSpeed.translation.x, maxSpeed.translation.x).limit(selectedSpeed.translation.x);
    selectedSpeed.translation.y = Range<>(-maxSpeed.translation.y, maxSpeed.translation.y).limit(selectedSpeed.translation.y);
    selectedSpeed.rotation = Range<>(-maxSpeed.rotation, maxSpeed.rotation).limit(selectedSpeed.rotation);
  }
  if(usingWalkTarget)
  {
    selectedSpeed.translation.x = Range<>(-maxSpeed.translation.x * theMotionSelection.walkRequest.speed.translation.x, maxSpeed.translation.x * theMotionSelection.walkRequest.speed.translation.x).limit(selectedSpeed.translation.x);
    selectedSpeed.translation.y = Range<>(-maxSpeed.translation.y * theMotionSelection.walkRequest.speed.translation.y, maxSpeed.translation.y * theMotionSelection.walkRequest.speed.translation.y).limit(selectedSpeed.translation.y);
    selectedSpeed.rotation = Range<>(-maxSpeed.rotation * theMotionSelection.walkRequest.speed.rotation, maxSpeed.rotation * theMotionSelection.walkRequest.speed.rotation).limit(selectedSpeed.rotation);
  }

  // clip speed to max speed change
  const Pose2D& maxSpeedChange(p.speedChangeMax);
  if(!kick && !preKick)
    selectedSpeed.translation.x = Range<>(lastSelectedSpeed.translation.x - maxSpeedChange.translation.x, lastSelectedSpeed.translation.x + maxSpeedChange.translation.x).limit(selectedSpeed.translation.x);
  selectedSpeed.translation.y = Range<>(lastSelectedSpeed.translation.y - maxSpeedChange.translation.y, lastSelectedSpeed.translation.y + maxSpeedChange.translation.y).limit(selectedSpeed.translation.y);
  selectedSpeed.rotation = Range<>(lastSelectedSpeed.rotation - maxSpeedChange.rotation, lastSelectedSpeed.rotation + maxSpeedChange.rotation ).limit(selectedSpeed.rotation);

  // convert speed into step sizes
  const float innerYLimit = (p.originFoot.y - 45.f);
  const float innerRotationLimit = 0.f;
  step.size.x = selectedSpeed.translation.x * 0.5f;
  step.size.y = selectedSpeed.translation.y * (((left && selectedSpeed.translation.y < 0.f) || (!left && selectedSpeed.translation.y > 0.f)) ? (innerYLimit / maxSpeed.translation.y) : (1.f - innerYLimit / maxSpeed.translation.y));
  step.rotation = selectedSpeed.rotation * (((left && selectedSpeed.rotation < 0.f) || (!left && selectedSpeed.rotation > 0.f)) ? (innerRotationLimit / maxSpeed.rotation) : (1.f - innerRotationLimit / maxSpeed.rotation));
  step.size.z = 0;
  if(!wasActive) // start without movement
    step.size.x = step.size.y = step.size.z = step.rotation = 0.;

  // generate parameters for lift and move phase
  Vector3<> liftOffset = left ? p.liftOffset : Vector3<>(p.liftOffset.x, -p.liftOffset.y, p.liftOffset.z);
  Vector3<> liftOffsetAtFullSpeedY = left ? p.liftOffsetAtFullSpeedY : Vector3<>(p.liftOffsetAtFullSpeedY.x, -p.liftOffsetAtFullSpeedY.y, p.liftOffsetAtFullSpeedY.z);
  step.liftOffset = liftOffset;
  step.liftPhases = p.liftPhases;
  step.movePhases = p.movePhases;
  Vector3<> liftOffsetSpeedFactor((liftOffsetAtFullSpeedY.x - liftOffset.x) / p.speedMax.translation.y,
    (liftOffsetAtFullSpeedY.y - liftOffset.y) / p.speedMax.translation.y,
    (liftOffsetAtFullSpeedY.z - liftOffset.z) / p.speedMax.translation.y);
  step.liftOffset.x += fabs(step.size.y) * liftOffsetSpeedFactor.x;
  step.liftOffset.y += fabs(step.size.y) * liftOffsetSpeedFactor.y;
  step.liftOffset.z += fabs(step.size.y) * liftOffsetSpeedFactor.z;
  Vector2<> liftPhasesSpeedFactor((p.liftPhasesAtFullSpeed.x - p.liftPhases.x) / (p.speedMax.translation.x * 0.5f),
    (p.liftPhasesAtFullSpeed.y - p.liftPhases.y) / (p.speedMax.translation.x * 0.5f));
  step.liftPhases.x += fabs(step.size.x) * liftPhasesSpeedFactor.x;
  step.liftPhases.y += fabs(step.size.x) * liftPhasesSpeedFactor.y;
 Vector2<> movePhasesSpeedFactor((p.movePhasesAtFullSpeed.x - p.movePhases.x) / (p.speedMax.translation.x * 0.5f),
    (p.movePhasesAtFullSpeed.y - p.movePhases.y) / (p.speedMax.translation.x * 0.5f));
  step.movePhases.x += fabs(step.size.x) * movePhasesSpeedFactor.x;
  step.movePhases.y += fabs(step.size.x) * movePhasesSpeedFactor.y;
  if(!wasActive)
    step.liftOffset = Vector3<>();

  // clip to walk target
  if(usingWalkTarget)
  {
    if((step.size.x > 0.f && walkTarget.translation.x > 0.f && step.size.x  > walkTarget.translation.x) || (step.size.x < 0.f && walkTarget.translation.x < 0.f && step.size.x < walkTarget.translation.x))
      step.size.x = walkTarget.translation.x;
    if((step.size.y > 0.f && walkTarget.translation.y > 0.f && step.size.y  > walkTarget.translation.y) || (step.size.y < 0.f && walkTarget.translation.y < 0.f && step.size.y < walkTarget.translation.y))
      step.size.y = walkTarget.translation.y;
    if((step.rotation > 0.f && walkTarget.rotation > 0.f && step.rotation  > walkTarget.rotation) || (step.rotation < 0.f && walkTarget.rotation < 0.f && step.rotation < walkTarget.rotation))
      step.rotation = walkTarget.rotation;
  }

  //
  step.kick = step.preKick = false;
  if(kickType != WalkRequest::none /*&& theMotionSelection.walkRequest.mode == WalkRequest::targetMode*/)
  {
    switch(kickState)
    {
    case 1:
      {
        switch(kickType)
        {
        case WalkRequest::left:
        case WalkRequest::right:
          {
            step.kick = true;
            step.kickStrikeOut = Vector3<>(p.kickStrikeOut.x, left ? p.kickStrikeOut.y : -p.kickStrikeOut.y, p.kickStrikeOut.z);
            step.kickDestination = Vector3<>(p.kickDestination.x, left ? p.kickDestination.y : -p.kickDestination.y, p.kickDestination.z);
            step.size = Vector3<>(p.kickStepSize.x, left ? p.kickStepSize.y : -p.kickStepSize.y, p.kickStepSize.z);
            kickState = 0;
            kickType = WalkRequest::none;
          }
          break;
        case WalkRequest::sidewardsLeft:
        case WalkRequest::sidewardsRight:
          {
            step.size = Vector3<>(p.kickSidewardsPreStepSize.x, left ? p.kickSidewardsPreStepSize.y : -p.kickSidewardsPreStepSize.y, p.kickSidewardsPreStepSize.z);
            step.preKick = true;
            kickState = 2;
          }
          break;
        default:
          ASSERT(false);
          break;
        }
      }
      break;
    case 2:
      {
        ASSERT(kickType == WalkRequest::sidewardsLeft || kickType == WalkRequest::sidewardsRight);
        step.kick = true;
        step.kickStrikeOut = Vector3<>(p.kickSidewardsStrikeOut.x, left ? p.kickSidewardsStrikeOut.y : -p.kickSidewardsStrikeOut.y, p.kickSidewardsStrikeOut.z);
        step.kickDestination = Vector3<>(p.kickSidewardsDestination.x, left ? p.kickSidewardsDestination.y : -p.kickSidewardsDestination.y, p.kickSidewardsDestination.z);
        step.size = Vector3<>(p.kickSidewardsStepSize.x, left ? p.kickStepSize.y : -p.kickSidewardsStepSize.y, p.kickSidewardsStepSize.z);
        kickState = 0;
        kickType = WalkRequest::none;
      }
      break;
    case 0:
      {
        //Pose2D nextWalkTarget(walkTarget - Pose2D(step.rotation, step.size.x, step.size.y));
        if(/*(theMotionSelection.walkRequest.mode != WalkRequest::targetMode || (fabs(nextWalkTarget.translation.x) < 1.f && fabs(nextWalkTarget.translation.y) < 1.f && fabs(nextWalkTarget.rotation) < 0.01f)) &&*/
          ((!left && (kickType == WalkRequest::left || kickType == WalkRequest::sidewardsRight)) || 
          (left && (kickType == WalkRequest::right || kickType == WalkRequest::sidewardsLeft))))
        {
          kickState = 1;
          if(kickType == WalkRequest::left || kickType == WalkRequest::right)
            step.preKick = true;
        }
      }
      break;
    default:
      ASSERT(false);
      break;
    }
  }
  else
    kickState = 0;

  // create copy of the step size 
  step.bodyRotation.y = (step.size.x /*- otherStep.size.x*/) * p.moveBodyRotationFactor.x;
  step.bodyRotation.x = (step.size.y /*- otherStep.size.y*/) * p.moveBodyRotationFactor.y;
  step.originalSize = step.size;
  
  BH_CHECK_NAN_AND_INF(step.size.x);
  BH_CHECK_NAN_AND_INF(step.size.y);

  //
  isLeavingPossible = fabs(leftStep.size.x) < 10. && fabs(leftStep.size.y) < 10. && fabs(leftStep.rotation) < 0.05 && 
    fabs(rightStep.size.x) < 10. && fabs(rightStep.size.y) < 10. && fabs(rightStep.rotation) < 0.05;

  lastSelectedSpeed = selectedSpeed;
}

void WalkingEngine::calculateRobotToFootVectors(JointRequest& jointRequest)
// => bodyShift, lastNewCoM, finalLeftFoot, finalRightFoot
{
  finalLeftFoot = Pose3D(world2Body).translate(tmpLeftFoot);
  finalRightFoot = Pose3D(world2Body).translate(tmpRightFoot);

  // set foot z-rotation and apply foot rotation correction
  Vector3<> leftRotation(p.originFootRotation.x, p.originFootRotation.y, p.originFootRotation.z + leftRotationZ);
  Vector3<> rightRotation(-p.originFootRotation.x, p.originFootRotation.y, -p.originFootRotation.z + rightRotationZ);
  if(p.balance)
  {
    Vector3<> correction(rotationError.x * -p.balanceFootRotation.x, rotationError.y * -p.balanceFootRotation.y, 0.f);
    leftRotation += correction;
    rightRotation += correction;
  }
  finalLeftFoot.rotate(RotationMatrix(leftRotation));
  finalRightFoot.rotate(RotationMatrix(rightRotation));

  Vector3<> newCoM((referenceCoMSet.leftFoot + referenceCoMSet.rightFoot) * 0.5);
  if(p.balance)
  {
    Vector3<> correction(coMError.x * -p.balanceCoM.x, coMError.y * -p.balanceCoM.y, coMError.z * -p.balanceCoM.z);
    newCoM += correction;
  }

  bool impressionHack = true;
  MODIFY("module:WalkingEngine:impressionHack", impressionHack);

  // update bodyShift
  if(wasActive)
    bodyShift -= world2Body * (newCoM - lastNewCoM) * 1.3f;
  if(impressionHack)
    bodyShift.z = 0;
  lastNewCoM = newCoM;

  // calculate foot positions using old bodyShift
  finalLeftFoot.translation -= bodyShift;
  finalRightFoot.translation -= bodyShift;

  //const float ratio = 1.f - (coMShape.getOscillation() + 1.f) * 0.5f;
  const float ratio = 0.5f;
  setLegJoints(finalLeftFoot, finalRightFoot, ratio, jointRequest);
  setArmJoints(leftArm, rightArm, jointRequest);
  setHeadJoints(jointRequest);
  for(int i = 0; i < 2; ++i)
  {
    const RobotModel tempRobotModel(jointRequest, theRobotDimensions, theMassCalibration); 
    const Vector3<> leftTempCoM(body2World * (tempRobotModel.limbs[RobotModel::footLeft].translation - tempRobotModel.centerOfMass));
    const Vector3<> rightTempCoM(body2World * (tempRobotModel.limbs[RobotModel::footRight].translation - tempRobotModel.centerOfMass));
    const Vector3<> tempCoM((leftTempCoM + rightTempCoM) * 0.5);
    Vector3<> bodyShiftOffset(world2Body * (newCoM - tempCoM) * 1.3f);
    if(impressionHack)
      bodyShiftOffset.z = 0;

    // improve body shift
    bodyShift -= bodyShiftOffset; 

    // apply new body shift
    finalLeftFoot.translation += bodyShiftOffset;
    finalRightFoot.translation += bodyShiftOffset;

    setLegJoints(finalLeftFoot, finalRightFoot, ratio, jointRequest);
  }

  PLOT("module:WalkingEngine:bodyShiftX", bodyShift.x);
  PLOT("module:WalkingEngine:bodyShiftY", bodyShift.y);
  PLOT("module:WalkingEngine:bodyShiftZ", bodyShift.z);

  PLOT("module:WalkingEngine:leftFootZ", finalLeftFoot.translation.z - p.originFoot.z);
  PLOT("module:WalkingEngine:rightFootZ", finalRightFoot.translation.z - p.originFoot.z);
  PLOT("module:WalkingEngine:leftFootX", finalLeftFoot.translation.x);
  PLOT("module:WalkingEngine:rightFootX", finalRightFoot.translation.x);
  PLOT("module:WalkingEngine:leftFootY", finalLeftFoot.translation.y);
  PLOT("module:WalkingEngine:rightFootY", finalRightFoot.translation.y);

#ifndef TARGET_ROBOT
  const RobotModel finalRobotModel(jointRequest, theRobotDimensions, theMassCalibration); 
  const Vector3<>& leftFinalCoM(body2World * (finalRobotModel.limbs[RobotModel::footLeft].translation - finalRobotModel.centerOfMass));
  const Vector3<>& rightFinalCoM(body2World * (finalRobotModel.limbs[RobotModel::footRight].translation - finalRobotModel.centerOfMass));
  const Vector3<>& finalCoM(leftFinalCoM * 0.5 + rightFinalCoM * 0.5);
  Vector3<> coMError2 = newCoM - finalCoM;
#endif
}

void WalkingEngine::calculateExpectedCoM()
// => CoMSet& expectedCoMSet
{  
  unsigned int motionCycleTime = (unsigned int)(theRobotDimensions.motionCycleTime * 1000.f); // in ms  
  const unsigned int oldCoMTimeStamp = theFrameInfo.time - p.measurementDelay;
  
  int newerCoMIndex = 1, numberOfEntries = oldCoMSets.getNumberOfEntries();
  for(; newerCoMIndex < numberOfEntries; ++newerCoMIndex)
  {
    const unsigned int currentTimeStamp = oldCoMSets.getEntry(newerCoMIndex).timeStamp;
    const unsigned int nextTimeStamp = (newerCoMIndex + 1) < numberOfEntries ? oldCoMSets.getEntry(newerCoMIndex + 1).timeStamp : (currentTimeStamp - motionCycleTime);
    if(oldCoMTimeStamp == currentTimeStamp)
    {
      expectedCoMSet = oldCoMSets.getEntry(newerCoMIndex);
      break;
    }
    else if(oldCoMTimeStamp < currentTimeStamp && oldCoMTimeStamp > nextTimeStamp)
    { // interpolate between two old CoMSets
      const CoMSet& newerCoMSet(oldCoMSets.getEntry(newerCoMIndex));
      const CoMSet& olderCoMSet((newerCoMIndex + 1) < numberOfEntries ? oldCoMSets.getEntry(newerCoMIndex + 1) : standCoMSet);
      const unsigned int olderCoMSetTimeStamp = (newerCoMIndex + 1) < numberOfEntries ? olderCoMSet.timeStamp : (newerCoMSet.timeStamp - motionCycleTime);
      
      const float newerRatio = newerCoMSet.timeStamp == olderCoMSetTimeStamp ? 0.f : (1.f - float(newerCoMSet.timeStamp - oldCoMTimeStamp) / float(newerCoMSet.timeStamp - olderCoMSetTimeStamp));
      const float olderRatio = 1.f - newerRatio;
      
      expectedCoMSet.leftFoot = olderCoMSet.leftFoot * olderRatio + newerCoMSet.leftFoot * newerRatio;
      expectedCoMSet.rightFoot = olderCoMSet.rightFoot * olderRatio + newerCoMSet.rightFoot * newerRatio;
      expectedCoMSet.bodyRotation = olderCoMSet.bodyRotation * olderRatio + newerCoMSet.bodyRotation * newerRatio;
      expectedCoMSet.velocity = olderCoMSet.velocity * olderRatio + newerCoMSet.velocity * newerRatio;
      expectedCoMSet.timeStamp = oldCoMTimeStamp;
      break;
    }
  }
  if(newerCoMIndex >= numberOfEntries)
  {
    expectedCoMSet = standCoMSet;
    expectedCoMSet.timeStamp = oldCoMTimeStamp;
  }

  PLOT("module:WalkingEngine:expectedAngleX", expectedCoMSet.bodyRotation.x);
  PLOT("module:WalkingEngine:expectedAngleY", expectedCoMSet.bodyRotation.y);
  PLOT("module:WalkingEngine:expectedCoMVelocityY", expectedCoMSet.velocity.y);
  PLOT("module:WalkingEngine:expectedCoMX", (expectedCoMSet.leftFoot.x + expectedCoMSet.rightFoot.x) * 0.5);
  PLOT("module:WalkingEngine:expectedCoMY", (expectedCoMSet.leftFoot.y + expectedCoMSet.rightFoot.y) * 0.5);
}

void WalkingEngine::calculateMeasuredCoM()
// => CoMSet& measuredCoMSet
{
  if(wasActive)
    lastMeasuredCoMSet = measuredCoMSet;

  if(p.measurementUseRobotModel)
  {
    measuredCoMSet.leftFoot = theTorsoMatrix.rotation * (theRobotModel.limbs[RobotModel::footLeft].translation - theRobotModel.centerOfMass);
    measuredCoMSet.rightFoot = theTorsoMatrix.rotation * (theRobotModel.limbs[RobotModel::footRight].translation - theRobotModel.centerOfMass);
  }
  else
  {
    Vector3<> rotationErrorAxis(
      (theFilteredSensorData.data[SensorData::angleX] + expectedCoMSet.bodyRotation.x),
      (theFilteredSensorData.data[SensorData::angleY] + expectedCoMSet.bodyRotation.y), 0.);
    RotationMatrix rotationError(rotationErrorAxis);
    
    measuredCoMSet.leftFoot = rotationError * expectedCoMSet.leftFoot;
    measuredCoMSet.rightFoot = rotationError * expectedCoMSet.rightFoot;
  }

  measuredCoMSet.bodyRotation.x = -theFilteredSensorData.data[SensorData::angleX];
  measuredCoMSet.bodyRotation.y = -theFilteredSensorData.data[SensorData::angleY];
  measuredCoMSet.timeStamp = expectedCoMSet.timeStamp;

  if(wasActive)
  {
    measuredCoMSet.velocity = ((measuredCoMSet.leftFoot - lastMeasuredCoMSet.leftFoot) + (measuredCoMSet.rightFoot - lastMeasuredCoMSet.rightFoot)) * 0.5f;
    measuredCoMSet.velocity /= float(measuredCoMSet.timeStamp - lastMeasuredCoMSet.timeStamp) * 0.001f;
  }
  else
    measuredCoMSet.velocity = Vector3<>();

  PLOT("module:WalkingEngine:measuredAngleX", measuredCoMSet.bodyRotation.x);
  PLOT("module:WalkingEngine:measuredAngleY", measuredCoMSet.bodyRotation.y);
  PLOT("module:WalkingEngine:measuredCoMVelocityY", measuredCoMSet.velocity.y);
  PLOT("module:WalkingEngine:measuredCoMX", (measuredCoMSet.leftFoot.x + measuredCoMSet.rightFoot.x) * 0.5);
  PLOT("module:WalkingEngine:measuredCoMY", (measuredCoMSet.leftFoot.y + measuredCoMSet.rightFoot.y) * 0.5);
}

void WalkingEngine::calculateError()
// => Vector3<>& coMError, Vector3<>& rotationError
{
  coMError = ((measuredCoMSet.leftFoot - expectedCoMSet.leftFoot) + (measuredCoMSet.rightFoot - expectedCoMSet.rightFoot)) * 0.5;
  instability.add(coMError.squareAbs()); // using  the unclipped error
  coMError.x = Range<>(-p.errorCoMMax.x, p.errorCoMMax.x).limit(coMError.x);
  coMError.y = Range<>(-p.errorCoMMax.y, p.errorCoMMax.y).limit(coMError.y);
  coMError.z = Range<>(-p.errorCoMMax.z, p.errorCoMMax.z).limit(coMError.z);

  coMVelocityError = measuredCoMSet.velocity - expectedCoMSet.velocity;
  coMVelocityError.x = Range<>(-p.errorCoMVelocityMax.x, p.errorCoMVelocityMax.x).limit(coMVelocityError.x);
  coMVelocityError.y = Range<>(-p.errorCoMVelocityMax.y, p.errorCoMVelocityMax.y).limit(coMVelocityError.y);
  coMVelocityError.z = Range<>(-p.errorCoMVelocityMax.z, p.errorCoMVelocityMax.z).limit(coMVelocityError.z);

  rotationError.x = theFilteredSensorData.data[SensorData::angleX] + expectedCoMSet.bodyRotation.x;
  rotationError.y = theFilteredSensorData.data[SensorData::angleY] + expectedCoMSet.bodyRotation.y;
  rotationError.z = 0.;
  rotationError.x = Range<>(-p.errorRotationMax.x, p.errorRotationMax.x).limit(rotationError.x);
  rotationError.y = Range<>(-p.errorRotationMax.y, p.errorRotationMax.y).limit(rotationError.y);

  PLOT("module:WalkingEngine:instability", instability.getAverage());
}

void WalkingEngine::setArmJoints(const Vector2<>& leftRotation, const Vector2<>& rightRotation, JointRequest& jointRequest)
{
  for(int side = 0; side < 2; side++)
  {
    const Vector2<>& rotation(side == 0 ? leftRotation : rightRotation);
    JointData::Joint firstJoint = side == 0 ? JointData::LShoulderPitch  : JointData::RShoulderPitch;
    jointRequest.angles[firstJoint + 0] = -pi_2 + rotation.y;
    jointRequest.angles[firstJoint + 1] = rotation.x;
    jointRequest.angles[firstJoint + 2] = 0;
    jointRequest.angles[firstJoint + 3] = -jointRequest.angles[firstJoint + 1];
  }
}

void WalkingEngine::setLegJoints(const Pose3D& leftTarget, const Pose3D& rightTarget, float ratio, JointRequest& jointRequest)
{
  InverseKinematic::calcLegJoints(leftTarget, rightTarget, jointRequest, theRobotDimensions, ratio);

  jointRequest.jointHardness.hardness[JointData::LHipYawPitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::LHipRoll] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::LHipPitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::LKneePitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::LAnklePitch] = p.hardnessAnklePitch;
  jointRequest.jointHardness.hardness[JointData::LAnkleRoll] = p.hardnessAnkleRoll;
  jointRequest.jointHardness.hardness[JointData::RHipYawPitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::RHipRoll] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::RHipPitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::RKneePitch] = p.hardness;
  jointRequest.jointHardness.hardness[JointData::RAnklePitch] = p.hardnessAnklePitch;
  jointRequest.jointHardness.hardness[JointData::RAnkleRoll] = p.hardnessAnkleRoll;
}

void WalkingEngine::setHeadJoints(JointRequest& jointRequest)
{
  jointRequest.angles[JointData::HeadYaw] = theHeadJointRequest.pan;
  jointRequest.angles[JointData::HeadPitch] = theHeadJointRequest.tilt;
}

void WalkingEngine::setHeadJoints(float pan, float tilt, JointRequest& jointRequest)
{
  jointRequest.angles[JointData::HeadYaw] = pan;
  jointRequest.angles[JointData::HeadPitch] = tilt;
}

void WalkingEngine::update(WalkingEngineOutput& walkingEngineOutput)
{
  BH_CHECK_NAN_AND_INF(theFilteredSensorData.data[SensorData::angleX]);
  BH_CHECK_NAN_AND_INF(theFilteredSensorData.data[SensorData::angleY]);
  BH_CHECK_NAN_AND_INF(theFilteredSensorData.data[SensorData::gyroX]);
  BH_CHECK_NAN_AND_INF(theFilteredSensorData.data[SensorData::gyroY]);

  ASSERT(theFilteredSensorData.data[SensorData::angleX] != SensorData::off);
  ASSERT(theFilteredSensorData.data[SensorData::angleY] != SensorData::off);
  ASSERT(theFilteredSensorData.data[SensorData::gyroX] != SensorData::off);
  ASSERT(theFilteredSensorData.data[SensorData::gyroY] != SensorData::off);

  MODIFY("module:WalkingEngine:parameters", p);

  DECLARE_PLOT("module:WalkingEngine:expectedAngleX");
  DECLARE_PLOT("module:WalkingEngine:expectedAngleY");
  DECLARE_PLOT("module:WalkingEngine:expectedCoMVelocityY");
  DECLARE_PLOT("module:WalkingEngine:expectedCoMX");
  DECLARE_PLOT("module:WalkingEngine:expectedCoMY");

  DECLARE_PLOT("module:WalkingEngine:measuredAngleX");
  DECLARE_PLOT("module:WalkingEngine:measuredAngleY");
  DECLARE_PLOT("module:WalkingEngine:measuredCoMVelocityY");
  DECLARE_PLOT("module:WalkingEngine:measuredCoMX");
  DECLARE_PLOT("module:WalkingEngine:measuredCoMY");

  DECLARE_PLOT("module:WalkingEngine:referenceAngleX");
  DECLARE_PLOT("module:WalkingEngine:referenceAngleY");
  DECLARE_PLOT("module:WalkingEngine:referenceCoMVelocityY");
  DECLARE_PLOT("module:WalkingEngine:referenceCoMX");
  DECLARE_PLOT("module:WalkingEngine:referenceCoMY");

  DECLARE_PLOT("module:WalkingEngine:phase");
  DECLARE_PLOT("module:WalkingEngine:oscillation");

  DECLARE_PLOT("module:WalkingEngine:referenceLeftFootX");
  DECLARE_PLOT("module:WalkingEngine:referenceLeftFootY");
  DECLARE_PLOT("module:WalkingEngine:referenceRightFootX");
  DECLARE_PLOT("module:WalkingEngine:referenceRightFootY");

  DECLARE_PLOT("module:WalkingEngine:bodyShiftX");
  DECLARE_PLOT("module:WalkingEngine:bodyShiftY");
  DECLARE_PLOT("module:WalkingEngine:bodyShiftZ");

  DECLARE_PLOT("module:WalkingEngine:leftFootZ");
  DECLARE_PLOT("module:WalkingEngine:rightFootZ");
  DECLARE_PLOT("module:WalkingEngine:leftFootX");
  DECLARE_PLOT("module:WalkingEngine:rightFootX");
  DECLARE_PLOT("module:WalkingEngine:leftFootY");
  DECLARE_PLOT("module:WalkingEngine:rightFootY");

  DECLARE_PLOT("module:WalkingEngine:instability");

  DECLARE_PLOT("module:WalkingEngine:stepSizeX");
  DECLARE_PLOT("module:WalkingEngine:stepSizeY");

  DECLARE_PLOT("module:WalkingEngine:leftLift");
  DECLARE_PLOT("module:WalkingEngine:rightLift");

  // update stand output
  if(theMotionSelection.ratios[MotionRequest::stand] > 0. || theMotionSelection.ratios[MotionRequest::walk] > 0.)
  {
    if(lastBodyOriginTilt != p.originBodyTilt || // detect origin changes
      lastFootOrigin != p.originFoot ||
      lastFootOriginRotation != p.originFootRotation ||
      lastArmOriginRotation != p.originArmRotation ||
      lastHardness != Vector3<int>(p.hardness, p.hardnessAnklePitch, p.hardnessAnkleRoll))
    {
      // generate stand joint request
      Pose3D leftLeg(p.originFoot.x, p.originFoot.y, p.originFoot.z);
      Pose3D rightLeg(p.originFoot.x, -p.originFoot.y, p.originFoot.z);
      leftLeg.rotation = RotationMatrix(Vector3<>(p.originFootRotation.x, p.originFootRotation.y, p.originFootRotation.z));
      rightLeg.rotation = RotationMatrix(Vector3<>(-p.originFootRotation.x, p.originFootRotation.y, -p.originFootRotation.z));

      Pose3D left(RotationMatrix(Vector3<>(0., p.originBodyTilt, 0.)));
      Pose3D right(RotationMatrix(Vector3<>(0., p.originBodyTilt, 0.)));
      left.conc(leftLeg);
      right.conc(rightLeg);

      setLegJoints(left, right, 0.5, standOutput);
      setArmJoints(p.originArmRotation, p.originArmRotation, standOutput);
      setHeadJoints(0.f, 0.523599f, standOutput);

      // calculate the stand robot model      
      standRobotModel.setJointData(standOutput, theRobotDimensions, theMassCalibration);

      standCoMSet.leftFoot = Pose3D(standRobotModel.centerOfMass * -1.).conc(standRobotModel.limbs[RobotModel::footLeft]).translation;
      standCoMSet.rightFoot = Pose3D(standRobotModel.centerOfMass * -1.).conc(standRobotModel.limbs[RobotModel::footRight]).translation;

      coMShape.setConstants(p.originFoot.z - standRobotModel.centerOfMass.z - theRobotDimensions.heightLeg5Joint);

      lastBodyOriginTilt = p.originBodyTilt;
      lastFootOrigin = p.originFoot;
      lastFootOriginRotation = p.originFootRotation;
      lastArmOriginRotation = p.originArmRotation;
      lastHardness = Vector3<int>(p.hardness, p.hardnessAnklePitch, p.hardnessAnkleRoll);
    }

    setHeadJoints(standOutput);
  }

  // update walking engine output
  if(theFilteredSensorData.data[SensorData::gyroX] != SensorData::off && theFilteredSensorData.data[SensorData::gyroX] != 0.f)
    calibrated = true;
  if(!calibrated) // wait for autocalibration
    enforceStandTime = theFrameInfo.time + 1;
  if(theMotionSelection.ratios[MotionRequest::walk] > 0. && calibrated)
  {
    // reset?
    if(!wasActive)
    {
      coMShape.init(Vector2<>(0, p.modelWidth), -(p.modelWidth - p.modelSwingWidth));
      lastSupportLeg = none;
      oldCoMSets.init();
      instability.init();
      kickType = WalkRequest::none;
      kickState = 0;
      bodyShift = Vector3<>();
      leftStep = rightStep = Step();
      lastSelectedSpeed = Pose2D();
      oldOdometryOffsets.init();
      lastActivationTime = theFrameInfo.time;
    }

    // store walking target
    if(theMotionSelection.walkRequest.mode < WalkRequest::targetMode)
      walkTarget = Pose2D();
    else if(theMotionSelection.walkRequest.target != Pose2D())
    {
      walkTarget = theMotionSelection.walkRequest.target;
      if(MVTools::isNearInf(walkTarget.translation.x) || MVTools::isNaN(walkTarget.translation.x))
        walkTarget.translation.x = 0.;
      if(MVTools::isNearInf(walkTarget.translation.y) || MVTools::isNaN(walkTarget.translation.y))
        walkTarget.translation.y = 0.;
      if(MVTools::isNearInf(walkTarget.rotation) || MVTools::isNaN(walkTarget.rotation))
        walkTarget.rotation = 0.;

      walkTarget.translation.x /= p.odometryScale.translation.x;
      walkTarget.translation.y /= p.odometryScale.translation.y;
      walkTarget.rotation /= p.odometryScale.rotation;

      for(int i = oldOdometryOffsets.getNumberOfEntries() - 1; i >= 0; --i)
        walkTarget -= oldOdometryOffsets.getEntry(i);
    }
    if(theMotionSelection.walkRequest.kickType != WalkRequest::none && kickType == WalkRequest::none && theFrameInfo.getTimeSince(lastKickTime) > int(p.kickDuration))
    {
      kickType = theMotionSelection.walkRequest.kickType;
      kickState = 0;
      lastKickTime = theFrameInfo.time;
    }

    // determine error!
    calculateExpectedCoM();
    calculateMeasuredCoM();
    calculateError();
    
    // update com position
    if(wasActive)
    {
      Vector2<> cx(coMError.x, coMError.y);
      Vector2<> cxv(coMVelocityError.x, coMVelocityError.y);
      coMShape.update(cx, cxv, float(theFrameInfo.time - lastIterationTime) * 0.001f);
    }
    Step& step(coMShape.getRefZMP().y > 0.f ? rightStep : leftStep);
    step.size = coMShape.getStepSize(); // for the case that the controller changed the step size  
    phase = coMShape.getPhase();
    walkingEngineOutput.speed.translation = Vector2<>(step.size.x, step.size.y) / coMShape.getStepDuration();
    walkingEngineOutput.speed.rotation = step.rotation /  coMShape.getStepDuration();
    supportLeg = coMShape.getRefZMP().y > 0.f ? left : right;

    PLOT("module:WalkingEngine:phase", phase);
    PLOT("module:WalkingEngine:oscillation", coMShape.getOscillation());
    PLOT("module:WalkingEngine:stepSizeX", coMShape.getStepSize().x);
    PLOT("module:WalkingEngine:stepSizeY", coMShape.getStepSize().y);

    //
    calculateShapes();
    calculateTmpRobotFootVectors();
    calculateBodyRotation();
    calculateReferenceCoM();
    calculateRobotToFootVectors(walkingEngineOutput);

    // calculate odometry offset
    Pose2D calculatedOdometryOffset;
    if(wasActive)
    {
      const Pose3D& activeTarget(supportLeg == right ? finalLeftFoot : finalRightFoot);
      const Pose3D& activateLastTarget(supportLeg == right ? lastFinalLeftFoot : lastFinalRightFoot);
      const Pose3D& inactiveTarget(supportLeg == right ? finalRightFoot : finalLeftFoot);
      const Pose3D& inactivateLastTarget(supportLeg == right ? lastFinalRightFoot : lastFinalLeftFoot);
      calculatedOdometryOffset.translation.x = (activeTarget.translation.x - activateLastTarget.translation.x - (inactiveTarget.translation.x - inactivateLastTarget.translation.x)) * .5f;
      calculatedOdometryOffset.translation.y = (activeTarget.translation.y - activateLastTarget.translation.y - (inactiveTarget.translation.y - inactivateLastTarget.translation.y)) * .5f;
      calculatedOdometryOffset.rotation = (activeTarget.rotation.getZAngle() - activateLastTarget.rotation.getZAngle() - (inactiveTarget.rotation.getZAngle() - inactivateLastTarget.rotation.getZAngle())) * .5f;
    }
    lastFinalLeftFoot = finalLeftFoot;
    lastFinalRightFoot = finalRightFoot;
    oldOdometryOffsets.add(calculatedOdometryOffset);
    if(p.odometryUseTorsoMatrix)
    { // "measured"
      if(lastTorsoMatrix.translation.z != 0.)
      {
        Pose3D odometryOffset3D(lastTorsoMatrix);
        odometryOffset3D.conc(theTorsoMatrix.offset);
        odometryOffset3D.conc(theTorsoMatrix.invert());            
        walkingEngineOutput.odometryOffset.translation.x = odometryOffset3D.translation.x;
        walkingEngineOutput.odometryOffset.translation.y = odometryOffset3D.translation.y;
        walkingEngineOutput.odometryOffset.rotation = odometryOffset3D.rotation.getZAngle();
      }
    }
    else // use theoretical odometry offset
      walkingEngineOutput.odometryOffset = calculatedOdometryOffset;

    // remove odometry offset from walking target
    if(theMotionSelection.walkRequest.mode >= WalkRequest::targetMode)
      walkTarget -= calculatedOdometryOffset;

    // apply odometry scale
    walkingEngineOutput.odometryOffset.translation.x *= p.odometryScale.translation.x;
    walkingEngineOutput.odometryOffset.translation.y *= p.odometryScale.translation.y;
    walkingEngineOutput.odometryOffset.rotation *= p.odometryScale.rotation;

    // generate executed walk request
    walkingEngineOutput.executedWalk.mode = theMotionSelection.walkRequest.mode;
    walkingEngineOutput.executedWalk.speed = walkingEngineOutput.speed;
    if(walkingEngineOutput.executedWalk.mode == WalkRequest::percentageSpeedMode)
    {
      walkingEngineOutput.executedWalk.speed.translation.x /= walkingEngineOutput.executedWalk.speed.translation.x >= 0.f ? p.speedMax.translation.x : -p.speedBackwardMax;
      walkingEngineOutput.executedWalk.speed.translation.y /= p.speedMax.translation.y;
      walkingEngineOutput.executedWalk.speed.rotation /= p.speedMax.rotation;
    }
    walkingEngineOutput.executedWalk.target = walkTarget;
    walkingEngineOutput.executedWalk.pedantic = theMotionSelection.walkRequest.pedantic;

    // cutout
    walkingEngineOutput.instability = instability.getAverage();
    if(p.cutout && theFrameInfo.time - lastActivationTime > 3000 && walkingEngineOutput.instability >= p.cutoutInstability) // capitulation
      enforceStandTime = theFrameInfo.time + p.cutoutDuration;

    // store last activation state
    wasActive = true;
    lastSupportLeg = supportLeg;
  }
  else
    wasActive = false;    

  // set further informations
  walkingEngineOutput.positionInWalkCycle = phase;
  walkingEngineOutput.isLeavingPossible = isLeavingPossible;
  walkingEngineOutput.enforceStand = theFrameInfo.time < enforceStandTime;

  //
  lastIterationTime = theFrameInfo.time;
  lastTorsoMatrix = theTorsoMatrix;
}


void WalkingEngine::CoMShape::setConstants(float z)
{
  const float g = -9806.65f;
  k = sqrt(g / z);
}

void WalkingEngine::CoMShape::init(const Vector2<>& r, float x0)
{
  ASSERT(sgn(-r.y) == sgn(x0));
  this->r = r;
  this->x0 = x0;
  s = Vector2<>();  
  generateNextStepSize();  
  generateNextReferenceZMP();
  generateNextCycleDuration();
  tsl = tn;
  tso = -tn;
  calculateSwapTime();
  calculateXs();
  t = tso;
  xso = 0;
  xvso = 0;
  calculateX();
}

void WalkingEngine::CoMShape::update(const Vector2<>& cx, const Vector2<>& cxv, float dt)
{
  ASSERT(sgn(r.y) == sgn(x0n));
  applyCorrection(cx, cxv);
  t += dt;  
  calculateSwapTime();
  calculateXs();
  if(t >= ts)
  {
    t = tson + (t - ts);
    r = rn;
    x0 = x0n;
    tsl = ts;
    tso = tson;
    s = sn;
    xso = xson;
    xvso = xvson;    
    generateNextStepSize();
    generateNextReferenceZMP();
    generateNextCycleDuration();
    calculateSwapTime();
    calculateXs();
  }
  calculateX();
}

void WalkingEngine::CoMShape::applyCorrection(const Vector2<>& cx, const Vector2<>& cxv)
{
  if(!walkingEngine.p.balance)
    return;

  // step size balancing
  float newSnX = sn.x - cx.x * walkingEngine.p.balanceStepSize.x;
  if(fabs(newSnX) < fabs(sn.x))
    sn.x = newSnX;
  sn.y -= cx.y * walkingEngine.p.balanceStepSize.y;

  // reference zmp balancing
  float correctedRY = r.y - cx.y * walkingEngine.p.balanceRefZMP.y;
  if(sgn(r.y) != sgn(correctedRY))
  {
    // ignore
  }
  else
    r.y = correctedRY;
  r.x -= cx.x * walkingEngine.p.balanceRefZMP.x;

  // swing width and time balancing
  float xy(x.y + cx.y * walkingEngine.p.balanceModelCoM.y);
  float xvy(xv.y + cxv.y * walkingEngine.p.balanceModelCoMVelocity.y);
  if(xy * sgn(-r.y) < 0.f || sqr(xvy) / sqr(k) >= sqr(xy))
  {
    // ignore
  }
  else
  {
    x.y = xy;
    xv.y = xvy;

    // update swing width
    x0 = sqrt(sqr(x.y) - sqr(xv.y) / sqr(k)) * sgn(-r.y);  
    if(x0 * sgn(-r.y) <= 0.001f || sqr(x.y) <= sqr(xv.y) / sqr(k))
      x0 = 0.001f * sgn(-r.y);
    //ASSERT(x.y * sgn(-r.y) >= x0 * sgn(-r.y));

    // update time
    t = 1.f / k * saveAsinh(xv.y / (k * x0));
  }

}

void WalkingEngine::CoMShape::calculateSwapTime()
{
  x0n = -rn.y / cosh(k * tn);

  float xy, xvy, xvyn, xyn, dt;  

  tson = -tn;
  xvyn = k * x0n * sinh(k * tson);
  xvy = xvyn;
  ts = saveAsinh(xvy / (k * x0)) / k;
  ASSERT(ts > 0.f);
  xy = x0 * cosh(k * ts);    
  xyn = x0n * cosh(k * tson);

  for(int i = 0; i < 15; ++i)
  {
    dt = (sn.y - ((xy + r.y) + (-xyn - rn.y))) / (2 * xvy);    
    ts += dt;
    if(ts > 1.f)
      ts = 1.f;
    if(ts < 0.001f)
      ts = 0.001f;
    tson -= dt;
    if(tson > -0.001f)
      tson = -0.001f;
    if(tson < -2.f * tn + 0.001f)
      tson = -2.f * tn + 0.001f;
    if(fabs(dt) <= 0.0001)
      return;

    xy = x0 * cosh(k * ts);    
    xvy = k * x0 * sinh(k * ts);
    xvyn = xvy;
    ASSERT(-xvyn / (k * x0n) >= 0.f);
    tson = -1.f / k * saveAsinh(-xvyn / (k * x0n));
    ASSERT(tson < 0.f);
    if(tson < -2.f * tn + 0.001f)
      tson = -2.f * tn + 0.001f;
    xyn = x0n * cosh(k * tson);
  }

  /*
  float xy, xvy, xvyn, xyn, dt;  

  x0n = -rn.y / cosh(k * tn);
  ASSERT(sgn(x0n) != sgn(x0));
  //ts = tn;
  ts = saveAcosh(-r.y / x0) / k;
  
  for(int i = 0; i < 15; ++i)
  {
    xy = x0 * cosh(k * ts);    
    xvy = k * x0 * sinh(k * ts);
    xvyn = xvy;
    ASSERT(-xvyn / (k * x0n) >= 0.f);
    tson = -1.f / k * saveAsinh(-xvyn / (k * x0n));
    ASSERT(tson < 0.f);
    xyn = x0n * cosh(k * tson);

    dt = (sn.y - ((xy + r.y) + (-xyn - rn.y))) / (2 * xvy);    
    ts += dt;
    if(ts > 1.f)
      ts = 1.f;
    if(ts < 0.001f)
      ts = 0.001f;
    tson -= dt;
    if(tson > -0.001f)
      tson = -0.001f;
    if(tson < -2.f * tn + 0.001f)
      tson = -2.f * tn + 0.001f;
    if(fabs(dt) <= 0.0001)
      return;

    ASSERT(2 * tn + tson > 0.f);
    x0n = -rn.y / cosh(k * (2.f * tn + tson));
    ASSERT(sgn(x0n) != sgn(x0));
  }

  */
}

void WalkingEngine::CoMShape::calculateXs()
{
  float xv0, xv0n;
  xv0 = k * (-r.x + sn.x + rn.x) / (sinh(k * ts) - cosh(k * ts) * tanh(k * tson));
  xv0n = k * (-r.x + sn.x + rn.x) / (cosh(k * tson) * tanh(k * ts) - sinh(k * tson));
  xs = 1.f / k * xv0 * sinh(k * ts);
  xvs = xv0 * cosh(k * ts);
  xson = 1.f / k * xv0n * sinh(k * tson);
  xvson = xv0n * cosh(k * tson);

  ASSERT(fabs((xs - xson) - (-r.x + sn.x + rn.x)) < 0.01f || fabs(r.x) > 1000.f || fabs(rn.x) > 1000.f);
  BH_CHECK_NAN_AND_INF(xs);
  BH_CHECK_NAN_AND_INF(xvs);
  BH_CHECK_NAN_AND_INF(xson);
  BH_CHECK_NAN_AND_INF(xvson);
}

void WalkingEngine::CoMShape::calculateX()
{
  const float as1 = cosh(k * ts);
  const float sinhkts = sinh(k * ts);
  const float as2 = 1.f / k * sinhkts;
  const float as3 = k * sinhkts;
  const float as4 = as1;
  const float ao1 = cosh(k * tso);
  const float sinhktso = sinh(k * tso);
  const float ao2 = 1.f / k * sinhktso;
  const float ao3 = k * sinhktso;
  const float ao4 = ao1;
  const float tt = ts - tso;

  const float b1 = xvs * tt - (xs - xso);
  const float b2 = as3 * tt - (as1 - ao1);
  const float b3 = as4 * tt - (as2 - ao2);
  const float b4 = xvso * tt - (xs - xso);
  const float b5 = ao3 * tt - (as1 - ao1);
  const float b6 = ao4 * tt - (as2 - ao2);

  const float b2b6mb3b5 = b2 * b6 - b3 * b5;
  const float c1 = b2b6mb3b5 != 0.f ? (b1 * b6 - b3 * b4) / b2b6mb3b5 : (
    (b2 != 0.f && fabs(b3) < 0.00001) ? (b1 / b2) : (
    (b5 != 0.f && fabs(b6) < 0.00001) ? (b4 / b5) : 0.f));
  const float c2 = b3 != 0.f ? (b1 - c1 * b2) / b3 : (
    (b6 != 0.f && fabs(b5) < 0.00001) ? (b4 / b6) : 0.f);
  const float c3 = xvs - c1 * as3 - c2 * as4;
  const float c4 = xs - c1 * as1 - c2 * as2 - c3 * ts;

#ifndef NDEBUG
  float testXs = c1 * cosh(k * ts) + 1.f / k * c2 * sinh(k * ts) + c3 * ts + c4;
  float testXvs = k * c1 * sinh(k * ts) + c2 * cosh(k * ts) + c3;
  float testXso = c1 * cosh(k * tso) + 1.f / k * c2 * sinh(k * tso) + c3 * tso + c4;
  float testXvso = k * c1 * sinh(k * tso) + c2 * cosh(k * tso) + c3;
  ASSERT(fabs(testXs - xs) < 0.01f);
  ASSERT(fabs(testXvs - xvs) < 1.f);
  ASSERT(fabs(testXso - xso) < 0.01f);
  ASSERT(fabs(testXvso - xvso) < 1.f);
#endif

  x.x = c1 * cosh(k * t) + 1.f / k * c2 * sinh(k * t) + c3 * t + c4;
  x.y = x0 * cosh(k * t);
  xv.x = k * c1 * sinh(k * t) + c2 * cosh(k * t) + c3;
  xv.y = k * x0 * sinh(k * t);

  BH_CHECK_NAN_AND_INF(x.x);
  BH_CHECK_NAN_AND_INF(x.y);
  BH_CHECK_NAN_AND_INF(xv.x);
  BH_CHECK_NAN_AND_INF(xv.y);
}

void WalkingEngine::CoMShape::generateNextStepSize()
{
  walkingEngine.generateNextStepSize(r.y < 0.f ? WalkingEngine::right : WalkingEngine::left);
  const Vector3<>& stepSize(r.y < 0.f ? walkingEngine.leftStep.size : walkingEngine.rightStep.size);
  sn = Vector2<>(stepSize.x, stepSize.y);
}

void WalkingEngine::CoMShape::generateNextReferenceZMP()
{
  float modelWidth = walkingEngine.p.modelWidth;
  float yxSpeedFactor = (walkingEngine.p.modelWidthAtFullSpeed.x - modelWidth) / (walkingEngine.p.speedMax.translation.x * 0.5f);
  float yySpeedFactor = (walkingEngine.p.modelWidthAtFullSpeed.y - modelWidth) / walkingEngine.p.speedMax.translation.y;
  modelWidth += fabs(sn.x) * yxSpeedFactor + fabs(sn.y) * yySpeedFactor;

  float modelShift = 0;
  float modelShiftSpeedFactor = (walkingEngine.p.modelShiftAtFullSpeed - 0.f) / (walkingEngine.p.speedMax.translation.x * 0.5f);
  modelShift += fabs(sn.x) * modelShiftSpeedFactor;

  rn.x = modelShift;
  rn.y = modelWidth * sgn(-r.y);
}

void WalkingEngine::CoMShape::generateNextCycleDuration()
{
  float modelSwingWidth = walkingEngine.p.modelSwingWidth;
  float yxSpeedFactor = (walkingEngine.p.modelSwingWidthAtFullSpeed.x - modelSwingWidth) / (walkingEngine.p.speedMax.translation.x * 0.5f);
  float yySpeedFactor = (walkingEngine.p.modelSwingWidthAtFullSpeed.y - modelSwingWidth) / walkingEngine.p.speedMax.translation.y;
  modelSwingWidth += fabs(sn.x) * yxSpeedFactor + fabs(sn.y) * yySpeedFactor;

  const Step& step(r.y < 0.f ? walkingEngine.leftStep : walkingEngine.rightStep);
  if(step.preKick)
    modelSwingWidth = walkingEngine.p.kickSwingWidth;

  float xn = rn.y - modelSwingWidth * sgn(rn.y);
  tn = 1.f / k  * saveAcosh(rn.y / xn);
}
