/** 
* @file Modules/MotionControl/WalkingEngine.h
* Declaration a module that creates the walking motions
* @author Colin Graf
*/

#ifndef WalkingEngine_H
#define WalkingEngine_H

#include "Tools/Module/Module.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/Configuration/MassCalibration.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Sensing/RobotModel.h"
#include "Representations/MotionControl/MotionSelection.h"
#include "Representations/MotionControl/WalkingEngineStandOutput.h"
#include "Representations/MotionControl/WalkingEngineOutput.h"
#include "Representations/MotionControl/HeadJointRequest.h"
#include "Representations/Sensing/TorsoMatrix.h"
#include "Tools/Math/Pose3D.h"
#include "Tools/RingBuffer.h"
#include "Tools/RingBufferWithSum.h"

MODULE(WalkingEngine)
  REQUIRES(RobotDimensions)
  REQUIRES(MassCalibration)
  REQUIRES(FrameInfo)
  REQUIRES(MotionSelection)
  REQUIRES(RobotModel)
  REQUIRES(HeadJointRequest)
  REQUIRES(FilteredSensorData)
  REQUIRES(TorsoMatrix)
  PROVIDES_WITH_MODIFY(WalkingEngineOutput)
  REQUIRES(WalkingEngineOutput)
  PROVIDES_WITH_MODIFY(WalkingEngineStandOutput)
END_MODULE

class WalkingEngine : public WalkingEngineBase
{
public:
  /**
  * Default constructor
  */
  WalkingEngine();

  /**
  * A collection of walking engine parameters
  */
  class Parameters : public Streamable
  {
  public:
    /** Default constructor */
    Parameters() {}

    Vector3<> originFoot;
    Vector3<> originFootRotation;
    float originBodyTilt;
    Vector2<> originArmRotation;

    float modelWidth;
    Vector2<> modelWidthAtFullSpeed;
    float modelSwingWidth;    
    Vector2<> modelSwingWidthAtFullSpeed;
    float modelShiftAtFullSpeed;

    float bodyRotation;
    
    Vector2<> liftPhases;
    Vector2<> liftPhasesAtFullSpeed;
    Vector3<> liftOffset;
    Vector3<> liftOffsetAtFullSpeedY;

    Vector2<> movePhases;
    Vector2<> movePhasesAtFullSpeed;
    Vector2<> moveArmRotation;
    Vector2<> moveBodyRotationFactor;

    int hardness;
    int hardnessAnklePitch;
    int hardnessAnkleRoll;

    Pose2D speedMax;
    float speedBackwardMax;
    Pose2D speedDashMax;
    Pose2D speedChangeMax;
    bool speedUseEllipsoidClipping;

    unsigned int measurementDelay;
    bool measurementUseRobotModel;

    Vector3<> errorCoMMax;
    Vector3<> errorCoMVelocityMax;
    Vector2<> errorRotationMax;

    bool balance;
    Vector2<> balanceStepSize;
    Vector2<> balanceRefZMP;
    Vector2<> balanceModelCoM;
    Vector2<> balanceModelCoMVelocity;
    Vector3<> balanceCoM;
    Vector2<> balanceBodyRotation;
    Vector2<> balanceFootRotation;
    Vector2<> balanceArm;

    bool cutout;
    float cutoutInstability;
    unsigned int cutoutDuration;

    Pose2D odometryScale;
    bool odometryUseTorsoMatrix;

    Vector3<> kickStrikeOut;
    Vector3<> kickDestination;
    float kickSwingWidth;
    Vector3<> kickStepSize;
    Vector3<> kickSidewardsPreStepSize;
    Vector3<> kickSidewardsStrikeOut;
    Vector3<> kickSidewardsDestination;
    Vector3<> kickSidewardsStepSize;
    unsigned int kickDuration;

  private:
    /**
    * Makes the object streamable
    * @param in The stream from which the object is read
    * @param out The stream to which the object is written.
    */
    void serialize(In* in, Out* out)
    {  
      STREAM_REGISTER_BEGIN();

        STREAM(originFoot);
        STREAM(originFootRotation);
        STREAM(originBodyTilt);
        STREAM(originArmRotation);

        STREAM(modelWidth);
        STREAM(modelWidthAtFullSpeed);
        STREAM(modelSwingWidth);
        STREAM(modelSwingWidthAtFullSpeed);
        STREAM(modelShiftAtFullSpeed);

        STREAM(bodyRotation);

        STREAM(liftPhases); 
        STREAM(liftPhasesAtFullSpeed);
        STREAM(liftOffset);
        STREAM(liftOffsetAtFullSpeedY);

        STREAM(movePhases);
        STREAM(movePhasesAtFullSpeed);
        STREAM(moveArmRotation);
        STREAM(moveBodyRotationFactor);

        STREAM(hardness);
        STREAM(hardnessAnklePitch);
        STREAM(hardnessAnkleRoll);

        STREAM(speedMax);
        STREAM(speedBackwardMax);
        STREAM(speedDashMax);
        STREAM(speedChangeMax);
        STREAM(speedUseEllipsoidClipping);
        
        STREAM(measurementDelay);
        STREAM(measurementUseRobotModel);

        STREAM(errorCoMMax);
        STREAM(errorCoMVelocityMax);
        STREAM(errorRotationMax);

        STREAM(balance);
        STREAM(balanceStepSize);
        STREAM(balanceRefZMP);
        STREAM(balanceModelCoM);
        STREAM(balanceModelCoMVelocity);
        STREAM(balanceCoM);
        STREAM(balanceBodyRotation);
        STREAM(balanceFootRotation);
        STREAM(balanceArm);

        STREAM(cutout);
        STREAM(cutoutInstability);
        STREAM(cutoutDuration);

        STREAM(odometryScale);
        STREAM(odometryUseTorsoMatrix);

        STREAM(kickStrikeOut);
        STREAM(kickDestination);
        STREAM(kickSwingWidth);
        STREAM(kickStepSize);
        STREAM(kickSidewardsPreStepSize);
        STREAM(kickSidewardsStrikeOut);
        STREAM(kickSidewardsDestination);
        STREAM(kickSidewardsStepSize);
        STREAM(kickDuration)

      STREAM_REGISTER_FINISH();
    }
  };

private:

  enum SupportLeg
  {
    none,
    right,
    left,
  };  

  class Step
  {
  public:
    Step() : rotation(0.f), preKick(false), kick(false) {}
    Vector3<> size;
    Vector3<> originalSize;
    float rotation;
    Vector2<> liftPhases;
    Vector3<> liftOffset;
    Vector2<> movePhases;
    Vector2<> bodyRotation;

    bool preKick;
    bool kick;
    Vector3<> kickStrikeOut;
    Vector3<> kickDestination;
  };
  
  class CoMSet
  {
  public:
    CoMSet(const Vector3<>& leftFoot = Vector3<>(), const Vector3<>& rightFoot = Vector3<>(),
      const Vector3<>& velocity = Vector3<>(), const Vector2<>& bodyRotation = Vector2<>(), unsigned int timeStamp = 0) : 
      leftFoot(leftFoot), rightFoot(rightFoot), velocity(velocity), bodyRotation(bodyRotation), timeStamp(timeStamp) {}
    Vector3<> leftFoot; /**< CoM to foot */
    Vector3<> rightFoot; /**< CoM to foot */
    Vector3<> velocity; /**< CoM velocity */
    Vector2<> bodyRotation;
    unsigned int timeStamp;
  };

  class CoMShape
  {
  public:

    CoMShape(WalkingEngine& walkingEngine) : walkingEngine(walkingEngine) {}

    void setConstants(float z);

    void init(const Vector2<>& r, float x0);

    void update(const Vector2<>& cx, const Vector2<>& cxv, float dt);

    inline float getPhase() const {return 0.5f * ((t - tso) / (ts - tso)) + (r.y > 0.f ? 0.5f : 0.f);}

    inline float getOscillation() const
    {
      if(t > 0)
        return sin((t / (ts - tson) * 2.f - 1.f) * pi_2) * sgn(-x0);
      else
        return sin((t / (tso - tsl) * 2.f - 1.f) * pi_2) * sgn(-x0);
    }

    inline Vector3<> getPosition() const {return Vector3<>(x.x, x.y, 0);}

    inline Vector3<> getRefZMP() const {return Vector3<>(r.x, r.y, 0);}

    inline Vector3<> getStepSize() const {return Vector3<>(sn.x, sn.y, 0);}

    inline float getStepDuration() const {return 2.f * tn;}

  private:
    WalkingEngine& walkingEngine;

    float k; /** sqrt(g / z) */

    Vector2<> x; /**< Current position */
    Vector2<> xv; /**< Current velocity */
    Vector2<> r; /**< Current reference ZMP */
    Vector2<> rn; /**< Next reference ZMP */
    float t; /**< Current time */
    float tn; /**< (cycle duration) / 4 */
    float ts; /**< Swap time */
    float tsl; /**< Last swap time */
    float tso; /**< Old swap time */
    float tson; /**< Next old swap time */
    Vector2<> sn; /**< Step offset */
    Vector2<> s; /**< Old step offset */
    float x0; /**< Current swing width */
    float x0n; /**< Next swing width */
    float xs; /* x-Position at swap time */
    float xvs; /* x-Velocity at swap time */
    float xso; /* x-Position at old swap time */
    float xvso; /* x-Velocity at old swap time */
    float xson; /* x-Position at next old swap time */
    float xvson; /* x-Velocity at next old swap time */
    
    void applyCorrection(const Vector2<>& cx, const Vector2<>& cxv);
    void calculateSwapTime();
    void calculateXs();
    void calculateX();
    void generateNextStepSize();
    void generateNextReferenceZMP();
    void generateNextCycleDuration();
  };

  Parameters p;

  void init();

  /**
  * The central update method to generate the walking motion
  * @param walkingEngineOutput The WalkingEngineOutput (mainly the resulting joint angles)
  */
  void update(WalkingEngineOutput& walkingEngineOutput);
  bool calibrated;
  unsigned int enforceStandTime;
  WalkingEngineStandOutput standOutput;
  RobotModel standRobotModel;
  float lastBodyOriginTilt;
  Vector3<> lastFootOrigin;
  Vector3<> lastFootOriginRotation;
  Vector2<> lastArmOriginRotation;
  Vector3<int> lastHardness;
  CoMSet standCoMSet;
  SupportLeg supportLeg;
  SupportLeg lastSupportLeg;
  Pose3D lastTorsoMatrix;
  Pose3D lastFinalLeftFoot;
  Pose3D lastFinalRightFoot;
  float phase; /**< The current position in walk phase. [0-1[ */
  bool wasActive; /**< Was this module active in the previous frame? */
  unsigned int lastIterationTime; /**< The time stamp of last execution */
  Pose2D walkTarget;
  CoMShape coMShape;
  RingBuffer<Pose2D, 5> oldOdometryOffsets;
  unsigned int lastActivationTime;
  unsigned int lastKickTime;

  /**
  * The update method to generate the standing motion from walkingParameters
  * @param standOutput The WalkingEngineStandOutput (mainly the resulting joint angles)
  */
  void update(WalkingEngineStandOutput& standOutput) { standOutput = this->standOutput; }

  /**
  * The method sets the joint angles (independently from the legs)
  * @param leftRotation The angle of LShoulderPitch and LShoulderRoll
  * @param rightRotation The angle of RShoulderPitch  and RShoulderRoll
  * @param jointRequest The JointRequest in which the arm joints are set
  */
  void setArmJoints(const Vector2<>& leftRotation, const Vector2<>& rightRotation, JointRequest& jointRequest);

  void setLegJoints(const Pose3D& leftTarget, const Pose3D& rightTarget, float ratio, JointRequest& jointRequest);

  void setHeadJoints(JointRequest& jointRequest);
  void setHeadJoints(float pan, float tilt, JointRequest& jointRequest);

  void calculateShapes();
  float leftFadeIn;
  float rightFadeIn;
  float leftLift;
  float rightLift;
  float leftStrikeOut;
  float rightStrikeOut;
  float leftKick;
  float rightKick;
  float leftMoveFadeIn; 
  float rightMoveFadeIn;

  void calculateTmpRobotFootVectors();
  Vector3<> tmpLeftFoot;
  Vector3<> tmpRightFoot;
  float leftRotationZ;
  float rightRotationZ;
  Vector2<> leftArm;
  Vector2<> rightArm;
  Vector2<> bodyStepRotation;

  void calculateBodyRotation();
  Vector2<> bodyRotation;
  RotationMatrix world2Body;
  RotationMatrix body2World;

  void calculateReferenceCoM();
  CoMSet referenceCoMSet;
  CoMSet lastReferenceCoMSet;
  RingBuffer<CoMSet, 20> oldCoMSets;

  void generateNextStepSize(SupportLeg supportLeg);
  Pose2D lastSelectedSpeed;
  Step leftStep;
  Step rightStep;
  bool isLeavingPossible;
  unsigned int kickState;
  WalkRequest::KickType kickType;

  void calculateRobotToFootVectors(JointRequest& jointRequest);
  Vector3<> bodyShift;
  Vector3<> lastNewCoM;
  Pose3D finalLeftFoot;
  Pose3D finalRightFoot;

  void calculateExpectedCoM();
  CoMSet expectedCoMSet;

  void calculateMeasuredCoM();
  CoMSet measuredCoMSet;
  CoMSet lastMeasuredCoMSet;

  void calculateError();
  Vector3<> coMError;
  Vector3<> coMVelocityError;
  Vector3<> rotationError;
  RingBufferWithSum<float, 100> instability;

  friend class CoMShape;
};

#endif // WalkingEngine_H
