#include "RobotPerceptor.h"
#include "Representations/Infrastructure/GameInfo.h"
#include "RobotPerceptorHelper.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Debugging/Stopwatch.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Streams/ConfigHelper.h"
#include "Tools/Boundary.h"
#include "Tools/Range.h"
#include <numeric>
#include <functional>
#include <cstdio>

RobotPerceptor::RobotPerceptor()
{
  ConfigHelper::loadConfig(params, "robotPerceptor.cfg", "perceptor");
  pointExplorer.initFrame(&theImage, &theColorTable64, params.exploreStepSize,
    params.gridStepSize, params.skipOffset, params.minSegSize);
}

void RobotPerceptor::update(RobotPercept& robotPercept)
{
  DECLARE_DEBUG_DRAWING("module:RobotPerceptor:regions", "drawingOnImage");
  DECLARE_DEBUG_DRAWING("module:RobotPerceptor:scanlines", "drawingOnImage");
  DECLARE_DEBUG_DRAWING("module:RobotPerceptor:orientation", "drawingOnImage");

  prepareRobotPercept(robotPercept);

  ASSERT(robotPercept.robots.empty());
  ASSERT(robotPercept.unlocalizableRobots.empty());

  for(const RegionPercept::Region* region = theRegionPercept.regions, * end = region + theRegionPercept.regionsCounter; region < end; ++region)
  {
    // TODO functions
    switch(region->color)
    {
      case ColorClasses::blue:
      case ColorClasses::red:
      {
        float areaTruth = areaMembership(region->size);
        if(areaTruth < 0.0f)
          break;

        const int m00 = region->calcMoment00(),
                  m10 = region->calcMoment10(),
                  m01 = region->calcMoment01();
        if(m00 == 0.0f)
          break;

        const Vector2<int> cog(m10/m00, m01/m00);
        CROSS("module:RobotPerceptor:regions", cog.x, cog.y, 2, 2, Drawings::ps_solid, ColorClasses::black);

        const float m20 = (float) region->calcCMoment20(cog.x),
                    m02 = (float) region->calcCMoment02(cog.y),
                    m11 = (float) region->calcCMoment11(cog.x, cog.y);
        const float cm20 = m20/m00,
                    cm02 = m02/m00,
                    cm11 = m11/m00;
        if(cm20 == cm02)
          break;

        const float cmsum = cm20 + cm02;
        const float cmdiff = cm20 - cm02;
        const float eccroot = sqrtf(cmdiff*cmdiff + 4.0f * cm11*cm11);
        const float eccentricity = (cmsum + eccroot) / (cmsum - eccroot);
        float eccentricityTruth = eccentricityMembership(eccentricity);
        if(eccentricityTruth < 0.0f)
          break;

        const float alpha = 0.5f * atan2f(-2.0f * m11, m02 - m20) + pi_2;
        float len_max = 0, len_min = 0;
        principleAxisLength(len_min, len_max, region, cog,
            Vector2<>(cosf(alpha), sinf(alpha)),
            Vector2<>(cosf(alpha + pi_2), sinf(alpha + pi_2)));
        const float sinAlpha = sinf(alpha);
        const float cosAlpha = cosf(alpha);
        const float lenDiff = len_max - len_min;

        Vector2<int> axis = Vector2<int>((int) (cosAlpha * lenDiff), (int) (sinAlpha * lenDiff)) * params.axisND.x / params.axisND.y;
        if(axis.x < 0) axis = axis * -1;
        Vector2<int> left = cog - axis;
        Vector2<int> right = cog + axis;
        Vector2<int> perpendicular = Matrix2x2<int>(0, -1, 1, 0) * axis;

        Vector2<int> slLeftUp = left - perpendicular * 5;
        Vector2<int> slMiddleUp = cog - perpendicular * 5;
        Vector2<int> slRightUp = right - perpendicular * 5;
        int whitePixelsAbove = 0;
        int totalPixelsAbove = 0;
        totalPixelsAbove += scanline(theImage, theColorTable64, region->color, left, slLeftUp, whitePixelsAbove);
        totalPixelsAbove += scanline(theImage, theColorTable64, region->color, cog, slMiddleUp, whitePixelsAbove);
        totalPixelsAbove += scanline(theImage, theColorTable64, region->color, right, slRightUp, whitePixelsAbove);
        float aboveWhiteRatio = (float) whitePixelsAbove / (float) totalPixelsAbove;
        float aboveTruth(0.0f);
        if(aboveWhiteRatio > 0.5f && aboveWhiteRatio < 0.6f)
          aboveTruth = (aboveWhiteRatio - 0.5f) / 0.1f;
        else if(aboveWhiteRatio >= 0.6f)
          aboveTruth = 1.0f;
        ASSERT(aboveTruth >= 0.0f);
        ASSERT(aboveTruth <= 1.0f);

        Vector2<int> slLeftDown = left + perpendicular * 5;
        Vector2<int> slMiddleDown = cog + perpendicular * 1;
        Vector2<int> slRightDown = right + perpendicular * 5;
        int whitePixelsBelow = 0;
        int totalPixelsBelow = 0;
        totalPixelsBelow += scanline(theImage, theColorTable64, region->color, left, slLeftDown, whitePixelsBelow);
        totalPixelsBelow += scanline(theImage, theColorTable64, region->color, cog, slMiddleDown, whitePixelsBelow);
        totalPixelsBelow += scanline(theImage, theColorTable64, region->color, right, slRightDown, whitePixelsBelow);
        float belowWhiteRatio = (float) whitePixelsBelow / (float) totalPixelsBelow;
        float belowTruth(0.0f);
        if(belowWhiteRatio > 0.4f && belowWhiteRatio < 0.5f)
          belowTruth = (belowWhiteRatio - 0.4f) / 0.1f;
        else if(belowWhiteRatio >= 0.5f)
          belowTruth = 1.0f;
        ASSERT(belowTruth >= 0.0f);
        ASSERT(belowTruth <= 1.0f);

        if(areaTruth * eccentricityTruth * aboveTruth * belowTruth < 0.5f)
          break;

        Vector2<int> lowestPx(cog.x, cog.y);
        while(lowestPx.y < theImage.cameraInfo.resolutionHeight
              && (ColorClasses::green != getColor(theColorTable64, theImage, lowestPx.x, lowestPx.y)
              || pointExplorer.colorRatio(ColorClasses::green,
                Boundary<int>(Range<int>(int(lowestPx.x - len_max), int(lowestPx.x + len_max)),
                              Range<int>(lowestPx.y, lowestPx.y + 2)), 2) < params.greenGroundRatio))
        {
          lowestPx.y++;
        }

        if(((alpha >= 0.5f && alpha <= 1.2f) || (alpha >= 2.0f && alpha <= 2.5f)))
          break;
        bool standing = ((alpha >= 0.0f && alpha <= 0.5f) || (alpha >= pi - 0.5f && alpha <= pi)) ? true : false;

        int pixelHeight = lowestPx.y - cog.y;
        // this is too high to be a horizontal robot (e. g. a goal post)
        if(!standing && pixelHeight > 100) // TODO params
          break;
        // small high team markers are not relevant
        if(region->size < 100 && pixelHeight > 100) // TODO params
          break;

        COMPLEX_DRAWING("module:RobotPerceptor:orientation",
        {
          const int axisLength = 30;
          CIRCLE("module:RobotPerceptor:orientation", cog.x, cog.y, axisLength,
              1, Drawings::ps_solid, ColorClasses::white, Drawings::ps_null, ColorClasses::blue);
          LINE("module:RobotPerceptor:orientation", cog.x, cog.y, cog.x + axisLength, cog.y,
              1, Drawings::ps_solid, ColorClasses::white);
          Vector2<> normalizedAxis = Vector2<>(cosf(0.5f), sinf(0.5f)) * (float) axisLength;
          LINE("module:RobotPerceptor:orientation", cog.x, cog.y,
              cog.x + (int) normalizedAxis.x, cog.y + (int) normalizedAxis.y,
              1, Drawings::ps_solid, ColorClasses::white);
          normalizedAxis = Vector2<>(cosf(pi-0.5f), sinf(pi-0.5f)) * (float) axisLength;
          LINE("module:RobotPerceptor:orientation", cog.x, cog.y,
              cog.x + (int) normalizedAxis.x, cog.y + (int) normalizedAxis.y,
              1, Drawings::ps_solid, ColorClasses::white);
          normalizedAxis = Vector2<>(cosf(pi), sinf(pi)) * (float) axisLength;
          LINE("module:RobotPerceptor:orientation", cog.x, cog.y,
              cog.x + (int) normalizedAxis.x, cog.y + (int) normalizedAxis.y,
              1, Drawings::ps_solid, ColorClasses::white);
          normalizedAxis = Vector2<>(cosAlpha, sinAlpha) * (float) (axisLength/2);
          LINE("module:RobotPerceptor:orientation", cog.x, cog.y,
              cog.x + (int) normalizedAxis.x, cog.y + (int) normalizedAxis.y,
              1, Drawings::ps_solid, region->color);
          char BUFFER[50];
          std::sprintf(BUFFER, "%.3f", alpha);
          DRAWTEXT("module:RobotPerceptor:orientation", cog.x - 5, cog.y - 10,
              10, ColorClasses::black, BUFFER);
        });

        COMPLEX_DRAWING("module:RobotPerceptor:scanlines",
        {
          Vector2<int> p1 = Vector2<int>((int)(cog.x + cosAlpha * len_max), (int)(cog.y + sinAlpha * len_max));
          Vector2<int> p2 = Vector2<int>((int)(cog.x + cosAlpha * len_min), (int)(cog.y + sinAlpha * len_min));
          LINE("module:RobotPerceptor:scanlines", cog.x, cog.y, p1.x, p1.y,
              2, Drawings::ps_solid, region->color);
          LINE("module:RobotPerceptor:scanlines", cog.x, cog.y, p2.x, p2.y,
              2, Drawings::ps_solid, region->color);
          LINE("module:RobotPerceptor:scanlines", slLeftDown.x, slLeftDown.y, slLeftUp.x, slLeftUp.y,
              2, Drawings::ps_solid, ColorClasses::white);
          LINE("module:RobotPerceptor:scanlines", slMiddleDown.x, slMiddleDown.y, slMiddleUp.x, slMiddleUp.y,
              2, Drawings::ps_solid, ColorClasses::white);
          LINE("module:RobotPerceptor:scanlines", slRightDown.x, slRightDown.y, slRightUp.x, slRightUp.y,
              2, Drawings::ps_solid, ColorClasses::white);

          char BUFFER[50];
          std::sprintf(BUFFER, "region = %d", region->size);
          DRAWTEXT("module:RobotPerceptor:scanlines", cog.x + 20, cog.y + 10,
              10, ColorClasses::black, BUFFER);
          std::sprintf(BUFFER, "white = %d/%d, %d/%d", whitePixelsAbove, totalPixelsAbove,
                                                       whitePixelsBelow, totalPixelsBelow);
          DRAWTEXT("module:RobotPerceptor:scanlines", cog.x + 20, cog.y + 20,
              10, ColorClasses::black, BUFFER);
          std::sprintf(BUFFER, "eccentricity = %.3f", eccentricity);
          DRAWTEXT("module:RobotPerceptor:scanlines", cog.x + 20, cog.y + 30,
              10, ColorClasses::black, BUFFER);
          std::sprintf(BUFFER, "%.3f, %d (standing=%s)", alpha, lowestPx.y - cog.y, standing ? "true" : "false");
          DRAWTEXT("module:RobotPerceptor:scanlines", cog.x + 20, cog.y + 40,
              10, ColorClasses::black, BUFFER);
          CROSS("module:RobotPerceptor:scanlines", lowestPx.x, lowestPx.y,
              2, 2, Drawings::ps_solid, ColorClasses::green);
        });

        if(exactLocalisation(theImage, lowestPx))
        {
          robotPercept.robots.push_back(RobotPercept::Robot(lowestPx, standing, region->color));
        }
        else
        {
          robotPercept.unlocalizableRobots.push_back(RobotPercept::Robot(lowestPx, true, region->color));
        }

        break;
      }
      default:
        break;
    }
  }
}

void RobotPerceptor::prepareRobotPercept(RobotPercept& robotPercept)
{
  robotPercept.robots.clear();
  robotPercept.unlocalizableRobots.clear();
  const static std::size_t MAX_PIXELS(theImage.cameraInfo.resolutionWidth / params.exploreStepSize);
  robotPercept.robots.reserve(MAX_PIXELS);
  robotPercept.unlocalizableRobots.reserve(MAX_PIXELS);
}

float RobotPerceptor::areaMembership(int area)
{
  float areaTruth(0.0f);
  if(area <= params.teamMarkerRegionMinSize)
    areaTruth = -1.0f;
  else if(area <= 500)
    areaTruth = 1.0f;
  else if(area <= 1500)
    areaTruth = (1500.0f - (float) area) / 1000.0f;
  ASSERT(areaTruth >= 0.0f || areaTruth == -1.0f);
  ASSERT(areaTruth <= 1.0f || areaTruth == -1.0f);
  return areaTruth;
}

float RobotPerceptor::eccentricityMembership(float eccentricity)
{
  float eccentricityTruth(0.0f);
  // ensures that the eccentricity is within the boundaries and not inf or NaN
  if(!(eccentricity > 1.5f && eccentricity < params.eccentricityBoundaries.y))
    eccentricityTruth = -1.0f;
  if(eccentricity >= 1.5f && eccentricity <= params.eccentricityBoundaries.x)
    eccentricityTruth = (2.0f - eccentricity) / 0.5f;
  if(eccentricity >= 2.0f && eccentricity < 12.5f)
    eccentricityTruth = 1.0f;
  else if(eccentricity >= 12.5 && eccentricity <= params.eccentricityBoundaries.y)
    eccentricityTruth = (eccentricity - 12.5f) / 0.5f;
  ASSERT(eccentricityTruth >= 0.0f || eccentricityTruth == -1.0f);
  ASSERT(eccentricityTruth <= 1.0f || eccentricityTruth == -1.0f);
  return eccentricityTruth;
}

void RobotPerceptor::principleAxisLength(float& len_min, float& len_max, const RegionPercept::Region* region, const Vector2<int>& cog, const Vector2<>& achse1, const Vector2<>& achse2)
{
  float len_max2 = 0, len_min2 = 0;

  typedef std::vector<RegionPercept::Segment*>::const_iterator CSI;
  for(CSI child = region->childs.begin(); child != region->childs.end(); child++)
  {
    const RegionPercept::Segment* seg = *child;
    const Vector2<> childRelToSwp(float(seg->x - cog.x), float(seg->y - cog.y)),
                    child2RelToSwp(float(seg->x - cog.x), float(seg->y + seg->length - cog.y));

    const float dist = childRelToSwp * achse1;
    if(dist > len_max)
      len_max = dist;
    else if(dist < len_min)
      len_min = dist;

    const float dist2 = child2RelToSwp * achse1;
    if(dist2 > len_max)
      len_max = dist2;
    else if(dist2 < len_min)
      len_min = dist2;

    const float dist3 = childRelToSwp * achse2;
    if(dist3 > len_max2)
      len_max2 = dist3;
    else if(dist3 < len_min2)
      len_min2 = dist3;

    const float dist4 = child2RelToSwp * achse2;
    if(dist4 > len_max2)
      len_max2 = dist4;
    else if(dist4 < len_min2)
      len_min2 = dist4;
  }
}

void RobotPerceptor::Parameters::serialize(In* in, Out* out)
{
  STREAM_REGISTER_BEGIN();
  STREAM(teamMarkerRegionMinSize);
  STREAM(eccentricityBoundaries);
  STREAM(aboveWhiteRatio);
  STREAM(belowWhiteRatio);
  STREAM(axisND);
  STREAM(greenGroundRatio);
  STREAM(exploreStepSize);
  STREAM(gridStepSize);
  STREAM(skipOffset);
  STREAM_ARRAY(minSegSize);
  STREAM_REGISTER_FINISH();
}

MAKE_MODULE(RobotPerceptor, Perception)

