 /**
* @file FallDownStateDetector.cpp
*
* This file implements a module that provides information about the current state of the robot's body.
*
* @author <a href="mailto:maring@informatik.uni-bremen.de">Martin Ring</a>
*/

#include "FallDownStateDetector.h"
#include "Representations/Infrastructure/JointData.h"
#include "Platform/SoundPlayer.h"


FallDownStateDetector::FallDownStateDetector(){
  InConfigFile stream(Global::getSettings().expandLocationFilename("fallDownStateDetector.cfg"), "default");
  if(stream.exists())
    stream >> parameters;
  else
  {
    InConfigFile stream("fallDownStateDetector.cfg", "default");
    ASSERT(stream.exists());
    stream >> parameters;
  }
  parameters.fallDownAngleX *= pi_180;
  parameters.fallDownAngleY *= pi_180;
  parameters.onGroundAngle  *= pi_180;
  parameters.staggeringAngleX *= pi_180;
  parameters.staggeringAngleY *= pi_180;  
   
  lastFallDetected = (unsigned int)-parameters.fallTime; 
  calibrated = false;  
}

void FallDownStateDetector::update(FallDownState& fallDownState)
{
  //Buffer data:
  buffers[accX].add(theFilteredSensorData.data[SensorData::accX]);
  buffers[accY].add(theFilteredSensorData.data[SensorData::accY]);
  buffers[accZ].add(theFilteredSensorData.data[SensorData::accZ]);

  //Compute average acceleration values and angles:
  float accXaverage(buffers[accX].getAverage());
  float accYaverage(buffers[accY].getAverage());
  float accZaverage(buffers[accZ].getAverage());
  float accelerationAngleXZ(atan2f(accZaverage, accXaverage));
  float accelerationAngleYZ(atan2f(accZaverage, accYaverage));
  MODIFY("module:FallDownStateDetector:accX",  accXaverage);
  MODIFY("module:FallDownStateDetector:accY",  accYaverage);
  MODIFY("module:FallDownStateDetector:accZ",  accZaverage);
  MODIFY("module:FallDownStateDetector:accAngleXZ", accelerationAngleXZ);
  MODIFY("module:FallDownStateDetector:accAngleYZ", accelerationAngleYZ);


  if (isCalibrated() && !specialSpecialAction())
  {        
    if (theFrameInfo.getTimeSince(lastFallDetected) <= parameters.fallTime)
    {
      fallDownState.state = FallDownState::falling;
    }
    else if((fabs(theFilteredSensorData.data[SensorData::angleX]) <= parameters.staggeringAngleX - pi_180
          && fabs(theFilteredSensorData.data[SensorData::angleY]) <= parameters.staggeringAngleY - pi_180)
              || (fallDownState.state == FallDownState::upright
              && !isStaggering()))
    {
      fallDownState.state = FallDownState::upright;
    }
    else if (fallDownState.state == FallDownState::staggering && 
             isFalling())
    {
      //SoundPlayer::play("doh.wav");
      lastFallDetected = theFrameInfo.time;
      fallDownState.state = FallDownState::falling;
    }
    else if((isUprightOrStaggering(fallDownState) && 
             isStaggering())
         || (fallDownState.state == FallDownState::staggering 
          && fabs(theFilteredSensorData.data[SensorData::angleX]) <= parameters.staggeringAngleX - pi_180
          && fabs(theFilteredSensorData.data[SensorData::angleY]) <= parameters.staggeringAngleY - pi_180))
    {
      fallDownState.state = FallDownState::staggering;
    }
    else
    {     
      fallDownState.state = FallDownState::undefined;
     
      if(sensorAvailable(accXaverage, accYaverage, accZaverage) && !isGettingUp())
      {
        if (fabs(accelerationAngleXZ) < 0.5f)
        {
          fallDownState.state = FallDownState::lyingOnFront;
        }
        else if(fabs(accelerationAngleXZ) > 2.5)
        {
          fallDownState.state = FallDownState::lyingOnBack;
        }
        else if (fabs(accelerationAngleYZ) < 0.5f)
        {
          fallDownState.state = FallDownState::lyingOnLeftSide;
        }
        else if(fabs(accelerationAngleYZ) > 2.5)
        {
          fallDownState.state = FallDownState::lyingOnRightSide;
        }
      }
    }
  }
  else 
  {
    fallDownState.state = FallDownState::undefined;
  }
}

bool FallDownStateDetector::sensorAvailable(float avgX, float avgY, float avgZ) const
{
  // If the sensor is not available, all values are set to SensorData::off.
  // Therefore, we need to check, if all values have been seet to this value:
  float diff(fabs(avgX - float(SensorData::off)) + fabs(avgY - float(SensorData::off)) + fabs(avgZ - float(SensorData::off)));
  return (diff > 1.0f);
}

bool FallDownStateDetector::isUprightOrStaggering(FallDownState& fallDownState){
  return fallDownState.state == FallDownState::upright
    || fallDownState.state == FallDownState::staggering;    
}

bool FallDownStateDetector::isGettingUp(){  
  return theMotionInfo.motion == MotionRequest::specialAction
     && (theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::standUpFrontNao
      || theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::standUpBackNao);
}

bool FallDownStateDetector::specialSpecialAction(){  
  return (theMotionInfo.motion == MotionRequest::specialAction
    && (theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::playDead ||
              theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::keeperJumpLeft ||
              theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::sitDown2 ||
              theMotionInfo.specialActionRequest.specialAction == SpecialActionRequest::goUp2))
      || (theMotionInfo.motion == MotionRequest::sike
          && theMotionInfo.sikeRequest.motion == SIKERequest::sitDown2);
}


bool FallDownStateDetector::isStaggering(){
  return fabs(theFilteredSensorData.data[SensorData::angleX]) >= parameters.staggeringAngleX + pi_180
    || fabs(theFilteredSensorData.data[SensorData::angleY]) >= parameters.staggeringAngleY + pi_180;
}

bool FallDownStateDetector::isFalling(){
  return fabs(theFilteredSensorData.data[SensorData::angleX]) >= parameters.fallDownAngleX
      || fabs(theFilteredSensorData.data[SensorData::angleY]) >= parameters.fallDownAngleY;
}

bool FallDownStateDetector::isCalibrated(){
  if (!calibrated && theFilteredSensorData.data[SensorData::gyroX] != SensorData::off 
      && theFilteredSensorData.data[SensorData::gyroX] != 0.f)
  {
    calibrated = true;
  }
  return calibrated;
}

MAKE_MODULE(FallDownStateDetector, Sensing)
