/**
* @file FallDownStateDetector.h
*
* This file declares a module that provides information about the current state of the robot's body.
*
* @author <a href="mailto:maring@informatik.uni-bremen.de">Martin Ring</a>
*/

#ifndef __FallDownStateDetector_h_
#define __FallDownStateDetector_h_

#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/MotionControl/MotionInfo.h"
#include "Representations/Modeling/FallDownState.h"
#include "Tools/Module/Module.h"
#include "Tools/RingBufferWithSum.h"


MODULE(FallDownStateDetector)
  REQUIRES(FilteredSensorData)
  USES(MotionInfo)
  REQUIRES(FrameInfo)
  PROVIDES_WITH_MODIFY(FallDownState)
END_MODULE


/**
* @class FallDownStateDetector
*
* A module for computing the current body state from sensor data
*/
class FallDownStateDetector: public FallDownStateDetectorBase
{
private:
  /**
   * @class Parameters
   * The parameters for FallDownStateDetector
   */
  class Parameters : public Streamable
  {
    private:
      void serialize(In* in, Out* out)
      {
        STREAM_REGISTER_BEGIN();
        STREAM(staggeringAngleX);
        STREAM(staggeringAngleY);
        STREAM(fallDownAngleX);
        STREAM(fallDownAngleY);
        STREAM(fallTime);
        STREAM(onGroundAngle);
        STREAM_REGISTER_FINISH();        
      }

    public:
      int   fallTime; /**< The time (in ms) to remain in state 'falling' after a detected fall */        
      float staggeringAngleX, /**< The threshold angle which is used to detect the robot staggering to the back or front*/
            staggeringAngleY, /**< The threshold angle which is used to detect the robot staggering sidewards*/
            fallDownAngleY, /**< The threshold angle which is used to detect a fall to the back or front*/
            fallDownAngleX, /**< The threshold angle which is used to detect a sidewards fall */
            onGroundAngle; /**< The threshold angle which is used to detect the robot lying on the ground */
  };

  Parameters parameters; /**< The parameters of this module. */

  /** Executes this module
  * @param fallDownState The data structure that is filled by this module
  */
  void update(FallDownState& fallDownState);

  bool isGettingUp();
  bool isFalling();  
  bool isStaggering();
  bool isCalibrated();
  bool specialSpecialAction();
  bool isUprightOrStaggering(FallDownState& fallDownState);

  bool calibrated;
  unsigned int lastFallDetected;  

  /** Checks, if the sensor data is valid
  * @param avgX average acceleration in x direction
  * @param avgY average acceleration in y direction
  * @param avgZ average acceleration in z direction
  * @return true, if data is valid
  */
  bool sensorAvailable(float avgX, float avgY, float avgZ) const;

  /** Indices for buffers of sensor data */
  enum BufferEntries{accX = 0, accY, accZ, numOfBuffers};

  /** Buffers for averaging sensor data */
  RingBufferWithSum<float,5> buffers[numOfBuffers];

public:
  /**
  * Default constructor.
  */
  FallDownStateDetector();
};

#endif// __FallDownStateDetector_h_
