/**
* @file GroundContactDetector.cpp
* Implementation of module GroundContactDetector.
* @author Colin Graf
*/

#include "GroundContactDetector.h"
#include "Tools/Debugging/DebugDrawings.h"

MAKE_MODULE(GroundContactDetector, Sensing)

GroundContactDetector::GroundContactDetector() : 
  framesWithContact(0), framesWithoutContact(0), contactStartTime(0), contact(false), wasContacted(false)
{
  p.fsrFactor = 1.7f;
  p.loadFactor = 2.5f;
  p.threshold = 4000.f;
  p.minCycleContactUnity = 10;
  p.safeContactTime = 3300;
  p.noContactTime = 1000;
}

void GroundContactDetector::init()
{
  InConfigFile stream(Global::getSettings().expandRobotFilename("groundContact.cfg"));
  if(stream.exists())
    stream >> p;
  else
  {
    InConfigFile stream("groundContact.cfg");
    if(stream.exists())
      stream >> p;
  }
}

void GroundContactDetector::update(GroundContactState& groundContactState)
{
  MODIFY("module:GroundContactDetector:parameters", p);
  
  bool lastContact = groundContactState.contact;
  //if(theSensorData.data[SensorData::fsrLFL] != SensorData::off && theSensorData.data[SensorData::fsrRFL] != SensorData::off)
  {
    float fsrSum = 0.f;
    fsrSum += theSensorData.data[SensorData::fsrLFL];
    fsrSum += theSensorData.data[SensorData::fsrLFR];
    fsrSum += theSensorData.data[SensorData::fsrLBL];
    fsrSum += theSensorData.data[SensorData::fsrLBR];
    fsrSum += theSensorData.data[SensorData::fsrRFL];
    fsrSum += theSensorData.data[SensorData::fsrRFR];
    fsrSum += theSensorData.data[SensorData::fsrRBL];
    fsrSum += theSensorData.data[SensorData::fsrRBR];

    float loadSum = 0.f;
    loadSum += theSensorData.currents[JointData::LHipPitch];
    loadSum += theSensorData.currents[JointData::LKneePitch];
    loadSum += theSensorData.currents[JointData::LAnklePitch];
    loadSum += theSensorData.currents[JointData::RHipPitch];
    loadSum += theSensorData.currents[JointData::RKneePitch];
    loadSum += theSensorData.currents[JointData::RAnklePitch];

    float totalWeightedSum = p.fsrFactor * fsrSum + p.loadFactor * loadSum;
    PLOT("module:GroundContactDetector:totalWeightedSum", totalWeightedSum);

    bool newContact = totalWeightedSum > p.threshold;
    if(newContact) // contact 
    {
      framesWithContact++;
      framesWithoutContact = 0;
    }
    else
    {
      framesWithContact = 0;
      framesWithoutContact++;
    }

    if(contact)
      contact = framesWithoutContact <= p.minCycleContactUnity;
    else
      contact = framesWithContact > p.minCycleContactUnity;

    if(contact)
      wasContacted = true;

    groundContactState.contact = contact || !wasContacted || theMotionRequest.motion == MotionRequest::specialAction || theMotionInfo.motion == MotionRequest::specialAction;
    if(theRobotInfo.penalty != PENALTY_NONE)
      groundContactState.contact = false;
  }
  //else
    //groundContactState.contact = true;

  if((groundContactState.contact && !lastContact) || (groundContactState.contact && contactStartTime == 0))
    contactStartTime = theFrameInfo.time;
  groundContactState.contactSafe = groundContactState.contact && theFrameInfo.getTimeSince(contactStartTime) >= p.safeContactTime;

  if((!groundContactState.contact && lastContact) || (!groundContactState.contact && noContactStartTime == 0))
    noContactStartTime = theFrameInfo.time;
  groundContactState.noContactSafe = !groundContactState.contact && theFrameInfo.getTimeSince(noContactStartTime) >= p.noContactTime;

  PLOT("module:GroundContactDetector:groundContact", groundContactState.contact ? 0.75 : 0.25);
  PLOT("module:GroundContactDetector:noGroundContactSafe", groundContactState.noContactSafe ? 100 : 50);
}
