/**
* @file InertiaSensorCalibrator.cpp
* Implementation of module InertiaSensorCalibrator.
* @author Colin Graf
*/

#include "InertiaSensorCalibrator.h"
#include "Tools/Math/Pose3D.h"

MAKE_MODULE(InertiaSensorCalibrator, Sensing)

InertiaSensorCalibrator::InertiaSensorCalibrator()
{
  reset();

  p.maxGyroOffset = Vector2<>(1.f, 1.f);
  p.maxAccOffset = Vector3<>(10.f, 10.f, 10.f);
  p.timeFrame = 1500;
  p.gyroBiasProcessNoise = Vector2<>(0.05f, 0.05f);
  p.gyroBiasStandMeasurementNoise = Vector2<>(0.01f, 0.01f);
  p.gyroBiasWalkMeasurementNoise = Vector2<>(0.1f, 0.1f);
  p.accBiasProcessNoise = Vector3<>(0.01f, 0.01f, 0.01f);
  p.accBiasStandMeasurementNoise = Vector3<>(0.1f, 0.1f, 0.1f);
  p.accBiasWalkMeasurementNoise = Vector3<>(1.f, 1.f, 1.f);
}

void InertiaSensorCalibrator::reset()
{
  lastTime = 0;
  lastMotion = MotionRequest::specialAction;
  calibrated = false;
  collectionStartTime = 0;
  cleanCollectionStartTime = 0;
  lastGyro = safeGyro = Vector2<>();
  lastAcc = safeAcc = Vector3<>(0.f, 0.f, -9.80665f);
  inertiaSensorDrops = 1000;
}

void InertiaSensorCalibrator::update(InertiaSensorData& inertiaSensorData)
{
  MODIFY("module:InertiaSensorCalibrator:parameters", p);

  // frame time check
  if(theFrameInfo.time <= lastTime)
  {
    if(theFrameInfo.time == lastTime)
      return; // done!
     reset();
  }

  // detect changes in joint calibration
#ifndef RELEASE
  bool jointCalibrationChanged = false;
  for(int i = JointData::LHipYawPitch; i <= JointData::LAnkleRoll; ++i)
    if(theJointCalibration.joints[i].offset != lastJointCalibration.joints[i].offset)
    {
      jointCalibrationChanged = true;
      lastJointCalibration.joints[i].offset = theJointCalibration.joints[i].offset;
    }
  for(int i = JointData::RHipYawPitch; i <= JointData::RAnkleRoll; ++i)
    if(theJointCalibration.joints[i].offset != lastJointCalibration.joints[i].offset)
    {
      jointCalibrationChanged = true;
      lastJointCalibration.joints[i].offset = theJointCalibration.joints[i].offset;
    }
  if(jointCalibrationChanged)
    reset();
#endif

  // drop corrupted sensor readings
  Vector2<> newGyro(theSensorData.data[SensorData::gyroX], theSensorData.data[SensorData::gyroY]);
  Vector3<> newAcc(theSensorData.data[SensorData::accX], theSensorData.data[SensorData::accY], theSensorData.data[SensorData::accZ]);
  newAcc *= 9.80665f; // strange unit => metric unit :)
  if(fabs(newGyro.x - lastGyro.x) > p.maxGyroOffset.x ||
     fabs(newGyro.y - lastGyro.y) > p.maxGyroOffset.y ||
     fabs(newAcc.x - lastAcc.x) > p.maxAccOffset.x ||
     fabs(newAcc.y - lastAcc.y) > p.maxAccOffset.y ||
     fabs(newAcc.z - lastAcc.z) > p.maxAccOffset.z )
  {    
    if(++inertiaSensorDrops > 3)
    {
      lastGyro = newGyro;
      lastAcc = newAcc;
    }
    newGyro.x = newGyro.y = newAcc.x = newAcc.y = newAcc.z = SensorData::off;
  } 
  else
  {
    safeGyro = newGyro;
    safeAcc = newAcc;
    inertiaSensorDrops = 0;
  }

  // it's prediction time!
  if(lastTime && calibrated)
  {
    float timeDiff = float(theFrameInfo.time - lastTime) * 0.001f; // in seconds
    accXBias.predict(0.f, sqr(p.accBiasProcessNoise.x * timeDiff));
    accYBias.predict(0.f, sqr(p.accBiasProcessNoise.y * timeDiff));
    accZBias.predict(0.f, sqr(p.accBiasProcessNoise.z * timeDiff));
    gyroXBias.predict(0.f, sqr(p.gyroBiasProcessNoise.x * timeDiff));
    gyroYBias.predict(0.f, sqr(p.gyroBiasProcessNoise.y * timeDiff));
  }

  // detect unstable stuff...
  const MotionRequest::Motion& currentMotion(theMotionSelection.targetMotion);
  bool unstable = false;
  if(currentMotion != lastMotion || // motion change
     currentMotion != theMotionInfo.motion ||  // interpolating
     !theGroundContactState.contact)
    unstable = true;
/*  else if(currentMotion == MotionRequest::stand && newAcc.z != SensorData::off && newAcc.z >= -9000 )
    unstable = true;*/
  else if(currentMotion == MotionRequest::walk &&
     (fabs(theWalkingEngineOutput.speed.translation.y) > 10 ||
      fabs(theWalkingEngineOutput.speed.translation.x) > 20 ||
      fabs(theWalkingEngineOutput.speed.rotation) > 0.15 /*|| (newAcc.z != SensorData::off && newAcc.z >= -8000)*/))
    unstable = true;
  else if(inertiaSensorDrops >= 2)
    unstable = true;
  else if(currentMotion != MotionRequest::walk && currentMotion != MotionRequest::stand)
    unstable = true;

  // update cleanCollectionStartTime
  if(unstable)
    cleanCollectionStartTime = 0;
  else if(!cleanCollectionStartTime)
    cleanCollectionStartTime = theFrameInfo.time;

  // restart sensor value collecting?
  if(unstable || (currentMotion == MotionRequest::walk && theWalkingEngineOutput.positionInWalkCycle < lastPositionInWalkCycle) ||
    (currentMotion == MotionRequest::stand && theFrameInfo.time - collectionStartTime > 1000))
  {
    // add collection within the time frame to the collection buffer
    if(cleanCollectionStartTime && theFrameInfo.time - cleanCollectionStartTime > p.timeFrame && 
      accValues.getNumberOfEntries() && gyroValues.getNumberOfEntries())
    {
      ASSERT(collections.getNumberOfEntries() < collections.getMaxEntries());
      collections.add(Collection(accValues.getSum() / float(accValues.getNumberOfEntries()), 
        gyroValues.getSum() / float(gyroValues.getNumberOfEntries()), 
        collectionStartTime + (theFrameInfo.time - collectionStartTime) / 2, currentMotion));
    }

    // restart collecting
    accValues.init();
    gyroValues.init();
    collectionStartTime = 0;

    // look if there are any useful buffered collections
    for(int i = collections.getNumberOfEntries() - 1; i >= 0; --i)
    {
      const Collection& collection(collections.getEntry(i));
      if(theFrameInfo.time - collection.timeStamp < p.timeFrame)
        break;
      if(cleanCollectionStartTime && cleanCollectionStartTime < collection.timeStamp)
      { // use this collection
        if(!calibrated)
        {
          calibrated = true;
          accXBias.init(collection.accAvg.x, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.x : p.accBiasWalkMeasurementNoise.x));
          accYBias.init(collection.accAvg.y, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.y : p.accBiasWalkMeasurementNoise.y));
          accZBias.init(collection.accAvg.z, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.z : p.accBiasWalkMeasurementNoise.z));
          gyroXBias.init(collection.gyroAvg.x, sqr(collection.motion == MotionRequest::stand ? p.gyroBiasStandMeasurementNoise.x : p.gyroBiasWalkMeasurementNoise.x));
          gyroYBias.init(collection.gyroAvg.y, sqr(collection.motion == MotionRequest::stand ? p.gyroBiasStandMeasurementNoise.y : p.gyroBiasWalkMeasurementNoise.y));
        }
        else
        {
          accXBias.update(collection.accAvg.x, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.x : p.accBiasWalkMeasurementNoise.x));
          accYBias.update(collection.accAvg.y, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.y : p.accBiasWalkMeasurementNoise.y));
          accZBias.update(collection.accAvg.z, sqr(collection.motion == MotionRequest::stand ? p.accBiasStandMeasurementNoise.z : p.accBiasWalkMeasurementNoise.z));
          gyroXBias.update(collection.gyroAvg.x, sqr(collection.motion == MotionRequest::stand ? p.gyroBiasStandMeasurementNoise.x : p.gyroBiasWalkMeasurementNoise.x));
          gyroYBias.update(collection.gyroAvg.y, sqr(collection.motion == MotionRequest::stand ? p.gyroBiasStandMeasurementNoise.y : p.gyroBiasWalkMeasurementNoise.y));
        }
      }
      collections.removeFirst();
    }
  }

  // collecting....
  if(!unstable)
  {
    // calculate rotation based on foot - torso transformation
    const Pose3D& footLeft(theRobotModel.limbs[RobotModel::footLeft]);
    const Pose3D& footRight(theRobotModel.limbs[RobotModel::footRight]);
    const Pose3D footLeftInvert(footLeft.invert());
    const Pose3D footRightInvert(footRight.invert());
    if(fabs(footLeftInvert.translation.z - footRightInvert.translation.z) < 3.f/* magic number */)
    { // use average of the calculated rotation of each leg
      calculatedRotation = RotationMatrix(Vector3<>(
        (atan2(footLeftInvert.rotation.c[1].z, footLeftInvert.rotation.c[2].z) + atan2(footRightInvert.rotation.c[1].z, footRightInvert.rotation.c[2].z)) * 0.5f,
        (atan2(-footLeftInvert.rotation.c[0].z, footLeftInvert.rotation.c[2].z) + atan2(-footRightInvert.rotation.c[0].z, footRightInvert.rotation.c[2].z)) * 0.5f,
        0.f));
    }
    else if(footLeftInvert.translation.z > footRightInvert.translation.z)
    { // use left foot
      calculatedRotation = footLeftInvert.rotation;
    }
    else
    { // use right foot
      calculatedRotation = footRightInvert.rotation;
    }

    // calculate expected acceleration sensor reading
    Vector3<> accGravOnly(calculatedRotation.c[0].z, calculatedRotation.c[1].z, calculatedRotation.c[2].z);
    accGravOnly *= -9.80665f;

    // add sensor reading to the collection
    ASSERT(accValues.getNumberOfEntries() < accValues.getMaxEntries());
    ASSERT(gyroValues.getNumberOfEntries() < gyroValues.getMaxEntries());
    accValues.add(safeAcc - accGravOnly);
    gyroValues.add(safeGyro);
    if(!collectionStartTime)
      collectionStartTime = theFrameInfo.time;
  }

  // provide calibrated inertia readings
  if(!calibrated || newGyro.x == SensorData::off)
  {
    inertiaSensorData.gyro.x = inertiaSensorData.gyro.y = InertiaSensorData::off;
    inertiaSensorData.acc.x = inertiaSensorData.acc.y = inertiaSensorData.acc.z = InertiaSensorData::off;
  }
  else
  {
    inertiaSensorData.gyro.x = newGyro.x - gyroXBias.value;
    inertiaSensorData.gyro.y = newGyro.y - gyroYBias.value;
    inertiaSensorData.acc.x = newAcc.x - accXBias.value;
    inertiaSensorData.acc.y = newAcc.y - accYBias.value;
    inertiaSensorData.acc.z = newAcc.z - accZBias.value;
  }

  MODIFY("module:InertiaSensorCalibrator:calibrated", calibrated);
  MODIFY("module:InertiaSensorCalibrator:gyroXBias", gyroXBias.value);
  MODIFY("module:InertiaSensorCalibrator:gyroYBias", gyroYBias.value);
  MODIFY("module:InertiaSensorCalibrator:accXBias", accXBias.value);
  MODIFY("module:InertiaSensorCalibrator:accYBias", accYBias.value);
  MODIFY("module:InertiaSensorCalibrator:accZBias", accZBias.value);

  // store some values for the next iteration
  lastTime = theFrameInfo.time;
  lastMotion = theMotionSelection.targetMotion;
  if(newGyro.x != SensorData::off)
  {
    lastGyro = newGyro;
    lastAcc = newAcc;
  }
  lastPositionInWalkCycle = theWalkingEngineOutput.positionInWalkCycle;
}
