/**
 * @file Platform/SimRobotQt/Oracle.cpp
 * Implementation of class Oracle for SimRobotQt.
 * @author Max Risler
 * @author Colin Graf
 */

#include "Simulation/SimObject.h"
#include "Oracle.h"
#include "RoboCupCtrl.h"
#include "Platform/SystemCall.h"
#include "Platform/hash_map.h"
#include "Representations/Infrastructure/Image.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Modeling/BallModel.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Sensing/OrientationData.h"

static const char* translate[] =
{
  "HeadYaw", "HeadYaw",
  "HeadPitch", "HeadPitch",
  "RShoulderPitch", "RShoulderPitch",
  "RShoulderRoll", "RShoulderRoll",
  "RElbowYaw", "RElbowYaw",
  "RElbowRoll", "RElbowRoll",
  "LShoulderPitch", "LShoulderPitch",
  "LShoulderRoll", "LShoulderRoll",
  "LElbowYaw", "LElbowYaw",
  "LElbowRoll", "LElbowRoll",
  "RHipYawPitch", "RHipYawPitch",
  "RHipRoll", "RHipRoll",
  "RHipPitch", "RHipPitch",
  "RKneePitch", "RKneePitch",
  "RAnklePitch", "RAnklePitch",
  "RAnkleRoll", "RAnkleRoll",
  "LHipYawPitch", "LHipYawPitch",
  "LHipRoll", "LHipRoll",
  "LHipPitch", "LHipPitch",
  "LKneePitch", "LKneePitch",
  "LAnklePitch", "LAnklePitch",
  "LAnkleRoll", "LAnkleRoll",
  "accelerometer", "accelerometer",
  "gyroscope", "gyro",
  "sonarCenterRight", "US-TopRight",
  "sonarRight", "US-BottomRight",
  "sonarCenterLeft", "US-TopLeft",
  "sonarLeft", "US-BottomLeft",
  "camera", "camera",
  "rFSR1", "RFsrFL",
  "rFSR2", "RFsrFR",
  "rFSR3", "RFsrBR",
  "rFSR4", "RFsrBL",
  "lFSR1", "LFsrFL",
  "lFSR2", "LFsrFR",
  "lFSR3", "LFsrBR",
  "lFSR4", "LFsrBL",
  "rFSRBump", "RFoot",
  "lFSRBump", "LFoot",
  "rFoot", "STRUCTURE_APP",
  "lFoot", "STRUCTURE_APP"
};

Oracle::Oracle() :
blue(false), me(0), leftFoot(0), rightFoot(0), ball(0), imageYUV422x2(0), naoV3(false), lastUsMeasurement(0) 
{}

Oracle::~Oracle()
{
  if(imageYUV422x2)
    delete[] imageYUV422x2;
}

void Oracle::init()
{
  ASSERT(me == 0);

  stdext::hash_map<std::string, const char*> names;
  for(size_t i = 0; i < sizeof(translate) / sizeof(translate[0]); i += 2)
    names[translate[i]] = translate[i + 1];
  const std::string& robotName(RoboCupCtrl::getRobotFullName());

  // get controller
  VERIFY(ctrl = RoboCupCtrl::getController());
  
  // get the robot
  VERIFY(me = ctrl->getObjectReference(robotName));
  blue = me->getName()[5] < '5';

  me->findObject(rightFoot, "rFoot", false);
  if(!rightFoot)
  {
    naoV3 = true;
    SimObject* rAnkleRoll;
    me->findObject(rAnkleRoll, names["RAnkleRoll"], false);
    rAnkleRoll->findObject(rightFoot, names["rFoot"], false);
  }

  me->findObject(leftFoot, "lFoot", false);
  if(!leftFoot)
  {
    SimObject* lAnkleRoll;
    me->findObject(lAnkleRoll, names["LAnkleRoll"], false);
    lAnkleRoll->findObject(leftFoot, names["lFoot"], false);
  }
  ASSERT(rightFoot && leftFoot);

  // get the ball
  const SimObject* balls = ctrl->getObjectReference("RoboCup.balls");
  if(balls->getNumberOfChildNodes() > 0)
    ball = balls->getChildNode(0);

  // get joint-ids
  std::vector<std::string> parts;
  parts.resize(2);
  parts[0] = robotName;
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    std::string angleName(JointData::getJointName(JointData::Joint(i)));
    parts[1] = angleName;
    spJoints[i] = ctrl->getSensorportId(parts);
    if(spJoints[i] == -1)
    {
      angleName = names[angleName];
      parts[1] = angleName;
      spJoints[i] = ctrl->getSensorportId(parts);
    }
    parts[1] += ".angle";
    apJoints[i] = ctrl->getActuatorportId(parts);
    parts[1] = angleName + ".switch";
    psJoints[i] = ctrl->getActuatorportId(parts);
  }

  // get sensor-ids
  parts.resize(3);
  parts[1] = "camera";
  parts[2] = "image";
  spCamera = ctrl->getSensorportId(parts);
  if(spCamera == -1)
  {
    parts[1] = names["camera"];
    spCamera = ctrl->getSensorportId(parts);
  }
  parts.resize(2);
  parts[1] = "accelerometer";
  spAcc = ctrl->getSensorportId(parts);
  if(spAcc == -1)
  {
    parts[1] = names["accelerometer"];
    spAcc = ctrl->getSensorportId(parts);
  }
  parts[1] = "gyroscope";
  spGyro = ctrl->getSensorportId(parts);
  if(spGyro == -1)
  {
    parts[1] = names["gyroscope"];
    spGyro = ctrl->getSensorportId(parts);
  }
  parts[1] = "sonarLeft";
  spUsLeft = ctrl->getSensorportId(parts);
  if(spUsLeft == -1)
  {
    parts[1] = names["sonarLeft"];
    spUsLeft = ctrl->getSensorportId(parts);
  }
  parts[1] = "sonarRight";
  spUsRight = ctrl->getSensorportId(parts);
  if(spUsRight == -1)
  {
    parts[1] = names["sonarRight"];
    spUsRight = ctrl->getSensorportId(parts);
  }
  parts[1] = "sonarCenterLeft";
  spUsCenterLeft = ctrl->getSensorportId(parts);
  if(spUsCenterLeft == -1)
  {
    parts[1] = names["sonarCenterLeft"];
    spUsCenterLeft = ctrl->getSensorportId(parts);
  }
  parts[1] = "sonarCenterRight";
  spUsCenterRight = ctrl->getSensorportId(parts);
  if(spUsCenterRight == -1)
  {
    parts[1] = names["sonarCenterRight"];
    spUsCenterRight = ctrl->getSensorportId(parts);
  }

  // get FSR-ids
  static const char* fsrs[8] =
  {
    "lFSR1", "lFSR2", "lFSR3", "lFSR4",
    "rFSR1", "rFSR2", "rFSR3", "rFSR4"
  };
  for(int i = 0; i < 8; ++i)
  {
    parts[1] = fsrs[i];
    spFSR[i] = ctrl->getSensorportId(parts);
    if(spFSR[i] == -1)
    {
      parts[1] = names[fsrs[i]];
      spFSR[i] = ctrl->getSensorportId(parts);
    }
  }

  static const char* bumpers[2] = {"lFSRBump", "rFSRBump"};
  for(int i = 0; i < 2; ++i)
  {
    parts[1] = bumpers[i];
    spFSRBump[i] = ctrl->getSensorportId(parts);
    if(spFSRBump[i] == -1)
    {
      parts[1] = names[bumpers[i]];
      spFSRBump[i] = ctrl->getSensorportId(parts);
    }
  }

  parts[1] = "CameraSelect.angle";
  int apCameraSelect = ctrl->getActuatorportId(parts);
  if(apCameraSelect != -1)
    ctrl->setActuatorport(apCameraSelect, fromDegrees(40));

  // load calibration
  InConfigFile stream(Global::getSettings().expandRobotFilename("jointCalibration.cfg"));
  ASSERT(stream.exists());
  stream >> jointCalibration;
}

void Oracle::getRobotPose(RobotPose& robotPose) const
{
  if(rightFoot && leftFoot)
  {
    Pose2D leftPose, rightPose;
    getPose2D(leftFoot, leftPose);
    getPose2D(rightFoot, rightPose);
    getPose2D(me, (Pose2D&)robotPose);
    robotPose.translation = (leftPose.translation + rightPose.translation) / 2.;
    robotPose.translation *= 1000.;
    if(blue)
      robotPose = Pose2D(pi) + robotPose;
  }
  else
  {
    ASSERT(me);
    getPose2D(me, robotPose);
    robotPose.translation *= 1000.;
    if(blue)
      robotPose = Pose2D(pi) + robotPose;
  }
  robotPose.validity = 1.f;
  robotPose.deviation = 1.f;
}

void Oracle::getOdometryData(const RobotPose& robotPose, OdometryData& odometryData) const
{
  ASSERT(me);
  (Pose2D&)odometryData = blue ? (Pose2D(pi) + robotPose) : (const Pose2D&)robotPose;
}

void Oracle::getBallModel(const RobotPose& robotPose, BallModel& ballModel)
{
  ASSERT(me);
  if(ball)
  {
    unsigned int timeWhenLastSeen = SystemCall::getCurrentSystemTime();
    Pose2D pose;
    getPose2D(ball, pose);
    pose.translation *= 1000.;
    if(blue)
      pose = Pose2D(pi) + pose;
    Vector2<float> velocity((pose.translation - lastBallPosition) / float(timeWhenLastSeen - ballModel.timeWhenLastSeen) * 1000.0f);
    ballModel.lastPerception.setPositionAndVelocityInFieldCoordinates(pose.translation, velocity, robotPose);
    ballModel.estimate = ballModel.lastPerception;
    ballModel.lastSeenEstimate = ballModel.lastPerception;
    ballModel.timeWhenLastSeen = timeWhenLastSeen;
    lastBallPosition = pose.translation;
  }
}

void Oracle::getImage(Image& image)
{
  ASSERT(me);
  if(spCamera != -1)
  {
    const SensorReading& imageReading = ctrl->getSensorReading(spCamera);
    const int w = imageReading.dimensions[0];
    const int h = imageReading.dimensions[1];
    if(imageYUV422x2 == 0)
      imageYUV422x2 = new unsigned char[w*h*8];

    const int w3 = w * 3, w2 = w * 2;
    unsigned char* src = (unsigned char*)imageReading.data.byteArray;
    unsigned char* srcLineEnd = src;
    Image::Pixel* destBegin = (Image::Pixel*)imageYUV422x2;
    Image::Pixel* dest;
    int r, g, b, yy, uu, vv;
    for(int y = 0; y < h; ++y)
    {
      for(srcLineEnd += w3, dest = destBegin + y * w2; src < srcLineEnd; )
      {
        b = *src++;
        g = *src++;
        r = *src++;
        yy =          306 * r,
        uu = 130560 + 512 * r - 429 * g,
        vv = 130560 - 173 * r - 339 * g + 512 * b;
        yy += 601 * g + 117 * b;
        uu -= 83 * b;
        // yy =          306 * r + 601 * g + 117 * b // && 0<=r<=255, 0<=g<=255, 0<=b<=255
        // => max = 261120, min = 0
        // uu = 130048 + 512 * r - 429 * g -  83 * b // && 0<=r<=255, 0<=g<=255, 0<=b<=255
        // => max = 260 608, min = -512
        // vv = 130048 - 173 * r - 339 * g + 512 * b // && 0<=r<=255, 0<=g<=255, 0<=b<=255
        // => max = 260 608, min = -512
        // I suggest using 130560 instead of 130048 to get rid of these tests:
        //if(yy < 0) yy = 0; else if(yy > 261120) yy = 261120;
        //if(uu < 0) uu = 0; else if(uu > 261120) uu = 261120;
        //if(vv < 0) vv = 0; else if(vv > 261120) vv = 261120;
        dest->y = dest->yCbCrPadding = (unsigned char)(yy >> 10);
        dest->cr = (unsigned char)(uu >> 10);
        dest->cb = (unsigned char)(vv >> 10);
        ++dest;
      }
    }
    image.setImage(imageYUV422x2);
  }
  image.timeStamp = SystemCall::getCurrentSystemTime();
}

void Oracle::getAndSetJointData(const JointData& jointRequest, JointData& jointData) const
{
  ASSERT(me);
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    // Get angles      
    if(spJoints[i] != -1)
    {
      const SensorReading& actualAngle = ctrl->getSensorReading(spJoints[i]);
      jointData.angles[i] = float(actualAngle.data.doubleValue * jointCalibration.joints[i].sign - jointCalibration.joints[i].offset);
    }

    // Set angles
    const float& targetAngle(jointRequest.angles[i]);
    if(targetAngle != JointData::off && 
       targetAngle != JointData::ignore &&
       spJoints[i] != -1) // if joint does exist
      ctrl->setActuatorport(apJoints[i], (targetAngle + jointCalibration.joints[i].offset) * jointCalibration.joints[i].sign);
  }
  jointData.timeStamp = SystemCall::getCurrentSystemTime();
}

void Oracle::getSensorData(SensorData& sensorData)
{
  ASSERT(me);

  sensorData.timeStamp = SystemCall::getCurrentSystemTime();

  // Battery
  sensorData.data[SensorData::batteryLevel] = 1.0f;

  // Acceleration sensors
  const SensorReading& acceleration = ctrl->getSensorReading(spAcc);
  sensorData.data[SensorData::accX] = float(acceleration.data.doubleArray[0] / 9.81) + 0.1f; // 0.1f = bias
  sensorData.data[SensorData::accY] = float(acceleration.data.doubleArray[1] / 9.81) + 0.2f; // 0.2f = bias
  sensorData.data[SensorData::accZ] = float(acceleration.data.doubleArray[2] / 9.81) + 0.05f; // 0.05f = bias

  // Gyro
  const SensorReading& rotationSpeed = ctrl->getSensorReading(spGyro);
  sensorData.data[SensorData::gyroX] = float(rotationSpeed.data.doubleArray[0]) + 12.f; // 12.f = bias
  sensorData.data[SensorData::gyroY] = float(rotationSpeed.data.doubleArray[1]) + 6.f; // 6.f = bias
  sensorData.data[SensorData::gyroZ] = 0.; //float(rotationSpeed.data.doubleArray[2]); // nao style :(
  if(blue)
  {
    sensorData.data[SensorData::gyroX] *= -1.f;
    sensorData.data[SensorData::gyroY] *= -1.f;
  }


  if(spFSR[0] != -1)
  {
    const SensorReading& pressure0 = ctrl->getSensorReading(spFSR[0]);
    sensorData.data[SensorData::fsrLFL] = float(pressure0.data.doubleArray[2]);
    const SensorReading& pressure1 = ctrl->getSensorReading(spFSR[1]);
    sensorData.data[SensorData::fsrLFR] = float(pressure1.data.doubleArray[2]);
    const SensorReading& pressure2 = ctrl->getSensorReading(spFSR[2]);
    sensorData.data[SensorData::fsrLBL] = float(pressure2.data.doubleArray[2]);
    const SensorReading& pressure3 = ctrl->getSensorReading(spFSR[3]);
    sensorData.data[SensorData::fsrLBR] = float(pressure3.data.doubleArray[2]);
    const SensorReading& pressure4 = ctrl->getSensorReading(spFSR[4]);
    sensorData.data[SensorData::fsrRFL] = float(pressure4.data.doubleArray[2]);
    const SensorReading& pressure5 = ctrl->getSensorReading(spFSR[5]);
    sensorData.data[SensorData::fsrRFR] = float(pressure5.data.doubleArray[2]);
    const SensorReading& pressure6 = ctrl->getSensorReading(spFSR[6]);
    sensorData.data[SensorData::fsrRBL] = float(pressure6.data.doubleArray[2]);
    const SensorReading& pressure7 = ctrl->getSensorReading(spFSR[7]);
    sensorData.data[SensorData::fsrRBR] = float(pressure7.data.doubleArray[2]);
  }
  else if(spFSRBump[0] != -1)
  {
    const SensorReading& contactLeft = ctrl->getSensorReading(spFSRBump[0]);
    const SensorReading& contactRight = ctrl->getSensorReading(spFSRBump[1]);
    if(contactLeft.data.boolValue && contactRight.data.boolValue)
      for(int i = 0; i < 8; i++)
        sensorData.data[SensorData::fsrLFL+i] = 0.8f;
    else if(contactLeft.data.boolValue)
      for(int i = 0; i < 4; i++)
      {
        sensorData.data[SensorData::fsrLFL+i] = 1.6f;
        sensorData.data[SensorData::fsrRFL+i] = 0.0f;
      }
    else if(contactRight.data.boolValue)
      for(int i = 0; i < 4; i++)
      {
        sensorData.data[SensorData::fsrLFL+i] = 0.0f;
        sensorData.data[SensorData::fsrRFL+i] = 1.6f;
      }
    else
      for(int i = 0; i < 8; i++)
        sensorData.data[SensorData::fsrLFL+i] = 0.0f; 
  }

  // ultrasonic (model approximation. not absolutly correct in reality)
  static const float scale = 1000; //meter to millimeter
  if(sensorData.timeStamp - lastUsMeasurement >= 70 - 5)
  {
    lastUsMeasurement = sensorData.timeStamp;
    if(sensorData.usActuatorMode == SensorData::rightToLeft || sensorData.usActuatorMode == SensorData::leftToRight)
    {
      const SensorReading& usLDistance = ctrl->getSensorReading(spUsLeft);
      const SensorReading& usRDistance = ctrl->getSensorReading(spUsRight);
      sensorData.data[SensorData::usL] = float(usLDistance.data.doubleValue * scale);
      sensorData.data[SensorData::usR] = float(usRDistance.data.doubleValue * scale);
      sensorData.usActuatorMode = sensorData.usActuatorMode == SensorData::rightToLeft ? SensorData::leftToLeft : SensorData::rightToRight;
      sensorData.usTimeStamp = sensorData.timeStamp;
    }
    else if(sensorData.usActuatorMode == SensorData::leftToLeft || sensorData.usActuatorMode == SensorData::rightToRight)
    {
      const SensorReading& usLDistance = ctrl->getSensorReading(spUsCenterRight);
      const SensorReading& usRDistance = ctrl->getSensorReading(spUsCenterLeft);
      sensorData.data[SensorData::usL] = float(usLDistance.data.doubleValue * scale);
      sensorData.data[SensorData::usR] = float(usRDistance.data.doubleValue * scale);
      sensorData.usActuatorMode = sensorData.usActuatorMode == SensorData::leftToLeft ? SensorData::leftToRight : SensorData::rightToLeft;
      sensorData.usTimeStamp = sensorData.timeStamp;
    }
  }

  // angle
  const Matrix3d& world2robot(me->getRotation());
  if(naoV3)
  {
    sensorData.data[SensorData::angleX] = float(atan2(world2robot.col[1].v[2], world2robot.col[2].v[2]));
    sensorData.data[SensorData::angleY] = -float(atan2(world2robot.col[0].v[2], world2robot.col[2].v[2]));
  }
  else
  {
    sensorData.data[SensorData::angleX] = float(atan2(world2robot.col[0].v[2], world2robot.col[2].v[2]));
    sensorData.data[SensorData::angleY] = float(atan2(world2robot.col[1].v[2], world2robot.col[2].v[2]));
  }

  // simulate random imu malfunctions
  if(rand() % 300 == 1) 
  {
    sensorData.data[SensorData::accX] += rand() % 10;
    sensorData.data[SensorData::accY] += rand() % 10;
    sensorData.data[SensorData::accZ] += rand() % 10;
    sensorData.data[SensorData::gyroX] += rand() % 1;
    sensorData.data[SensorData::gyroY] += rand() % 1;
    sensorData.data[SensorData::angleX] += rand() % 20;
    sensorData.data[SensorData::angleY] += rand() % 20;
  }
}

void Oracle::getOrientationData(const SensorData& sensorData, OrientationData& orientationData)
{
  ASSERT(me);
  getPose3D(me, robotPose3D);
  const Pose3D offset(lastRobotPose3D.invert().conc(robotPose3D));
  orientationData.orientation.x = sensorData.data[SensorData::angleX];
  orientationData.orientation.y = sensorData.data[SensorData::angleY];
  float timeScale = 1.f / (float(sensorData.timeStamp - lastTimeStamp) * 0.001f);
  if(naoV3)
  {
    orientationData.velocity.x = float(offset.translation.y * timeScale);
    orientationData.velocity.y = -float(offset.translation.x * timeScale);
  }
  else
  {
    orientationData.velocity.x = float(offset.translation.x * timeScale);
    orientationData.velocity.y = float(offset.translation.y * timeScale);
  }
  orientationData.velocity.z = float(offset.translation.z * timeScale);
  lastRobotPose3D = robotPose3D;
  lastTimeStamp = sensorData.timeStamp;
}

SimObject* Oracle::getRobot() const
{
  return me;
}

SimObject* Oracle::getBall() const
{
  return ball;
}

void Oracle::getPose2D(const SimObject* obj, Pose2D& pose2D) const
{
  pose2D.rotation = normalize((float) obj->getRotation().getZAngle() - (naoV3 ? 0 : pi_2));
  const Vector3d& position(obj->getPosition());
  pose2D.translation.x = (float) position.v[0];
  pose2D.translation.y = (float) position.v[1];
}

void Oracle::getPose3D(const SimObject* obj, Pose3D& pose3D) const
{
  const Matrix3d& rotation(obj->getRotation());
  const Vector3d& translation(obj->getPosition());
  pose3D.rotation.c[0].x = float(rotation.col[0].v[0]);
  pose3D.rotation.c[0].y = float(rotation.col[0].v[1]);
  pose3D.rotation.c[0].z = float(rotation.col[0].v[2]);
  pose3D.rotation.c[1].x = float(rotation.col[1].v[0]);
  pose3D.rotation.c[1].y = float(rotation.col[1].v[1]);
  pose3D.rotation.c[1].z = float(rotation.col[1].v[2]);
  pose3D.rotation.c[2].x = float(rotation.col[2].v[0]);
  pose3D.rotation.c[2].y = float(rotation.col[2].v[1]);
  pose3D.rotation.c[2].z = float(rotation.col[2].v[2]);
  pose3D.translation.x = -float(translation.v[1]);
  pose3D.translation.y = float(translation.v[0]);
  pose3D.translation.z = float(translation.v[2]);
  pose3D.translation *= 1000.;
}
