/** 
* @file  Platform/SimRobotQt/SystemCall.cpp
* @brief static class for system calls
* @attention this is the SimRobotQt implementation
*/

#include "SystemCall.h"
#include "ControllerQt/ConsoleRoboCupCtrl.h"
#ifndef _WIN32
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <netdb.h>
#include <arpa/inet.h>
#endif

unsigned SystemCall::getCurrentSystemTime() 
{
  if(RoboCupCtrl::getController())
    return RoboCupCtrl::getController()->getTime();
  else
    return getRealSystemTime();
}

unsigned SystemCall::getRealSystemTime() 
{
#ifdef _WIN32
  unsigned time = timeGetTime();
#else
  struct timeval tv;
  gettimeofday(&tv, 0);
  unsigned int time = (unsigned int) (tv.tv_sec * 1000 + tv.tv_usec / 1000);
#endif
  static unsigned base = 0;
  if(!base)
    base = time - 10000; // avoid time == 0, because it is often used as a marker
  return time - base;
}

unsigned long long SystemCall::getCurrentThreadTime()
{
#ifdef _WIN32
  FILETIME ct, et, kernelT, userT;
  /*VERIFY(!*/GetThreadTimes(GetCurrentThread(),&ct, &et, &kernelT, &userT)/*)*/;

  unsigned long long userTime = userT.dwHighDateTime;
  userTime = (userT.dwLowDateTime + (userTime << 32)) / 10;
  unsigned long long kernelTime = userT.dwHighDateTime;
  kernelTime = (userT.dwLowDateTime + (kernelTime << 32)) / 10;
  unsigned long long time = userTime + kernelTime;

  static unsigned long long base = 0;
  if(!base)
    base = time - 1000000;
  return time - base;
#else
  clockid_t cid;
  struct timespec ts;

  VERIFY(pthread_getcpuclockid(pthread_self(), &cid) == 0);
  VERIFY(clock_gettime(cid, &ts) == 0);

  unsigned long long time = ts.tv_sec * 1000000 + ts.tv_nsec/1000;

  static unsigned long long base = 0;
  if(!base)
    base = time - 1000000;
  return time - base;
#endif
}

const char* SystemCall::getHostName()
{
  static char buf[100] = {0};
  if(!buf[0])
    VERIFY(!gethostname(buf, sizeof(buf)));
  return buf;
}

const char* SystemCall::getHostAddr()
{
  static char buf[100] = {0};
  if(!buf[0])
  {
    hostent* hostAddr = (hostent*) gethostbyname(getHostName());
    if(hostAddr && *hostAddr->h_addr_list)
      strcpy(buf, inet_ntoa(*(in_addr*) *hostAddr->h_addr_list));
    else
      strcpy(buf, "127.0.0.1");
  }
  return buf;
}

SystemCall::Mode SystemCall::getMode()
{
  return ConsoleRoboCupCtrl::getMode();
}

void SystemCall::sleep(unsigned int ms)
{
#ifdef _WIN32
  Sleep(ms);
#else
  usleep(ms * 1000);
#endif
}

void SystemCall::getLoad(float& mem, float load[3])
{
#ifdef _WIN32
  load[0] = load[1] = load[2] = -1.f; //Not implemented yet
  MEMORYSTATUS memStat;
  memStat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memStat);
  mem = float(memStat.dwMemoryLoad) / 100.f;
#else
  struct sysinfo info;
  if(sysinfo(&info) == -1)
    load[0] = load[1] = load[2] = mem = -1.f;
  else
  {
    load[0] = float(info.loads[0]) / 65536.f;
    load[1] = float(info.loads[1]) / 65536.f;
    load[2] = float(info.loads[2]) / 65536.f;
    mem = float(info.totalram - info.freeram) / float(info.totalram);
  }
#endif
}

