/**
* @file Platform/linux/GTAssert.h
* This file contains macros for low level debugging.
* @author Colin Graf
*/
#ifndef GTAssert_H
#define GTAssert_H

#undef ASSERT
#undef VERIFY
#undef TRACE

#if defined(RELEASE) || defined(NOASSERTIONS)

#if defined(TRACEASSERTIONS)

#include "NaoBody.h"
#define ASSERT(cond) BH_ASSERT_TRACE(cond)
#define VERIFY(cond) BH_VERIFY_TRACE(cond)
#define TRACE(...) BH_TRACE_TRACE(__VA_ARGS__)

#else

#define ASSERT(cond) ((void)0)
#define VERIFY(cond) ((void)(cond))
#define TRACE(...) ((void)0)

#endif

#else

class Assert
{
public:
  static void trace(const char* format, ...);
  static void verifyFailed(const char* file, unsigned int line, const char* condition);
  static void assertFailed(const char* file, unsigned int line, const char* condition);
};

/**
 * ASSERT prints a message if cond is false and RELEASE is not defined.
 * ASSERT does not evaluate cond if RELEASE is defined.
 * @param cond The condition to be checked.
 */
#define ASSERT(cond) ((void)((cond) ? 0 : (Assert::assertFailed(__FILE__, __LINE__, #cond), 0)))

/**
 * VERIFY prints a message if cond is false and RELEASE is not defined.
 * VERIFY does evaluate cond even if RELEASE is defined.
 * @param cond The condition to be checked.
 */
#define VERIFY(cond) ((void)((cond) ? 0 : (Assert::verifyFailed(__FILE__, __LINE__, #cond), 0)))

/**
 * TRACE prints a message if RELEASE is not defined.
 */
#define TRACE(...) Assert::trace(__VA_ARGS__)

#endif

#endif // GTAssert_H

