/**
* @file Platform/linux/NaoBody.cpp
* Declaration of a class for accessing the body of the nao via NaoQi/libbhuman.
* @author Colin Graf
* @author jeff
*/

#include <sys/mman.h>
#include <fcntl.h>
#include <semaphore.h>
#include <unistd.h>
#include <pthread.h>
#include <cerrno>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>

#include "NaoBody.h"
#include "GTAssert.h"

#include "libbhuman/bhuman.h"

int NaoBody::fd = -1;
void* NaoBody::sem = SEM_FAILED;
LBHData* NaoBody::lbhData = (LBHData*)MAP_FAILED;
int NaoBody::allocations = 0;

unsigned int NaoBody::cognitionThreadId = 0;
unsigned int NaoBody::motionThreadId = 0;

NaoBody::NaoBody() : opened(false), writingActuators(-1), fdCpuTemp(-1), fdMbTemp(-1)
{
  if(lbhData != MAP_FAILED)
  {
    opened = true;
    ++allocations;
  }
}

NaoBody::~NaoBody()
{
  if(opened)
    if(--allocations == 0)
    {
      if(lbhData != MAP_FAILED)
        munmap(lbhData, sizeof(LBHData));
      if(fd != -1)
        close(fd);
      if(sem != SEM_FAILED)
        sem_close((sem_t*)sem);

      fd = -1;
      sem = SEM_FAILED;
      lbhData = (LBHData*)MAP_FAILED;
    }
  if(fdCpuTemp != -1)
    close(fdCpuTemp);
  if(fdMbTemp != -1)
    close(fdMbTemp);
}

bool NaoBody::init()
{
  if(opened)
    return true; // already initialized

  if(lbhData != MAP_FAILED)
  {
    opened = true;
    ++allocations;
    return true;
  }

  fd = shm_open(LBH_MEM_NAME, O_RDWR, S_IRUSR | S_IWUSR);
  if(fd == -1)
    return false;

  sem = sem_open(LBH_SEM_NAME, O_RDWR, S_IRUSR | S_IWUSR, 0);
  if(sem == SEM_FAILED)
  {
    close(fd);
    fd = -1;
    return false;
  }

  VERIFY((lbhData = (LBHData*)mmap(NULL, sizeof(LBHData), PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)) != MAP_FAILED);
  lbhData->state = okState;

  opened = true;
  ++allocations;
  return true;
}

void NaoBody::setCrashed(int termSignal)
{
  lbhData->state = BHState(termSignal);
}

bool NaoBody::wait()
{
  do
  {
    ASSERT(opened);
    if(sem_wait((sem_t*)sem) == -1)
    {
      bool success = false;
      while(errno == 516)
      {
        if(sem_wait((sem_t*)sem) == -1)
        {
          ASSERT(false);
          continue;
        }
        else
        {
          success = true;
          break;
        }
      }
      if(!success)
      {
        ASSERT(false);
        return false;
      }
    }
  }
  while(lbhData->readingSensors == lbhData->newestSensors);
  lbhData->readingSensors = lbhData->newestSensors;

  static bool shout = true;
  if(shout)
  {
    shout = false;
    printf("NaoQi is working\n");
  }

  return true;
}

const char* NaoBody::getName() const
{
  ASSERT(opened);
  printf("Hi, I am %s.\n", lbhData->robotName);
  return lbhData->robotName;
}

float* NaoBody::getSensors()
{
  ASSERT(opened);
  return lbhData->sensors[lbhData->readingSensors];
}

void NaoBody::getTemperature(float& cpu, float& mb)
{
  cpu = mb = 0;

#ifdef DEBUG
  if(fdCpuTemp == -1)
  {
    fdCpuTemp = open("/sys/class/i2c-adapter/i2c-1/1-004c/temp2_input", O_RDONLY);
    ASSERT(fdCpuTemp > -1);
  }

  if(fdMbTemp == -1)
  {
    fdMbTemp = open("/sys/class/i2c-adapter/i2c-1/1-004c/temp1_input", O_RDONLY);
    ASSERT(fdMbTemp > -1);
  }

  char buf[32];

  if(fdCpuTemp != -1)
  {
    VERIFY(lseek(fdCpuTemp, 0, SEEK_SET) == 0);
    int i = read(fdCpuTemp, buf, sizeof(buf)); 
    ASSERT(i != -1);
    if(i >= 0)
    {
      buf[i < int(sizeof(buf) - 1) ? i : int(sizeof(buf) - 1)] = '\0';
      cpu = float(atoi(buf)) / 1000.f;
    }
  }

  if(fdMbTemp != -1)
  {
    VERIFY(lseek(fdMbTemp, 0, SEEK_SET) == 0);
    int i = read(fdMbTemp, buf, sizeof(buf)); 
    ASSERT(i != -1);
    if(i >= 0)
    {
      buf[i < int(sizeof(buf) - 1) ? i : int(sizeof(buf) - 1)] = '\0';
      mb = float(atoi(buf)) / 1000.f;
    }
  }
#endif
}

void NaoBody::openActuators(float*& actuators)
{
  ASSERT(opened);
  ASSERT(writingActuators == -1);
  writingActuators = 0;
  if(writingActuators == lbhData->newestActuators)
    ++writingActuators;
  if(writingActuators == lbhData->readingActuators)
    if(++writingActuators == lbhData->newestActuators)
      ++writingActuators;
  ASSERT(writingActuators != lbhData->newestActuators);
  ASSERT(writingActuators != lbhData->readingActuators);
  actuators = lbhData->actuators[writingActuators];
}

void NaoBody::closeActuators()
{
  ASSERT(writingActuators >= 0);
  lbhData->newestActuators = writingActuators;
  writingActuators = -1;
}

bool NaoBody::initProcessTrace(const char* proc)
{
  if(!strcmp(proc, "Cognition"))
    cognitionThreadId = pthread_self();
  else if(!strcmp(proc, "Motion"))
    motionThreadId = pthread_self();
  else
    return false;
  return true;
}

void NaoBody::processTraceMsg(const char* file, const int line, const char* msg)
{
  unsigned int threadId = pthread_self();
  if(threadId == cognitionThreadId) 
  { 
    if(lbhData->cognitionTraceCounter == LBH_TRACE_SIZE)
      lbhData->cognitionTraceCounter = 0;
    snprintf(lbhData->cognitionTrace[lbhData->cognitionTraceCounter],
        LBH_TRACE_MSG_LENGTH,
        "%s:%d %s", file, line, msg);
    lbhData->cognitionTraceCounter++;
  }
  else if(threadId == motionThreadId)
  {
    if(lbhData->motionTraceCounter == LBH_TRACE_SIZE)
      lbhData->motionTraceCounter = 0;
    snprintf(lbhData->motionTrace[lbhData->motionTraceCounter],
        LBH_TRACE_MSG_LENGTH,
        "%s:%d %s", file, line, msg);
    lbhData->motionTraceCounter++;
  }
}

void NaoBody::traceMsg(const char* format, ...)
{
  if(lbhData == MAP_FAILED)
    return;

  // TODO: mutex!

  if(lbhData->assertTraceCounter == LBH_TRACE_SIZE)
    lbhData->assertTraceCounter = 0;

  va_list ap;
  va_start(ap, format);
  vsnprintf(lbhData->assertTrace[lbhData->assertTraceCounter], LBH_TRACE_MSG_LENGTH, format, ap);
  va_end(ap);

  lbhData->assertTraceCounter++;
}

void NaoBody::assertFailed(const char* file, const int line, const char* condition)
{
  traceMsg("%s:%d ASSERT(%s) failed", file, line, condition);
}

void NaoBody::verifyFailed(const char* file, const int line, const char* condition)
{
  traceMsg("%s:%d VERIFY(%s) failed", file, line, condition);
}

void NaoBody::dumpAndFlushAssertTrace()
{
  FILE *fp = NULL;

  //dump
  for(int i = 0; i < LBH_TRACE_SIZE; i++)
  {
    if(lbhData->assertTrace[(lbhData->assertTraceCounter + i) % LBH_TRACE_SIZE][0] != 0)
    {
      if(!fp)
      {
        printf("---- ASSERT TRACE ----\n");

        fp = fopen("/media/userdata/assertTrace", "a");
        if(fp)
          fprintf(fp, "---- ASSERT TRACE ----\n");
      }

      if(fp)
        fprintf(fp, "%s\n", lbhData->assertTrace[(lbhData->assertTraceCounter + i) % LBH_TRACE_SIZE]);
      printf("%s\n", lbhData->assertTrace[(lbhData->assertTraceCounter + i) % LBH_TRACE_SIZE]);
    }
  }

  //flush
  lbhData->assertTraceCounter = 0;
  memset(lbhData->assertTrace, 0, sizeof(lbhData->assertTrace));
}

void NaoBody::dumpTrace(bool toStdout) const
{
  FILE* fp = toStdout ? stderr : fopen("/var/log/bhuman", "w");
  if(fp == 0)
    return;

  fprintf(fp, "---- COGNITION TRACE ----\n");
  for(int i = 0; i < LBH_TRACE_SIZE; i++)
    fprintf(fp, "%d: %s\n", i, lbhData->cognitionTrace[(lbhData->cognitionTraceCounter + i) % LBH_TRACE_SIZE]);

  fprintf(fp, "---- MOTION TRACE ----\n");
  for(int i = 0; i < LBH_TRACE_SIZE; i++)
    fprintf(fp, "%d: %s\n", i, lbhData->motionTrace[(lbhData->motionTraceCounter + i) % LBH_TRACE_SIZE]);

  fprintf(fp, "---- %s ----\n", getBHStateName(int(lbhData->state)));

  if(fp != stderr)
    fclose(fp);
}

const char* NaoBody::getBHStateName(int bhState)
{
  switch(bhState)
  {
  case okState:
    return "ok";
  case abnormalTerminationState:
    return "abnormalTermination";
  case sigINTState:
    return "sigINT";
  case sigQUITState:
    return "sigQUIT";
  case sigILLState:
    return "sigILL";
  case sigABRTState:
    return "sigABRT";
  case sigFPEState:
    return "sigFPE";
  case sigKILLState:
    return "sigKILL";
  case sigSEGVState:
    return "sigSEGV";
  case sigPIPEState:
    return "sigPIPE";
  case sigALRMState:
    return "sigALRM";
  case sigTERMState:
    return "sigTERM";
  default:
    return "unknown";
  }
}

