/**
* @file Platform/linux/NaoBody.cpp
* Declaration of a class for accessing the body of the nao via NaoQi/libbhuman.
* @author Colin Graf
*/

#ifndef NaoBody_H
#define NaoBody_H

struct LBHData;

/**
* @class NaoBody
* Encapsulates communication with libbhuman.
*/
class NaoBody
{
public:

  /** Default constructor. */
  NaoBody();

  /** Destructor. */
  ~NaoBody();

  /** Establishes the connection to libbhuman.
  * @return Whether the connection was established correctly. */
  bool init();

  /** Waits for a new set of sensor data. */
  bool wait();

  /** Activates the eye-blinking mode to indicate a crash.
  * @param termSignal The termination signal that was raised by the crash. */
  void setCrashed(int termSignal);

  /** Accesses the name of the robot's body. 
  * @return The name. */
  const char* getName() const;

  /** Accesses the sensor value buffer. 
  * @return An array of sensor values. Ordered corresponding to \c lbhSensorNames of \c bhuman.h. */
  float* getSensors();

  void getTemperature(float& cpu, float& mb);

  /** Accesses the actuator value buffer for writing. 
  * @param actuators A reference to a variable to store a pointer to the actuator value buffer in. Ordered corresponding to \c lbhActuatorNames of \c bhuman.h. */
  void openActuators(float*& actuators);
  
  /** Commits the actuator value buffer. */
  void closeActuators();

  /** Initializes low-level flow tracing for the calling process.
  * @param proc The name of the calling process. (should be "Cognition" or "Motion".)
  * @return Whether the initialization was successful.
  */
  static bool initProcessTrace(const char* proc);

  /** Adds a process tracing entry.
  * @param file The name of the current file. (Use the __FILE__ macro.)
  * @param line The current file line. (Use the __LINE__ macro.)
  * @param msg A additional message. */
  static void processTraceMsg(const char* file, const int line, const char* msg);

  /** Adds an assert tracing entry.
  * @param format The format of the message (printf style). */
  static void traceMsg(const char* format, ...);

  /** Adds an assert tracing entry.
  * @param file The name of the current file. (Use the __FILE__ macro.)
  * @param line The current file line. (Use the __LINE__ macro.)
  * @param condition The condition of the assert. */
  static void assertFailed(const char* file, const int line, const char* condition);

  /** Adds an assert tracing entry.
  * @param file The name of the current file. (Use the __FILE__ macro.)
  * @param line The current file line. (Use the __LINE__ macro.)
  * @param condition The condition of the verify. */
  static void verifyFailed(const char* file, const int line, const char* condition);

  /** Prints the trace-entries to /var/log/bhuman or stderr. 
  * @param toStderr whether the trace-entries should be printed to stderr or to /var/log/bhuman
  */
  void dumpTrace(bool toStderr = false) const;

  /** Prints the assert trace-entries to stderr. */
  void dumpAndFlushAssertTrace();

private:
  static int fd; /**< The file descriptor for the shared memory block. */
  static void* sem; /**< The semaphore used for synchronizing to the NaoQi DCM. */ 
  static LBHData* lbhData; /**< The pointer to the mapped shared memory block. */
  static int allocations; /**< The amount of instances of this class. */

  static unsigned int cognitionThreadId; /**< The thread id of the coginition process. */
  static unsigned int motionThreadId; /**< The thread id of the motion process. */

  bool opened; /**< Whether the shared memory block and the semaphore was opened. */
  int writingActuators; /**< The index of the opened exclusive actuator writing buffer. */

  int fdCpuTemp;
  int fdMbTemp;

  /** Converts a BHState into a string.
  * @param bhState The BHState to convert.
  * @return The name of the BHState. */
  static const char* getBHStateName(int bhState);
};

#if defined(TARGET_ROBOT) && !defined(RELEASE)

#define BH_TRACE_INIT(proc) \
  NaoBody::initProcessTrace(proc)

#define BH_TRACE_MSG(msg) \
  NaoBody::processTraceMsg(__FILE__, __LINE__, msg)

#define BH_TRACE \
  NaoBody::processTraceMsg(__FILE__, __LINE__, "")

#else

#define BH_TRACE ((void)0)
#define BH_TRACE_INIT(x) ((void)0)
#define BH_TRACE_MSG(x) ((void)0)

#endif //TARGET_ROBOT && RELEASE

#if defined(TARGET_ROBOT) && defined(TRACEASSERTIONS)

#define BH_VERIFY_TRACE(c) ((void)((c) ? 0 : (NaoBody::verifyFailed(__FILE__, __LINE__, #c), 0)))
#define BH_ASSERT_TRACE(c) ((void)((c) ? 0 : (NaoBody::assertFailed(__FILE__, __LINE__, #c), 0)))
#define BH_TRACE_TRACE(...) NaoBody::traceMsg(__VA_ARGS__)

#else

#define BH_ASSERT_TRACE(c) ((void)0)
#define BH_VERIFY_TRACE(c) ((void)(c))
#define BH_TRACE_TRACE(...) ((void)0)

#endif //TARGET_ROBOT && TRACEASSERTIONS

#endif //NaoBody_H

