/** 
* @file  Platform/linux/SystemCall.cpp
* @brief static class for system calls, Linux implementation
* @attention this is the Linux implementation
*
* @author <A href=mailto:brunn@sim.informatik.tu-darmstadt.de>Ronnie Brunn</A>
* @author <A href=mailto:martin@martin-loetzsch.de>Martin Ltzsch</A>
* @author <A href=mailto:risler@sim.informatik.tu-darmstadt.de>Max Risler</A>
* @author <a href=mailto:dueffert@informatik.hu-berlin.de>Uwe Dffert</a>
*/

#include "SystemCall.h"
#include "GTAssert.h"
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <cstring>
#include <pthread.h>
#include <ctime>

unsigned SystemCall::getCurrentSystemTime() 
{
  return getRealSystemTime();
}

unsigned SystemCall::getRealSystemTime() 
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  unsigned int time = (unsigned int) (tv.tv_sec * 1000 + tv.tv_usec / 1000);
  static unsigned base = 0;
  if(!base)
    base = time - 10000; // avoid time == 0, because it is often used as a marker
  return time - base;
}

unsigned long long SystemCall::getCurrentThreadTime()
{
  clockid_t cid;
  struct timespec ts;

  VERIFY(pthread_getcpuclockid(pthread_self(), &cid) == 0);
  VERIFY(clock_gettime(cid, &ts) == 0);

  unsigned long long time = ts.tv_sec * 1000000 + ts.tv_nsec/1000;

  static unsigned long long base = 0;
  if(!base)
    base = time - 10000 * 1000;
  return time - base;
}

const char* SystemCall::getHostName()
{
  static char buf[100] = {0};
  if(!buf[0])
    {VERIFY(!gethostname(buf, sizeof(buf)));}
  return buf;
}

const char* SystemCall::getHostAddr()
{
  static char buf[100] = {0};
  if(!buf[0])
  {
    hostent* hostAddr = (hostent*) gethostbyname(getHostName());
    if(*hostAddr->h_addr_list)
      strcpy(buf, inet_ntoa(*(in_addr*) *hostAddr->h_addr_list));
    else
      strcpy(buf, "127.0.0.1");
  }
  return buf;
}

SystemCall::Mode SystemCall::getMode()
{
  return physicalRobot;
}

void SystemCall::sleep(unsigned long ms)
{
  usleep(1000 * ms);
}

void SystemCall::getLoad(float& mem, float load[3])
{
  struct sysinfo info;
  if(sysinfo(&info) == -1)
    load[0] = load[1] = load[2] = mem = -1.f;
  else
  {
    load[0] = float(info.loads[0]) / 65536.f;
    load[1] = float(info.loads[1]) / 65536.f;
    load[2] = float(info.loads[2]) / 65536.f;
    mem = float(info.totalram - info.freeram) / float(info.totalram);
  }
}

