/**
* @file Representations/BehaviorControl/BehaviorData.h
* The file declares a class that containts data about the current behavior state.
* @author Colin Graf
*/ 

#ifndef BehaviorData_H
#define BehaviorData_H

#include "Tools/Streams/Streamable.h"
#include "Tools/Math/Vector2.h"

/**
* @class BehaviorData
* A class that containts data about the current behavior state.
*/ 
class BehaviorData : public Streamable  
{
public:
  enum Role
  {
    undefined,
    firstRole=1,
    keeper=1,
    supporter,
    striker,
    defender,
    offensiveSupporter,
    defensiveSupporter,
    numOfRoles
  } role; /**< A dynamically chosen role. */
  
  enum Action
  {
    unknown,
    dribble,
    goToBall,
    searchForBall,
    goToTarget,
    prepareKick,
    kick,
    kickSidewards,
    pass,
    block,
    hold,
    standUp,
    patrol,
    passBeforeGoal,
    kickoff,
    goAround,
    waitForPass,
    preparePass,
    parry,
    guard,
    defending,
    numOfActions
  } action; /**< What is the robot doing in general? */

  unsigned int estimatedTimeToReachBall; /**< The estimated time to reach the the kick pose (in ms) */
  bool kickoffInProgress; /**< Whether the kickoff is currently in progress or not. */

  /**
  * Default constructor.
  */
  BehaviorData() : 
    role(undefined), 
    action(unknown),
    kickoffInProgress(false)
  {}

  /**
  * The function returns the name as a string of a role.
  * @param role The role
  * @return The name of the role
  */
  static const char* getRoleName(Role role)
  {
    static const char* name[] =
    {
      "undefined",
      "keeper",
      "supporter",
      "striker",
      "defender",
      "offensiveSupporter",
      "defensiveSupporter"
    };
    return name[role > 0 && (size_t) role < sizeof(name) / sizeof(*name) ? role : 0];
  };

  /**
  * The function returns the name as a string of a behavior action.
  * @param action The behavior action
  * @return The name of the behavior action
  */
  static const char* getActionName(Action action)
  {
    static const char* name[] =
    {
      "unknown",
      "dribble",
      "goToBall",
      "searchForBall",
      "goToTarget",
      "prepareKick",
      "kick",
      "kickSidewards",
      "pass",
      "block",
      "hold",
      "standUp",
      "patrol",
      "passBeforeGoal",
      "kickoff",
      "goAround",
      "waitForPass",
      "preparePass",
      "defending"
    };
    return name[action > 0 && (size_t) action < sizeof(name) / sizeof(*name) ? action : 0];
  };
  
private:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read (if in != 0).
  * @param out The stream to which the object is written (if out != 0).
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    STREAM_ENUM(role, numOfRoles, BehaviorData::getRoleName);
    STREAM_ENUM(action, numOfActions, BehaviorData::getActionName);
    STREAM(estimatedTimeToReachBall);
    STREAM(kickoffInProgress);
    STREAM_REGISTER_FINISH();
  }
};

#endif // BehaviorData_H
