/**
* @file Representations/Configuration/BH2010Parameters.h
* The file declares a class that containts frequently used parameters of the behavior which can be modified.
* @author <a href="mailto:dhonsel@informatik.uni-bremen.de">Daniel Honsel</a>
*/

#ifndef BH2010Parameters_H
#define BH2010Parameters_H

#include "Tools/Math/Pose2D.h"

class BH2010Parameters : public Streamable  
{
public:  
  // supporter
  float supporterStayBehindStrikerThreshhold; /**< threshhold for the striker position at which the supporter waits behind the striker */
  float supporterWaitingTimeBeforeLookingForBall; /**< time the supporter waits before starting to look for the ball */
  enum Tactic
  {
    defensive,
    normal,
    offensive
  } tactic; /**< tactic the supporter uses */

  //defender
  bool disableDefender; /**< decides whether the defender is used*/

  //nativeRoles
  bool nativeRolesOnly; /**< decides whether role selector will determine role by nativeRole or dynamically*/

  // kick pose parameters
  float kickPoseKickInaccuracy; /**< The inaccuracy of the kick direction angle according to the intended direction */
  Pose2D kickPoseAssumedWalkingSpeed; /**< The walking speed assumed for estimating the time to reach the ball for shooting a goal */
  float kickPoseSidewardsKickRange; /**< The range of sidwards kicks */
  float kickPoseBackwardsKickRange; /**< The range of backwards kicks */

  // walk to point method
  enum WalkToPointMethod
  {
    matroc,
    cgraf,
  } walkToPointMethod;

  // obstacle avoidance
  float obstacleMinDistanceToLeft;
  float obstacleMinDistanceToRight;
  float obstacleMinDistanceToCenterLeft;
  float obstacleMinDistanceToCenterRight;
  float obstacleMinDistanceToClosest;
  float obstacleMinDistanceToClosestLeft;
  float obstacleMinDistanceToClosestRight;

  // kickoff parameters
  Vector2<> kickoffDribbleDistance; /**< Distance to begin dribbling in kickoff. */
  float kickoffDribbleDistanceTolerance; /**< Hysteresis used in kickoff dribble. */

  //dribble parameters
  float aligningMinDistanceToBall;
  float dcMinShootDistanceX;
  float dcDribbleTimeout;
  float dcUseGamecontroller;
  float dcInitialLocalisationTime;
  float dcBasicDribbleSpeed;

  /** Default constructor. */
  BH2010Parameters() : 
    supporterStayBehindStrikerThreshhold(0),
    supporterWaitingTimeBeforeLookingForBall(0),
    tactic(defensive),
    disableDefender(0),
    nativeRolesOnly(0),
    kickPoseKickInaccuracy(0),
    kickPoseSidewardsKickRange(999999.f),
    walkToPointMethod(matroc),
    obstacleMinDistanceToLeft(0),
    obstacleMinDistanceToRight(0),
    obstacleMinDistanceToCenterLeft(0),
    obstacleMinDistanceToCenterRight(0),
    obstacleMinDistanceToClosest(0),
    obstacleMinDistanceToClosestLeft(0),
    obstacleMinDistanceToClosestRight(0),
    kickoffDribbleDistance(0, 0),
    kickoffDribbleDistanceTolerance(0),
    aligningMinDistanceToBall(0),
    dcMinShootDistanceX(0),
    dcDribbleTimeout(15),
    dcUseGamecontroller(0),
    dcInitialLocalisationTime(8),
    dcBasicDribbleSpeed(35) {}

private:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read (if in != 0).
  * @param out The stream to which the object is written (if out != 0).
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
      STREAM(supporterStayBehindStrikerThreshhold);
      STREAM(supporterWaitingTimeBeforeLookingForBall);
      STREAM_ENUMASINT(tactic);
      STREAM(disableDefender);
      STREAM(nativeRolesOnly);
      STREAM(kickPoseKickInaccuracy);
      STREAM(kickPoseAssumedWalkingSpeed);
      STREAM(kickPoseSidewardsKickRange);
      STREAM(kickPoseBackwardsKickRange);
      STREAM_ENUMASINT(walkToPointMethod);
      STREAM(obstacleMinDistanceToLeft);
      STREAM(obstacleMinDistanceToRight);
      STREAM(obstacleMinDistanceToCenterLeft);
      STREAM(obstacleMinDistanceToCenterRight);
      STREAM(obstacleMinDistanceToClosest);
      STREAM(obstacleMinDistanceToClosestLeft);
      STREAM(obstacleMinDistanceToClosestRight);
      STREAM(kickoffDribbleDistance);
      STREAM(kickoffDribbleDistanceTolerance);
      STREAM(aligningMinDistanceToBall);
      STREAM(dcMinShootDistanceX);
      STREAM(dcDribbleTimeout);
      STREAM(dcUseGamecontroller);
      STREAM(dcInitialLocalisationTime);
      STREAM(dcBasicDribbleSpeed);
    STREAM_REGISTER_FINISH();
  }
};

#endif // BH2010Parameters_H
