/**
 * @file RobotPose.h
 *
 * The file contains the definition of the class RobotPose.
 *
 * @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
 */

#ifndef __RobotPose_h_
#define __RobotPose_h_

#include "Tools/Math/Pose2D.h"
#include "Tools/Debugging/DebugDrawings.h"

/**
* @class RobotPose
* A Pose2D with validity.
*/
class RobotPose : public Pose2D
{
protected:
  virtual void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM_BASE(Pose2D);
    STREAM(validity);
    STREAM(deviation);
    STREAM_REGISTER_FINISH();
  }

public:
  float validity;                   /**< The validity of the robot pose. (0 = invalid, 1 = perfect) */
  float deviation;                  /**< The deviation of the robot pose. */
  ColorRGBA ownTeamColorForDrawing;  /**< As the name says... */
 
  enum
  {
    unknownDeviation = 100000,
  };

  /** Constructor */
  RobotPose() : validity(0.0f), deviation(unknownDeviation) {}

  /** Assignment operator
  * @param other Another RobotPose
  * @return A reference to the object after the assignment
  */
  const RobotPose& operator=(const RobotPose& other)
  {
    (Pose2D&) *this = (const Pose2D&) other;
    validity = other.validity;
    deviation = other.deviation;
    ownTeamColorForDrawing = other.ownTeamColorForDrawing;
    return *this;
  }

  /** Cast Contructor.
  * @param otherPose A Pose2D object
  */
  RobotPose (const Pose2D& otherPose) 
  {
    (Pose2D&) *this = otherPose;
    validity = 0;
    deviation = 100000.f;
  }

  /** Assignment operator for Pose2D objects
  * @param other A Pose2D object
  * @return A reference to the object after the assignment
  */
  const RobotPose& operator=(const Pose2D& other)
  {
    (Pose2D&) *this = other;
    //validity is not set
    return *this;
  }

  /** Draws the robot pose to the field view*/
  void draw();
};

class GroundTruthRobotPose : public RobotPose 
{
public:
  /** Draws the robot pose to the field view*/
  void draw();
};

class PotentialRobotPose : public RobotPose
{
public:
  /** Draws the robot pose to the field view*/
  void draw() {};
};

/**
* @class RobotPoseInfo
* Encapsulates further information about the robot pose
*/
class RobotPoseInfo : public Streamable
{
public:
  unsigned int timeLastPoseReset; /**< The time when the robot pose was reset. */

private:
  /**
  * Makes the object streamable
  * @param in The stream from which the object is read
  * @param out The stream to which the object is written
  */
  virtual void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(timeLastPoseReset);
    STREAM_REGISTER_FINISH();
  }
};

#endif //__RobotPose_h_
