/**
* @file Representations/MotionControl/HeadMotionRequest.h
* This file declares a class that represents the requested head motion.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#ifndef HeadMotionRequest_H
#define HeadMotionRequest_H

#include "Tools/Streams/Streamable.h"
#include "Tools/Math/Vector3.h"

/**
* @class HeadMotionRequest
* A class that represents the requested head motion.
*/
class HeadMotionRequest : public Streamable
{
public:
  enum Mode
  {
    panTiltMode,        /**< Use \c pan, \c tilt and \c speed. */
    targetMode,         /**< (A target relative to the center of hip.) Use \c target and \c speed. */
    targetOnGroundMode, /**< Use \c target and \c speed. */
    numOfModes,
  };

  Mode mode;        /**< The active head motion mode. */
  float pan,        /**< Head pan target angle in radians. */
        tilt,       /**< Head tilt target angle in radians. */
        speed;      /**< Maximum joint speed to reach target angles in radians/s. */
  Vector3<> target; /**< Look at target relative to the robot. */

  /** Default constructor. */
  HeadMotionRequest() : mode(panTiltMode), pan(0), tilt(0), speed(1), target(1,0,0) {}

protected:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read 
  * @param out The stream to which the object is written 
  */
  virtual void serialize( In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    STREAM_ENUM(mode, numOfModes, getModeName);
    STREAM(pan);
    STREAM(tilt);
    STREAM(speed);    
    STREAM(target);
    STREAM_REGISTER_FINISH();
  }

private:
  /** 
  * The function returns names of modes.
  * @param mode The mode the name of which is returned.
  * @return The corresponding name.
  */
  static const char* getModeName(Mode mode)
  {
    switch(mode)
    {
    case panTiltMode: return "panTilt";
    case targetMode: return "target";
    case targetOnGroundMode: return "targetOnGround";
    default: return "unknown";
    }
  }
};

#endif // HeadMotionRequest_H
