/**
* @file Representations/MotionControl/MotionRequest.h
* This file declares a class that represents the motions that can be requested from the robot.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#ifndef MotionRequest_H
#define MotionRequest_H

#include "SpecialActionRequest.h"
#include "WalkRequest.h"
#include "BikeRequest.h"
#include "SIKERequest.h"

/**
* @class MotionRequest
* A class that represents the motions that can be requested from the robot.
*/
class MotionRequest : public Streamable
{
public:
  enum Motion
  {
    walk,
    bike,
    specialAction,
    stand,
    sike,
    numOfMotions
  };

  Motion motion; /**< The selected motion. */
  SpecialActionRequest specialActionRequest; /**< The special action request, if it is the selected motion. */
  WalkRequest walkRequest; /**< The walk request, if it is the selected motion. */
  BikeRequest bikeRequest; /**< The kick request, if it is the selected motion. */
  SIKERequest sikeRequest;

  /** 
  * The function returns names of motions.
  * @param motion The motion the name of which is returned.
  * @return The corresponding name.
  */
  static const char* getMotionName(Motion motion)
  {
    switch(motion)
    {
    case walk: return "walk";
    case bike: return "bike";
    case specialAction: return "specialAction";
    case stand : return "stand";
    case sike : return "sike";
    default: return "unknown";
    }
  }

  /** 
  * Default constructor.
  */
  MotionRequest() : motion(specialAction) {}

  /**
  * Prints the motion request to a readable string. (E.g. "walk: 100mm/s 0mm/s 0/s")
  * @param destination The string to fill
  */
  void printOut(char* destination) const;

  /** Draws something*/
  void draw() const; 

protected:
  /**
  * Makes the object streamable
  * @param in The stream from which the object is read
  * @param out The stream to which the object is written
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    STREAM_ENUM(motion, numOfMotions, MotionRequest::getMotionName);
    STREAM(specialActionRequest);
    STREAM(walkRequest);
    STREAM(bikeRequest);
    STREAM(sikeRequest);
    STREAM_REGISTER_FINISH();
  }
};

#endif // MotionRequest_H
