/**
* @file Representations/MotionControl/WalkRequest.h
* This file declares a class that represents a walk request.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
* @author Colin Graf
*/

#ifndef WalkRequest_H
#define WalkRequest_H

#include "Tools/Math/Pose2D.h"

/**
* @class WalkRequest
* A class that represents a walk request.
*/
class WalkRequest : public Streamable
{
public:
  enum Mode
  {
    speedMode, /**< Use absolute \c speed. */
    percentageSpeedMode, /**< Use percentage \c speed. */
    targetMode, /**< Use \c target as walking target relative to the current position of the robot. */
    numOfModes,
  };

  Mode mode; /**< The walking mode. */
  Pose2D speed; /**< Walking speeds, in percentage or mm/s and radian/s. */
  Pose2D target; /**< Walking target, in mm and radians, relative to the robot. Use either a speed or a target. */
  bool pedantic; /**< Allows to disable the step size stabilization. set it when precision is indispensable. */

  enum KickType
  {
    none, /**< do not kick */
    left, /**< kick using the left foot */
    right, /**< kick using the right foot */
    sidewardsLeft,
    sidewardsRight,
    numOfKickTypes,
  };

  KickType kickType;
  Pose2D kickTarget;

  /** Default constructor. */
  WalkRequest() : mode(speedMode), pedantic(false), kickType(none) {}

private:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read 
  * @param out The stream to which the object is written 
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN( );
    STREAM_ENUM(mode, numOfModes, getModeName);
    STREAM(speed);
    STREAM(target);
    STREAM(pedantic);
    STREAM_ENUM(kickType, numOfKickTypes, getKickTypeName);
    STREAM(kickTarget);
    STREAM_REGISTER_FINISH();
  }

  /** 
  * The function returns names of modes.
  * @param mode The mode the name of which is returned.
  * @return The corresponding name.
  */
  static const char* getModeName(Mode mode)
  {
    switch(mode)
    {
    case speedMode: return "speed";
    case percentageSpeedMode: return "percentageSpeed";
    case targetMode : return "target";
    default: return "unknown";
    }
  }

  /** 
  * The function returns names of kick types.
  * @param kickType The kick type the name of which is returned.
  * @return The corresponding name.
  */
  static const char* getKickTypeName(KickType kickType)
  {
    switch(kickType)
    {
    case none: return "none";
    case left: return "left";
    case right: return "right";
    case sidewardsLeft: return "sidewardsLeft";
    case sidewardsRight: return "sidewardsRight";
    default: return "unknown";
    }
  }
};

#endif // WalkRequest_H
