/**
* @file RobotModel.cpp
* Implementation of class RobotModel.
* @author Alexander Hrtl
*/

#include "RobotModel.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Debugging/DebugDrawings3D.h"

RobotModel::RobotModel(const JointData& joints, const RobotDimensions& robotDimensions, const MassCalibration& massCalibration)
{
  setJointData(joints, robotDimensions, massCalibration);
}

void RobotModel::setJointData(const JointData& joints, const RobotDimensions& robotDimensions, const MassCalibration& massCalibration)
{
  // compute unilateral limbs
  limbs[neck] = Pose3D(0, 0, robotDimensions.zLegJoint1ToHeadPan)
    .rotateZ(joints.angles[JointData::HeadYaw]);
  limbs[head] = Pose3D(limbs[neck])
    .rotateY(-joints.angles[JointData::HeadPitch]);

  // compute bilateral limbs 
  for(int side = 0; side < 2; side++)
  {
    bool left = side == 0;
    int sign = left ? -1 : 1;
    Limb pelvis = left ? pelvisLeft : pelvisRight;
    JointData::Joint leg0 = left ? JointData::LHipYawPitch : JointData::RHipYawPitch;
    Limb shoulder = left ? shoulderLeft : shoulderRight;
    JointData::Joint arm0 = left ? JointData::LShoulderPitch : JointData::RShoulderPitch;

    limbs[shoulder + 0] = Pose3D(robotDimensions.armOffset.x, robotDimensions.armOffset.y * -sign, robotDimensions.armOffset.z)
      .rotateY(-joints.angles[arm0 + 0]);
    limbs[shoulder + 1] = Pose3D(limbs[shoulder + 0])
      .rotateZ(joints.angles[arm0 + 1] * -sign);
    limbs[shoulder + 2] = Pose3D(limbs[shoulder + 1])
      .translate(robotDimensions.upperArmLength, 0, 0)
      .rotateX(joints.angles[arm0 + 2] * -sign);
    limbs[shoulder + 3] = Pose3D(limbs[shoulder + 2])
      .rotateZ(joints.angles[arm0 + 3] * -sign);
    limbs[pelvis + 0] =  Pose3D(0, robotDimensions.lengthBetweenLegs / 2.0f * -sign, 0)
      .rotateX(-pi_4 * sign)
      .rotateZ(joints.angles[leg0 + 0] * sign);
    limbs[pelvis + 1] = Pose3D(limbs[pelvis + 0])
      .rotateX((joints.angles[leg0 + 1] + pi_4) * sign);
    limbs[pelvis + 2] = Pose3D(limbs[pelvis + 1])
      .rotateY(joints.angles[leg0 + 2]);
    limbs[pelvis + 3] = Pose3D(limbs[pelvis + 2])
      .translate(0, 0, -robotDimensions.upperLegLength)
      .rotateY(joints.angles[leg0 + 3]);
    limbs[pelvis + 4] = Pose3D(limbs[pelvis + 3])
      .translate(0, 0, -robotDimensions.lowerLegLength)
      .rotateY(joints.angles[leg0 + 4]);
    limbs[pelvis + 5] = Pose3D(limbs[pelvis + 4])
      .rotateX(joints.angles[leg0 + 5] * sign);
  }

  // calculate center of mass
  const MassCalibration::MassInfo& torso(massCalibration.masses[MassCalibration::torso]);
  // initialize accumulators
  centerOfMass = torso.offset * torso.mass;
  totalMass = torso.mass;
  for(int i = 0; i < numOfLimbs; i++)
  {
    const MassCalibration::MassInfo& limb(massCalibration.masses[i]);
    totalMass += limb.mass;
    centerOfMass += (limbs[i] * limb.offset) * limb.mass;
  }
  centerOfMass /= totalMass;
}

void RobotModel::draw()
{
  DECLARE_DEBUG_DRAWING3D("representation:RobotModel", "origin");
  COMPLEX_DRAWING3D("representation:RobotModel",
    for(int i = 0; i < numOfLimbs; ++i)
    {
      const Pose3D& p = limbs[i];
      const Vector3<>& v = p.translation;
      const Vector3<> v1 = p * Vector3<>(200, 0, 0);
      const Vector3<> v2 = p * Vector3<>(0, 200, 0);
      const Vector3<> v3 = p * Vector3<>(0, 0, 200);
      LINE3D("representation:RobotModel", v.x, v.y, v.z, v1.x, v1.y, v1.z, 1, ColorRGBA(255, 0, 0));
      LINE3D("representation:RobotModel", v.x, v.y, v.z, v2.x, v2.y, v2.z, 1, ColorRGBA(0, 255, 0));
      LINE3D("representation:RobotModel", v.x, v.y, v.z, v3.x, v3.y, v3.z, 1, ColorRGBA(0, 0, 255));
    }
  );
  DECLARE_DEBUG_DRAWING3D("representation:RobotModel:centerOfMass", "origin");
  SPHERE3D("representation:RobotModel:centerOfMass", centerOfMass.x, centerOfMass.y, centerOfMass.z, 25, ColorRGBA(255, 0, 0));
}
