/**
 * @file Slider.cpp
 * 
 * Implementation of class Slider
 *
 * @author <A href="mailto:Tim.Laue@dfki.de">Tim Laue</A>
 * @author <A href="mailto:kspiess@informatik.uni-bremen.de">Kai Spiess</A>
 */ 

#include "Slider.h"
#include "Motor.h"
#include "MotorController.h"
#include "Simulation/Actuatorport.h"
#include "Simulation/Simulation.h"
#include "PhysicalObjects/PhysicalObject.h"
#include "Tools/SimGraphics.h"
#include "Platform/Assert.h"


Slider::Slider(const AttributeSet& attributes)
{
  name = ParserUtilities::getValueFor(attributes, "name");
  jointPositionReading.data.doubleValue = 0.0;
  jointPositionReading.minMaxRangeValid = false;
  jointPositionReading.unitForDisplay = ("m");
  jointPositionReading.dimensions.push_back(1);
  jointPositionReading.sensorType = SensorReading::doubleSensor;
  jointPositionReading.portName = "position";
  motorVelocityReading.data.doubleValue = 0.0;
  motorVelocityReading.minMaxRangeValid = false;
  motorVelocityReading.unitForDisplay = ("m/s");
  motorVelocityReading.dimensions.push_back(1);
  motorVelocityReading.sensorType = SensorReading::doubleSensor;
  motorVelocityReading.portName = "velocity";
}


bool Slider::parseAttributeElements(const std::string& name, 
                                    const AttributeSet& attributes,
                                    int line, int column,
                                    ErrorManager* errorManager)
{
  if(name == "Axis")
  {
    parseAxis(name, attributes);
    axis = &axes[0];
  }
  else if(name == "AngularMotor")
  {
    parseMotor(name, attributes);
  }
  else if(name == "VelocityMotor")
  {
    parseMotor(name, attributes);
  }
  else if(name == "Deflection")
  {
    axes.back().minValue = (ParserUtilities::toDouble(ParserUtilities::getValueFor(attributes, "min")));
    axes.back().maxValue = (ParserUtilities::toDouble(ParserUtilities::getValueFor(attributes, "max")));
    axes.back().limited = true;
  }
  else if(name == "PController")
  {
    axis->motor->parseController(name, attributes);
  }
  else if(name == "PIDController")
  {
    axis->motor->parseController(name, attributes);
    ((PIDController*)axis->motor->getMotorController())->setTimeStep(simulation->getStepLength());
  }
  else if(name == "Friction")
  {
    if(!axes.empty())
      axis->friction = ParserUtilities::toDouble(ParserUtilities::getValueFor(attributes, "value"));
  }
  return true;
}

SimObject* Slider::clone() const
{
  Slider* newSlider = new Slider();
  //copy SimObject members
  newSlider->copyStandardMembers(this);
  newSlider->jointPositionReading = jointPositionReading;
  newSlider->motorVelocityReading = motorVelocityReading;
  bool createsNewMovID = newSlider->adjustMovableID();
  //copy joint members
    //axes
  std::vector<AxisDescription>::const_iterator axesIter;
  for(axesIter = axes.begin(); axesIter != axes.end(); ++axesIter)
  {
    newSlider->axes.push_back(*axesIter);
    if(axesIter->motor != NULL)
    {
      newSlider->motors.push_back(axesIter->motor->clone());
      newSlider->axes.back().motor = newSlider->motors.back();
    }

  }
    //currentMotor
  if(!newSlider->motors.empty())
    newSlider->currentMotor = newSlider->motors.back();

  //copy slider members
  newSlider->axis = &(newSlider->axes[0]);

  //copy child nodes
  newSlider->childNodes.clear();
  std::list<SimObject*>::const_iterator pos;
  for(pos = childNodes.begin(); pos != childNodes.end(); ++pos)
  {
    SimObject* childNode = (*pos)->clone();
    newSlider->addChildNode(childNode, false);
  }

  if(createsNewMovID)
    simulation->removeCurrentMovableID();

  SimObject* newObject = newSlider;
  return newObject;
}

void Slider::addToLists(std::vector<SensorPort>& sensorPortList,
                        std::vector<Actuatorport*>& actuatorportList,
                        std::vector<Actuator*>& actuatorList) 
{
  // Complete sensor port structs:
  jointPositionReading.fullSensorName = fullName;
  if(axes.back().limited)
  {
    jointPositionReading.minMaxRangeValid = true;
    jointPositionReading.minValue = axes.back().minValue;
    jointPositionReading.maxValue = axes.back().maxValue;
  }
  motorVelocityReading.fullSensorName = fullName;
  if(axis->motor)
  {
    motorVelocityReading.minMaxRangeValid = true;
    motorVelocityReading.maxValue = axis->motor->getMaxVelocity();
    motorVelocityReading.minValue = -motorVelocityReading.maxValue;
  }
  // Add to sensor port list:
  sensorPortList.push_back(SensorPort(fullName+"."+jointPositionReading.portName, JOINT_POSITION, this));
  if(axis->motor && axis->motor->getMotorType() == VELOCITY)
  {
    sensorPortList.push_back(SensorPort(fullName+"."+motorVelocityReading.portName, MOTOR_VELOCITY, this));
  }
  // Add to actuator (port) lists:
  if(axis->motor) // No motor, no actuation...
  {
    actuatorList.push_back(this);
    if(axis->motor->getMotorType() == VELOCITY)
    {
      Actuatorport* velocityPort = new Actuatorport("velocity", this, MOTOR_VELOCITY, 
        -axis->motor->getMaxVelocity(), axis->motor->getMaxVelocity());
      actuatorportList.push_back(velocityPort);
    }
    else // (type == ANGULAR)
    {
      Actuatorport* anglePort = new Actuatorport("position", this, JOINT_POSITION, 
        axis->minValue, axis->maxValue);
      actuatorportList.push_back(anglePort);
    }
  }
}

void Slider::addToDescriptions(std::vector<ObjectDescription>& objectDescriptionTree,
                               int depth) 
{
  Sensor::addToDescriptions(objectDescriptionTree, depth);
  ObjectDescription valueDesc;
  // Slider position can always be sensed:
  valueDesc.name = jointPositionReading.portName;
  valueDesc.fullName = fullName + "." + jointPositionReading.portName;
  valueDesc.depth = depth + 1;
  valueDesc.type = OBJECT_TYPE_SENSORPORT;
  objectDescriptionTree.push_back(valueDesc);

  if(axis->motor)
  {
    std::string actuatorDescription;
    if(axis->motor->getMotorType() == VELOCITY)
    {
      actuatorDescription = motorVelocityReading.portName;
      //Add sensor port for motor velocity:
      valueDesc.name = motorVelocityReading.portName;
      valueDesc.fullName = fullName + "." + motorVelocityReading.portName;
      valueDesc.depth = depth + 1;
      valueDesc.type = OBJECT_TYPE_SENSORPORT;
      objectDescriptionTree.push_back(valueDesc);

    }
    else if(axis->motor->getMotorType() == ANGULAR)
    {
      actuatorDescription = jointPositionReading.portName;
    }
    //Add one actuator port (for velocity or for angle):
    valueDesc.name = actuatorDescription;
    valueDesc.fullName = fullName + "." + actuatorDescription;
    valueDesc.depth = depth + 1;
    valueDesc.type = OBJECT_TYPE_ACTUATORPORT;
    objectDescriptionTree.push_back(valueDesc);
  }
}

void Slider::setValue(double value, int port)
{
  /* TEMPORARY FIX */
  double newValue = value;
  if(axis->motor->getMotorType() == ANGULAR)
  {
    if(value == axis->maxValue || value == axis->minValue)
    {
      if(value == 0.0 && axis->minValue == 0.0)
      {
        newValue = -0.01;
      }
      else if(value == 0.0 && axis->maxValue == 0.0)
      {
        newValue = 0.01;
      }
      else
        newValue -= (value *dReal(0.01));
    }
  }
  axis->motor->setValue(newValue, port);
}

void Slider::act(bool initial)
{
  if(!initial)
  {
    if(axis->motor->getMotorType() == VELOCITY)
    {
      dReal desiredVelocity = dReal(axis->motor->getDesiredVelocity());
      dJointSetSliderParam(physicalJoint, dParamFMax, dReal(axis->motor->getMaxForce()));
      dJointSetSliderParam(physicalJoint, dParamVel, desiredVelocity);
    }
    else if(axis->motor->getMotorType() == ANGULAR)
    {
      double controllerOutput = axis->motor->getControllerOutput(dJointGetSliderPosition(physicalJoint));
      dJointSetSliderParam(physicalJoint, dParamFMax, dReal(axis->motor->getMaxForce()));
      dJointSetSliderParam(physicalJoint, dParamVel, dReal(controllerOutput));
    }
  }
}

SensorReading& Slider::getSensorReading(int localPortId)
{
  if(localPortId == JOINT_POSITION)
  {
    jointPositionReading.data.doubleValue = dJointGetSliderPosition(physicalJoint);
    return jointPositionReading;
  }
  else //if(localPortId == MOTOR_VELOCITY)
  {
    motorVelocityReading.data.doubleValue = dJointGetSliderParam(physicalJoint, dParamVel);
    return motorVelocityReading;
  }
}

void Slider::createJointPhysics()
{
  Vector3d physAxis = axis->direction;
  physAxis.rotate(rotation);

  physicalJoint = dJointCreateSlider(*simulation->getPhysicalWorld(), 0);
  PhysicalObject* o1 = getPhysicalParentObject();
  PhysicalObject* o2 = getPhysicalChildObject(this);
  ASSERT(o1 != 0 && o2 != 0);
  if(o1 && o2)
  {
    dJointAttach(physicalJoint, *(o2->getBody()), *(o1->getBody()));
    //set slider joint parameter
    dJointSetSliderAxis(physicalJoint, dReal(axis->direction.v[0]), dReal(axis->direction.v[1]), dReal(axis->direction.v[2]));

    if(axis->cfm != -1.0)
      dJointSetSliderParam(physicalJoint, dParamCFM, dReal(axis->cfm));

    if(axis->limited || (axis->motor != 0 && axis->motor->getMotorType() == ANGULAR))
    {
      dJointSetSliderParam(physicalJoint, dParamLoStop, dReal(axis->minValue));
      dJointSetSliderParam(physicalJoint, dParamHiStop, dReal(axis->maxValue));
      // this has to be done due to the way ODE sets joint stops
      dJointSetSliderParam(physicalJoint, dParamLoStop, dReal(axis->minValue));

      if(axis->fullScaleDeflectionCFM != -1.0)
        dJointSetSliderParam(physicalJoint, dParamStopCFM, dReal(axis->fullScaleDeflectionCFM));
      if(axis->fullScaleDeflectionERP != -1.0)
        dJointSetSliderParam(physicalJoint, dParamStopERP, dReal(axis->fullScaleDeflectionERP));
    }
  }
}

void Slider::applyFriction(bool initial)
{
  if(!initial)
  {
    if(motors.size()== 0)
    {
      double fmax = axis->friction * fabs((simulation->getPhysicsParameters())->getGravity());
      double velocity = 0.0;
      dJointSetSliderParam(physicalJoint, dParamFMax, dReal(fmax));
      dJointSetSliderParam(physicalJoint, dParamVel, dReal(velocity));
    }
  }
}

void Slider::drawObject(const VisualizationParameterSet& visParams)
{
  if(visParams.surfaceStyle == VisualizationParameterSet::PHYSICS_WIREFRAME ||
     visParams.surfaceStyle == VisualizationParameterSet::PHYSICS_FLAT ||
     visParams.surfaceStyle == VisualizationParameterSet::PHYSICS_SMOOTH ||
     visParams.surfaceStyle == VisualizationParameterSet::MIXED_APPEARANCE ||
     visParams.surfaceStyle == VisualizationParameterSet::MIXED_PHYSICS)
  {
    dVector3 tempA;
    dJointGetSliderAxis(physicalJoint, tempA);
    Vector3d tempAxis(tempA[0], tempA[1], tempA[2]);
    tempAxis *= 0.1;
    dBodyID b2 = dJointGetBody(physicalJoint, 0);
    if(b2 == NULL)
      b2 = dJointGetBody(physicalJoint, 1);
    const dReal* b2Pos = dBodyGetPosition(b2);
    Vector3d pos(b2Pos[0], b2Pos[1], b2Pos[2]);

    if(visParams.surfaceStyle == VisualizationParameterSet::MIXED_APPEARANCE)
    {
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glShadeModel(GL_FLAT);
      GLHelper::getGLH()->drawSlider(pos, tempAxis, true);
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glShadeModel(GL_SMOOTH);
    }
    else
      GLHelper::getGLH()->drawSlider(pos, tempAxis, false);
  }
}
