/**
* @file Math/Common.h
*
* This contains some often used mathematical definitions and functions.
*
* @author <a href="mailto:martin.kallnik@gmx.de">Martin Kallnik</a>
* @author Max Risler
*/

#ifndef Common_H
#define Common_H

#include <cmath>
#include <cstdlib>

/**
* Returns the sign of a value.
* \param a The value.
* \return The sign of \c a.
*/
template <class V> inline int sgn(const V& a) {return a < 0 ? -1 : ((a == 0) ? 0 : 1);}

/**
* Calculates the square of a value.
* \param a The value.
* \return The square of \c a.
*/
template <class V> inline V sqr(const V& a) {return a * a;}

/**
* Calculates the sec of a value.
* \param a The value.
* \return The sec of \c a.
*/
template <class V> inline V sec(const V& a) {return 1/cosf(a);}

/**
* Calculates the cosec of a value.
* \param a The value.
* \return The cosec of \c a.
*/
template <class V> inline V cosec(const V& a) {return 1/sinf(a);}

/** @name constants for some often used angles */
///@{
/** constant for a half circle*/
const float pi = 3.1415926535897932384626433832795f;
/** constant for a full circle*/
const float pi2 = 2.0f*pi;
/** constant for three quater circle*/
const float pi3_2 = 1.5f*pi;
/** constant for a quarter circle*/
const float pi_2 = 0.5f*pi;
/** constant for a 1 degree*/
const float pi_180 = pi/180.0f;
/** constant for a 1/8 circle*/
const float pi_4 = pi*0.25f;
/** constant for a 3/8 circle*/
const float pi3_4 = pi*0.75f;
/** constant for an expression used by the gaussian function*/
const float sqrt2pi = sqrtf(2.0f*pi);
/** constant for converting radians to degrees */
const float _180_pi = 180.0f / pi;
///@}

/** 
* Converts angle from rad to degrees.
* \param angle code in rad
* \return angle coded in degrees
*/
template <class V> inline V toDegrees(const V& angle){return angle * V(_180_pi);}

/** Converts angle from degrees to rad.
* \param degrees angle coded in degrees
* \return angle coded in rad
*/
inline float fromDegrees(int degrees){return float(degrees) * float(pi_180);}

/** Converts angle from degrees to rad.
 * \param degrees angle coded in degrees
 * \return angle coded in rad
 */
template <class V> inline V fromDegrees(const V& degrees){return degrees * V(pi_180);}

/** 
* reduce angle to [-pi..+pi[
* \param data angle coded in rad
* \return normalized angle coded in rad
*/
template <class V> inline V normalize(const V& data)
{
  if(data < V(pi) && data >= -V(pi)) return data;
  V ndata = data - ((int )(data / V(pi2))) * V(pi2);
  if(ndata >= V(pi))
  {
    ndata -= V(pi2); 
  }
  else if(ndata < -V(pi))
  {
    ndata += V(pi2);
  }
  return ndata;
}


// TODO: move the random functions into a separated file?

/** constant, cast before execution*/
const float RAND_MAX_FLOAT = static_cast<float>(RAND_MAX);

/**
* The function returns a random number in the range of [0..1].
* @return The random number.
*/
inline float randomFloat() {return float(rand()) / RAND_MAX_FLOAT;}

/**
* The function returns a random integer number in the range of [0..n-1].
* @param n the number of possible return values (0 ... n-1)
* @return The random number.
*/
inline int random(int n) {return (int)(randomFloat()*n*0.999999);}

/**
* The function returns a random integer number in the range of [0..n].
* @param n the number of possible return values (0 ... n)
* @return The random number.
*/
inline int randomFast(int n)
{
  return static_cast<int>((rand()*n) / RAND_MAX_FLOAT);
}

/**
* Round to the next integer
* @param d A number
* @return The number as integer
*/
inline int roundNumberToInt(float d)
{
  return static_cast<int>(floor(d+0.5f));
}

/**
* Round to the next integer but keep type
* @param d A number
* @return The number 
*/
inline float roundNumber(float d)
{
  return floor(d+0.5f);
}

#endif // Common_H
