/** 
* @file Tools/ProcessFramework/TeamHandler.cpp
* The file implements a class for team communication between simulated robots.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#include "TeamHandler.h"
#include "Platform/GTAssert.h"
#include "Platform/SystemCall.h"
#include "Tools/Streams/OutStreams.h"
#include "Tools/Streams/InStreams.h"

TeamHandler::TeamHandler(MessageQueue& in, MessageQueue& out) :
in(in),
out(out),
port(0)
{
}

void TeamHandler::start(int port, const char* subnet)
{
  ASSERT(!this->port);
  this->port = port;
  socket.setBlocking(false);

  if(port == LOCAL_TEAM_COMM)
  {
    VERIFY(socket.setBroadcast(false));
    std::string group = SystemCall::getHostAddr();
    group = "239" + group.substr(group.find('.'));
    VERIFY(socket.bind("0.0.0.0", port));
    VERIFY(socket.setTTL(0)); //keep packets off the network. non-standard(?), may work.
    VERIFY(socket.joinMulticast(group.c_str()));
    VERIFY(socket.setTarget(group.c_str(), port));
  }
  else
  {
    VERIFY(socket.setBroadcast(true));
    VERIFY(socket.bind("0.0.0.0", port));
    socket.setTarget(subnet, port);
    socket.setLoopback(false);
  }
}

void TeamHandler::send()
{
  if(!out.isEmpty())
  {
    OutBinarySize size;
    size << out;
    char* package = new char[size.getSize() + 4];
    ASSERT(package);
    OutBinaryMemory memory(package + 4);
    memory << out;
    ((unsigned short*) package)[1] = size.getSize() + 4;
    ((unsigned short*) package)[0] = calcCRC16(package + 2, size.getSize() + 2);

    if(socket.write(package, size.getSize() + 4))
      out.clear();

    delete [] package;
  }
}

void TeamHandler::receive()
{
  if(!port)
    return; // not started yet

  char buffer[1400];
  int size;
  do
  {
    size = port != LOCAL_TEAM_COMM 
           ? socket.read(buffer, sizeof(buffer))
           : socket.readLocal(buffer, sizeof(buffer));
    if(size >= 4 && size == ((unsigned short*) buffer)[1] && 
       ((unsigned short*) buffer)[0] == calcCRC16(buffer + 2, size - 2))
    {
      in.out.bin << SystemCall::getCurrentSystemTime();
      in.out.finishMessage(idReceiveTimeStamp);
      InBinaryMemory memory(buffer + 4, size - 4);
      memory >> in;
      ASSERT(memory.eof());
    }
  }
  while(size > 0);
}

unsigned short TeamHandler::calcCRC16(const char* address, unsigned size) const
{
  unsigned short crc16 = 0;
  unsigned char lastChar = 0;
  while(size--)
  {
    if(crc16 & 0x8000)
      crc16 = (crc16 << 1) ^ 0x8005;
    else
      crc16 <<= 1;
    unsigned char c = *address++;
    crc16 ^= (unsigned short) (c | (lastChar << 8));
    lastChar = c;
  }
  return crc16;
}
