/** 
* @file Tools/ProcessFramework/TeamHandler.h
* The file declares a class for the team communication between robots.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#ifndef __TeamHandler_h_
#define __TeamHandler_h_

#include "Platform/UdpComm.h"
#include "Tools/MessageQueue/MessageQueue.h"

#define LOCAL_TEAM_COMM 9876

#define TEAM_COMM \
  MessageQueue theTeamReceiver; \
  MessageQueue theTeamSender; \
  TeamHandler theTeamHandler;

#define INIT_TEAM_COMM \
  theTeamHandler(theTeamReceiver, theTeamSender)

#ifdef TARGET_SIM
#define START_TEAM_COMM \
  theTeamHandler.start(LOCAL_TEAM_COMM, "");
#else
#define START_TEAM_COMM \
  theTeamHandler.start(Global::getSettings().teamPort, "255.255.255.255");
#endif

#define RECEIVE_TEAM_COMM \
  theTeamHandler.receive()

#define SEND_TEAM_COMM \
  theTeamHandler.send()

/** 
* @class TeamHandler
* A class for team communication by broadcasting.
*/
class TeamHandler
{
public:
  /**
  * Constructor.
  * @param in Incoming debug data is stored here.
  * @param out Outgoing debug data is stored here.
  */
  TeamHandler(MessageQueue& in, MessageQueue& out);

  /**
  * The method starts the actual communication on the given port.
  * @param port The UDP port this handler is listening to.
  * @param subnet The subnet the handler is broadcasting to.
  */
  void start(int port, const char* subnet);

  /**
  * The method sends the outgoing message queue if possible.
  */
  void send();

  /**
  * The method receives packages if available.
  */
  void receive();

private:
  MessageQueue& in, /**< Incoming debug data is stored here. */
              & out; /**< Outgoing debug data is stored here. */
  int port; /**< The UDP port this handler is listening to. */
  UdpComm socket; /**< The socket used to communicate. */

  /**
  * The method calculates a CRC16 checksum of a memory area.
  * @param address The begin of the area.
  * @param size The length of the area in bytes.
  */
  unsigned short calcCRC16(const char* address, unsigned size) const;
};

#endif 
